/**
 * @author Stanislav Lapitsky
 * @version 1.0
 * 
 * Edited by Trung, N.V
 * 21 June, 2008
 * Modified: 29 Jul, 2008
 * Modified: 13 Aug, 2008
 */

import java.awt.*;

import javax.swing.*;
import javax.swing.text.*;

import com.Ostermiller.Syntax.HighlightedDocument;
/**
 * An extended class of JEditorPane that
 * shows line numbers in this editor's view.
 * It also supports syntax highlighting for C language.
 */
public class LNTextPane extends JTextPane {

	public LNTextPane() {
        setEditorKit(new NumberedEditorKit());
    }
    
    public boolean isEmpty() {
    	if (this.getText() == null)
    		return true;
    	if (this.getText().equals(""))
    		return true;
    	
    	return false;
    }
}

class NumberedEditorKit extends StyledEditorKit {

	public Document createDefaultDocument() {
		return new HighlightedDocument();
	}
	
    public ViewFactory getViewFactory() {
    	return new NumberedViewFactory();
    }
}

class NumberedViewFactory implements ViewFactory {
    public View create(Element elem) {
        String kind = elem.getName();
    	//return new NumberedParagraphView(elem);
    	
        if (kind != null)
            if (kind.equals(AbstractDocument.ContentElementName)) {
                return new LabelView(elem);
            }
            else if (kind.equals(AbstractDocument.
                             ParagraphElementName)) {
//              return new ParagraphView(elem);
                return new NumberedParagraphView(elem);
            }
            else if (kind.equals(AbstractDocument.
                     SectionElementName)) {
                return new BoxView(elem, View.Y_AXIS);
            }
            else if (kind.equals(StyleConstants.
                     ComponentElementName)) {
                return new ComponentView(elem);
            }
            else if (kind.equals(StyleConstants.IconElementName)) {
                return new IconView(elem);
            }
        // default to text display
        return new LabelView(elem);
    }
}

class NumberedParagraphView extends ParagraphView {
    public static short NUMBERS_WIDTH=30;

    public NumberedParagraphView(Element e) {
        super(e);
        short top = 0;
        short left = 0;
        short bottom = 0;
        short right = 0;
        this.setInsets(top, left, bottom, right);
    }

    protected void setInsets(short top, short left, short bottom,
                             short right) {
    	super.setInsets(top,(short)(left+NUMBERS_WIDTH),
                             bottom,right);
    }

    public void paintChild(Graphics g, Rectangle r, int n) {
        super.paintChild(g, r, n);
        int previousLineCount = getPreviousLineCount();
        int numberX = r.x - getLeftInset();
        int numberY = r.y + r.height - 5;
        if (n == 0 && count != 1) {
        	Color oldColor = g.getColor();
        	g.setColor(new Color(85,107,47));
        	g.drawString(Integer.toString(previousLineCount + n ),
                                      numberX, numberY);
        	g.setColor(oldColor);
        }	
    }

    public int getPreviousLineCount() {
        int lineCount = 0;
        View parent = this.getParent();
        View prev = null, curr = null;
        int count = parent.getViewCount();
        
        this.count = count;
        for (int i = 0; i < count; i++) {
            if (parent.getView(i) == this) {
                break;
            }
            else {
                curr = parent.getView(i);
                if (!curr.equals(prev))
                	lineCount += 1;
                prev = curr;
            }
        }
        return lineCount;
    }
    
    public int count;
}