
/**
 * A panel that displays source tree browser
 */

import javax.swing.*;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreeSelectionModel;

import java.io.*;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.*;
/**
 * TreePanel displays a tree. 
 * Used to display source tree and dot file tree
 * 
 * @author TRUNG, N.V
 *
 */
public class TreePanel extends JPanel {
    protected JTree tree;
    
    public TreePanel() {
        super(new GridLayout(1,0));
        if (MainFrame.projectStarted) { 
        	tree = ProjectResource.getTree(MainFrame.graphMode);
            // added
            tree.setCellRenderer(new MyRenderer());
        }	
        else 
        	tree = ProjectResource.getTree();
        
        tree.getSelectionModel().setSelectionMode
                (TreeSelectionModel.SINGLE_TREE_SELECTION);

        //Create the scroll pane and add the tree to it. 
        JScrollPane treeView = new JScrollPane(tree);
        tree.addMouseListener(new MyMouseListener());
        
        treeView.setMinimumSize(new Dimension(160, 500));
        treeView.setPreferredSize(new Dimension(200, 500));

        add(treeView);
    }

	// handle double-click on a tree's node
    class MyMouseListener extends MouseAdapter {
    	public void mouseClicked(MouseEvent e) {
    		if (e.getClickCount() != 2)
    			return;
    	    // it's a double click
    		DefaultMutableTreeNode node 
    	        = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();

    	    if (node == null) return;
    	    
    	    // get node info object
    	    Object nodeInfo =  node.getUserObject(); 
    	    if (nodeInfo instanceof SourceNodeInfo) {
    	       	SourceNodeInfo sni = (SourceNodeInfo) nodeInfo;
    	       	MainFrame.displaySource(sni.sourceFile);
    	       	// if there is a dot file associated with this source
    	       	if (sni.dotFile != null)
    	       		MainFrame.displayGraph(sni.dotFile, sni.toString());
    	    } else if (nodeInfo instanceof FuncNodeInfo) {
    	    	FuncNodeInfo fni = (FuncNodeInfo) nodeInfo;
	    	    DefaultMutableTreeNode parent 
	    	    	= (DefaultMutableTreeNode) node.getParent();
	    	    File src = ((SourceNodeInfo) parent.getUserObject()).sourceFile;
	    	    MainFrame.displaySource(src);
	    	    MainFrame.displayGraph(fni.dotFile, fni.name);
    	    } else  if (nodeInfo instanceof DotNodeInfo ){
    	    	DotNodeInfo dni = (DotNodeInfo) nodeInfo;
    	    	MainFrame.displayGraph(dni.dotFile, dni.toString());
    	    }  	
    	}
    }

    private class MyRenderer extends DefaultTreeCellRenderer {
        public Component getTreeCellRendererComponent(
                            JTree tree,
                            Object value,
                            boolean sel,
                            boolean expanded,
                            boolean leaf,
                            int row,
                            boolean hasFocus) {

            super.getTreeCellRendererComponent(
                            tree, value, sel,
                            expanded, leaf, row,
                            hasFocus);
            setIcon(leafIcon);
            // root icon should be open icon
            if (isRoot(value.toString()))
            	setIcon(openIcon);
            return this;
        }
        
        boolean isRoot(String str) {
        	for (int i = 0; i < ProjectResource.graphs.length; i++)
        		if (ProjectResource.graphs[i].equals(str))
        			return true;
        	return false;
        }
    }
}
