import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.InputStream;
import java.io.PrintWriter;

import javax.swing.JComponent;
import javax.swing.JPanel;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;


import att.grappa.Element;
import att.grappa.Graph;
import att.grappa.GraphEnumeration;
import att.grappa.GrappaPanel;
import att.grappa.Parser;
/**
 * Some utilities to process general tasks.
 * 
 * @author NGUYEN, T.V
 *
 */
public class Utils {
	/**
	 * Returns a grappa panel for the dot file
	 * @param fileName: name of the dot file
	 * @return a grappa panel
	 */
	public static GrappaPanel getGrappaPanel(File file) {
		try {
			// run this command to convert .dot file
			// to .gdot file (grappa compatibility)
			Utils.runCommand("dot", "dot -Tdot " + file.getAbsolutePath()
					+ " -o " + file.getAbsolutePath() + ".g");
			File gdotFile = new File(file.getAbsolutePath() + ".g");
			InputStream input = new FileInputStream(gdotFile);
	    	Parser program = new Parser(input, System.err);
			try {
			    program.parse();
			} catch(Exception ex) {
			    System.err.println("Exception: " + ex.getMessage());
			    System.exit(1);
			}
			
			Graph graph = program.getGraph();
			graph.setEditable(true);
			graph.setErrorWriter(new PrintWriter(System.err,true));
			
			GrappaPanel gp = new GrappaPanel(graph);
			gp.addGrappaListener(new MyGrappaAdapter());
			gp.setScaleToFit(false);
			
			input.close();
			return gp;
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}
		// !!
		return null;
	}
	
	/**
	 * Return a component that can be used for
	 * holding space. For example, it is used to
	 * reserve space for the grappa panel in this 
	 * application. 
	 * It also set the preferred size of this component
	 * and the minimum size. Minimum size is calculated
	 * as (w * 0.5, h) 
	 * @param w : preferred width of this component
	 * @param h : preferred height of this component
	 */
	public static JComponent getComp(int w, int h) {
		JComponent comp = new JPanel();
		comp.setPreferredSize(new Dimension(w, h));
		comp.setMinimumSize(new Dimension((int) (w * 0.5), h));
		// make it the same as menu bar's background color
		comp.setBackground(new Color(238, 238, 238)); 
		return comp;
	}
	
	/**
	 * Returns the content of a file as string
	 */
	public static String getText(File file) {
		String res = "";
		try {
			BufferedReader in = new BufferedReader(
					new FileReader(file));
			String line;
			while ((line = in.readLine()) != null) {
				res += line + "\n";
			}
			in.close();
		}	catch (Exception e) {
			MainFrame.printMessage(e.getMessage());
		}

		return res;
	}

	/**
	 * Run the clap.byte file command when a file is double clicked
	 * Requires the path to the directory that contains "clap.byte"
	 * is set in the PATH evironment 
	 * @param file
	 */
	public static void runCommand(String cmd, String execStmt) {
        try {
            Runtime rt = Runtime.getRuntime();
            Process pr = rt.exec(execStmt);
            
            int exitVal = pr.waitFor();
            if (exitVal != 0)
            	throw new Exception(execStmt);
        } catch(Exception e) {
            MainFrame.printMessage("Command : " + execStmt + " failed...");
        }
	}
	
	/**
	 * Read all sources of this project from PROJECT.XML file
	 * and return as array of strings. Each element of the array
	 * is path to a source file.
	 * @return
	 */
	public static String[] getSources() {
		String path = ProjectResource.root.getAbsolutePath() 
		+ "/" + GUIConstants.PROJECT_FILE;
		Document doc = XMLReader.getDocument(path); 
		NodeList sources = doc.getElementsByTagName("source");
		String[] res = new String[sources.getLength()];
		for (int i = 0; i < sources.getLength(); i++) {
			org.w3c.dom.Element sourceElement = (org.w3c.dom.Element) sources.item(i);
			res[i] = sourceElement.getAttribute("pathToSource");
		}
		
		return res;
	}
	
}