(* la/slicing_path.ml                                            *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open LogicUtil;;
open LogicUtil.Logic;;
open Cfa;;
open Art;;
open La_common;;

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let add_var_to_remaining_vars (var_to_add : expression)
    (remaining_vars : expression list)
: expression list =
begin
    let is_mem : bool = is_mem_of_expr_list (var_to_add) (remaining_vars) in
    if is_mem = false then begin
        let new_remaining_vars = var_to_add :: remaining_vars in
        new_remaining_vars
    end else remaining_vars
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* if a_var remaining_vars haven't contained, we add it to remaining_vars  *)
let add_vars_to_remaining_vars (vars_to_add : expression list)
    (remaining_vars : expression list)
: expression list =
begin
    let new_remaining_vars = ref remaining_vars in
    List.iter
        (fun a_var_to_add ->
            let is_mem = is_mem_of_expr_list (a_var_to_add) (!new_remaining_vars) in
            if is_mem = false then
              new_remaining_vars := a_var_to_add :: !new_remaining_vars
        )
        vars_to_add;
    (* return *)
    !new_remaining_vars
end

(* ~~~~~~~~~~~~~~~~~~~~ slicing eq_predicate ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let slice_eq_predicate (left_expr : expression)
    (right_expr : expression)
    (remaining_vars : expression list)
    (remaining_preds : predicate list)
    (is_replaced : bool)
: (expression list) * (predicate option) * (predicate list) =
begin
    (* debug *)
    (*debug ("---------------------------slice eq predicate: left_var: " ^ 
      (string_of_expression left_expr));
    debug ("remaining_vars: " ^ (string_of_expression (Sum remaining_vars)));*)
    
    (* find all vars in left expr *)
    let left_var = (*match find_all_vars_in_expression (left_expr) with
                   | a_var :: _ -> a_var
                   | [] -> assert false*) left_expr
    in
    (*debug ("left_var: " ^ (string_of_expression left_var));*)
    (* 20090620. If left expr contains memory which is in remaining_vars: *)
    let memories_of_leftarrays = find_arrays_in_expr (left_expr) in
    (* 20090620. check if var_of_left is remaining or not *)
    (*let is_remaining = is_mem_of_expr_list (left_var) (remaining_vars) in*)
    let is_remaining : bool = 
        an_element_of_exprs_is_mem (left_var :: memories_of_leftarrays) 
          (remaining_vars) 
    in
    (* find all vars of right_expr *)
    let vars_of_right_expr = find_all_vars_in_expression (right_expr) in
    (* 20090619. Deal with arrays: index of array in the left of assignment
    need to remain, too                                                     *)
    let index_of_leftarrays = find_indexes_of_arrays_in_expr (left_expr) in
    (* if left_expr contains var to remain, update vars_of_right_expr *)
    if is_remaining = true then begin
        let new_remaining_vars =
            add_vars_to_remaining_vars (*vars_of_right_expr*) 
              (vars_of_right_expr @ index_of_leftarrays) (remaining_vars) in
        if is_replaced = false then begin
            (* remove left_var in new_remaining_vars *)
            (*debug "remove left_var in new_remaining_vars";*)
            let remaining_vars_to_return = ref [] in
            List.iter
                (fun a_remaining_var ->
                    (*debug ("a_remaining_var: " ^ (string_of_expression a_remaining_var));
                    debug ("left_var: " ^ (string_of_expression left_var));*)
                    let is_mem_of_exprs : bool =
                        is_mem_of_expr_list (a_remaining_var) [left_var]
                    in
                    (*debug ("is_mem_of_remaining_var: " ^ (string_of_bool is_mem_of_exprs));*)
                    if is_mem_of_exprs = false then
                    remaining_vars_to_return :=
                        !remaining_vars_to_return @ [a_remaining_var]
                )
                new_remaining_vars;
            let eq = Eq(left_expr, right_expr) in
            (*(!remaining_vars_to_return, Some (eq), eq :: remaining_preds)*)
            (new_remaining_vars, Some (eq), eq :: remaining_preds)
        end else begin
            if remaining_preds = [] then begin
                let eq = Eq(left_expr, right_expr) in
                (* return *)
                (new_remaining_vars, Some (eq),[eq])
            end else begin
                let vars_of_remaining_preds =
                    find_all_vars_in_predicate_list (remaining_preds) in
                (* check if left_var is mem of remaining_preds or not *)
                let flag_to_get_weakest_precondition =
                    is_mem_of_expr_list (left_var) (vars_of_remaining_preds) in
                if flag_to_get_weakest_precondition = true then begin
                    let new_remaining_preds = ref [] in
                    List.iter
                        (fun a_remaining_pred ->
                            let a_new_remaining_pred =
                                replace_var_in_predicate (a_remaining_pred)
                                  (left_var) (right_expr)
                            in
                            new_remaining_preds := !new_remaining_preds @
                                                   [a_new_remaining_pred]
                        )
                        remaining_preds;
                    (* remove left_var in new_remaining_vars *)
                    (*debug "remove left_var in new_remaining_vars";*)
                    let remaining_vars_to_return = ref [] in
                    List.iter
                        (fun a_remaining_var ->
                            (*debug ("a_remaining_var: " ^ (string_of_expression a_remaining_var));
                            debug ("left_var: " ^ (string_of_expression left_var));*)
                            let is_mem_of_exprs : bool =
                                is_mem_of_expr_list (a_remaining_var) [left_var]
                            in
                            (*debug ("is_mem_of_remaining_var: " ^ (string_of_bool is_mem_of_exprs));*)
                            if is_mem_of_exprs = false then
                                remaining_vars_to_return :=
                                    !remaining_vars_to_return @ [a_remaining_var]
                        )
                        new_remaining_vars;
                    (* return *)
                    (!remaining_vars_to_return, None, !new_remaining_preds)
                end else begin
                    (new_remaining_vars, Some (Eq (left_expr, right_expr)),
                    (Eq (left_expr, right_expr)) :: remaining_preds)
                end
            end
        end
    end else begin
        (* debug ("is_remaining = false"); *)
        (remaining_vars, None, remaining_preds)
    end
end

(* ~~~~~~~~~~~~~~~~~~~ slicing implication ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec slice_an_implication (cond_pred : predicate)
    (consequent_pred : predicate)
    (remaining_vars : expression list)
    (remaining_preds : predicate list)
: (expression list) * (predicate option) * (predicate list) =
begin
    match consequent_pred with
    | And pred_list -> begin
        let (new_remaining_vars, new_edge_option, _) =
            slice_predicate_of_assignment (consequent_pred) (remaining_vars)
              ([]) (false)
        in
        (match new_edge_option with
        | None -> (new_remaining_vars, None, remaining_preds)
        | Some new_edge -> begin
            let vars_of_cond_pred = find_all_vars_in_predicate (cond_pred) in
            let remaining_vars_to_return =
                add_vars_to_remaining_vars (vars_of_cond_pred) (new_remaining_vars)
            in
            (* return *)
            let new_implication = Or [Not cond_pred; new_edge] in
            (remaining_vars_to_return, Some (new_implication),
            new_implication :: remaining_preds)
          end
        )
      end
    | a_consequent_pred -> begin
        (* slide a_consequent_pred *)
        let (new_remaining_vars_of_consequence, new_remaining_pred_option, _) =
            slice_predicate_of_assignment (a_consequent_pred) (remaining_vars)
              ([]) (false)
        in
        (match new_remaining_pred_option with
        | Some p -> begin
            let vars_of_cond_pred = find_all_vars_in_predicate (cond_pred) in
            let new_remaining_vars_of_cond =
                add_vars_to_remaining_vars (vars_of_cond_pred)
                  (new_remaining_vars_of_consequence)
            in
            let new_implication = Or [Not cond_pred; a_consequent_pred] in
            (new_remaining_vars_of_cond, Some new_implication,
            new_implication :: remaining_preds)
          end
        | None -> (remaining_vars, None, remaining_preds)
        )
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
and slice_predicate_of_assume (remaining_vars : expression list)
    (remaining_preds : predicate list)
    (assume : predicate)
: (expression list) * (predicate option) * (predicate list) =
begin
    (* debug debug "slice_predicate_of_assume"; let _ = read_int() *)
    (* in Step1. Find vars of assume:                                      *)
    let vars_of_assume : expression list = find_all_vars_in_predicate (assume) in
    (* debug debug ("vars_of_assume: "); debug             *)
    (* (string_of_expression (Sum vars_of_assume)); debug          *)
    (* ("remaining_vars: "); debug (string_of_expression (Sum      *)
    (* remaining_vars)); let _ = read_int() in                             *)
    
    (* Check if one of vars_of_assume is in remaining_vars or not: *)
    let is_remaining : bool ref = ref false in
    List.iter
        (fun a_var ->
            if !is_remaining = false then begin
              is_remaining := is_mem_of_expr_list (a_var) (remaining_vars);
            end
        )
        vars_of_assume;
    (**)
    if !is_remaining = true then begin
        (* debug debug ("!is_remaining = true"); let _ =           *)
        (* read_int() in                                                   *)
        let new_remaining_vars : expression list =
            add_vars_to_remaining_vars (vars_of_assume) (remaining_vars) in
        let new_remaining_preds : predicate list = assume :: remaining_preds in
        (* return *)
        (new_remaining_vars, Some assume, new_remaining_preds)
    end else begin
        (* debug ("!is_remaining = false"); let _ = read_int() in *)
        (remaining_vars, None, remaining_preds)
    end
end

(* ~~~~~~~~~~~~~~~~~~~ slice_predicate_of_assignment ~~~~~~~~~~~~~~~~~~~~ *)
and slice_predicate_of_assignment (pred : predicate)
    (remaining_vars : expression list)
    (remaining_preds : predicate list)
    (is_replaced : bool)
: (expression list) * (predicate option) * (predicate list) =
begin
    (* debug *)
    (*debug ("slice_predicate_of_assignment:");             
    debug ("pred: " ^ string_of_predicate pred); debug  
     ("remaining_vars: " ^ string_of_expression (Sum remaining_vars));*)   
     (*debug ("remaining_preds: " ^ string_of_predicate (And       
     remaining_preds)); let _ = read_int() in                            *)
    
    match pred with
    | True -> (remaining_vars, None, remaining_preds)
    | False -> assert false
    (* 24.2.2009. Added the following to deal with "assume" *)
    | Not _ ->
        slice_predicate_of_assume (remaining_vars) (remaining_preds) (pred)
    | Eq (expr1, expr2) ->
        slice_eq_predicate (expr1) (expr2) (remaining_vars) (remaining_preds)
          (is_replaced)
    | Or [Not cond_pred; consequent_pred] ->
        slice_an_implication (cond_pred) (consequent_pred) (remaining_vars)
          (remaining_preds)
    | And pred_list -> begin
        let rev_pred_list = List.rev pred_list in
        let new_remaining_vars = ref remaining_vars in
        let new_remaining_preds = ref [] in
        let new_pred_list : predicate list ref = ref [] in
        List.iter
            (fun a_pred ->
                (* debug debug(string_of_predicate a_pred); let _  *)
                (* = read_int() in                                         *)
              
                (match a_pred with
                | Or _ -> begin
                    let (crvars, new_edge_option , _) =
                        slice_predicate_of_assignment (a_pred)
                          (!new_remaining_vars) ([]) (false)
                    in
                    (match new_edge_option with
                    | Some p -> begin
                        new_pred_list := p :: !new_pred_list;
                        new_remaining_preds := p :: !new_remaining_preds;
                        new_remaining_vars := crvars
                      end
                    |None -> ())
                  end
                | _ -> begin
                    let (crvars, new_edge_option , current_remaining_preds) =
                        slice_predicate_of_assignment (a_pred)
                          (!new_remaining_vars) (!new_remaining_preds)
                          (is_replaced)
                    in
                    (match new_edge_option with
                    | Some p -> new_pred_list := p :: !new_pred_list;
                        new_remaining_preds := current_remaining_preds;
                        new_remaining_vars := crvars;
                    | None -> new_pred_list := current_remaining_preds;
                        new_remaining_preds := current_remaining_preds;
                        new_remaining_vars := crvars;)
                  end)
            )
            rev_pred_list;
        (* return *)
        if !new_pred_list = [] then (remaining_vars, None, !new_remaining_preds)
        else if (List.length !new_pred_list) = 1 then begin
            let hd_of_new_preds = try List.hd !new_pred_list
                                  with
                                  | Failure _ -> assert false
            in
            (!new_remaining_vars, Some hd_of_new_preds, !new_remaining_preds)
        end else
            let new_edge = And !new_pred_list in
            (!new_remaining_vars, Some (new_edge), !new_remaining_preds)
        end
    | anything_else -> begin
        emerg (string_of_predicate pred);
        assert false
      end;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~ slicing path ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Note: index of old_edges_in_path are changed Replace edge by True if we *)
(* don't need it to prove infeasible path                                  *)
let rec slice_path (path : art_path) : art_path =
begin
    (* debug *)
    (*debug ("slice path"); read_int();*)

    let new_edges_in_path = ref [] in
    let remaining_vars : expression list ref = ref [] in
    
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = List.rev (path.art_nodes_in_path) in
    let rev_art_edges_in_path = List.rev (path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    
    (* foreach block in path, slice it: *)
    (try
        (List.iter2
            (fun a_node an_edge ->
                match a_node.art_type with
                | ART_COND -> begin
                    let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                    (* update information: *)
                    remaining_vars :=
                        add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                    (* remains the edge: *)
                    new_edges_in_path := an_edge :: !new_edges_in_path;
                  end
                | ART_ASSUME -> begin
                    (match an_edge with
                    | And [pred_of_assume; pred_of_reset] -> begin
                        let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                        (* Slice reset first *) 
                        let (new_remaining_vars, new_reset_option, _) =
                            slice_predicate_of_assignment (pred_of_reset)(!remaining_vars)([])
                              (*true*) (false)
                        in
                        let remaining_reset : predicate = 
                            (match new_reset_option with
                            | Some p ->
                                remaining_vars := new_remaining_vars;
                                p
                            | None -> True)
                        in
                        (* remains the edge: *)
                        new_edges_in_path := 
                            And [pred_of_assume; remaining_reset] :: !new_edges_in_path;
                      end
                    | pred_of_assume -> begin
                        let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                        (* remains the edge: *)
                        new_edges_in_path := pred_of_assume :: !new_edges_in_path;
                      end)
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    let (new_remaining_vars, new_edge_option, _) =
                        slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                          (*true*) (false)
                    in
                    (match new_edge_option with
                    | Some p ->
                        new_edges_in_path := p :: !new_edges_in_path;
                        remaining_vars := new_remaining_vars;
                    | None ->
                        new_edges_in_path := True :: !new_edges_in_path);
                  end
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* return *)
    let new_path : art_path = {
        art_nodes_in_path = path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~ slicing path ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Note: index of old_edges_in_path are changed Replace edge by True if we *)
(* don't need it to prove infeasible path                                  *)
(*let rec slice_path_with_criterion_vars (path : art_path)
                                       (criterion_vars : expression list) 
: art_path =
begin
    (* debug *)
    (*debug ("slice_path_with_criterion_vars"); read_int();
    debug ("criterion_vars: " ^ (string_of_expression (Sum criterion_vars)));
    let _ = read_int () in*)

    let new_edges_in_path = ref [] in
    let remaining_vars : expression list ref = ref criterion_vars in
    
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = List.rev (path.art_nodes_in_path) in
    let rev_art_edges_in_path = List.rev (path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    
    (* foreach block in path, slice it: *)
    (try
        (List.iter2
            (fun a_node an_edge ->
                match a_node.art_type with
                | ART_COND -> begin
                    (*let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                    (* update information: *)
                    remaining_vars :=
                        add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);*)
                    (* remains the edge: *)
                    new_edges_in_path := True :: !new_edges_in_path;
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    let (new_remaining_vars, new_edge_option, _) =
                        slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                          (*true*) (false)
                    in
                    (match new_edge_option with
                    | Some p ->
                        new_edges_in_path := p :: !new_edges_in_path;
                        remaining_vars := new_remaining_vars;
                    | None ->
                        new_edges_in_path := True :: !new_edges_in_path);
                  end
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* return *)
    let new_path : art_path = {
        art_nodes_in_path = path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end*)

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_min_subset_of_thread_conditions (thread_conditions : predicate list)
    (cond_labels : (string * int) list)
: (string * int) list =
begin
    (* debug *) 
    (*debug ("find_min_subset_of_thread_conditions");      
    List.iter 
        (fun a_tc -> debug (string_of_predicate a_tc);) 
        thread_conditions; 
    let _ = read_int () in*)
    
    match thread_conditions with
    | [] -> []
    | [a_tc] -> begin
        (* Check sat before return *)
        if (is_sat a_tc) = false then cond_labels
        else []
      end
    | _ -> begin
        (* Create subset to return *)
        let min_cl_subset : (string * int) list ref = ref cond_labels in
        (* Create some vars *)
        let flag : bool ref = ref false in
        let left_tcs : predicate list ref = ref [] in
        let right_tcs : predicate list ref = ref thread_conditions in
        let left_cond_labels : (string * int) list ref = ref [] in
        let right_cond_labels : (string * int) list ref = ref cond_labels in
        (* Remove an element of thread_conditions which isn't useful when  *)
        (* prove infeasible path, then call                                *)
        (* find_min_subset_of_thread_conditions for it                     *)
        List.iter
            (fun a_tc ->
                if !flag = false then begin
                    let removing_tc = List.hd !right_tcs in
                    let removing_cl = List.hd !right_cond_labels in
                    right_tcs := List.tl !right_tcs;
                    right_cond_labels := List.tl !right_cond_labels;
                    (* Create new_tc_subset and check sat *)
                    let new_tc_subset = !left_tcs @ !right_tcs in
                    let new_cl_subset = !left_cond_labels @ !right_cond_labels in
                    let sat : bool = is_sat (And new_tc_subset) in
                    if sat = false then begin
                        min_cl_subset :=
                            find_min_subset_of_thread_conditions (new_tc_subset)
                              (new_cl_subset);
                        flag := true;
                    end else begin
                        left_tcs := !left_tcs @ [removing_tc];
                        left_cond_labels := !left_cond_labels @ [removing_cl];
                    end
                end
            )
            thread_conditions;
        (* Return *)
        !min_cl_subset
    end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec update_value_of_thread_conditions (thread_conditions : predicate list ref)
    (an_edge : predicate)
: unit =
begin
    match an_edge with
    | True -> ()
    | False -> assert false
    | And pred_list -> begin
        let rev_pred_list : predicate list = List.rev pred_list in
        let _ = List.map (update_value_of_thread_conditions thread_conditions)
           (rev_pred_list) in
        ()
      end
    | Or pred_list -> begin
        (* debug debug (string_of_predicate an_edge); assert false *)
        (* Find all vars in assume                                         *)
        let vars_in_assume : expression list =
            find_all_vars_in_predicate (Or pred_list) in
        (* Find all var in thread_condition *)
        let new_thread_conditions : predicate list ref = ref [] in
        List.iter
            (fun a_tc ->
                (* Find all vars in a_tc *)
                let vars_in_tc : expression list =
                    find_all_vars_in_predicate (a_tc) in
                (* Check if any var of the assume is mem of vars_in_tc or  *)
                (* not                                                     *)
                let is_mem : bool =
                    an_element_of_exprs_is_mem (vars_in_assume) (vars_in_tc) in
                (* If true, then add assume to current thread-condition *)
                if is_mem = true then begin
                    let new_tc : predicate = And [an_edge; a_tc] in
                    new_thread_conditions := !new_thread_conditions @ [new_tc]
                end else begin
                    new_thread_conditions := !new_thread_conditions @ [a_tc]
                end
            )
            !thread_conditions;
        thread_conditions := !new_thread_conditions;
      end
    | Not pred (* an assume *) -> begin
        (* Find all vars in assume *)
        let vars_in_assume : expression list = 
            find_all_vars_in_predicate (pred) in
        (* Find all var in thread_condition *)
        let new_thread_conditions : predicate list ref = ref [] in
        List.iter
            (fun a_tc ->
                (* Find all vars in a_tc *)
                let vars_in_tc : expression list =
                    find_all_vars_in_predicate (a_tc) in
                (* Check if any var of the assume is mem of vars_in_tc or  *)
                (* not                                                     *)
                let is_mem : bool =
                    an_element_of_exprs_is_mem (vars_in_assume) (vars_in_tc) in
                (* If true, then add assume to current thread-condition *)
                if is_mem = true then begin
                    let new_tc : predicate = And [an_edge; a_tc] in
                    new_thread_conditions := !new_thread_conditions @ [new_tc]
                end else begin
                    new_thread_conditions := !new_thread_conditions @ [a_tc]
                end
            )
            !thread_conditions;
        thread_conditions := !new_thread_conditions;
      end
    | Eq (expr1, expr2) (* an assignment *) -> begin
        let new_thread_conditions : predicate list ref = ref [] in
        List.iter
            (fun a_tc ->
                (* Compute weakest-precondition for a thread condition *)
                let a_new_tc : predicate =
                    get_weakest_precondition (a_tc) (an_edge) in
                (* 20090619. Xuan check a_new_tc *)
                if (is_identical_predicates a_tc a_new_tc) = true then begin
                    (* find arrays in threadcondition & the left of assignment *)
                    let arrays_in_a_tc : expression list = find_arrays_in_pred a_tc in
                    let arrays_in_asgn : expression list = find_arrays_in_expr (expr1) in
                    (* Check if any array of the left of assignment is mem of 
                    arrays in the threadcondition:                            *)
                    let is_mem : bool = 
                        an_element_of_exprs_is_mem (arrays_in_asgn) (arrays_in_a_tc) in
                    (* If true, then add the assignment to current threadcond *)
                    if is_mem = true then begin
                        new_thread_conditions := 
                            !new_thread_conditions @ [And [Eq(expr1, expr2); a_tc]]
                    end else begin
                        new_thread_conditions := !new_thread_conditions @ [a_new_tc]
                    end
                end else begin
                    new_thread_conditions := !new_thread_conditions @ [a_new_tc]
                end
                (* 20090619. end *)
            )
            !thread_conditions;
        thread_conditions := !new_thread_conditions;
      end
    | Lt _ | Leq _ | Atom _ -> assert false
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in WeiBenbacher_paradigm to find real conditions  *)
(* which need to prove counterexample. Input: path has changed index &     *)
(* flatten Output: art labels of conditions in path which need to prove    *)
(* counterexample                                                          *)
let find_criterion_labels_to_slice_path (changed_index_path : art_path)
: (string * int) list =
begin
    (* debug debug "find criteria to slice path"; debug    *)
    (* ("flatten path: "); debug (string_of_predicate (And         *)
    (* changed_index_path.art_edges_in_path)); let _ = read_int () in      *)
    
    (* New thread_conditions *)
    let thread_conditions : predicate list ref = ref [] in
    (* New condition labels list *)
    let cond_labels : (string * int) list ref = ref [] in
    (* Step1. Create thread_conditions of the path *)
    let rev_nodes_in_path = List.rev (changed_index_path.art_nodes_in_path) in
    let rev_edges_in_path = List.rev (changed_index_path.art_edges_in_path) in
    (try
        (List.iter2
            (fun a_node an_edge ->
                match a_node.art_type with
                | ART_COND -> begin
                    (* Add new thread condition *)
                    (* Currently reverse thread_conditions to find 
                    sub-thread_conditions reach to first False-region *)
                    thread_conditions := !thread_conditions @ [an_edge];
                    cond_labels := !cond_labels @ [a_node.art_label];
                  end
                | ART_ASSUME -> begin
                    (match an_edge with
                    | And [pred_of_assume; pred_of_reset] -> begin
                        (* Consider pred_of_reset: *)
                        update_value_of_thread_conditions (thread_conditions) (pred_of_reset);
                        (* Add new thread condition *)
                        thread_conditions := !thread_conditions @ [pred_of_assume];
                        cond_labels := !cond_labels @ [a_node.art_label];
                      end
                    | pred_of_assume -> begin
                        (* Add new thread condition *)
                        thread_conditions := !thread_conditions @ [pred_of_assume];
                        cond_labels := !cond_labels @ [a_node.art_label];
                      end)
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    update_value_of_thread_conditions (thread_conditions) (an_edge)
                  end
            )
            rev_nodes_in_path
            rev_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* Step2. Remove true_thread_conditions *)
    let rm_true_thread_conditions : predicate list ref = ref [] in
    let rm_true_cond_labels : (string * int) list ref = ref [] in
    (try
        (List.iter2
            (fun a_cl a_tc ->
                (* debug debug ("a_tc: " ^ string_of_predicate     *)
                (* a_tc);                                                  *)
                let is_True : int = is_equivalent_predicates (True) (a_tc) in
                if is_True != 0 then begin
                    (* debug debug ("--> is not true"); *)
                    rm_true_cond_labels := !rm_true_cond_labels @ [a_cl];
                    rm_true_thread_conditions := !rm_true_thread_conditions @ [a_tc]
                end
            )
            !cond_labels
            !thread_conditions
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* debug debug ("rm_true_thread_conditions: "); List.iter2     *)
    (* (fun (l, v) a_tc -> debug (l ^"%" ^ (string_of_int v) ^ " " *)
    (* ^ (string_of_predicate a_tc)); ) !rm_true_cond_labels;              *)
    (* !rm_true_thread_conditions; debug (string_of_predicate (And *)
    (* !thread_conditions)); List.iter (fun a_tc -> debug          *)
    (* (string_of_predicate a_tc); ) !thread_conditions;                   *)
    
    (* debug *)
    (*debug ("thread_conditions: " ^ 
        (string_of_predicate (And !thread_conditions)));*) 
    (* Step3. Find min sub-set of the thread_condtitions which is invalid *)
    let is_feasible_path : bool = is_sat (And !thread_conditions) in
    if is_feasible_path = false then begin
        (* debug debug ("is_feasible_path = false"); *)
        let min_cl_sub_set : (string * int) list =
            find_min_subset_of_thread_conditions (!rm_true_thread_conditions)
              (!rm_true_cond_labels)
        (* find_min_subset_of_thread_conditions (!thread_conditions)               *)
        (* (!cond_labels)                                                          *)
        in
        (* Step4. Return min_cl_sub_set *)
        min_cl_sub_set
    end else []
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in WeiBenbacher_paradigm to find real conditions  *)
(* which need to prove counterexample.                                     *)
let find_criterion_labels_in_path_with_criterion_edge (changed_index_path : art_path)
      (criterion_pred : predicate)
      (criterion_label : string * int)
: (string * int) list =
begin
    (* New thread_conditions *)
    let thread_conditions : predicate list ref = ref [] in
    (* New condition labels list *)
    let cond_labels : (string * int) list ref = ref [] in
    let started_flag : bool ref = ref false in
    (* Step1. Create thread_conditions of the path *)
    let rev_nodes_in_path = List.rev (changed_index_path.art_nodes_in_path) in
    let rev_edges_in_path = List.rev (changed_index_path.art_edges_in_path) in
    (try
        (List.iter2
            (fun a_node an_edge ->
                if !started_flag = false then begin
                    (* Check current label if it is criterion_label or not: *)
                    if (compare_art_labels a_node.art_label criterion_label) = 0 then begin
                        thread_conditions := !thread_conditions @ [criterion_pred];
                        cond_labels := !cond_labels @ [("", -1)];
                        started_flag := true;
                    end;
                end else begin
                    match a_node.art_type with
                    | ART_COND -> begin
                        (* Add new thread condition *)
                        (* Currently reverse thread_conditions to find 
                        sub-thread_conditions reach to first False-region *)
                        thread_conditions := !thread_conditions @ [an_edge];
                        cond_labels := !cond_labels @ [a_node.art_label];
                     end
                    | ART_ASSUME -> begin
                        (match an_edge with
                        | And [pred_of_assume; pred_of_reset] -> begin
                            (* Consider pred_of_reset: *)
                            update_value_of_thread_conditions (thread_conditions) (pred_of_reset);
                            (* Add new thread condition *)
                            thread_conditions := !thread_conditions @ [pred_of_assume];
                            cond_labels := !cond_labels @ [a_node.art_label];
                          end
                        | pred_of_assume -> begin
                            (* Add new thread condition *)
                            thread_conditions := !thread_conditions @ [pred_of_assume];
                            cond_labels := !cond_labels @ [a_node.art_label];
                          end)
                      end
                    | _ (* maybe predicates of an assignment *) -> begin
                        update_value_of_thread_conditions (thread_conditions) (an_edge)
                      end
                end  
            )
            rev_nodes_in_path
            rev_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* Step2. Remove true_thread_conditions *)
    let rm_true_thread_conditions : predicate list ref = ref [] in
    let rm_true_cond_labels : (string * int) list ref = ref [] in
    (try
        (List.iter2
            (fun a_cl a_tc ->
                (* debug debug ("a_tc: " ^ string_of_predicate     *)
                (* a_tc);                                                  *)
                let is_True : int = is_equivalent_predicates (True) (a_tc) in
                if is_True != 0 then begin
                    (* debug debug ("--> is not true"); *)
                    rm_true_cond_labels := !rm_true_cond_labels @ [a_cl];
                    rm_true_thread_conditions := !rm_true_thread_conditions @ [a_tc]
                end
            )
            !cond_labels
            !thread_conditions
        )
    with
    | Invalid_argument _ -> assert false);

    (* Step3. Find min sub-set of the thread_condtitions which is invalid *)
    let is_feasible_path : bool = is_sat (And !rm_true_thread_conditions) in
    let subset_of_cond_labels : (string * int) list = 
        (if is_feasible_path = false then begin
            (* debug debug ("is_feasible_path = false"); *)
            let min_cl_sub_set : (string * int) list =
                find_min_subset_of_thread_conditions (!rm_true_thread_conditions)
                  (!rm_true_cond_labels)
            in
            (* Step4. Return min_cl_sub_set *)
            min_cl_sub_set
        end else [])
    in

    (* Remove ("", -1) label which we added before *)
    let final_cond_labels : (string * int) list ref = ref [] in
    List.iter
        (fun a_cl ->
            if (compare_art_labels a_cl ("", -1)) != 0 then begin
                final_cond_labels := !final_cond_labels @ [a_cl]
            end
        )
        subset_of_cond_labels;
    (* Return *)
    !final_cond_labels
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let slice_path_with_criteria_labels (not_flatten_path : art_path)
    (criteria_labels : (string * int) list)
: art_path =
begin
    (* debug *)
    (*debug "slice_path_with_criteria_labels"; 
    let _ = read_int () in*)
    
    let new_edges_in_path = ref [] in
    let remaining_vars : expression list ref = ref [] in
    
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = List.rev (not_flatten_path.art_nodes_in_path) in
    let rev_art_edges_in_path = List.rev (not_flatten_path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    
    (* foreach block in path, slice it: *)
    (try
        (List.iter2
            (fun a_node an_edge ->
                match a_node.art_type with
                | ART_COND -> begin
                    let is_mem_of_cl : bool =
                        is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                    if is_mem_of_cl = true then begin
                        let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                        (* remains the edge: *)
                        new_edges_in_path := an_edge :: !new_edges_in_path;
                    end else begin
                        let (new_remaining_vars, new_edge_option, _) =
                            slice_predicate_of_assume (!remaining_vars)
                                                          ([])
                                                          (an_edge)
                        in
                        (match new_edge_option with
                        | Some p ->
                            new_edges_in_path := p :: !new_edges_in_path;
                            remaining_vars := new_remaining_vars;
                        | None ->
                            new_edges_in_path := True :: !new_edges_in_path);
                    end
                  end
                | ART_ASSUME -> begin 
                    (match an_edge with
                    | And [pred_of_assume; pred_of_reset] -> begin
                        (* Slice reset first *)
                        let (new_remaining_vars, new_reset_option, _) = 
                            slice_predicate_of_assignment (pred_of_reset) (!remaining_vars)
                              ([]) (*true*) (false)
                        in
                        let new_reset : predicate = 
                            (match new_reset_option with
                            | Some p -> remaining_vars := new_remaining_vars;
                                        p
                            | None -> True)
                        in 
                        (* Second, slice assume *)
                        let new_assume : predicate ref = ref True in
                        let is_mem_of_cl : bool = 
                            is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                        if is_mem_of_cl = true then begin
                            let vars_in_assume = find_all_vars_in_predicate (pred_of_assume) in
                            (* Update information *)
                            remaining_vars := 
                                add_vars_to_remaining_vars (vars_in_assume) (!remaining_vars);
                            (* Update new_assume *)
                            new_assume := pred_of_assume;
                        end;
                        (* Third, save the edge to path *)
                        new_edges_in_path := 
                            (And [!new_assume; new_reset]) :: !new_edges_in_path;
                      end
                    | pred_of_assume -> begin
                        let is_mem_of_cl : bool = 
                            is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                        if is_mem_of_cl = true then begin
                            let vars_in_assume = find_all_vars_in_predicate (pred_of_assume) in
                            (* Update information: *)
                            remaining_vars := 
                                add_vars_to_remaining_vars (vars_in_assume) (!remaining_vars);
                            (* remaining the edge *)
                            new_edges_in_path := pred_of_assume :: !new_edges_in_path;
                        end else begin
                            new_edges_in_path := True :: !new_edges_in_path;
                        end
                      end)
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    (* 20090718. Assignments maybe criteria *)
                    let is_mem_of_cl : bool = 
                        is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                    (if is_mem_of_cl = true then begin
                        let vars_in_asgn = find_all_vars_in_predicate (an_edge) in
                        (* update information *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_asgn) (!remaining_vars);
                        (* remains the edge: *)
                        new_edges_in_path := an_edge :: !new_edges_in_path;
                    (* 20090718 --End *)
                    end else begin
                        let (new_remaining_vars, new_edge_option, _) =
                            slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                              (*true*) (false)
                        in
                        (match new_edge_option with
                        | Some p ->
                            new_edges_in_path := p :: !new_edges_in_path;
                            remaining_vars := new_remaining_vars;
                        | None ->
                            new_edges_in_path := True :: !new_edges_in_path);
                    end);
                  end
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);

    (* return *)
    let new_path : art_path = {
        art_nodes_in_path = not_flatten_path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let slice_path_with_criteria_labels_from_criterion (not_flatten_path : art_path)
    (criteria_labels : (string * int) list)
    (started_criterion_pred : predicate)
    (started_criterion_label : string * int)
: art_path =
begin
    (* Initialize remaining_vars & new_edges_in_path *)
    let vars_in_criterion_pred : expression list = 
        find_all_vars_in_predicate (started_criterion_pred) in
    let new_edges_in_path = ref [] in
    let remaining_vars : expression list ref = ref vars_in_criterion_pred in
    
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = List.rev (not_flatten_path.art_nodes_in_path) in
    let rev_art_edges_in_path = List.rev (not_flatten_path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    
    (* foreach block in path, slice it: *)
    let started_flag : bool ref = ref false in
    (try
        (List.iter2
            (fun (a_node : art_node) (an_edge : predicate) ->
                if !started_flag = false then (begin
                    let is_started : int =
                        compare_art_labels started_criterion_label a_node.art_label
                    in 
                    if is_started = 0 then begin
                        started_flag := true;
                    end
                end) else (begin
                    (match a_node.art_type with
                    | ART_COND -> begin
                        let is_mem_of_cl : bool =
                            is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                        if is_mem_of_cl = true then begin
                            let vars_in_cond : expression list = 
                                find_all_vars_in_predicate (an_edge) in
                            (* update information: *)
                            remaining_vars :=
                                add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                            (* remains the edge: *)
                            new_edges_in_path := an_edge :: !new_edges_in_path;
                        end else begin
                            new_edges_in_path := True :: !new_edges_in_path;
                        end
                      end
                    | ART_ASSUME -> begin 
                        (match an_edge with
                        | And [pred_of_assume; pred_of_reset] -> begin
                            (* Slice reset first *)
                            let (new_remaining_vars, new_reset_option, _) = 
                                slice_predicate_of_assignment (pred_of_reset) (!remaining_vars)
                                  ([]) (*true*) (false)
                            in
                            let new_reset : predicate = 
                                (match new_reset_option with
                                | Some p -> remaining_vars := new_remaining_vars;
                                            p
                                | None -> True)
                            in 
                            (* Second, slice assume *)
                            let new_assume : predicate ref = ref True in
                            let is_mem_of_cl : bool = 
                                is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                            if is_mem_of_cl = true then begin
                                let vars_in_assume = find_all_vars_in_predicate (pred_of_assume) in
                                (* Update information *)
                                remaining_vars := 
                                    add_vars_to_remaining_vars (vars_in_assume) (!remaining_vars);
                                (* Update new_assume *)
                                new_assume := pred_of_assume;
                            end;
                            (* Third, save the edge to path *)
                            new_edges_in_path := 
                                (And [!new_assume; new_reset]) :: !new_edges_in_path;
                          end
                        | pred_of_assume -> begin
                            let is_mem_of_cl : bool = 
                                is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                            if is_mem_of_cl = true then begin
                                let vars_in_assume = find_all_vars_in_predicate (pred_of_assume) in
                                (* Update information: *)
                                remaining_vars := 
                                    add_vars_to_remaining_vars (vars_in_assume) (!remaining_vars);
                                (* remaining the edge *)
                                new_edges_in_path := pred_of_assume :: !new_edges_in_path;
                            end else begin
                                new_edges_in_path := True :: !new_edges_in_path;
                            end
                          end)
                      end
                    | _ (* maybe predicates of an assignment *) -> begin
                        (* 20090718. Assignments maybe criteria *)
                        let is_mem_of_cl : bool = 
                            is_mem_of_art_labels (a_node.art_label) (criteria_labels) in
                        (if is_mem_of_cl = true then begin
                            let vars_in_asgn = find_all_vars_in_predicate (an_edge) in
                            (* update information *)
                            remaining_vars :=
                                add_vars_to_remaining_vars (vars_in_asgn) (!remaining_vars);
                            (* remains the edge: *)
                            new_edges_in_path := an_edge :: !new_edges_in_path;
                        (* 20090718 --End *) 
                        end else begin
                            let (new_remaining_vars, new_edge_option, _) =
                                slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                                  (*true*) (false)
                            in
                            (match new_edge_option with
                            | Some p ->
                                new_edges_in_path := p :: !new_edges_in_path;
                                remaining_vars := new_remaining_vars;
                            | None ->
                                new_edges_in_path := True :: !new_edges_in_path);
                        end);
                      end)
                end)
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);

    (* return *)
    let new_path : art_path = {
        art_nodes_in_path = not_flatten_path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end
(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* Note: Function-call has type: 
   "Application ("$function_call$", index :: funcname :: paras)"           *)
let find_possible_function_pointer_in_fc_call (fc_call_pred : predicate)
      (cutlist : Sexpr.function_entry_t list)
: expression option =
begin
    (* debug *)
    (*debug "find_possible_function_pointer_in_fc_call";
    debug ("fc_call_pred: " ^ (string_of_predicate fc_call_pred));*)
    
    (* find variable function call *)
    match fc_call_pred with
    | Eq (left_expr, Application("$function_call$", index :: funcexpr :: paras)) -> 
       begin
        (* Check type of funcname *)
        (match funcexpr with 
        | Variable name_level_star ->
            let index_of_star = (try String.index name_level_star (*name_of_fc_not_dolar*) '*'
                                with Not_found -> String.length name_level_star (*name_of_fc_not_dolar*))
            in
            let name_of_fc_not_star : string = 
                (try String.sub name_level_star (*name_of_fc_not_dolar*) 0 index_of_star
                with Invalid_argument _ -> assert false)
            in
            (* cut from 0 to rindex of "%" *)
            let index_of_pc : int = 
                (try String.rindex name_of_fc_not_star (*name_of_fc*) '%'
                with Not_found -> assert false)
            in
            let name_of_fc : string = 
                (try String.sub name_of_fc_not_star (*name_of_fc*) 0 index_of_pc
                with Invalid_argument _ -> assert false)
            in
            (* Check if the function is in cutting_funcs or not *)
            let member (id : string)
                       (cutlist : Sexpr.function_entry_t list) : bool =
                let rec find id cutlist =
                    match cutlist with
                    | [] -> false
                    | h :: t ->
                        if (id = h.Sexpr.fe_funcname) then true
                        else find id t
                in
                find id cutlist
            in
            if (member name_of_fc cutlist) then 
               None
            else
               Some funcexpr
        | Application ("sref", _) | Application ("pref", _) -> (Some funcexpr) 
        | _ -> assert false)
      end
    | anything_else -> None
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let rec find_function_call_asgn_in_art_call_edge (art_call_edge : predicate)
: predicate (* function_call_asgn *)
  * predicate (* the rest *) = 
begin
    (* debug *)
    (*debug ("find_function_call_asgn_in_art_call_edge: ");
    debug (string_of_predicate art_call_edge);*)

    match art_call_edge with
    | And [Eq (_, Application("$function_call$", _))]
    | Eq (_, Application("$function_call$", _)) -> (art_call_edge, True)
    | And [Or [Not p1; And asgns]]
    | Or [Not p1; And asgns] -> (begin
        (* Get tail of asgns, because func_call_asgn must be at last *)
        let maybe_fc_asgn : predicate = List.hd (List.rev asgns) in
        let rest1 : predicate list = List.rev (List.tl (List.rev asgns)) in
        (* Check maybe_fc_asgn *)
        let (fc_asgn, rest2) = find_function_call_asgn_in_art_call_edge (maybe_fc_asgn) in
        (* Check result & return *)
        if fc_asgn = True then (True, (Or [Not p1; And asgns]))
        else (fc_asgn, (Or [Not p1; And (rest1 @ [rest2])]))
      end)
    | And [] -> (True, And [])
    | And preds -> (begin
        (* Get tail of preds, because func_call_asgn must be at last *)
        let maybe_fc_asgn : predicate = List.hd (List.rev preds) in
        let rest1 : predicate list = List.rev (List.tl (List.rev preds)) in
        (* Check maybe_fc_asgn *)
        let (fc_asgn, rest2) = find_function_call_asgn_in_art_call_edge (maybe_fc_asgn) in
        (* Check result & return *)
        if fc_asgn = True then (True, And preds)
        else (fc_asgn, And (rest1 @ [rest2]))
      end)
    | Eq (left_expr, right_expr) -> begin
        (True, Eq (left_expr, right_expr))
      end
    | anything_else -> begin
        emerg (string_of_predicate art_call_edge); 
        assert false
      end
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let sliced_art_call_with_function_pointers (an_edge : predicate)
    (a_node : art_node)
    (cutlist : Sexpr.function_entry_t list)
    (remaining_vars : expression list ref)
    (fc_pointer_criteria : ((string * int) * expression list) list ref)
    (new_edges_in_path : predicate list ref)
: unit =
begin
    (* debug *)
    (*debug "sliced_art_call_with_function_pointers";
    debug (string_of_predicate an_edge);*)
    
    (* Find function-call-assignment *)
    let (fc_asgn, rest) = find_function_call_asgn_in_art_call_edge (an_edge) in
    (* debug *)
    (*debug ("fc_asgn: " ^ (string_of_predicate fc_asgn));
    debug ("rest: " ^ (string_of_predicate rest));*)
    
    (* Create vars *)
    let new_edge : predicate ref = ref True in
    
    (* check if fc_asgn is function call or not *)
    (match fc_asgn (*an_edge*) with
    | And [Eq (left_expr, Application("$function_call$", index :: name_of_fc :: paras))]
    | Eq (left_expr, Application("$function_call$", index :: name_of_fc :: paras)) -> 
      begin
        (* Step1. Remove left_expr if it is in remaining_vars *)
        let is_mem : bool = is_mem_of_expr_list (left_expr) (!remaining_vars) in
        if is_mem = true then begin
            remaining_vars := remove_expr_from_expr_list (left_expr) (!remaining_vars)
        end;
        (* Step2. Find possible function pointer *)
        let possible_fp : expression option =
            find_possible_function_pointer_in_fc_call
              (Eq (left_expr, Application("$function_call$", index :: name_of_fc :: paras)))
              (cutlist)
        in
        (* Check possible_fp *)
        (match possible_fp with
        | Some fp -> begin
            (* Step3a. Add to remaining_vars & update fc_pointer_criteria *)
            remaining_vars := add_var_to_remaining_vars (fp) (!remaining_vars);
          end
        | None -> ());
        (*fc_pointer_criteria :=
                (a_node.art_label, !remaining_vars) :: !fc_pointer_criteria;*)
        (* Step4. Update information to new_edge *)
        if is_mem = true then new_edge := fc_asgn;
      end
    | True -> ()
    | anything_else -> emerg (string_of_predicate fc_asgn); assert false);

    (* Process the rest: *)
    let (new_remaining_vars, new_rest_option, _) =
        slice_predicate_of_assignment (rest)(!remaining_vars)([])
          (*true*) (false)
    in
    (match new_rest_option with
    | Some p ->
        new_edges_in_path := (And [p; !new_edge]) :: !new_edges_in_path;
        remaining_vars := new_remaining_vars;
        (*fc_pointer_criteria :=
            (a_node.art_label, !remaining_vars) :: !fc_pointer_criteria;*)
    | None ->
        new_edges_in_path := !new_edge :: !new_edges_in_path
    );
    fc_pointer_criteria :=
        (a_node.art_label, !remaining_vars) :: !fc_pointer_criteria;
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* 14.05.2009. Xuan has just added to deal with function pointers          *)
let slice_path_with_function_pointers (based_not_flatten_path : art_path)
    (cutlist : Sexpr.function_entry_t list)
: art_path (*sliced path *) * (((string * int) * expression list) list)
(* list of vertexes in the path, each vertex saves list of names of        *)
(* function pointers                                                       *)
=
begin
    (* debug*)
    (*debug "slice_path_with_function_pointers";
    debug "based_flatten_path: ";
    debug (get_string_of_art_path based_not_flatten_path);
    debug "cutting_funcs: ";
    List.iter  
        (fun (name, level) -> debug (name ^ "%" ^ (string_of_int level)))
        cutting_funcs;*)
    
    let new_edges_in_path : predicate list ref = ref [] in
    let remaining_vars : expression list ref = ref [] in
    let fc_pointer_criteria : ((string * int) * expression list) list ref =
      ref [] in
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = 
        List.rev (based_not_flatten_path.art_nodes_in_path) in
    let rev_art_edges_in_path = 
        List.rev (based_not_flatten_path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    (* foreach block in path, slice it: *)
    (try
        (List.iter2
            (fun a_node an_edge ->
                (* Debug *)
                (*debug ("slice edge: " ^ (string_of_predicate an_edge));*)
                match a_node.art_type with
                | ART_CALL -> begin
                    sliced_art_call_with_function_pointers (an_edge)
                      (a_node) (cutlist) (remaining_vars)
                      (fc_pointer_criteria) (new_edges_in_path)
                  end
                | ART_COND -> begin (* ignore conditions *)
                    new_edges_in_path := True :: !new_edges_in_path;
                    fc_pointer_criteria :=
                        (a_node.art_label, !remaining_vars) :: !fc_pointer_criteria;
                  end
                | ART_ASSUME -> begin (* ignore assumes *)
                    new_edges_in_path := True :: !new_edges_in_path;
                    fc_pointer_criteria :=
                        (a_node.art_label, !remaining_vars) :: !fc_pointer_criteria;
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    (* slice assignment *)
                    let (new_remaining_vars, new_edge_option, _) =
                        slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                          (*true*) (false)
                    in
                    (match new_edge_option with
                    | Some p ->
                        let flatten_ps = flatten_implications p in
                        new_edges_in_path := (And flatten_ps) :: !new_edges_in_path;
                        remaining_vars := new_remaining_vars;
                    | None ->
                        new_edges_in_path := True :: !new_edges_in_path);
                    (* Update fc_pointer_criteria *)
                    fc_pointer_criteria :=
                        (a_node.art_label, !remaining_vars) :: !fc_pointer_criteria;
                  end
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* Create new path *)
    let new_path : art_path = {
        art_nodes_in_path = based_not_flatten_path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    (* debug *)
    (*debug ("after slice_path_with_function_pointers: ");
    debug (get_string_of_art_path new_path);
    debug ("fc_pointer_criteria:");
    List.iter
        (fun (alabel, expr_list) ->
            debug ((string_of_art_label alabel) ^ ": " ^ 
              (string_of_expression (Sum expr_list)));
        )
        !fc_pointer_criteria;
    let _ = read_int () in*)
    (* Return *)
    (new_path, !fc_pointer_criteria)
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let slice_path_backward_from_label_with_criterion_var (path : art_path)
      (criterion_var : expression)
      (criterion_label : string * int)
: art_path = 
begin
    (* debug *)
    (*debug ("slice_path_backward_from_label_with_criterion_var"); 
    read_int();*)
    (*debug ("criterion_vars: " ^ (string_of_expression (Sum criterion_vars)));
    debug ("criterion_label: " ^ (string_of_art_label criterion_label));
    let _ = read_int () in*)

    let new_edges_in_path = ref [] in
    let remaining_vars : expression list ref = ref [criterion_var] in
    
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = List.rev (path.art_nodes_in_path) in
    let rev_art_edges_in_path = List.rev (path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    
    (* foreach block in path, slice it: *)
    let start : bool ref = ref false in
    (try
        (List.iter2
            (fun a_node an_edge ->
                if !start = false then begin
                    (* Check labels to mark flag "start" *)
                    (if (compare_art_labels a_node.art_label criterion_label) = 0 then begin
                        (* Mark flag "start" = true to start slice the path *)
                        start := true;
                        (* Consider the edge *)
                        (*let vars_in_criteria_edge = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_criteria_edge) (!remaining_vars);*)
                        (* Remains the edge *)
                        new_edges_in_path := True(*an_edge*) :: !new_edges_in_path;
                    end else new_edges_in_path := True :: !new_edges_in_path)
                end else begin
                    (match a_node.art_type with
                    | ART_COND -> begin
                        (*let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                        (* Remains the edge *)
                        new_edges_in_path := an_edge :: !new_edges_in_path;*)
                        let (new_remaining_vars, new_cond_option, _) = 
                            slice_predicate_of_assume (!remaining_vars) ([])
                              (an_edge)
                        in
                        (match new_cond_option with
                        | Some c -> begin
                            remaining_vars := new_remaining_vars;
                            new_edges_in_path := c :: !new_edges_in_path;
                          end
                        | None -> new_edges_in_path := True :: !new_edges_in_path;);
                      end
                    | ART_ASSUME -> begin
                        (match an_edge with
                        | And [pred_of_assume; pred_of_reset] -> begin
                            (* Slice reset first *) 
                            let (new_remaining_vars, new_reset_option, _) =
                                slice_predicate_of_assignment (pred_of_reset)(!remaining_vars)([])
                                  (*true*) (false)
                            in
                            let remaining_reset : predicate = 
                                (match new_reset_option with
                                | Some p ->
                                    remaining_vars := new_remaining_vars;
                                    p
                                | None -> True)
                            in
                            (* Second, slice assume *)
                            let (new_remaining_vars_2, new_assume_option, _) = 
                                slice_predicate_of_assume (!remaining_vars) ([]) (pred_of_assume)
                            in
                            let remaining_assume : predicate = 
                                (match new_assume_option with
                                | Some p -> remaining_vars := new_remaining_vars_2;
                                            p
                                | None -> True)
                            in
                            (* Thirt, remains the edge: *)
                            new_edges_in_path := 
                                And [remaining_assume; remaining_reset] :: !new_edges_in_path;
                          end
                        | pred_of_assume -> begin
                            (* First, slice assume *)
                            let (new_remaining_vars, new_assume_option, _) = 
                                slice_predicate_of_assume (!remaining_vars) ([]) (pred_of_assume)
                            in
                            let remaining_assume : predicate = 
                                (match new_assume_option with
                                | Some p -> remaining_vars := new_remaining_vars;
                                            p
                                | None -> True)
                            in
                            (* Thirt, remains the edge: *)
                            new_edges_in_path := remaining_assume :: !new_edges_in_path;
                          end)
                      end
                    | _ (* maybe predicates of an assignment *) -> begin
                        let (new_remaining_vars, new_edge_option, _) =
                            slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                              (*true*) (false)
                        in
                        (match new_edge_option with
                        | Some p ->
                            new_edges_in_path := p :: !new_edges_in_path;
                            remaining_vars := new_remaining_vars;
                        | None ->
                            new_edges_in_path := True :: !new_edges_in_path);
                      end)
                end
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* return *)
    let new_path : art_path = {
        art_nodes_in_path = path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end


(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
let slice_path_backward_from_label_with_criterion_vars (path : art_path)
      (criterion_vars : expression list)
      (criterion_label : string * int)
: art_path = 
begin
    (* debug *)
    (*debug ("slice_path_backward_from_label_with_criterion_vars"); 
    (*read_int(); *)
    debug ("criterion_vars: " ^ (string_of_expression (Sum criterion_vars)));
    debug ("criterion_label: " ^ (string_of_art_label criterion_label));*)
    (*let _ = read_int () in*)

    let new_edges_in_path = ref [] in
    let remaining_vars : expression list ref = ref criterion_vars in
    
    (* reverse path to traverse *)
    let rev_art_nodes_in_path = List.rev (path.art_nodes_in_path) in
    let rev_art_edges_in_path = List.rev (path.art_edges_in_path) in
    let l1 = List.length rev_art_nodes_in_path in
    let l2 = List.length rev_art_edges_in_path in
    (if l1 != l2 then assert false);
    
    (* foreach block in path, slice it: *)
    let start : bool ref = ref false in
    (try
        (List.iter2
            (fun a_node an_edge ->
                if !start = false then begin
                    (* Check labels to mark flag "start" *)
                    (if (compare_art_labels a_node.art_label criterion_label) = 0 then begin
                        (* Mark flag "start" = true to start slice the path *)
                        start := true;
                        (*let vars_in_criterion_edge = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_criterion_edge) (!remaining_vars);*)
                        (* Remains the edge *)
                        new_edges_in_path := True(*an_edge*) :: !new_edges_in_path;
                    end else begin
                        (* Update information *)
                        new_edges_in_path := True :: !new_edges_in_path;
                    end);
                end else begin
                    (match a_node.art_type with
                    | ART_COND -> begin
                        (*let vars_in_cond = find_all_vars_in_predicate (an_edge) in
                        (* update information: *)
                        remaining_vars :=
                            add_vars_to_remaining_vars (vars_in_cond) (!remaining_vars);
                        (* Remains the edge *)
                        new_edges_in_path := an_edge :: !new_edges_in_path;*)
                        let (new_remaining_vars, new_cond_option, _) = 
                            slice_predicate_of_assume (!remaining_vars) ([])
                              (an_edge)
                        in
                        (match new_cond_option with
                        | Some c -> begin
                            remaining_vars := new_remaining_vars;
                            new_edges_in_path := c :: !new_edges_in_path;
                          end
                        | None -> new_edges_in_path := True :: !new_edges_in_path;);
                      end
                    | ART_ASSUME -> begin
                        (match an_edge with
                        | And [pred_of_assume; pred_of_reset] -> begin
                            (* Slice reset first *) 
                            let (new_remaining_vars, new_reset_option, _) =
                                slice_predicate_of_assignment (pred_of_reset)(!remaining_vars)([])
                                  (*true*) (false)
                            in
                            let remaining_reset : predicate = 
                                (match new_reset_option with
                                | Some p ->
                                    remaining_vars := new_remaining_vars;
                                    p
                                | None -> True)
                            in
                            (* Second, slice assume *)
                            let (new_remaining_vars_2, new_assume_option, _) = 
                                slice_predicate_of_assume (!remaining_vars) ([]) (pred_of_assume)
                            in
                            let remaining_assume : predicate = 
                                (match new_assume_option with
                                | Some p -> remaining_vars := new_remaining_vars_2;
                                            p
                                | None -> True)
                            in
                            (* Thirt, remains the edge: *)
                            new_edges_in_path := 
                                And [remaining_assume; remaining_reset] :: !new_edges_in_path;
                          end
                        | pred_of_assume -> begin
                            (* First, slice assume *)
                            let (new_remaining_vars, new_assume_option, _) = 
                                slice_predicate_of_assume (!remaining_vars) ([]) (pred_of_assume)
                            in
                            let remaining_assume : predicate = 
                                (match new_assume_option with
                                | Some p -> remaining_vars := new_remaining_vars;
                                            p
                                | None -> True)
                            in
                            (* Thirt, remains the edge: *)
                            new_edges_in_path := remaining_assume :: !new_edges_in_path;
                          end)    
                      end
                    | _ (* maybe predicates of an assignment *) -> begin
                        let (new_remaining_vars, new_edge_option, _) =
                            slice_predicate_of_assignment (an_edge)(!remaining_vars)([])
                              (*true*) (false)
                        in
                        (match new_edge_option with
                        | Some p ->
                            new_edges_in_path := p :: !new_edges_in_path;
                            remaining_vars := new_remaining_vars;
                        | None ->
                            new_edges_in_path := True :: !new_edges_in_path);
                      end)
                end
            )
            rev_art_nodes_in_path
            rev_art_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    (* debug *)
    (*let ni : int = List.length (path.art_nodes_in_path) in
    let ei : int = List.length (!new_edges_in_path) in
    debug ("ni = " ^ (string_of_int ni));
    debug ("ei = " ^ (string_of_int ei));
    debug ("END slice_path_backward_from_label_with_criterion_vars.");
    let _ = read_int () in*)
    (* return *)
    let new_path : art_path = {
        art_nodes_in_path = path.art_nodes_in_path;
        art_edges_in_path = !new_edges_in_path;
    } in
    new_path
end

(* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
(* This function is used in WeiBenbacher_paradigm to find real conditions  *)
(* which need to prove counterexample.                                     *)
(*let find_criterion_labels_backward_from_label (changed_index_path : art_path)
                                              (criterion_label : string * int)
                                              (criteria_conditions : predicate list)
: (string * int) list =
begin
    (* debug debug "find criteria to slice path"; debug    *)
    (* ("flatten path: "); debug (string_of_predicate (And         *)
    (* changed_index_path.art_edges_in_path)); let _ = read_int () in      *)
    
    (* New thread_conditions *)
    let thread_conditions : predicate list ref = ref criteria_conditions in
    (* New condition labels list *)
    let cond_labels : (string * int) list ref = ref [] in
    (* Step1. Create thread_conditions of the path *)
    let rev_nodes_in_path = List.rev (changed_index_path.art_nodes_in_path) in
    let rev_edges_in_path = List.rev (changed_index_path.art_edges_in_path) in
    (try
        (List.iter2
            (fun a_node an_edge ->
                match a_node.art_type with
                | ART_COND -> begin
                    (* Add new thread condition *)
                    (* Currently reverse thread_conditions to find 
                    sub-thread_conditions reach to first False-region *)
                    thread_conditions := !thread_conditions @ [an_edge];
                    cond_labels := !cond_labels @ [a_node.art_label];
                  end
                | ART_ASSUME -> begin
                    (match an_edge with
                    | And [pred_of_assume; pred_of_reset] -> begin
                        (* Consider pred_of_reset: *)
                        update_value_of_thread_conditions (thread_conditions) (pred_of_reset);
                        (* Add new thread condition *)
                        thread_conditions := !thread_conditions @ [pred_of_assume];
                        cond_labels := !cond_labels @ [a_node.art_label];
                      end
                    | pred_of_assume -> begin
                        (* Add new thread condition *)
                        thread_conditions := !thread_conditions @ [pred_of_assume];
                        cond_labels := !cond_labels @ [a_node.art_label];
                      end)
                  end
                | _ (* maybe predicates of an assignment *) -> begin
                    update_value_of_thread_conditions (thread_conditions) (an_edge)
                  end
            )
            rev_nodes_in_path
            rev_edges_in_path
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* Step2. Remove true_thread_conditions *)
    let rm_true_thread_conditions : predicate list ref = ref [] in
    let rm_true_cond_labels : (string * int) list ref = ref [] in
    (try
        (List.iter2
            (fun a_cl a_tc ->
                (* debug debug ("a_tc: " ^ string_of_predicate     *)
                (* a_tc);                                                  *)
                let is_True : int = is_equivalent_predicates (True) (a_tc) in
                if is_True != 0 then begin
                    (* debug debug ("--> is not true"); *)
                    rm_true_cond_labels := !rm_true_cond_labels @ [a_cl];
                    rm_true_thread_conditions := !rm_true_thread_conditions @ [a_tc]
                end
            )
            !cond_labels
            !thread_conditions
        )
    with
    | Invalid_argument _ -> assert false);
    
    (* debug debug ("rm_true_thread_conditions: "); List.iter2     *)
    (* (fun (l, v) a_tc -> debug (l ^"%" ^ (string_of_int v) ^ " " *)
    (* ^ (string_of_predicate a_tc)); ) !rm_true_cond_labels;              *)
    (* !rm_true_thread_conditions; debug (string_of_predicate (And *)
    (* !thread_conditions)); List.iter (fun a_tc -> debug          *)
    (* (string_of_predicate a_tc); ) !thread_conditions;                   *)
    
    (* debug *)
    (*debug ("thread_conditions: " ^ 
        (string_of_predicate (And !thread_conditions)));*) 
    (* Step3. Find min sub-set of the thread_condtitions which is invalid *)
    let is_feasible_path : bool = is_sat (normalize (And !thread_conditions)) in
    if is_feasible_path = false then begin
        (* debug debug ("is_feasible_path = false"); *)
        let min_cl_sub_set : (string * int) list =
            find_min_subset_of_thread_conditions (!rm_true_thread_conditions)
              (!rm_true_cond_labels)
        (* find_min_subset_of_thread_conditions (!thread_conditions)               *)
        (* (!cond_labels)                                                          *)
        in
        (* Step4. Return min_cl_sub_set *)
        min_cl_sub_set
    end else []
end*)
