(* parser/ccommon.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

let raise_itermax_err_fn = "__raise_itermax__"

let null_struct = {
  Cexpr.strc_struct_or_union_p = true;
  strc_tag = "";
  strc_unique_id = "" ;
  strc_typebody_table = (Hashtbl.create 10)
}

let (post_condition_hashtbl : (string, Cexpr.expr_t list)Hashtbl.t)
        = Hashtbl.create 10

let (contract_of_fun_pointer :
    (string, (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.function_contract_t *
              Cexpr.expr_t list) Hashtbl.t)
        = Hashtbl.create 10

let (contract_of_field_hashtbl :
    ((string * string), (((Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.function_contract_t)
                         * Cexpr.expr_t list)) Hashtbl.t)
       = Hashtbl.create 10
let (assign_in_func : (string, Cexpr.expr_t list)Hashtbl.t) = Hashtbl.create 10

let current_func_scope : ((string, Cexpr.binding_t) Hashtbl.t *
                              (* name scope stack *)
                              (string, Cexpr.binding_t) Hashtbl.t
                              (* hash scope stack *)) ref =
    ref (Hashtbl.create 1, Hashtbl.create 1)
let current_loop_scope : ((string, Cexpr.binding_t) Hashtbl.t *
                              (* name scope stack *)
                              (string, Cexpr.binding_t) Hashtbl.t
                              (* hash scope stack *)) ref =
    ref (Hashtbl.create 1, Hashtbl.create 1)
 
(*
  Check typenode_t is point-to or not
*)
let is_pointer ( typ : Cexpr.typenode_t ) : bool = (
   match typ with
    | Cexpr.POINTER_TO _ -> true
    | _ -> false
)

let rec get_function_type (ty : Cexpr.typenode_t) : Cexpr.typenode_t option = (
    match ty with
    | Cexpr.FUNCTION_OF _ -> Some ty
    | Cexpr.POINTER_TO tyn -> get_function_type tyn
    | Cexpr.ATTRIBUTED_TYPE (tyn, _) -> get_function_type tyn
    | _ -> None
)

let get_return_function_type (type_node : Cexpr.typenode_t)
: Cexpr.typenode_t =
begin
    match type_node with
    | Cexpr.FUNCTION_OF (ty, _) -> ty
    | _ -> assert false
end

let rec is_function_type (ty : Cexpr.typenode_t) : bool = (
    match ty with
    | Cexpr.FUNCTION_OF _ -> true
    | Cexpr.ATTRIBUTED_TYPE (tyn,_ ) -> is_function_type tyn
    | Cexpr.POINTER_TO tyn -> is_function_type tyn
    | _ -> false
)

(** Check a stmt to be either a variable declaration or not
        @return True: the stmt is variable declaration
                False: the stmt is not variable declaration
 *)
let is_stmt_var (stmt : Cexpr.stmt_t) : bool = (
    match stmt with
    | Cexpr.DEFVAR decl -> (
        let tyn = decl.Cexpr.decl_tn in
        if (is_function_type tyn) then false
        else true
    )
    | anything_else -> false 
)

let get_function_name (decl : Cexpr.declaration_t): string option = (
    match (is_function_type decl.Cexpr.decl_tn) with
    | false -> None
    | true -> (
        let (id,lev) = decl.Cexpr.decl_id in
        Some id
     )
)

let get_container_file (decl : Cexpr.declaration_t) : string option = (
    match (is_function_type decl.Cexpr.decl_tn) with
    | false -> None
    | true -> (
        match decl.Cexpr.decl_location with
        | Some (fn, _, _) -> Some fn
        | None -> None))

let is_function_pointer (ty : Cexpr.typenode_t) (expr : Cexpr.expr_t) : bool = (
    let rec check ex = (
        match ex with
        | Cexpr.VREF _
        | Cexpr.SREF _
                      -> true
        | Cexpr.OPER (_, _, e, _, _) -> (
            match e with
            | Cexpr.SREF _ -> true
            | _ -> false
        )
        | Cexpr.PREF (_, _, e, _) -> check e
        | Cexpr.ADDROF (_, e) -> check e
        | _ -> false
    )in
    let boo = check expr in
    if ((is_function_type ty) && (is_pointer ty) && boo) then true
    else false
)

let rec is_struct_or_union (ty : Cexpr.typenode_t) : bool = (
    match ty with
    | Cexpr.POINTER_TO t -> is_struct_or_union t
    | Cexpr.QUALIFIED_TYPE (t, _) -> is_struct_or_union t
    | Cexpr.TYPEDEF_TYPE (t, _ ) -> is_struct_or_union t
    | Cexpr.STRCDEF_TYPE (t, _) -> is_struct_or_union t
    | Cexpr.STRUCTURE_TYPE  _ -> true
    | _ -> false
)

let run_id : string = (
    Random.self_init ();
    string_of_int (Random.bits ()))
let (return_label : string) = "__return_t_" ^ run_id
let (return_expr_name : string) = "__RET_" ^ run_id

let make_return_variable (decl : Cexpr.declaration_t) : Cexpr.expr_t option = (
    let func_type : Cexpr.typenode_t option =
        get_function_type decl.Cexpr.decl_tn
    in
    match func_type with
    | Some (Cexpr.FUNCTION_OF (return_tn, _)) -> (
        match return_tn with
          Cexpr.NOTYPE
        (* Pseudo Types *)
        | Cexpr.SPECIFR_TYPE _
        | Cexpr.QUALIFIED_TYPE _
        | Cexpr.ATTRIBUTED_TYPE _
        | Cexpr.TYPEDEF_TYPE _
        | Cexpr.STRCDEF_TYPE _
        | Cexpr.ENUMDEF_TYPE _
        | Cexpr.ELLIPSIS_TYPE
        | Cexpr.PSEUDO_VA_LIST_TYPE -> None
        | Cexpr.GCC_VA_ARG -> assert false
        | _ -> Some (Cexpr.VREF(return_tn, return_expr_name, 3, 0)))
    | _ ->
        None
)

let current_components = ref ("", Cexpr.NOTYPE)

let (implicit_type_invariants_hashtbl :
    (string, Sexpr.ltype_invariant_t) Hashtbl.t) = Hashtbl.create 10
let (current_list_components : (string * Cexpr.typenode_t) list ref) = ref []

(*
    params :
             header : declaration of variable/struct/function prototype
    output :
             (name_struct, true) neu day la khai bao mot struct
             ("###", false) if other
*)
let get_struct_or_union_name (header : Cexpr.declaration_t) = (
    let rec get_name (tyn : Cexpr.typenode_t ) = (
        match tyn with
        | Cexpr.STRUCTURE_TYPE stype ->
            Some stype.Cexpr.strc_tag
        | Cexpr.STRCDEF_TYPE (ty,_) -> get_name ty
        | Cexpr.POINTER_TO ty -> get_name ty
        | _ -> None
    ) in
    let id,level = header.Cexpr.decl_id in
    let tyn = header.Cexpr.decl_tn in
    match get_name tyn with
    | Some name ->
        (begin
        if (Misc.string_equal id "") then
            Some name
        else
            None
        end)
    | _ -> None
)

let get_struct_type_from_binding (id: string) = (
    let res = ref Cexpr.NOTYPE in
    (List.iter(
        fun scope ->
            try
                let temp =  (Hashtbl.find scope id) in
                res := temp.Cexpr.binding_tn;
            with
            | Not_found -> ()
    )(!Cexpr.tag_scope_stack));
    !res
)

(**************************************************************************)
(*
   get Cexpr.VREF of a expr
*)
let rec get_deep_expr_from_expr (expr : Cexpr.expr_t) =
   match expr with
   | Cexpr.OPER (_,_,ex,_,_)   -> get_deep_expr_from_expr ex
   | Cexpr.SETF (_, _, ex, _)  -> get_deep_expr_from_expr ex
   | Cexpr.VREF _              -> expr
   | Cexpr.PREF (_,_, ex, _)     -> get_deep_expr_from_expr ex
   | Cexpr.SREF (_,_,ex, _)    -> get_deep_expr_from_expr ex
   | Cexpr.CALL (_, ex, _, _)     -> get_deep_expr_from_expr ex
   | Cexpr.ADDROF (_, ex)      -> get_deep_expr_from_expr ex
   | _
                              -> Cexpr.NIL

(*   params :
           field : sref
   output :
           get expression of field
*)
let get_constant_field_of (field_name : string)
                          (field_type : Cexpr.typenode_t)
                          (expr: Cexpr.expr_t) : Cexpr.expr_t = (
    Cexpr.SREF (field_type, true,
                (get_deep_expr_from_expr expr),
                (field_name^"_compare_constant"))
)


(*
    get type of struct
*)
let rec get_struct_param_info1 (tyn : Cexpr.typenode_t )
                             (struct_name : string) = (
    match tyn with
    | Cexpr.STRUCTURE_TYPE struct_type -> (
        let name = struct_type.Cexpr.strc_tag in
        if name == struct_name then
            (true, tyn)
        else
            (false, Cexpr.NOTYPE)
        )
    | Cexpr.POINTER_TO typ ->
        let boo, typ = get_struct_param_info1 typ struct_name in
        (boo, Cexpr.POINTER_TO (typ))
    | _ ->
        (false, Cexpr.NOTYPE)
)

(*
   params :
           fun_header : header of function declaration
           name_struct : name of struct
   output :
           get type of name_struct in parameters of fun_header
*)
let get_struct_param_info (fun_header : Cexpr.declaration_t)
                          (struct_name : string)  = (
    let tn = fun_header.Cexpr.decl_tn in
    let params = fun_header.Cexpr.decl_params in
    match tn with
    | Cexpr.FUNCTION_OF ( _, param_types) -> (
        match param_types with
        | None -> ("###", Cexpr.NOTYPE)
        | Some types -> (
            let name = ref "###"
            and s_t = ref Cexpr.NOTYPE in
            let temp_list = List.combine types params in
            List.iter
                (fun (param_type, param_name) ->
                    let boo, typ =
                        get_struct_param_info1 param_type
                                               struct_name in
                    if boo == true then (
                        name := param_name;
                        s_t := typ
                    )
             )temp_list;
            (!name, !s_t)))
    | _ -> assert false
)

(***************************************************************)
(*
   params :
            typ : Cexpr.typenode_t
   output :
           Get type of typenode_t (not point-to)

*)
let rec get_type_from_typenode (typ : Cexpr.typenode_t) = (
    match typ with
    | Cexpr.POINTER_TO ty ->
        get_type_from_typenode ty
    | _ -> typ
)

(*
   params :
           expr : Cexpr.VREF expression
   output :
          typ, expr : if typ is not point-to
          ty, Cexpr.PREF(ty,expr) : if typ is point-to
*)
let change_vref_expr (expr : Cexpr.expr_t) =
    match expr with
    | Cexpr.VREF(tyn, _, _, _) -> (
        if (is_pointer tyn) == true then
            (begin
            let ty = get_type_from_typenode tyn in
            (ty, Cexpr.PREF(ty, true, expr, expr))
            end)
        else
            (tyn, expr)
        )
    | _ ->
        assert false
(***************************************************************)
(* 
   params : 
            ty : type of C expression
  output :
            string : name of ty
*)
let rec name_of_type (ty : Cexpr.typenode_t) = (
   match ty with
   | Cexpr.NOTYPE -> "NOTYPE"
   | Cexpr.QUALIFIED_TYPE (t, q) -> name_of_type t      
   | Cexpr.ELLIPSIS_TYPE -> ""
   | Cexpr.STRCDEF_TYPE (t, l) -> name_of_type t       
   | Cexpr.ENUMDEF_TYPE (t, l) -> name_of_type t 
   | Cexpr.TYPEDEF_TYPE (t, s) -> name_of_type t       
   | Cexpr.SPECIFR_TYPE l -> "SPECIFR"
   | Cexpr.VOID_TYPE -> "VOID"
   | Cexpr.CHARACTER -> "CHAR"
   | Cexpr.SIGNED_CHARACTER -> "SIGNED_CHAR"
   | Cexpr.UNSIGNED_CHARACTER -> "UNSIGNED_CHAR"
   | Cexpr.SHORT_INTEGER -> "SHORT_INT"
   | Cexpr.UNSIGNED_SHORT_INTEGER -> "UNSIGNED_SHORT_INT"
   | Cexpr.INTEGER -> "INT"
   | Cexpr.UNSIGNED_INTEGER -> "UNSIGNED_INT"
   | Cexpr.LONG_INTEGER -> "LONG_INT"
   | Cexpr.UNSIGNED_LONG_INTEGER -> "UNSIGNED_LONG_INT"
   | Cexpr.LONG_LONG_INTEGER -> "LONG_LONG_INT"
   | Cexpr.UNSIGNED_LONG_LONG_INTEGER -> "UNSIGNED_LONG_LONG_INT"
   | Cexpr.SINGLE_FLOAT -> "SINGLE_FLOAT"
   | Cexpr.DOUBLE_FLOAT -> "DOUBLE_FLOAT"
   | Cexpr.LONG_DOUBLE_FLOAT -> "LONG_DOUBLE_FLOAT"
   | Cexpr.BOOLEAN -> "BOOLEAN"
   | Cexpr.SINGLE_COMPLEX -> "SINGLE_COMPLEX"
   | Cexpr.DOUBLE_COMPLEX -> "DOUBLE_COMPLEX"
   | Cexpr.LONG_DOUBLE_COMPLEX -> "LONG_DOUBLE_COMPLEX"
   | Cexpr.SINGLE_IMAGINARY -> "SINGLE_IMAGINARY"
   | Cexpr.DOUBLE_IMAGINARY -> "DOUBLE_IMAGINARY"
   | Cexpr.LONG_DOUBLE_IMAGINARY -> "LONG_DOUBLE_IMAGINARY"
   | Cexpr.POINTER_TO t -> name_of_type t     
   | Cexpr.ARRAY_OF (t, _) -> "ARRAY"      
   | Cexpr.FUNCTION_OF (t, _) -> "FUNCTION"
   | Cexpr.BITFIELD_TYPE (t, e) -> name_of_type t      
   | Cexpr.STRUCTURE_TYPE {
                             Cexpr.strc_tag = tag ;
                          } -> tag     
   | Cexpr.ENUMERATION_TYPE {
                              Cexpr.enum_tag = tag ;
                          } -> tag
   | _ -> ""
)

(*
   Returns the name of the "main" variable in expr
*)
let rec get_name_of_expr (expr : Cexpr.expr_t) =
    match expr with
    | Cexpr.OPER (_,_, ex,_,_)  -> get_name_of_expr ex
    | Cexpr.SETF (_, _, ex, _)  -> get_name_of_expr ex
    | Cexpr.VREF (_, str, _, _) -> str
    | Cexpr.PREF (_, _, ex, _)    -> get_name_of_expr ex
    | Cexpr.SREF (_,_,ex, str)  -> (get_name_of_expr ex) ^ "." ^ str
    | Cexpr.CALL (_, ex, _, _)     -> get_name_of_expr ex
    | Cexpr.ADDROF (_, ex)      -> get_name_of_expr ex
    | Cexpr.CONST_I (_, _, str) -> str
    | Cexpr.CONST_F (_, _, str) -> str
    | Cexpr.STRING (_, str, _)  -> str
    | _                         -> ""


(***************************************************************)
let find_type_in_current_components (id : string): Cexpr.typenode_t = (
    let rest_list = (
        List.map(
            fun a -> (
                let name , ty = a in
                if (Misc.string_equal name id) then [ty]
                else [])
        )(!current_list_components)
    )in
    match List.flatten(rest_list) with
    | [typ] -> typ
    | [] -> Cexpr.NOTYPE
    | _ -> assert false
)

(*************************************************************)
(*
   params :
           stmts    : list of C statement
           position : position for split
   output :
           head , tail  ( head @ tail = stmts )
   NOTE: written in a stupid way, to be improved.
*)
let rec split_stmts (stmts : Cexpr.stmt_t list)
                    (position : int)
                    (curr_pos : int)
                    (head : Cexpr.stmt_t list)
                    (tail : Cexpr.stmt_t list) = (
    match stmts with
    | [] -> head, tail
    | s::rest ->
        if curr_pos < position then
            split_stmts rest position (curr_pos + 1) (head @ [s]) tail
        else
            split_stmts rest position (curr_pos + 1) head (tail @ [s])
)

(***************************************************************)
(* translate :
         return x --> ACC = x
                      goto __return__;

         return   --> goto __return__;
   End of function:
                      __return__:
                      return ACC / return
*)


let get_return_type (decl : Cexpr.declaration_t) = (
    let perhaps_error d =
        (begin
            let decl_params_to_string p : string = (
                if (p != []) then
                    ("/(" ^ (String.concat "," p) ^ ")")
                else
                    "") in
            let string_of_fun = (
                 "(defun " ^ (Cexpr.varid_to_string d.Cexpr.decl_id)
                 ^ " : " ^ (Cexpr.type_to_string_body 1 d.Cexpr.decl_tn)
                 ^ "/" ^ (Cexpr.scls_to_string d.Cexpr.decl_scls)
                 ^ (decl_params_to_string d.Cexpr.decl_params) ^ ")") in
            raise (Sexpr.Specs_error
                      (string_of_fun ^ " is not a function", None))
        end) in
    let ty = decl.Cexpr.decl_tn in
    match ty with
    | Cexpr.FUNCTION_OF (tyn, _) -> tyn
    | Cexpr.ATTRIBUTED_TYPE (t, _) ->
        (begin
            match t with
            | Cexpr.FUNCTION_OF (t, _) -> t
            | _ -> perhaps_error decl
        end)
    | _ ->
        perhaps_error decl
)

let rec normalize_returns (curr_return_type : Cexpr.typenode_t) 
                          (defs : Cexpr.stmt_t list) 
                          : Cexpr.stmt_t list =
    List.flatten (List.map (normalize_return curr_return_type) defs)

and normalize_return (curr_return_type : Cexpr.typenode_t)
                     (stmt : Cexpr.stmt_t) 
                     : Cexpr.stmt_t list = 
    (begin
    match stmt with
    | Cexpr.RETURN None -> 
        [Cexpr.GOTO(return_label)]
    | Cexpr.RETURN Some(ex) -> 
        let temp_var = Cexpr.VREF(curr_return_type, return_expr_name, 3, 0) in
        [Cexpr.STMT(Cexpr.SETF(curr_return_type, 
                                Cexpr.SET, 
                                temp_var, ex)); 
         Cexpr.GOTO(return_label)]
    | Cexpr.DEFUN(header, body) -> (
        let return_type = get_return_type header in
        let temp_var = Cexpr.VREF(return_type, return_expr_name, 3, 0) in
        let decl = {
            Cexpr.decl_id = (return_expr_name ,3) ;
            Cexpr.decl_tn = return_type;
            Cexpr.decl_scls = Cexpr.NILSCLS;
            Cexpr.decl_inits = [];
            Cexpr.decl_params = [];
            Cexpr.decl_struct_defined_p = false;
            Cexpr.decl_specs = None;
            Cexpr.decl_location = header.Cexpr.decl_location;
            } in
        let defvars = (
            if return_type != Cexpr.VOID_TYPE then 
                [Cexpr.DEFVAR (decl)]
            else 
                [])
        in
        let label_return = Cexpr.LABEL(Cexpr.NAME_LABEL(return_label)) in
        let return_statement = (
            match return_type with
            | Cexpr.VOID_TYPE -> Cexpr.RETURN (None)
            | _ -> Cexpr.RETURN (Some temp_var)
        ) in
        let new_body = (
            match body with
            | Cexpr.BEGIN(stmts) -> (
                (*let tail_stmts = 
                    try 
                        let first_stmt = (List.hd stmts) in
                        (match first_stmt with
                         | Cexpr.NOTE _ -> Hashtbl.add 
                         | _ -> ());                       
                        (List.tl stmts)
                    with
                    | _ -> stmts
                in *)
                let new_stmts = normalize_returns return_type stmts in
                Cexpr.BEGIN(defvars @ new_stmts 
                            @ [label_return] 
                            @ [return_statement])
                )
            | _ -> (
                let name_op = get_function_name header in
                let name = (
                    match name_op with
                    | None -> ""
                    | Some str -> str)
		in
                raise (Sexpr.Specs_error
                           ("Function " ^ name ^ " has no body", None)))
        ) in
        [Cexpr.DEFUN(header, new_body)]
    )
    | Cexpr.IF(ex, stmts_t, stmts_f) -> 
        [Cexpr.IF(ex, 
                  (normalize_returns curr_return_type stmts_t),
                  (normalize_returns curr_return_type stmts_f))]
    | Cexpr.FOR(stmts1, ex1,ex2, stmts) -> 
        [Cexpr.FOR ((normalize_returns curr_return_type stmts1), 
                    ex1, 
                    ex2,
                    (normalize_returns curr_return_type stmts))]
    | Cexpr.DOWHILE (boo, ex, stmts) -> 
        [Cexpr.DOWHILE (boo, 
                        ex,
                        (normalize_returns curr_return_type stmts))]
    | Cexpr.SWITCH(ex, stmts) -> 
        [Cexpr.SWITCH(ex, (normalize_returns curr_return_type stmts))]
    | Cexpr.BEGIN(stmts) -> 
        [Cexpr.BEGIN (normalize_returns curr_return_type stmts)]
    | _ -> [stmt]
    end)

(************************************************************)
(* Get note of Entry of each function *)
let note_of_entry_function = Hashtbl.create 10
let get_note_of_entry_function (stmts : Cexpr.stmt_t list) : unit = (
    List.iter(
        fun stmt ->
            match stmt with
            | Cexpr.DEFVAR _ -> ()
            | Cexpr.DEFUN(header, body) -> (
                let fun_name = (
                    match (get_function_name header) with
                    | Some str -> str
                    | None -> assert false
                ) in
                match body with
                | Cexpr.BEGIN(s::t) -> (
                    match s with
                    | Cexpr.NOTE _ ->
                        (Hashtbl.add note_of_entry_function fun_name s)
                    | _ -> ()
                )
                | _ -> ()
            )
            | _ -> ()
    )stmts
)
(************************************************************)
(* print type name of C statement - just for debug, to be removed *)
let rec print_name_stmt (stmt : Cexpr.stmt_t) = (
   match stmt with
   | Cexpr.SPEC _          -> "spec\n"
   | Cexpr.ASSUME _ -> "assumes\n"
   | Cexpr.DEFUN (_, body) -> (
       "fun: \n"^(print_name_stmt body)
     )
   | Cexpr.DEFVAR _       -> "fun\n"
   | Cexpr.STMT _         -> "stmt\n"
   | Cexpr.ASGN _         -> "ass\n"
   | Cexpr.IF _           -> "if\n"
   | Cexpr.BEGIN _        -> "begin\n"
   | Cexpr.SWITCH _       -> "switch\n"
   | Cexpr.DOWHILE _      -> "while\n"
   | Cexpr.FOR _          -> "for\n"
   | Cexpr.BREAK          -> "break\n"
   | Cexpr.RETURN _       -> "return\n"
   | Cexpr.GOTO _         -> "goto\n"
   | Cexpr.CONTINUE       -> "continue\n"
   | Cexpr.LABEL _        -> "label\n"
   | Cexpr.ASM _          -> "asm\n"
   | Cexpr.NOTE _         -> "note\n"
   | Cexpr.ANNOTATION _   -> "annotation\n"
   | Cexpr.CALLASGN _     -> "call-assign\n"
   | _ -> "\n"
)

(*************************************************************)
let right_expr_is_function_call
        (expr: Cexpr.expr_t)
        (fun_header_hashtbl : (string,Cexpr.declaration_t) Hashtbl.t)
        : string option = (
    match expr with
    | Cexpr.CALL _
    | Cexpr.OPER(_, _, Cexpr.CALL _, _, _) -> None
    | _ -> (
        let name = (get_name_of_expr(get_deep_expr_from_expr expr)) in
        if (Hashtbl.mem fun_header_hashtbl name) then Some name
        else None
    )
)

let left_expr_is_field (expr : Cexpr.expr_t)
                       : (string * string * string) option = (
    match expr with
    | Cexpr.SREF(_, _,ex, field_name) -> (
        let rec get_name_struct ty = (
            match ty with
            | Cexpr.STRUCTURE_TYPE (s_t) -> Some(s_t.Cexpr.strc_tag)
            | Cexpr.STRCDEF_TYPE (ty1, _) -> (get_name_struct ty1)
            | Cexpr.POINTER_TO (ty) -> (get_name_struct ty)
            | _ -> None
        ) in
        let var_name = (get_name_of_expr (get_deep_expr_from_expr ex)) in
        let ty = Cexpr.type_of_expr ex in
        let struct_name = get_name_struct ty in
        match struct_name with
        | Some s_name ->  Some (s_name, field_name, var_name)
        | None -> None
    )
    | _ -> None
)

let left_expr_is_function_pointer (expr : Cexpr.expr_t): string option = (
    let ty = Cexpr.type_of_expr expr in
    match expr with
    | Cexpr.SREF _ -> None
    | _ -> (
        let name = (get_name_of_expr(get_deep_expr_from_expr expr)) in
        if ((is_pointer ty) && (is_function_type ty)) then Some name
        else None
    )
)

let find_index_expr_from_expr name exs1= (
    let res = ref 1000 in
    let ind = ref 0 in
    List.iter(
        fun ex ->
            let name1 = get_name_of_expr(get_deep_expr_from_expr ex) in
            if (Misc.string_equal name name1) then res := !ind
            else (incr ind)
    )exs1;
    !res
)
let rec replace_expr_in_exprs  (expr : Cexpr.expr_t)
                               (param_exprs : Cexpr.expr_t list)
                               (param_expr_orgs : Cexpr.expr_t list)
                               (ret_expr : Cexpr.expr_t) = (
    match expr with
    | Cexpr.OPER (ty, op, ex1, ex2, ex3) -> (
        Cexpr.OPER(ty, op,
                   (replace_expr_in_exprs ex1
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr),
                   (replace_expr_in_exprs ex2
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr),
                   (replace_expr_in_exprs ex3
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr))
    )
    | Cexpr.SETF (ty,op,ex1,ex2) -> (
        Cexpr.SETF(ty, op,
                   (replace_expr_in_exprs ex1
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr),
                   (replace_expr_in_exprs ex2
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr))
    )
    | Cexpr.VREF (ty,name, level, _)-> (
        match (find_index_expr_from_expr name param_expr_orgs) with
        | 1000 -> (
            if (Misc.string_equal name return_expr_name) &&
               (level == 3) then ret_expr
            else expr
        )
        | n -> (List.nth param_exprs n)

    )
    | Cexpr.PREF (ty, boo, ex1, ex2) -> (
        Cexpr.PREF(ty, boo, (replace_expr_in_exprs ex1
                                                   param_exprs
                                                   param_expr_orgs
                                                   ret_expr),
                            (replace_expr_in_exprs ex2
                                                   param_exprs
                                                   param_expr_orgs
                                                   ret_expr))
    )
    | Cexpr.SREF (ty, boo, ex, str) -> (
        Cexpr.SREF(ty, boo, replace_expr_in_exprs ex
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr,
                   str)
    )
    | Cexpr.ADDROF (ty, ex) ->  (
        Cexpr.ADDROF(ty, replace_expr_in_exprs ex
                                          param_exprs
                                          param_expr_orgs
                                          ret_expr)
    )
    | Cexpr.DSGN_AREF _
    | Cexpr.PROGN _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.CLITERAL _
    | Cexpr.ASMSEQUENCE _
    | Cexpr.PHI _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.INIT_FILL _
    | Cexpr.PRAGMA _
    | Cexpr.QUANTIFIER _
    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE
    | Cexpr.NIL
    | Cexpr.PSEUDO_NONDET_VALUE _
    | Cexpr.CALL _
    | Cexpr.SIZEOF _
                  -> expr
    | Cexpr.PSEUDO_ACSL_RESULT (tn, ex) -> (
        Cexpr.PSEUDO_ACSL_RESULT (tn, replace_expr_in_exprs ex
                                                            param_exprs
                                                            param_expr_orgs
                                                            ret_expr))
    | Cexpr.PSEUDO_ACSL_OLD (tn, ex0, ex1) -> (
        Cexpr.PSEUDO_ACSL_OLD (tn, replace_expr_in_exprs ex0
                                                         param_exprs
                                                         param_expr_orgs
                                                         ret_expr,
                                   replace_expr_in_exprs ex1
                                                         param_exprs
                                                         param_expr_orgs
                                                         ret_expr))
    | Cexpr.PSEUDO_ACSL_UNIQUE (tn, ex) -> (
        Cexpr.PSEUDO_ACSL_UNIQUE (tn, replace_expr_in_exprs ex
                                                            param_exprs
                                                            param_expr_orgs
                                                            ret_expr))
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn, ex0, ex1) -> (
        Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn, replace_expr_in_exprs ex0
                                                                 param_exprs
                                                                 param_expr_orgs
                                                                ret_expr,
                                           replace_expr_in_exprs ex1
                                                                 param_exprs
                                                                 param_expr_orgs
                                                                 ret_expr))
    | Cexpr.PSEUDO_ACSL_VALID (tn, ex) -> (
        Cexpr.PSEUDO_ACSL_VALID (tn, replace_expr_in_exprs ex
                                                            param_exprs
                                                            param_expr_orgs
                                                            ret_expr))
    | Cexpr.ALL_FIELD (tn, ex) ->
        Cexpr.ALL_FIELD (tn,
                         replace_expr_in_exprs ex
                                               param_exprs
                                               param_expr_orgs
                                               ret_expr)
)
let replace_exprs_in_exprs (exprs : Cexpr.expr_t list)
                           (param_exprs : Cexpr.expr_t list)
                           (param_expr_orgs : Cexpr.expr_t list)
                           (ret_expr : Cexpr.expr_t) = (
    List.map(
        fun ex ->
            (replace_expr_in_exprs ex param_exprs param_expr_orgs ret_expr)
    )exprs
)
(***********************************************)
let params_of_fun_body (decls : Cexpr.stmt_t list) = (
    let res = Hashtbl.create 10 in
    List.iter(
        fun decl ->
            match decl with
            | Cexpr.DEFUN (header, _) -> (
                let (fun_name, _) = header.Cexpr.decl_id in
                let params = header.Cexpr.decl_params in
                Hashtbl.add res fun_name params
            )
            | _ -> ()
    )decls;
    res
)
(***********************************************)
let rec replace_special_string (str : string) (sub : string) (by : string) 
: string =
begin
    let (is_taken_place, new_str) = 
        ExtString.String.replace ~str:str ~sub:sub ~by:by in
    if is_taken_place then
        replace_special_string new_str sub by
    else str    
end

let replace_special_char_in_string (str : string) : string =
begin
    let replace_special_strings (str : string) 
                                (spe_lst : string list)
                                (rep_lst : string list)
    : string = (
        try 
            List.fold_left2 (replace_special_string) str spe_lst rep_lst     
        with _ -> assert false
    ) in 
    let special_str_3 = ["..."; "==>"; "<=>"; "<<="; ">>="] in
    let replace_str_3 = ["_ellipse_"; "_impl_"; "_equi_"; "_lsh_"; "_rsh_"] in
    let replaced_str_3 = 
        replace_special_strings str special_str_3 replace_str_3 
    in
    let special_str_2 = ["->"; "<<"; ">>"; "=="; "!="; ">="; "<="; "||"; "&&"; 
                         ":="; "++"; "--"; "+="; "-="; "*="; "/="; "%="; "&=";
                         "|="; "^="] in
    let replace_str_2 = ["_arw_"; "_ltlt_"; "_gtgt_"; "_eq_"; "_ne_"; "_ge_";
                         "_le_"; "_barbar_"; "_ampamp_"; "_set_"; "_plspls_"; 
                         "_mnsmns_"; "_add_"; "_sub_"; "_mul_"; "_diveq_"; 
                         "_modeq_"; "_and_"; "_ior_"; "_xor_"] in
    let replaced_str_2 = 
        replace_special_strings str special_str_2 replace_str_2
    in
    let special_str_1 = ["$"; "*"; "("; ")"; "."; "["; "]"; "&"; ";"; ":";
                         "\""; "{"; "}"; " "; "#"; ","; "-"; "~"; "+"; "/";
                         "%"; "!"; "|"; "^"; ">"; "<"; "?"; "="] in
    let replace_str_1 = ["_doll_"; "_str_"; "_lpa_"; "_rpa_"; "_ptr_"; "_lbk_";
                         "_rbk_"; "_amp_"; "_sem_"; "_col_"; "_qum_"; "_lbc_";
                         "_rbc_"; ""; "_sha_"; "_com_"; "_mns_"; "_tld_"; 
                         "_pls_"; "_div_"; "_mod_"; "_ban_"; "_bar_"; "_hat_";
                         "_gt_"; "_lt_"; "_qes_"; "_eql_"               
                        ] in
    let replaced_str_1 = 
        replace_special_strings str special_str_1 replace_str_1 
    in replaced_str_1
end

(** Make stmt for every expr in '$old'
        @param exprs : all exprs in '$old'
        @return : stmt list        
*) 
let rec make_old_exprs_stmts (exprs : Cexpr.expr_t list)
: Cexpr.stmt_t list = 
begin
    List.flatten(
        List.map(
            fun a_expr ->
                make_init_for_old_var a_expr) 
            exprs)
end

and make_init_for_old_var (var : Cexpr.expr_t) : Cexpr.stmt_t list =
begin
    let ty = Cexpr.type_of_expr var in
    let old_expr_name = 
        ("old_of_" ^ 
            (replace_special_char_in_string (Cexpr.expr_to_string var))) in
    let old_var = Cexpr.VREF (ty, old_expr_name, 3, 0) in
    let inits = [Cexpr.SETF (ty, Cexpr.SET, old_var, var)] in
    let header =
        {
          Cexpr.decl_id = (old_expr_name, 3);
          Cexpr.decl_tn = ty;
          Cexpr.decl_scls = Cexpr.NILSCLS;
          Cexpr.decl_inits = inits;
          Cexpr.decl_params = [];
          Cexpr.decl_struct_defined_p = false;
          Cexpr.decl_specs = None;
          Cexpr.decl_location = None;
        }
    in
    ([Cexpr.DEFVAR (header)])
end
(******************************)
let artificial_assert = (
    let header = {
         Cexpr.decl_id= ("artificial_assert", 1);
          Cexpr.decl_tn= Cexpr.FUNCTION_OF (
                              Cexpr.VOID_TYPE,Some ([ Cexpr.INTEGER ]));
          Cexpr.decl_scls= Cexpr.STATIC;
          Cexpr.decl_inits= [];
          Cexpr.decl_params= [ "a" ];
          Cexpr.decl_struct_defined_p= false ;
          Cexpr.decl_specs = None;
          Cexpr.decl_location = None;
    } in

    let body =
        Cexpr.BEGIN ([
           Cexpr.IF (
                 Cexpr.OPER (
                       Cexpr.INTEGER,
                       Cexpr.EQ,
                       Cexpr.VREF (Cexpr.INTEGER,"a",2,0),
                       Cexpr.CONST_I ((Cexpr.INTEGER, Num.Int (0), "0")),
                       Cexpr.NIL),
                  [Cexpr.BEGIN ([
                         Cexpr.LABEL (Cexpr.NAME_LABEL ("__ERROR__"));
                         Cexpr.STMT (Cexpr.SETF (Cexpr.INTEGER,Cexpr.SET,
                                            Cexpr.VREF (Cexpr.INTEGER,"a",2,0),
                                            Cexpr.OPER (Cexpr.INTEGER,Cexpr.CAST (false),
                                                  Cexpr.CONST_I ((Cexpr.UNSIGNED_INTEGER,
                                                            Num.Int (0), "0")),
                                                  Cexpr.NIL,Cexpr.NIL)));
                         Cexpr.GOTO ("#:return_t")]) ],
                   [Cexpr.BEGIN (
                        [Cexpr.STMT (Cexpr.SETF (Cexpr.INTEGER,Cexpr.SET,
                                            Cexpr.VREF (Cexpr.INTEGER,"a",2,0),
                                            Cexpr.CONST_I ((Cexpr.INTEGER, Num.Int (1), "1"))));
                        Cexpr.GOTO ("#:return_t")]) ]);
            Cexpr.LABEL (Cexpr.NAME_LABEL ("#:return_t"));
            Cexpr.RETURN (None) ])
     in
     Cexpr.DEFUN (header, body)
)

(**************************)
let rec field_names (expr : Cexpr.expr_t) : string list = (
    match expr with
    | Cexpr.SETF(_, Cexpr.SET, ex, _) -> (field_names ex)
    | Cexpr.SETF _ -> []
    | Cexpr.SREF (_, _, ex, s) -> (field_names ex) @[s]
    | Cexpr.VREF (_, s, _, _) -> [s]
    | Cexpr.PREF(_, _, ex, _) -> (field_names ex)
    | _ -> []
)
let rec flatten_fields (fields : string list)
                       : (string * string) list = (
    match fields with
    | [] -> []
    | s::[] -> []
    | pre::s::tail -> (pre,s)::(flatten_fields (s::tail))
)
let fields_has_init (expr :Cexpr.expr_t) : (string * string) list = (
    flatten_fields (field_names expr)
)
let fields_has_inits (exprs :Cexpr.expr_t list) (boo : bool)
                     : (string, string list) Hashtbl.t = (
    let fields = List.flatten(List.map fields_has_init exprs) in
    let res = Hashtbl.create 10 in
    List.iter(
        fun field ->
            let (pre, next) = field in
            try
                let t = Hashtbl.find res pre in
                Hashtbl.replace res pre (next::t)
            with
            | Not_found -> (Hashtbl.add res pre [next])
    )fields;
    res
)

(******************** Is a expr in spec? ********************)
let is_unique_expr (expr : Cexpr.expr_t) : bool = (
    match expr with
    | Cexpr.PSEUDO_ACSL_UNIQUE _
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK _ -> true
    | _ -> false
)

let rec is_expr_in_expr (is_expr : Cexpr.expr_t -> bool)
                    (expr : Cexpr.expr_t)
: bool = (
    if (is_expr expr) then true
    else (
         match expr with
         | Cexpr.NIL
         | Cexpr.PSEUDO_NONDET_VALUE _
         | Cexpr.SIZEOF _
         | Cexpr.CONST_I _
         | Cexpr.CONST_F _
         | Cexpr.STRING _
         | Cexpr.VREF _
         | Cexpr.PSEUDO_MEMORY _
         | Cexpr.PRAGMA _
         | Cexpr.PSEUDO_PROLOGUE
         | Cexpr.PSEUDO_EPILOGUE -> false
         | Cexpr.SREF (_, _, ex1, _)
         | Cexpr.ADDROF (_, ex1)
         | Cexpr.INIT_FILL (_, ex1)
         | Cexpr.PSEUDO_ACSL_UNIQUE (_, ex1)
         | Cexpr.PSEUDO_ACSL_RESULT (_, ex1)
         | Cexpr.PSEUDO_ACSL_VALID (_, ex1)
         | Cexpr.ALL_FIELD (_, ex1) ->
             is_expr_in_expr is_expr ex1
         | Cexpr.SETF (_, _, ex1, ex2)
         | Cexpr.PREF (_, _, ex1, ex2)
         | Cexpr.DSGN_AREF (_, ex1, ex2)
         | Cexpr.PSEUDO_ACSL_OLD (_, ex1, ex2)
         | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (_, ex1, ex2) ->
             (is_expr_in_expr is_expr ex1) or
             (is_expr_in_expr is_expr ex2)
         | Cexpr.OPER (_, _, ex1, ex2, ex3) ->
             (is_expr_in_expr is_expr ex1) or
             (is_expr_in_expr is_expr ex2) or
             (is_expr_in_expr is_expr ex3)
         | Cexpr.CALL (_, ex1, exs1, _) ->
             (is_expr_in_expr is_expr ex1) or
             (List.exists (is_expr_in_expr is_expr) exs1)
         | Cexpr.PROGN (_, stmts) -> false (* because is_unique_expr isn't in this expr, I don't review now *)
         | Cexpr.CLITERAL (_, initializers) -> false
         | Cexpr.ASMSEQUENCE exs1 ->
             List.exists (is_expr_in_expr is_expr) exs1
         | Cexpr.PHI (_, str_exs) ->
             List.exists (is_expr_in_expr is_expr) (List.map snd str_exs)
         | Cexpr.QUANTIFIER (_, exs1, ex1) ->
             (List.exists (is_expr_in_expr is_expr) exs1) or
             (is_expr_in_expr is_expr ex1)
    )
)

let is_expr_in_func_cond (is_expr : Cexpr.expr_t -> bool)
                         (func_cond : Cexpr.full_expr_t Sexpr.func_condition_t)
: bool = (
    let full_expr = snd func_cond in
    is_expr_in_expr is_expr full_expr.Cexpr.expr_node
) 

let is_expr_in_func_conds (is_expr : Cexpr.expr_t -> bool)
                          (func_conds : Cexpr.full_expr_t Sexpr.func_condition_t list)
: bool = (
    List.exists (fun func_cond ->
        is_expr_in_func_cond is_expr func_cond
    ) func_conds;
)

let is_expr_in_spec (is_expr : Cexpr.expr_t -> bool)
                    (cspec : Cexpr.cspec_t)
: bool = (
    match cspec with
    | Sexpr.SPEC_FUN func_spec -> (
        let func_conds = func_spec.Sexpr.fspec_conditions in
        is_expr_in_func_conds is_expr func_conds
      )
    | _ -> assert false
)

let is_expr_in_spec_op (is_expr : Cexpr.expr_t -> bool)
                       (cspec_op : Cexpr.cspec_t option)
: bool = (
    match cspec_op with
    | Some cspec -> is_expr_in_spec is_expr cspec
    | None -> false
)

(* initialize for pointer *)

let rec init_pointer_stmts (stmts : Cexpr.stmt_t list) : Cexpr.stmt_t list = (
    List.map init_pointer_stmt stmts
)
and init_pointer_stmt (stmt : Cexpr.stmt_t) : Cexpr.stmt_t = (
    match stmt with
    | Cexpr.DEFUN (d, b) -> (
        Cexpr.DEFUN(d, init_pointer_stmt b)
    )
    | Cexpr.DEFVAR (decl) -> (
        init_pointer_declaration decl
    )
    | Cexpr.STMT (ex) -> Cexpr.STMT(init_pointer_expr ex)
    | Cexpr.ASGN (e1, e2) -> (
        Cexpr.ASGN(init_pointer_expr e1,
                   init_pointer_expr e2)
    )
    | Cexpr.CALLASGN (ex, s, es) -> (
        Cexpr.CALLASGN(init_pointer_expr ex,
                      init_pointer_stmts s,
                      es)
    )
    | Cexpr.BEGIN (s) -> (
        Cexpr.BEGIN(init_pointer_stmts s)
    )
    | Cexpr.IF (ex, s1, s2) -> (
        Cexpr.IF((init_pointer_expr ex),
                 (init_pointer_stmts s1),
                 (init_pointer_stmts s2))
    )
    | Cexpr.SWITCH (ex, s) -> (
        Cexpr.SWITCH((init_pointer_expr ex),
                     (init_pointer_stmts s))
    )
    | Cexpr.DOWHILE (boo, ex, s) -> (
        Cexpr.DOWHILE(boo,
                     (init_pointer_expr ex),
                     (init_pointer_stmts s))
    )
    | Cexpr.FOR (s1, e1, e2, s2) -> (
        Cexpr.FOR((init_pointer_stmts s1),
                  (init_pointer_expr e1),
                  (init_pointer_expr e2),
                  (init_pointer_stmts s2))
     )
    | Cexpr.RETURN (ex_o) -> (
        match ex_o with
        | None -> Cexpr.RETURN (None)
        | Some ex -> Cexpr.RETURN(Some(init_pointer_expr ex))
    )
    | Cexpr.BREAK
    | Cexpr.CONTINUE
    | Cexpr.LABEL _
    | Cexpr.GOTO _
    | Cexpr.ANNOTATION _
    | Cexpr.ASM _
    | Cexpr.NOTE _
    | Cexpr.SPEC _
    | Cexpr.ASSUME _
    | Cexpr.RESET _
                     -> stmt

)
and init_pointer_declaration (decl : Cexpr.declaration_t)
                             : Cexpr.stmt_t = (
    let fill_nondet_value_for_basic_type decl = (
      let ty = decl.Cexpr.decl_tn in
      let (name, level) = decl.Cexpr.decl_id in
      let v = Cexpr.VREF (ty, name, level, 0) in
      let nondet = Cexpr.PSEUDO_NONDET_VALUE (ty, "__nondet_value") in
      let e = Cexpr.SETF (ty, Cexpr.SET, v, nondet) in
      let d = { decl with Cexpr.decl_inits = [e] } in
      Cexpr.DEFVAR d) in
    let inits = decl.Cexpr.decl_inits in
    let (name, level) = decl.Cexpr.decl_id in
    let ty = decl.Cexpr.decl_tn in
    let scls = decl.Cexpr.decl_scls in
    if (String.compare name "") == 0 || scls == Cexpr.TYPEDEF then
      Cexpr.DEFVAR decl
    else if Cexpr.integral_type_p ty then
      if List.length inits == 0 then
	fill_nondet_value_for_basic_type decl
      else
	Cexpr.DEFVAR decl
    else if Cexpr.floating_type_p ty then
      if List.length inits == 0 then
	fill_nondet_value_for_basic_type decl
      else
	Cexpr.DEFVAR decl
    else (begin
      let ty' = Cexpr.strip_off_pseudo_type ty true in
      match ty' with
      | Cexpr.NOTYPE -> assert false
      | Cexpr.VOID_TYPE
      | Cexpr.FUNCTION_OF _ ->
          Cexpr.DEFVAR decl
      | Cexpr.ARRAY_OF _
      | Cexpr.STRUCTURE_TYPE _ ->
          let inits' = fill_uninitialized_slots decl inits in
          let d = { decl with Cexpr.decl_inits = inits' } in
          Cexpr.DEFVAR d
      | Cexpr.POINTER_TO _
      | Cexpr.BITFIELD_TYPE _
      | Cexpr.ENUMERATION_TYPE _ ->
	  if List.length inits == 0 then
	    fill_nondet_value_for_basic_type decl
	  else
	    Cexpr.DEFVAR decl
      | _ -> assert false
    end)
)

and init_pointer_expr (expr : Cexpr.expr_t) : Cexpr.expr_t = (
    match expr with
    | Cexpr.PROGN (ty, s) -> Cexpr.PROGN(ty, init_pointer_stmts s)
    | _ -> expr
)

(* Fills slots in a structure or an array which are not explicitly
   initialized.  It first makes initializers for the whole part, then
   removes redundant ones.  It appends the original one after the
   fillers, preferring the initializers when redundancy is not
   eliminated.  Note that removing redundant assignment cannot be
   perfect, when they include overlaps or have variables in array
   indexes.  IT IS FAR FROM PERFECT; IT WORK IN ONLY EASY CASES. *)

and fill_uninitialized_slots (decl : Cexpr.declaration_t) (inits : Cexpr.expr_t list) : Cexpr.expr_t list = (
  let (id, level) = decl.Cexpr.decl_id in
  let tn = decl.Cexpr.decl_tn in
  let vref = (Cexpr.VREF (tn, id, level, 0)) in
  let (subst, inits1) = (skip_non_initializer_part decl.Cexpr.decl_inits []) in
  match inits1 with
    [] -> 
      subst
  | (((Cexpr.INIT_FILL _) as hd) :: inits) ->
      (begin
	(* Drop the 1st of the initializer list, which is {INIT_FILL}. *)
	let wholefills = (make_slot_filler vref) in
	let fills = (filter_fills inits wholefills) in
  let remain_inits =
      init_fields_of_declaration { decl with Cexpr.decl_inits = subst @ inits1 } in
	(subst @ [hd] @ inits @ remain_inits @ fills)
      end)
  | _ -> (assert false))

(* Splits the initializer list into parts before and after
   {INIT_FILL}, with putting {INIT_FILL} in the after part.  The part
   before {INIT_FILL} is not initializers, but is substitutions of
   values by variables appearing in types.  It means a
   dynamically-sized array and should not have initializers. *)

and skip_non_initializer_part (inits : Cexpr.expr_t list) (subst : Cexpr.expr_t list) : (Cexpr.expr_t list * Cexpr.expr_t list) = (
  match inits with
    [] ->
      (subst, [])
  | ((Cexpr.INIT_FILL _) :: tl) ->
      assert (subst == []) ;
      (subst, inits)
  | (hd :: tl) ->
      (skip_non_initializer_part tl (subst @ [hd])))

and make_slot_filler (e : Cexpr.expr_t) : Cexpr.expr_t list = (
  let izero (tn : Cexpr.typenode_t) (e : Cexpr.expr_t) : Cexpr.expr_t = (
    (Cexpr.SETF (tn, Cexpr.SET, e, Cexpr.constant_zero))) in
  let fzero (tn : Cexpr.typenode_t) (e : Cexpr.expr_t) : Cexpr.expr_t = (
    (assert false)) in
  let azero (tn : Cexpr.typenode_t) (e : Cexpr.expr_t) : Cexpr.expr_t = (
    (Cexpr.SETF (tn, Cexpr.SET, e, (Cexpr.PSEUDO_MEMORY (tn, tn))))) in
  let tn = (Cexpr.type_of_expr e) in
  let nqtn = (Cexpr.strip_off_pseudo_type tn true) in
  if ((Cexpr.integral_type_p nqtn) || (Cexpr.pointer_type_p nqtn)) then
    [(izero tn e)]
  else if (Cexpr.complex_type_p nqtn) then
    [(fzero tn e)]
  else if (Cexpr.floating_type_p nqtn) then
    [(fzero tn e)]
  else
    (begin
      match nqtn with
	(Cexpr.ARRAY_OF (tn1, (Some e1))) ->
	  (begin
	    let sz = (Cexpr.fold_constant_integer e1) in
	    if (not (Cexpr.constant_integer_p sz)) then
	      (begin
		(* It is a warning, but it be an error. *)
		(Cexpr.warn_in_parse ("(cannot determine array size in"
				       ^ " initializer)")) ;
		[(azero tn e)]
	      end)
	    else
	      (begin
		let z = (Cexpr.constant_integer_value false sz) in
		(make_slot_filler_array e z (Num.num_of_int 0) [])
	      end)
	  end)
      | (Cexpr.STRUCTURE_TYPE _) ->
	  (begin
	    let body = (Cexpr.get_structure_body nqtn) in
	    let components = body.Cexpr.strc_components in
	    (List.flatten
		(List.map
		    (fun (s, tn1) ->
		      let sref = (Cexpr.SREF (tn1, false, e, s)) in
		      (make_slot_filler sref))
		    components))
	  end)
      | _ -> (assert false)
    end))

and make_slot_filler_array (e : Cexpr.expr_t) (ub : Num.num) (i : Num.num) (ee : Cexpr.expr_t list) : Cexpr.expr_t list = (
  assert (Num.le_num i ub) ;
  if (Num.eq_num i ub) then
    ee
  else
    let aref = (Cexpr.make_aref e (Cexpr.constant_integer i)) in
    let ff = (make_slot_filler aref) in
    (make_slot_filler_array e ub (Num.succ_num i) (ee @ ff)))

and filter_fills (inits : Cexpr.expr_t list) (fills : Cexpr.expr_t list) : Cexpr.expr_t list= (
  (List.filter
      (fun (f : Cexpr.expr_t) ->
	match f with
	  (Cexpr.SETF (tn, Cexpr.SET, f0, rhs0)) ->
	    (not (List.exists
		     (fun (i : Cexpr.expr_t) ->
		       match i with
			 (Cexpr.INIT_FILL _) -> (assert false)
		       | (Cexpr.SETF (tn, Cexpr.SET, i0, rhs1)) ->
			   (match_initializer_elements i0 f0)
		       | _ -> (assert false))
		     inits))
	| _ -> (assert false))
      fills))

(* Matches initializers to remove redundant assignments.  It returns
   true when {e0} covers {e1}, that is, for example, {e.a0} matches to
   {e.a0.b0}.  It assumes array initializers are of the form
   {*(x+o)=v} where the array part is the 1st operand and the offset
   part is the 2nd operand of the addition.  IT IS SLOW. *)

and match_initializer_elements (e0 : Cexpr.expr_t) (e1 : Cexpr.expr_t) : bool = (
  let c0 : int = (count_nests_in_initializer_elements e0) in
  let c1 : int = (count_nests_in_initializer_elements e1) in
  (match_initializer_elements_in_nest e0 e1 c0 c1))

and match_initializer_elements_in_nest (e0 : Cexpr.expr_t) (e1 : Cexpr.expr_t) (c0 : int) (c1 : int) : bool = (
  if (c0 > c1) then
    false
  else if (c0 == c1) then
    (match_initializer_elements_same_level e0 e1)
  else
    (begin
      let ex1 = (Cexpr.ignore_cast e1) in
      match ex1 with
	(Cexpr.SREF (tn, false, x, s)) ->
	  (match_initializer_elements_in_nest e0 x c0 (c1 - 1))
      | (Cexpr.PREF (tn, deref, x0, _)) ->
	  let (x, o) = (array_operands_of_pointer x0) in
	  (match_initializer_elements_in_nest e0 x c0 (c1 - 1))
      | (Cexpr.VREF _) -> (assert false)
      | _ -> (assert false)
    end))

and match_initializer_elements_same_level (e0 : Cexpr.expr_t) (e1 : Cexpr.expr_t) : bool = (
  let ex0 = (Cexpr.ignore_cast e0) in
  let ex1 = (Cexpr.ignore_cast e1) in
  match (ex0, ex1) with
    ((Cexpr.SREF (tn0, false, x0, s0)), (Cexpr.SREF (tn1, false, x1, s1))) ->
      ((match_initializer_elements_same_level x0 x1)
	&& (Misc.string_equal s0 s1))
  | ((Cexpr.PREF (tn0, deref0, xx0, _)), (Cexpr.PREF (tn1, deref1, xx1, _))) ->
      let (x0, o0) = (array_operands_of_pointer xx0) in
      let (x1, o1) = (array_operands_of_pointer xx1) in
      if ((Cexpr.constant_integer_p o0) && (Cexpr.constant_integer_p o1)) then
	(begin
	  let z0 = (Cexpr.constant_integer_value false o0) in
	  let z1 = (Cexpr.constant_integer_value false o1) in
	  ((match_initializer_elements_same_level x0 x1)
	    && (Num.eq_num z0 z1) && (deref0 == deref1))
	end)
      else
	(begin
	  (Cexpr.warn_in_parse ("(cannot determine array offset in"
				 ^ " initializer)")) ;
	  false
	end)
  | ((Cexpr.VREF (tn0, id0, level0, 0)), (Cexpr.VREF (tn1, id1, level1, 0))) ->
      assert (tn0 == tn1 && (Misc.string_equal id0 id1) && level0 == level1) ;
      true
  | _ -> (assert false))

and count_nests_in_initializer_elements (e : Cexpr.expr_t) : int = (
  let ex = (Cexpr.ignore_cast e) in
  match ex with
    (Cexpr.SREF (tn, false, x, s)) ->
      ((count_nests_in_initializer_elements x) + 1)
  | (Cexpr.PREF (tn, _, x0, _)) ->
      let (x, o) = (array_operands_of_pointer x0) in
      ((count_nests_in_initializer_elements x) + 1)
  | (Cexpr.VREF _) ->
      0
  | _ -> (assert false))

(* Returns the operands of the addition of the pointer, where the
   addition is on the array and the offset. *)

and array_operands_of_pointer (e : Cexpr.expr_t) : Cexpr.expr_t * Cexpr.expr_t = (
  match e with
    (Cexpr.OPER (tn, Cexpr.ADD, e0, e1, Cexpr.NIL)) ->
      assert (Cexpr.integral_type_p (Cexpr.type_of_expr e1)) ;
      let z1 = (Cexpr.fold_constant_integer e1) in
      (e0, z1)
  | _ -> (assert false))

and get_name_and_component (ty : Cexpr.typenode_t) = (
    let tn = (Cexpr.strip_off_pseudo_type ty true) in
    match tn with
    | Cexpr.STRUCTURE_TYPE s_t -> (
        let uniq_name = s_t.Cexpr.strc_unique_id in
        let name = s_t.Cexpr.strc_tag in
        let typebody = Hashtbl.find s_t.Cexpr.strc_typebody_table uniq_name in
        match typebody with
        | Cexpr.STRC_BODY s -> (name, s.Cexpr.strc_components)
        | Cexpr.ENUM_BODY _ -> assert false
    )
    | Cexpr.STRCDEF_TYPE (t, _) -> (get_name_and_component t)
    | Cexpr.POINTER_TO t -> (get_name_and_component t)
    | _ -> assert false
)

(*  *)
and init_fields_of_struct1 (ty : Cexpr.typenode_t)
                        (name : string)
                        (var : Cexpr.expr_t)
                        (components : (string * Cexpr.typenode_t)list)
                        (hash : (string, string list) Hashtbl.t)
                        (pri : int) (boo : bool)
                        : ((Cexpr.expr_t * int) list) = (
    List.flatten (
        List.map (
            fun com ->
                let field_name, field_type = com in
                (* find all fields of struct that are initiated *)
                let field_inits = (
                    try Hashtbl.find hash name
                    with
                        | Not_found -> []
                )in
                (* if a field is already initiated, don't anything
                   else initiate field:
                        if field is struct then init sub field of this struct
                        else if field is array then set field value equal to a memory
                        else set field value equal 0 *)
                if (List.mem field_name field_inits) then []
                else (begin
                    match field_type with
                    | Cexpr.STRUCTURE_TYPE _ -> (
                        let s_name, s_comp = get_name_and_component field_type
                        in
                        match ty with
                        | Cexpr.POINTER_TO (Cexpr.STRUCTURE_TYPE _) -> (
                            let expr = Cexpr.SREF (
                                           field_type, true, var, field_name) in
                            (init_fields_of_struct1 field_type s_name expr s_comp hash
                                                 (pri+1) boo)
                          )
                        | _ -> (
                            let expr = Cexpr.SREF (
                                           field_type, false, var, field_name) in
                            (init_fields_of_struct1 field_type s_name expr s_comp hash
                                                 (pri+1) boo)
                          )
                    )
                    | Cexpr.ARRAY_OF(ty1, _) -> (
                        match ty with
                        | Cexpr.POINTER_TO (Cexpr.STRUCTURE_TYPE _) -> (
                            let expr = Cexpr.SREF (
                                           field_type, true, var, field_name) in
                            let pseudo_memory = Cexpr.PSEUDO_MEMORY (
                                                    ty1, field_type) in
                            [(Cexpr.SETF (field_type, Cexpr.SET, expr,
                                              pseudo_memory), pri)]
                          )
                        | _ -> (
                            let expr = Cexpr.SREF (
                                           field_type, false, var, field_name) in
                            let pseudo_memory = Cexpr.PSEUDO_MEMORY (
                                                    ty1, field_type) in
                            [(Cexpr.SETF (field_type, Cexpr.SET, expr,
                                              pseudo_memory), pri)]
                          )
                    )
                    | _ -> (
                        match ty with
                        | Cexpr.POINTER_TO (Cexpr.STRUCTURE_TYPE _) -> (
                            let expr = Cexpr.SREF (
                                       field_type, true, var, field_name) in
                            let zero = Cexpr.CONST_I (
                                           Cexpr.INTEGER, Num.Int(0), "0") in
                             [(Cexpr.SETF (field_type, Cexpr.SET, expr, zero), pri)]
                          )
                        | _ -> (
                            let nondet = Cexpr.PSEUDO_NONDET_VALUE (ty, "__nondet_value") in
                            let expr = Cexpr.SREF (
                                       field_type, false, var, field_name) in
                            let zero = Cexpr.CONST_I (
                                           Cexpr.INTEGER, Num.Int(0), "0") in
                            [(Cexpr.SETF (field_type, Cexpr.SET, expr, nondet), pri)]
                          )
                    )
                end)
    ) components)
)

and init_fields_of_struct (decl : Cexpr.declaration_t)
: Cexpr.expr_t list = (
    (* function to check that whether a struct is set to another struct *)
    let is_init_struct_to_struct (var_name : string)
                                 (var_typ : Cexpr.typenode_t)
                                 (inits : Cexpr.expr_t list) : bool = (
        List.exists (fun init ->
            match init with
            | Cexpr.SETF (_, _, ex1, ex2) -> (
                (match ex2 with
                | Cexpr.OPER (_, _,Cexpr.CALL (_, _, _, cspec_op), _, _) -> (
                   let is_unique_in_spec =
                        is_expr_in_spec_op is_unique_expr cspec_op in
                     (not is_unique_in_spec)
                    )
                | Cexpr.OPER _ -> false
                | _ ->
                ( 
                (String.compare (get_name_of_expr ex1) var_name) == 0)
                & (Cexpr.compatible_type_p false (Cexpr.type_of_expr ex1) var_typ)
                & (Cexpr.compatible_type_p false (Cexpr.type_of_expr ex2) var_typ)
                )
              )
            | _ -> false
        ) inits
    )
    in
    let inits = decl.Cexpr.decl_inits in
    let (name, level) = decl.Cexpr.decl_id in
    let ty = Cexpr.strip_off_pseudo_type decl.Cexpr.decl_tn true in
    let scls = decl.Cexpr.decl_scls in
    (* if name = "" or struct is set to another struct, don't anything *)
    if ((String.compare name "") == 0 || (is_init_struct_to_struct name ty inits)) then
        []
    else (begin
        let boo = ((String.compare name "itc_fops") == 0) in
        (* collect all inits of struct *)
        let hash = fields_has_inits inits boo in
        (* get all fields of struct *)
        let _, s_comp = get_name_and_component ty in
        let var = Cexpr.VREF(ty, name, level, 0) in
        let init_exprs = init_fields_of_struct1 ty name var s_comp hash 0 boo in
        let len = List.length init_exprs in
        let (arr_exprs ) = Array.create len [Cexpr.NIL] in
        List.iter(
            fun int_expr ->
                let (expr, pri) = int_expr in
                let temp = (
                    try
                         let temp1 = Array.get arr_exprs pri in
                         if(List.mem expr temp1) then temp1
                         else (expr::temp1)
                   with
                       | _ -> []) in
                   try
                       (Array.set arr_exprs pri temp)
                   with
                       | _ -> (Array.set arr_exprs pri [expr])
        ) init_exprs;
        List.flatten(Array.to_list arr_exprs)
    end)
)

and init_fields_of_declaration (decl : Cexpr.declaration_t)
: Cexpr.expr_t list = (
    let ty = decl.Cexpr.decl_tn in
    let ty' = Cexpr.strip_off_pseudo_type ty true in
    match ty' with
    | Cexpr.STRUCTURE_TYPE _
    | Cexpr.POINTER_TO (Cexpr.STRUCTURE_TYPE _) -> (
        let new_inits = init_fields_of_struct decl in
        new_inits
      )
    | _ -> []
)

(* Check existence of construct '$result' in function contract of void function.
   Report error if this happens.
 *)
let check_result_in_specs_of_void_funcs
    (func_contracts : (string, Sexpr.lspec_t) Hashtbl.t) 
    (decls : Cexpr.stmt_t list) 
= (
    List.iter (
        fun decl ->
            let (func_name, func_type) = (
                match decl with
                | Cexpr.DEFUN(h, _)
                | Cexpr.DEFVAR(h) -> 
                  ((get_function_name h), 
                   (get_function_type h.Cexpr.decl_tn ))
                | _ -> None, None)
            in
            match func_name, func_type with 
            | Some name, Some type_ -> (
                match type_ with 
                | Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE, _) -> (
                  try
                      let spec_of_func = Hashtbl.find func_contracts name in
                      match spec_of_func with
                      | Sexpr.SPEC_FUN a_func_contract -> 
                        if (Sexpr.find_result_in_spec_func a_func_contract) then (
                            print_endline ("Found construct '$result' in the specs of function "
                                           ^ name);
                            print_endline ("Function contract: " 
                                           ^ (Sexpr.string_of_lspecs spec_of_func));
                            failwith ("Found construct '$result' in the specs of a void function")
                        );
                      | _ -> ()
                  with Not_found -> ()
                  )
                | _ -> ()
                )
            | _, _ -> ()
    )decls;
)

let check_using_var_declarationvar_following_var_spec
        (pre_stmts : Cexpr.stmt_t list)
        (decl_stmts : Cexpr.stmt_t list) : unit =
begin
    let last_op =
        try Some (ExtList.List.last pre_stmts)
        with _ -> None in
    match last_op with
    | Some Cexpr.SPEC Sexpr.SPEC_VAR var_spec -> (
        let len =
            List.fold_left
                (fun var_num a_stmt ->
                    if (is_stmt_var a_stmt) then
                        (var_num + 1)
                    else var_num
                ) 0 decl_stmts in
        if (len == 0) then (
            let pos_op =
                Sexpr.get_first_loc_var_spec (Sexpr.SPEC_VAR var_spec) in
            raise (Sexpr.Specs_error ("Variable spec without a variable",
                                      pos_op))
        )
        else if (len == 1) then ()
        else if (len > 1) then
            Cexpr.syntaxerror ("Following variable spec expects"
                              ^ " the declaration of only variable")
        else assert false
      )
    | anything_else -> ()
end

let get_position_of_id_in_ids id s_params = (
    let index = ref 0 in
    let res = ref 10000 in
    (List.iter(
        fun p ->
            if (Misc.string_equal p id) then
                res := !index
            else (incr index)
    )s_params);
    !res
)

let replace_id_by_id (id: string)
                       (s_params : string list)
                       (d_params : string list) : string = (
    match (get_position_of_id_in_ids id s_params) with
    | 10000 -> id
    | n -> (List.nth d_params n)
)

let get_position_of_expr_in_exprs (expr : Cexpr.expr_t)
                                  (s_params : Cexpr.expr_t list) : int = (
    let index = ref 0 in
    let res = ref 1000 in
    List.iter (
        fun p ->
            if (Cexpr.compare_expr p expr) then res := !index
            else (incr index)
    ) s_params;
    !res
)

let replace_ids_by_ids (ids: string list)
                         (s_params : string list)
                         (d_params : string list) : string list = (
    List.map(
       fun id ->
           replace_id_by_id id s_params d_params
    )ids
)

let rec replace_expr_by_expr (s_params : Cexpr.expr_t list) (d_params : Cexpr.expr_t list)
                         (expr : Cexpr.expr_t) : Cexpr.expr_t = (
    match (get_position_of_expr_in_exprs expr s_params) with
    | 1000 -> (
        match expr with
        | Cexpr.OPER (tn, op, expr1, expr2, expr3) ->
            Cexpr.OPER (tn, op,
                        replace_expr_by_expr s_params d_params expr1,
                        replace_expr_by_expr s_params d_params expr2,
                        replace_expr_by_expr s_params d_params expr3)
        | Cexpr.SETF (tn, op, expr1, expr2) ->
            Cexpr.SETF (tn, op,
                        replace_expr_by_expr s_params d_params expr1,
                        replace_expr_by_expr s_params d_params expr2)
        | Cexpr.CALL (tn, expr1, exprs1, spec_op) ->
            Cexpr.CALL (tn,
                        replace_expr_by_expr s_params d_params expr1,
                        List.map (replace_expr_by_expr s_params d_params) exprs1,
                        spec_op)
        | Cexpr.PREF (tn, b, expr1, expr2) ->
            Cexpr.PREF (tn, b, (replace_expr_by_expr s_params d_params expr1),
                               (replace_expr_by_expr s_params d_params expr2))
        | Cexpr.SREF (tn, b, expr1, id) ->
            Cexpr.SREF (tn, b,
                        replace_expr_by_expr s_params d_params expr1,
                        id)
        | Cexpr.DSGN_AREF (tn, expr1, expr2) ->
            Cexpr.DSGN_AREF (tn,
                             replace_expr_by_expr s_params d_params expr1,
                             replace_expr_by_expr s_params d_params expr2)
        | Cexpr.ADDROF (tn, expr1) ->
            Cexpr.ADDROF (tn,
                          replace_expr_by_expr s_params d_params expr1)
        | Cexpr.ASMSEQUENCE (exprs1) ->
            Cexpr.ASMSEQUENCE (List.map (replace_expr_by_expr s_params d_params) exprs1)
        | Cexpr.PHI (tn, id_exprs) ->
            let new_id_exprs =
                List.map (fun (id, expr1) ->
                    (id,
                     replace_expr_by_expr s_params d_params expr1)
                ) id_exprs in
            Cexpr.PHI (tn, new_id_exprs)
        | Cexpr.INIT_FILL (tn, expr1) ->
            Cexpr.INIT_FILL (tn,
                             replace_expr_by_expr s_params d_params expr1)
        | Cexpr.QUANTIFIER (id, exprs1, expr1) ->
            Cexpr.QUANTIFIER (id,
                              List.map (replace_expr_by_expr s_params d_params) exprs1,
                              replace_expr_by_expr s_params d_params expr1)
        | Cexpr.PSEUDO_ACSL_RESULT (tn, expr1) ->
            Cexpr.PSEUDO_ACSL_RESULT (tn,
                                      replace_expr_by_expr s_params d_params expr1)
        | Cexpr.PSEUDO_ACSL_UNIQUE (tn, expr) ->
            Cexpr.PSEUDO_ACSL_UNIQUE (tn,
                                      replace_expr_by_expr s_params d_params expr)
        | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn, expr1, expr2) ->
            Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn,
                                           replace_expr_by_expr s_params
                                                                d_params
                                                                expr1,
                                           replace_expr_by_expr s_params
                                                                d_params
                                                                expr2)
        | Cexpr.PSEUDO_ACSL_OLD (tn, expr1, expr2) ->
            Cexpr.PSEUDO_ACSL_OLD (tn,
                                   replace_expr_by_expr s_params d_params expr1,
                                   replace_expr_by_expr s_params d_params expr2)
        | Cexpr.PSEUDO_ACSL_VALID (tn, expr1) ->
            Cexpr.PSEUDO_ACSL_VALID (tn,
                                      replace_expr_by_expr s_params d_params expr1)
        | Cexpr.ALL_FIELD (tn, expr1) ->
            let new_expr = replace_expr_by_expr s_params d_params expr1 in
            Cexpr.ALL_FIELD (Cexpr.type_of_expr new_expr, new_expr)
        | Cexpr.NIL
        | Cexpr.PSEUDO_NONDET_VALUE _
        | Cexpr.PSEUDO_PROLOGUE
        | Cexpr.PSEUDO_EPILOGUE
        | Cexpr.PRAGMA _
        | Cexpr.CLITERAL _
        | Cexpr.VREF _
        | Cexpr.SIZEOF _
        | Cexpr.PROGN _
        | Cexpr.CONST_F _
        | Cexpr.STRING _
        | Cexpr.PSEUDO_MEMORY _
        | Cexpr.CONST_I _ -> expr
    )
    | n -> (
        List.nth d_params n)
)

let replace_full_expr_by_full_expr (full_expr : Cexpr.full_expr_t)
                                   (s_params : Cexpr.expr_t list)
                                   (d_params : Cexpr.expr_t list)
: Cexpr.full_expr_t = (
    {
        Cexpr.expr_node = replace_expr_by_expr s_params d_params
                                               full_expr.Cexpr.expr_node;
        Cexpr.expr_loc = full_expr.Cexpr.expr_loc
    }
)

let replace_exprs_in_conds (conds : 'pred Sexpr.func_condition_t list)
                           (s_params : Cexpr.expr_t list)
                           (d_params : Cexpr.expr_t list)
                           : 'pred Sexpr.func_condition_t list = (
    let replace_exprs_in_cond (cond : 'pred Sexpr.func_condition_t)
                              : 'pred Sexpr.func_condition_t =
      (begin
        let pos, expr = cond in
        let new_expr = replace_full_expr_by_full_expr expr s_params d_params in
        (pos, new_expr)
      end)
    in List.map replace_exprs_in_cond conds
)

let replace_exprs_in_assign (assign : Cexpr.full_expr_t Sexpr.assign_t)
                              (s_params : Cexpr.expr_t list)
                              (d_params : Cexpr.expr_t list)
: Cexpr.full_expr_t Sexpr.assign_t = (
    let zone, zones = assign in
    let replace_zone zone s_params d_params =
        (match zone with
        | Sexpr.Location expr ->
            Sexpr.Location (replace_full_expr_by_full_expr expr s_params d_params)
        | Sexpr.Nothing ->
            Sexpr.Nothing)
    in
    let replaced_zone = replace_zone zone s_params d_params in
    let replaced_zones = List.map (fun a_zone ->
                             replace_zone zone s_params d_params) zones in
    (replaced_zone, replaced_zones)
)
 
let replace_exprs_in_assigns (assigns : (Cexpr.full_expr_t Sexpr.assign_t) list)
                             (s_params : Cexpr.expr_t list)
                             (d_params : Cexpr.expr_t list)
: (Cexpr.full_expr_t Sexpr.assign_t) list = (
    List.map (fun assign ->
        replace_exprs_in_assign assign s_params d_params) assigns
)

let replace_exprs_in_behavior (behav : (Cexpr.full_expr_t,
                                        Cexpr.full_expr_t) Sexpr.behavior_t)
                               (s_params : Cexpr.expr_t list)
                               (d_params : Cexpr.expr_t list)
: (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.behavior_t = (
    let { Sexpr.b_name = id;
          Sexpr.b_assumes = exprs;
          Sexpr.b_conditions = conds;
          Sexpr.b_assigns = assigns; } = behav in
    let new_conds = replace_exprs_in_conds conds s_params d_params in
    {
      Sexpr.b_name = id;
      Sexpr.b_assumes = List.map (fun expr ->
                            replace_full_expr_by_full_expr expr s_params d_params
                        ) exprs;
      Sexpr.b_conditions = new_conds;
      Sexpr.b_assigns = replace_exprs_in_assigns assigns s_params d_params; }
)

let replace_exprs_in_behaviors (behavs : (Cexpr.full_expr_t,
                                          Cexpr.full_expr_t) Sexpr.behavior_t list)
                               (s_params : Cexpr.expr_t list)
                               (d_params : Cexpr.expr_t list)
: (Cexpr.full_expr_t, Cexpr.full_expr_t) Sexpr.behavior_t list = (
    List.map (fun behav ->
        replace_exprs_in_behavior behav s_params d_params
    ) behavs
)

let replace_exprs_in_spec
    (cspec : Cexpr.cspec_t)
    (s_params : Cexpr.expr_t list)
    (d_params : Cexpr.expr_t list)
: Cexpr.cspec_t = (
    match cspec with
    | Sexpr.SPEC_FUN
      {
        Sexpr.fspec_conditions = conds;
        Sexpr.fspec_behaviors = behavs;
        Sexpr.fspec_assigns = assigns;
        Sexpr.fspec_prohibits = prohibits;
        Sexpr.fspec_local_invs = invs;
        Sexpr.fspec_context = context;
        Sexpr.fspec_is_handler = handler
      } -> (
      Sexpr.SPEC_FUN
      {
        Sexpr.fspec_conditions = replace_exprs_in_conds conds s_params d_params;
        Sexpr.fspec_behaviors = replace_exprs_in_behaviors behavs s_params d_params;
        Sexpr.fspec_assigns = replace_exprs_in_assigns assigns s_params d_params;
        Sexpr.fspec_prohibits = prohibits;
        Sexpr.fspec_local_invs = List.map (fun (id, expr) ->
                               (id,
                                replace_full_expr_by_full_expr expr s_params d_params)
                           ) invs;
        Sexpr.fspec_context = context;
        Sexpr.fspec_is_handler = handler
      }
      )
    | _ -> assert false (* temporary have not processed yet *)
)

let compact_scope (scope : (string, Cexpr.binding_t) Hashtbl.t list)
: (string, Cexpr.binding_t) Hashtbl.t = (
    let copy_scope = List.map (Hashtbl.copy) scope in
    let compacted_scope : (string, Cexpr.binding_t) Hashtbl.t =
        Hashtbl.create 1 in
    List.iter (fun a_hash ->
        Hashtbl.iter (fun key value ->
            Hashtbl.add compacted_scope key value
        ) a_hash;
    ) copy_scope;
    compacted_scope
)     

let find_typedef_type (id : string) : Cexpr.typenode_t = (
    let bindings_of_id = List.flatten (List.map (fun a_hash ->
                                Hashtbl.find_all a_hash id
                           ) !Cexpr.name_scope_stack) in
    let bindings_of_typedef = List.filter (fun binding ->
                                  match binding.Cexpr.binding_scls with
                                  | Cexpr.TYPEDEF -> true
                                  | _ -> false
                              ) bindings_of_id in
    let res = try List.hd bindings_of_typedef
              with Failure _ -> assert false in
    res.Cexpr.binding_tn
)

let find_struct_type (id : string) : Cexpr.typenode_t = (
    let res = (
        try
            Hashtbl.find (compact_scope !Cexpr.tag_scope_stack) id
        with
            | Not_found -> assert false
    )
    in
    res.Cexpr.binding_tn
)

let rec replace_old_term_in_cexpr (expr : Cexpr.expr_t)
: Cexpr.expr_t = (
    match expr with
    | Cexpr.NIL
    | Cexpr.PSEUDO_NONDET_VALUE _
    | Cexpr.VREF _
    | Cexpr.SIZEOF _
    | Cexpr.PROGN _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.CLITERAL _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.PRAGMA _
    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE -> expr
    | Cexpr.PSEUDO_ACSL_OLD (tn, expr1, expr2) -> expr2
    | Cexpr.PREF (tn, b, expr1, expr2) ->
        Cexpr.PREF (tn, b, (replace_old_term_in_cexpr expr1),
                           (replace_old_term_in_cexpr expr2))
    | Cexpr.SREF (tn, b, expr1, str) ->
        Cexpr.SREF (tn, b, replace_old_term_in_cexpr expr1, str)
    | Cexpr.ADDROF (tn, expr1) ->
        Cexpr.ADDROF (tn, replace_old_term_in_cexpr expr1)
    | Cexpr.INIT_FILL (tn, expr1) ->
        Cexpr.INIT_FILL (tn, replace_old_term_in_cexpr expr1)
    | Cexpr.PSEUDO_ACSL_UNIQUE (tn, expr1) ->
        Cexpr.PSEUDO_ACSL_UNIQUE (tn, replace_old_term_in_cexpr expr1)
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn, expr1, expr2) ->
        Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (tn,
                                       replace_old_term_in_cexpr expr1,
                                       replace_old_term_in_cexpr expr2)
    | Cexpr.PSEUDO_ACSL_VALID (tn, expr1) ->
        Cexpr.PSEUDO_ACSL_VALID (tn, replace_old_term_in_cexpr expr1)
    | Cexpr.ALL_FIELD (tn, expr1) ->
        Cexpr.ALL_FIELD (tn, replace_old_term_in_cexpr expr1)
    | Cexpr.PSEUDO_ACSL_RESULT (tn, expr1) ->
        Cexpr.PSEUDO_ACSL_RESULT (tn, replace_old_term_in_cexpr expr1)
    | Cexpr.SETF (tn, op, expr1, expr2) ->
        Cexpr.SETF (tn,
                    op,
                    replace_old_term_in_cexpr expr1,
                    replace_old_term_in_cexpr expr2)
    | Cexpr.DSGN_AREF (tn, expr1, expr2) ->
        Cexpr.DSGN_AREF (tn,
                         replace_old_term_in_cexpr expr1,
                         replace_old_term_in_cexpr expr2)
    | Cexpr.OPER (tn, op, expr1, expr2, expr3) ->
        Cexpr.OPER (tn,
                    op,
                    replace_old_term_in_cexpr expr1,
                    replace_old_term_in_cexpr expr2,
                    replace_old_term_in_cexpr expr3)
    | Cexpr.ASMSEQUENCE exprs ->
        Cexpr.ASMSEQUENCE (
            List.map replace_old_term_in_cexpr exprs)
    | Cexpr.PHI (tn, str_exprs) ->
        Cexpr.PHI (tn,
                   List.map (fun (str, a_expr) ->
                       (str, replace_old_term_in_cexpr a_expr)
                   ) str_exprs)
    | Cexpr.QUANTIFIER (tn, exprs, expr1) ->
        Cexpr.QUANTIFIER (tn,
                          List.map replace_old_term_in_cexpr exprs,
                          replace_old_term_in_cexpr expr1)
    | Cexpr.CALL (tn, expr1, exprs, spec) ->
        Cexpr.CALL (tn,
                    replace_old_term_in_cexpr expr1,
                    List.map replace_old_term_in_cexpr exprs,
                    spec)
)

let replace_old_term_in_full_expr (full_expr : Cexpr.full_expr_t)
: Cexpr.full_expr_t = (
    {full_expr with Cexpr.expr_node = replace_old_term_in_cexpr
                                          full_expr.Cexpr.expr_node}
)

let replace_old_term_in_cond (cond : 'pred Sexpr.func_condition_t)
: 'pred Sexpr.func_condition_t = (
    let pos, pred = cond in
    (pos, replace_old_term_in_full_expr pred)
)

let replace_old_term_in_conds (conds : 'pred Sexpr.func_condition_t list)
: 'pred Sexpr.func_condition_t list = (
    List.map replace_old_term_in_cond conds
)

let replace_old_term_in_function_contract 
            (func_contract: ('pred, 'locs) Sexpr.function_contract_t)
: ('pred, 'locs) Sexpr.function_contract_t = (
    let { Sexpr.fspec_conditions = conds;
          Sexpr.fspec_behaviors = behavs;
          Sexpr.fspec_assigns = _;
          Sexpr.fspec_prohibits = _;
          Sexpr.fspec_local_invs = _;
          Sexpr.fspec_context = _;
          Sexpr.fspec_is_handler = _;
        } = func_contract in
    let behavs_replaced_old_term = (
        List.map (fun behav ->
            let { Sexpr.b_name = _;
                  Sexpr.b_assumes = _;
                  Sexpr.b_conditions = b_conds;
                  Sexpr.b_assigns = _
                } = behav in
            {behav with Sexpr.b_conditions = replace_old_term_in_conds b_conds})
        behavs)
    in
    {func_contract with Sexpr.fspec_conditions = replace_old_term_in_conds conds;
                        Sexpr.fspec_behaviors= behavs_replaced_old_term}
)

let rec get_old_term_from_cexpr (expr : Cexpr.expr_t)
: (Cexpr.typenode_t * Cexpr.expr_t * Cexpr.expr_t) list = (
    match expr with
    | Cexpr.NIL
    | Cexpr.PSEUDO_NONDET_VALUE _
    | Cexpr.VREF _
    | Cexpr.SIZEOF _
    | Cexpr.PROGN _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.CLITERAL _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.PRAGMA _
    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE -> []
    | Cexpr.PSEUDO_ACSL_OLD (tn, expr1, expr2) -> [(tn, expr1, expr2)]
    | Cexpr.PREF (_, _, expr1, _)
    | Cexpr.SREF (_, _, expr1, _)
    | Cexpr.ADDROF (_, expr1)
    | Cexpr.INIT_FILL (_, expr1)
    | Cexpr.PSEUDO_ACSL_UNIQUE (_, expr1)
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (_, expr1, _)
    | Cexpr.PSEUDO_ACSL_VALID (_, expr1)
    | Cexpr.ALL_FIELD (_, expr1)
    | Cexpr.PSEUDO_ACSL_RESULT (_, expr1) -> get_old_term_from_cexpr expr1
    | Cexpr.SETF (_, _, expr1, expr2)
    | Cexpr.DSGN_AREF (_, expr1, expr2) ->
        (get_old_term_from_cexpr expr1)
        @ (get_old_term_from_cexpr expr2)
    | Cexpr.OPER (_, _, expr1, expr2, expr3) ->
        (get_old_term_from_cexpr expr1)
        @ (get_old_term_from_cexpr expr2)
        @ (get_old_term_from_cexpr expr3)
    | Cexpr.ASMSEQUENCE exprs ->
        List.fold_left (fun old_terms a_expr ->
            old_terms @ (get_old_term_from_cexpr a_expr)
        ) [] exprs
    | Cexpr.PHI (_, str_exprs) ->
        List.fold_left (fun old_terms (_, a_expr) ->
            old_terms @ (get_old_term_from_cexpr a_expr)
        ) [] str_exprs
    | Cexpr.QUANTIFIER (_, exprs, expr1)
    | Cexpr.CALL (_, expr1, exprs, _) ->
        List.fold_left (fun old_terms a_expr ->
            old_terms @ (get_old_term_from_cexpr a_expr)
        ) [] (exprs @ [expr1])
)

let get_old_term_from_full_expr (full_expr : Cexpr.full_expr_t)
: (Cexpr.typenode_t * Cexpr.expr_t * Cexpr.expr_t) list = (
    get_old_term_from_cexpr full_expr.Cexpr.expr_node
)

let check_redundant_list (l : 'a list) : 'a list = (
    List.fold_left (fun remain_list a_mem ->
        if List.mem a_mem remain_list then remain_list
        else remain_list @ [a_mem]
    ) [] l;
)

let get_old_term_from_function_contract 
            (func_contract: ('pred, 'locs) Sexpr.function_contract_t)
: (Cexpr.typenode_t * Cexpr.expr_t * Cexpr.expr_t) list = (
    let { Sexpr.fspec_conditions = conds;
          Sexpr.fspec_behaviors = behavs;
          Sexpr.fspec_assigns = _;
          Sexpr.fspec_prohibits = _;
          Sexpr.fspec_local_invs = _;
          Sexpr.fspec_context = _;
          Sexpr.fspec_is_handler = _;
        } = func_contract in
    let b_func_conds = (
        List.flatten(
            List.map(
                function
                  {
                    Sexpr.b_name = _;
                    Sexpr.b_assumes = _;
                    Sexpr.b_conditions = b_conds;
                    Sexpr.b_assigns = _
                  } -> b_conds )
            behavs)) 
    in 
    let total_func_conds = (conds @ b_func_conds) in
    let full_exprs = List.map snd  total_func_conds in
    let res = 
        List.flatten (List.map get_old_term_from_full_expr full_exprs) in
    (check_redundant_list res)
)

let func_scopes : ((string, Cexpr.binding_t) Hashtbl.t *
                              (* name scope stack *)
                              (string, Cexpr.binding_t) Hashtbl.t
                              (* hash scope stack *)) Queue.t = Queue.create ()
let loop_scopes : ((string, Cexpr.binding_t) Hashtbl.t *
                              (* name scope stack *)
                              (string, Cexpr.binding_t) Hashtbl.t
                              (* hash scope stack *)) Queue.t = Queue.create ()
let global_spec_scopes : ((string, Cexpr.binding_t) Hashtbl.t *
                              (* name scope stack *)
                              (string, Cexpr.binding_t) Hashtbl.t
                              (* hash scope stack *)) Queue.t = Queue.create ()
let stmt_spec_scopes : (((string, Cexpr.binding_t) Hashtbl.t *
                              (* name scope stack *)
                              (string, Cexpr.binding_t) Hashtbl.t
                              (* hash scope stack *)) Queue.t) Stack.t = Stack.create ()

let push_func_scope () : unit = (
    Queue.push !current_func_scope func_scopes
)
let push_loop_scope () : unit = (
    Queue.push !current_loop_scope loop_scopes)

let push_global_spec_scopes () : unit = (
    Queue.push (compact_scope !Cexpr.name_scope_stack,
                compact_scope !Cexpr.tag_scope_stack) global_spec_scopes)
let push_stmt_spec_scopes () : unit = (
    Queue.push (compact_scope !Cexpr.name_scope_stack,
                compact_scope !Cexpr.tag_scope_stack) (Stack.top stmt_spec_scopes))
let pop_func_scope () : ((string, Cexpr.binding_t) Hashtbl.t *
                      (string, Cexpr.binding_t) Hashtbl.t) = (
    try Queue.pop func_scopes
    with Queue.Empty -> assert false
)
let pop_loop_scope () : ((string, Cexpr.binding_t) Hashtbl.t *
                      (string, Cexpr.binding_t) Hashtbl.t) = (
    try Queue.pop loop_scopes
    with Queue.Empty -> assert false
)
let pop_global_spec_scope () : ((string, Cexpr.binding_t) Hashtbl.t *
                      (string, Cexpr.binding_t) Hashtbl.t) = (
    try Queue.pop global_spec_scopes
    with Queue.Empty -> assert false
)
let pop_stmt_spec_scope () : ((string, Cexpr.binding_t) Hashtbl.t *
                      (string, Cexpr.binding_t) Hashtbl.t) = (
    try Queue.pop (Stack.top stmt_spec_scopes)
    with Queue.Empty -> assert false
)

let get_typenames_from_namescope () : unit = (
    let visited_names = Hashtbl.create 107 in
    List.iter
        (fun name_scope ->
             Hashtbl.iter
                 (fun name binding ->
                      let _ =
                          match binding.Cexpr.binding_scls with
                          | Cexpr.TYPEDEF ->
                              if not (Hashtbl.mem visited_names name) then
                                  if not (Scommon.typename_status name) then
                                      Scommon.add_typename name
                          | _ -> () in
                      Hashtbl.add visited_names name ())
                 name_scope)
        !Cexpr.name_scope_stack)

let add_location_to_decl (location : Sexpr.global_pos_t)
                         (decl : Cexpr.declaration_t)
: Cexpr.declaration_t = (
    { decl with Cexpr.decl_location = Some location }
)

let add_location_to_stmt (location : Sexpr.global_pos_t)
                         (stmt : Cexpr.stmt_t)
: Cexpr.stmt_t = (
    match stmt with
    | Cexpr.DEFUN (header, body) ->
        Cexpr.DEFUN (add_location_to_decl location header, body)
    | Cexpr.DEFVAR header ->
        Cexpr.DEFVAR (add_location_to_decl location header)
    | _ -> stmt
)

let find_called_stmts_in_ofile (ofile : Cexpr.ofile_t) (func_name : string)
: Cexpr.stmt_t list=
begin
    let def_lst = 
        List.filter
            (fun a_def -> 
                let (name_def, def) = a_def in
                ((String.compare name_def func_name) == 0)
            )
            !(ofile.Cexpr.ofile_defs)
    in
    (snd (List.split def_lst))
end

let find_called_stmts_in_ofiles1 (ofiles : Cexpr.ofile_t list)
                                 (func_name : string)
: Cexpr.stmt_t list =
begin
    List.flatten
        (List.map
            (fun a_ofile ->
                let raw_called_stmts =
                    find_called_stmts_in_ofile a_ofile func_name in
                List.fold_left
                    (fun tmp_stmts a_called_st ->
                        let is_local = Cexpr.is_local [a_called_st] in
                        if is_local then tmp_stmts
                        else (tmp_stmts @ [a_called_st])
                    ) [] raw_called_stmts
            )
            ofiles)
end

(** Find called stmt have name to be "func_name"
*)
let find_called_stmts_in_ofiles (ofiles : Cexpr.ofile_t list)
                                (current_ofile : Cexpr.ofile_t)
                                (func_name : string)
: Cexpr.stmt_t list =
begin
    let local_stmts = find_called_stmts_in_ofile current_ofile func_name in
    let is_local = Cexpr.is_local local_stmts in
    if is_local then
        (* the function is declarated as STATIC *)
        local_stmts
    else
        (* the function is declarated as NON-STATIC or isn't declarated
           in this file *)
        find_called_stmts_in_ofiles1 ofiles func_name
end

let check_standard_quantity_func_cont_of_a_func
    (ofiles : Cexpr.ofile_t list) : unit =
begin
    List.iter
        (fun a_ofile ->
            List.iter
                (fun a_def ->
                    let (name_def, def) = a_def in
                    let decl =
                        match def with
                        | Cexpr.DEFUN (de, st) -> de
                        | Cexpr.DEFVAR (de) -> de
                        | _ -> assert false
                    in
                    if (is_function_type decl.Cexpr.decl_tn) then (
                        let stmts =
                            find_called_stmts_in_ofiles ofiles a_ofile name_def
                        in
                        let cspec_lst =
                            List.fold_left
                                (fun tmp_spec_lst a_stmt ->
                                    let header =
                                    match a_stmt with
                                    | Cexpr.DEFUN (decl, st) -> decl
                                    | Cexpr.DEFVAR (decl) -> decl
                                    | _ -> assert false
                                    in
                                    if (is_function_type header.Cexpr.decl_tn) then (
                                        match header.Cexpr.decl_specs with
                                        | None -> tmp_spec_lst
                                        | Some spec -> tmp_spec_lst @ [spec]
                                        )
                                    else tmp_spec_lst
                                ) [] stmts in
                        if (List.length cspec_lst) > 1 then (
                            LogUtil.log LogUtil.PARSER LogUtil.ERROR
                                ("Function "
                                  ^ name_def
                                  ^ " has more than one function contract");
                            exit 100))
                    else () 
                ) !(a_ofile.Cexpr.ofile_defs)          
        ) ofiles
end

(** Find all fields of struct_var. 
    Replace struct_var.* in cspec by every field
 *)
let replace_fields_of_struct (struct_var : Cexpr.expr_t)
                             (cspec : Cexpr.cspec_t)
: Cexpr.cspec_t = (
    (* get fields of struct *)
    match Cexpr.type_of_expr struct_var with
    | Cexpr.POINTER_TO (Cexpr.STRUCTURE_TYPE struct_tag) -> (
        let struct_body =
            Cexpr.get_structure_body (Cexpr.STRUCTURE_TYPE struct_tag) in
        let fields = struct_body.Cexpr.strc_components in
        let replaced_specs =
            List.map (fun (field_name, field_type) ->
                let replaced_spec =
                    replace_exprs_in_spec
                        cspec
                        [Cexpr.ALL_FIELD ((Cexpr.type_of_expr struct_var),
                                           struct_var)]
                        [Cexpr.SREF (field_type, true, struct_var, field_name)]
                in
                replaced_spec
            ) fields in
         Sexpr.merge_fun_contracts replaced_specs
      )
    | _ -> assert false
)

let get_func_cond_in_decls (args : Cexpr.expr_t list)
                           (stmts : Cexpr.stmt_t list)
: Cexpr.cspec_t option =
begin
    let make_param_for_decl (decl : Cexpr.declaration_t)
    : Cexpr.expr_t list = (
        let param_names = decl.Cexpr.decl_params in
        let ty_op = get_function_type decl.Cexpr.decl_tn in
        let param_types =
            match (ty_op) with
            | Some Cexpr.FUNCTION_OF (func_ty, para_ty_op) -> (
                match para_ty_op with
                | None -> []
                | Some para_tys -> para_tys)
            | _ -> []
        in
        let params =
            try
                List.fold_left2
                    (fun curr_paras param_name param_type ->
                        match param_type with
                        | Cexpr.ELLIPSIS_TYPE ->
                            curr_paras
                        | _ -> 
                            (curr_paras @
                             [(Cexpr.VREF (param_type, param_name,
                                           Cexpr.parameter_scope_level, 0))])
                     ) [] param_names param_types
            with Invalid_argument _ -> assert false
        in params
    ) in

    let cspec_op =
        List.fold_left
            (fun tmp_spec_op a_stmt ->
                let header =
                    match a_stmt with
                    | Cexpr.DEFUN (decl, st) -> decl
                    | Cexpr.DEFVAR (decl) -> decl                        
                    | _ -> assert false
                in
                match header.Cexpr.decl_specs with
                | None -> tmp_spec_op
                | Some spec -> (
                    match tmp_spec_op with
                    | Some sp -> (
                        let (fun_name, _) = header.Cexpr.decl_id in
                        LogUtil.log LogUtil.PARSER LogUtil.ERROR
                            ("Function "
                             ^ fun_name
                             ^ " has more than one function contract");
                        exit 100);
                    | None -> (
                        (* replace all parameters in spec - function contract
                           by arguments of function caller and replace
                           struct_all_field by fields of struct correspondingly
                        *)
                        (* Make expr parameters of function declaration *)
                        let s_params =
                            make_param_for_decl header in
                        (* Get arguments correspondingly *)
                        let (d_params, _) =
                            ExtList.List.split_nth (List.length s_params) args
                        in
                        (* Replace parameters by arguments *)
                        let replaced_params_spec =
                            replace_exprs_in_spec spec
                                                  s_params
                                                  d_params in
                        let structs_get_field =
                            Sexpr.struct_get_field
                                Cexpr.struct_get_field_in_full_expr
                                replaced_params_spec in
                        let len = List.length structs_get_field in
                        (* Replace struct all field *)
                        let spec_replaced_fields = 
                            if len = 0 then
                                replaced_params_spec
                            else if len = 1 then (
                                let struct_var = List.hd structs_get_field in
                                replace_fields_of_struct struct_var
                                                         replaced_params_spec)
                            else assert false
                        in
                        (Some spec_replaced_fields))
                )
            ) None stmts
    in
    (cspec_op)
end

let find_contract_field (ofile : Cexpr.ofile_t)
                        (func_name : string)
                        (d_params : Cexpr.expr_t list)
: Cexpr.cspec_t option =
begin
    let lst = ExtString.String.nsplit func_name "." in
    match lst with
    | [] -> assert false
    | [x] -> None
    | _ -> (
        let str_var = List.hd lst in
        let str_field = List.nth lst ((List.length lst) - 1) in
        let spec_op =
            try
                let struct_name =
                    Hashtbl.find ofile.Cexpr.f_pointer_field_hashtbl
                                 (str_var, str_field) in
                let (func_cont, s_params) =
                    Hashtbl.find ofile.Cexpr.contract_of_field_hashtbl
                                 (struct_name, str_field) in
                let spec = Sexpr.SPEC_FUN func_cont in
                assert (List.length s_params = List.length d_params);
                let replaced_params_spec =
                replace_exprs_in_spec spec
                                      s_params
                                      d_params in
                let structs_get_field =
                    Sexpr.struct_get_field Cexpr.struct_get_field_in_full_expr
                                           replaced_params_spec in
                let len = List.length structs_get_field in
                let spec_replaced_fields = 
                    if len = 0 then
                        replaced_params_spec
                    else if len = 1 then (
                        let struct_var = List.hd structs_get_field in
                        replace_fields_of_struct struct_var
                                                 replaced_params_spec)
                    else assert false
                in (Some spec_replaced_fields)
            with Not_found -> None
        in spec_op
    )
end

(** Insert function contract into function caller (Cexpr.CALL).
    Function contracts is saved in 4th part of Cexpr.CALL. This function finds
    function contracts of function to which Cexpr.CALL actualy calls 
        @param current_expr : expression is being processed
        @param current_ofile : stmt ofile holds this caller
        @param ofiles : all of ofiles
*)
let rec insert_conds_expr (current_expr : Cexpr.expr_t)
                          (current_ofile : Cexpr.ofile_t)
                          (ofiles : Cexpr.ofile_t list)
: Cexpr.expr_t =
begin
    match current_expr with
    | Cexpr.CALL (ty, func_var, args, func_cond_op) -> (
        let new_args = insert_conds_exprs args current_ofile ofiles in
        match func_cond_op with
        | None -> (
            let func_name, _ =
                try ExtString.String.split (get_name_of_expr func_var) "%"
                with _ -> (get_name_of_expr func_var), ""
            in
            (* Finds function stmt which caller calls to *)
            let called_stmts =
                find_called_stmts_in_ofiles ofiles
                                            current_ofile
                                            func_name
            in
            let len = List.length called_stmts in
            let f_cond_op =
                if (len = 0) then (
                    (* If have no function stmt then find in function contract
                       of field *)
                    find_contract_field current_ofile
                                        func_name
                                        args
                )
                else
                    (* Else get function contract in function stmt *)
                    get_func_cond_in_decls args called_stmts in
            Cexpr.CALL (ty, func_var, new_args, f_cond_op)
        )
        | Some func_cond -> Cexpr.CALL (ty, func_var, new_args, func_cond_op)
    )
    | Cexpr.NIL
    | Cexpr.VREF _
    | Cexpr.SIZEOF _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.PRAGMA _
    | Cexpr.PSEUDO_PROLOGUE 
    | Cexpr.PSEUDO_EPILOGUE
    | Cexpr.PSEUDO_NONDET_VALUE _ -> current_expr
    | Cexpr.OPER (ty, op, expr1, expr2, expr3) -> (
        let new_expr1 = insert_conds_expr expr1 current_ofile ofiles in
        let new_expr2 = insert_conds_expr expr2 current_ofile ofiles in
        let new_expr3 = insert_conds_expr expr3 current_ofile ofiles in
        Cexpr.OPER (ty, op, new_expr1, new_expr2, new_expr3)
    )
    | Cexpr.SETF (ty, op, expr1, expr2) -> (
        let new_expr1 = insert_conds_expr expr1 current_ofile ofiles in
        let new_expr2 = insert_conds_expr expr2 current_ofile ofiles in
        Cexpr.SETF (ty, op, new_expr1, new_expr2)
    )
    | Cexpr.PREF (ty, bo, expr1, expr2) ->
        Cexpr.PREF (ty, bo, (insert_conds_expr expr1 current_ofile ofiles),
                            (insert_conds_expr expr2 current_ofile ofiles))
    | Cexpr.SREF (ty, bo, expr, str) ->
        Cexpr.SREF (ty, bo, (insert_conds_expr expr current_ofile ofiles), str)
    | Cexpr.DSGN_AREF (ty, expr1, expr2) -> (
        let new_expr1 = insert_conds_expr expr1 current_ofile ofiles in
        let new_expr2 = insert_conds_expr expr2 current_ofile ofiles in
        Cexpr.DSGN_AREF (ty, new_expr1, new_expr2)
    )
    | Cexpr.ADDROF (ty, expr) ->
        Cexpr.ADDROF (ty, (insert_conds_expr expr current_ofile ofiles))
    | Cexpr.PROGN (ty, sts) -> 
        Cexpr.PROGN (ty, insert_conds_stmts sts current_ofile ofiles)
    | Cexpr.CLITERAL (ty, init) ->
        Cexpr.CLITERAL (ty, (insert_conds_init init current_ofile ofiles))
    | Cexpr.ASMSEQUENCE exprs ->
        Cexpr.ASMSEQUENCE (insert_conds_exprs exprs current_ofile ofiles)
    | Cexpr.PHI (ty, phi_lst) -> (
        let new_phi_lst =
            List.map
                (fun a_phi ->
                    let (str, expr) = a_phi in
                    (str, (insert_conds_expr expr current_ofile ofiles)) 
                )
                phi_lst
        in Cexpr.PHI (ty, new_phi_lst)
    )    
    | Cexpr.INIT_FILL (ty, expr) ->
        Cexpr.INIT_FILL (ty, (insert_conds_expr expr current_ofile ofiles))
    | Cexpr.QUANTIFIER (str, exprs, expr) -> (
        let new_exprs = insert_conds_exprs exprs current_ofile ofiles in
        let new_expr = insert_conds_expr expr current_ofile ofiles in
        Cexpr.QUANTIFIER (str, new_exprs, new_expr)
    )
    | Cexpr.PSEUDO_ACSL_RESULT (ty, expr) ->
        Cexpr.PSEUDO_ACSL_RESULT (ty, (insert_conds_expr expr current_ofile ofiles))
    | Cexpr.PSEUDO_ACSL_OLD (ty, expr1, expr2) -> (
        let new_expr1 = insert_conds_expr expr1 current_ofile ofiles in
        let new_expr2 = insert_conds_expr expr2 current_ofile ofiles in
        Cexpr.PSEUDO_ACSL_OLD (ty, new_expr1, new_expr2)
    )
    | Cexpr.PSEUDO_ACSL_UNIQUE (ty, expr) ->
        Cexpr.PSEUDO_ACSL_UNIQUE (ty, (insert_conds_expr expr current_ofile ofiles))
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (ty, expr1, expr2) -> (
        let new_expr1 = insert_conds_expr expr1 current_ofile ofiles in
        let new_expr2 = insert_conds_expr expr2 current_ofile ofiles in
        Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (ty, new_expr1, new_expr2)
    )
    | Cexpr.PSEUDO_ACSL_VALID (ty, expr) ->
        Cexpr.PSEUDO_ACSL_VALID (ty, (insert_conds_expr expr current_ofile ofiles))
    | Cexpr.ALL_FIELD (ty, expr) ->
        Cexpr.ALL_FIELD (ty, (insert_conds_expr expr current_ofile ofiles))
end
 
and insert_conds_init (init : Cexpr.initializers_t)
                      (current_ofile : Cexpr.ofile_t)
                      (ofiles : Cexpr.ofile_t list)
: Cexpr.initializers_t =
begin
    match init with
    | Cexpr.INIT_EXPR expr -> Cexpr.INIT_EXPR (insert_conds_expr expr current_ofile ofiles)
    | Cexpr.INIT_LIST inits -> (
        let new_inits =
            List.map
                (fun a_init ->
                    insert_conds_init a_init current_ofile ofiles
                )
                inits
        in (Cexpr.INIT_LIST new_inits)
    )
    | Cexpr.INIT_DSGN (exprs, init) -> (
        let new_exprs = insert_conds_exprs exprs current_ofile ofiles in
        let new_init = insert_conds_init init current_ofile ofiles in
        Cexpr.INIT_DSGN (new_exprs, new_init)
    )
end

and insert_conds_stmts (sts : Cexpr.stmt_t list) 
                       (current_ofile : Cexpr.ofile_t) 
                       (ofiles : Cexpr.ofile_t list)
: Cexpr.stmt_t list =
begin
    List.map
        (fun a_st ->
            insert_conds_stmt a_st current_ofile ofiles   
        ) sts
end

and insert_conds_exprs (exs : Cexpr.expr_t list) 
                       (current_ofile : Cexpr.ofile_t)
                       (ofiles : Cexpr.ofile_t list)
: Cexpr.expr_t list =
begin
    List.map
        (fun a_ex ->
            insert_conds_expr a_ex current_ofile ofiles 
        ) exs
end
 
and insert_conds_stmt (current_stmt : Cexpr.stmt_t) 
                      (current_ofile : Cexpr.ofile_t)
                      (ofiles : Cexpr.ofile_t list) : Cexpr.stmt_t =
begin
    match current_stmt with
    | Cexpr.DEFUN (decl, st) ->
        Cexpr.DEFUN (decl, (insert_conds_stmt st current_ofile ofiles))
    | Cexpr.BREAK
    | Cexpr.GOTO _
    | Cexpr.NOTE _
    | Cexpr.CONTINUE
    | Cexpr.SPEC _ ->
         current_stmt
    | Cexpr.DEFVAR decl ->
         let new_inits =
             List.map (fun init ->
                 insert_conds_expr init current_ofile ofiles
             ) decl.Cexpr.decl_inits in
         Cexpr.DEFVAR { decl with Cexpr.decl_inits = new_inits }
    | Cexpr.STMT ex ->
        Cexpr.STMT (insert_conds_expr ex current_ofile ofiles)
    | Cexpr.ASGN (ex1, ex2) ->
        Cexpr.ASGN (insert_conds_expr ex1 current_ofile ofiles, insert_conds_expr ex1 current_ofile ofiles)
    | Cexpr.CALLASGN (ex, sts, exs) -> (
        let new_ex = insert_conds_expr ex current_ofile ofiles in
        let new_sts = insert_conds_stmts sts current_ofile ofiles in
        let new_exs = insert_conds_exprs exs current_ofile ofiles in
        Cexpr.CALLASGN (new_ex, new_sts, new_exs)
    )
    | Cexpr.BEGIN sts -> 
        Cexpr.BEGIN (insert_conds_stmts sts current_ofile ofiles)
    | Cexpr.IF (ex, sts_1, sts_2) -> (
        let new_ex = insert_conds_expr ex current_ofile ofiles in
        let new_sts_1 = insert_conds_stmts sts_1 current_ofile ofiles in
        let new_sts_2 = insert_conds_stmts sts_2 current_ofile ofiles in
        Cexpr.IF (new_ex, new_sts_1, new_sts_2)
    )
    | Cexpr.SWITCH (ex, sts) -> (
        let new_ex = insert_conds_expr ex current_ofile ofiles in
        let new_sts = insert_conds_stmts sts current_ofile ofiles in
        Cexpr.SWITCH (new_ex, new_sts)
    )
    | Cexpr.DOWHILE (bo, ex, sts) -> (
        let new_ex = insert_conds_expr ex current_ofile ofiles in
        let new_sts = insert_conds_stmts sts current_ofile ofiles in
        Cexpr.DOWHILE (bo, new_ex, new_sts)
    )
    | Cexpr.FOR (sts_1, ex1, ex2, sts_2) -> (
        let new_sts_1 = insert_conds_stmts sts_1 current_ofile ofiles in
        let new_ex1 = insert_conds_expr ex1 current_ofile ofiles in
        let new_ex2 = insert_conds_expr ex2 current_ofile ofiles in
        let new_sts_2 = insert_conds_stmts sts_2 current_ofile ofiles in
        Cexpr.FOR (new_sts_1, new_ex1, new_ex2, new_sts_2)
    )
    | Cexpr.RETURN ex_op -> (
        match ex_op with
        | None -> current_stmt
        | Some ex ->
            Cexpr.RETURN ( Some (insert_conds_expr ex current_ofile ofiles))
    )
    | Cexpr.ANNOTATION (str, ex) ->
        Cexpr.ANNOTATION (str, (insert_conds_expr ex current_ofile ofiles))
    | Cexpr.ASM ex ->
        Cexpr.ASM (insert_conds_expr ex current_ofile ofiles)
    | Cexpr.ASSUME ex ->
        Cexpr.ASSUME (insert_conds_expr ex current_ofile ofiles)
    | Cexpr.RESET exs ->
        Cexpr.RESET (insert_conds_exprs exs current_ofile ofiles)    
    | Cexpr.LABEL label -> (
        match label with
        | Cexpr.NAME_LABEL _
        | Cexpr.DFLT_LABEL -> current_stmt
        | Cexpr.CASE_LABEL (ex1, ex2) -> (
            let new_ex1 = insert_conds_expr ex1 current_ofile ofiles in
            let new_ex2 = insert_conds_expr ex2 current_ofile ofiles in
            Cexpr.LABEL (Cexpr.CASE_LABEL (new_ex1, new_ex2))
        )
    )
end

(* Remove all field spec in declaration *)
and remove_all_field_spec_decl (current_stmt : Cexpr.stmt_t)
: Cexpr.stmt_t =
begin
    let length_all_field_spec (spec_op : Cexpr.cspec_t option) : int = (
        match spec_op with
        | None -> 0
        | Some spec -> (
            let all_fields =
                Sexpr.struct_get_field Cexpr.struct_get_field_in_full_expr
                                       spec in
            List.length all_fields
        )
    ) in
    match current_stmt with
    | Cexpr.DEFUN (decl, st) -> (
        let len = length_all_field_spec decl.Cexpr.decl_specs in 
        if len = 0 then
            current_stmt
        else Cexpr.DEFUN ({decl with Cexpr.decl_specs = None}, st)
    )
    | Cexpr.DEFVAR decl -> (
        let len = length_all_field_spec decl.Cexpr.decl_specs in 
        if len = 0 then
            current_stmt
        else Cexpr.DEFVAR ({decl with Cexpr.decl_specs = None})
    )
end

(** Insert function contract into each function caller
*)
and insert_conds_when_calling_func (ofiles : Cexpr.ofile_t list) : Cexpr.ofile_t list =
begin
    let inserted_ofiles =
        List.map 
            (fun a_ofile ->
                let new_ofile_defs =
                    List.map
                        (fun a_def ->
                            let (name_def, def) = a_def in
                            (name_def, insert_conds_stmt def a_ofile ofiles) 
                        )
                        !(a_ofile.Cexpr.ofile_defs)
                in {a_ofile with Cexpr.ofile_defs = ref new_ofile_defs}                
            )
            ofiles in
    (* Remove all field spec in declaration *)
    List.map 
        (fun a_ofile ->
            let new_ofile_defs =
                List.map
                    (fun a_def ->
                        let (name_def, def) = a_def in
                        (name_def, remove_all_field_spec_decl def) 
                    )
                    !(a_ofile.Cexpr.ofile_defs)
            in {a_ofile with Cexpr.ofile_defs = ref new_ofile_defs}                
        )
        inserted_ofiles
    
end

(** initialize old var (saved in Cexpr.PSEUDO_ACSL_OLD) in function condition list
*)
let rec make_stmt_old_expr_from_funct_cont
            (func_cont : ('pred, 'locs) Sexpr.function_contract_t)
: Cexpr.stmt_t list =
begin
    let old_infos = get_old_term_from_function_contract func_cont in
    make_stmt_old_exprs old_infos
end

(** initialize old var (saved in Cexpr.PSEUDO_ACSL_OLD) in expression list
*)
and make_stmt_old_exprs
            (old_infos : (Cexpr.typenode_t * Cexpr.expr_t * Cexpr.expr_t) list)
: Cexpr.stmt_t list = 
begin
    List.flatten(
        List.map(
            fun a_old_info ->
                make_init_for_old_expr a_old_info) 
            old_infos)
end

(** initialize old var (saved in Cexpr.PSEUDO_ACSL_OLD) in an expression
*)
and make_init_for_old_expr
            (old_info : Cexpr.typenode_t * Cexpr.expr_t * Cexpr.expr_t)
: Cexpr.stmt_t list =
begin
    let (ty, inside_expr, actual_expr) = old_info in
    let inits = [Cexpr.SETF (ty, Cexpr.SET, actual_expr, inside_expr)] in
    let name_var = get_name_of_expr actual_expr in
    let header =
      {
        Cexpr.decl_id = (name_var, 3);
        Cexpr.decl_tn = ty;
        Cexpr.decl_scls = Cexpr.NILSCLS;
        Cexpr.decl_inits = inits;
        Cexpr.decl_params = [];
        Cexpr.decl_struct_defined_p = false;
        Cexpr.decl_specs = None;
        Cexpr.decl_location = None;
      }
    in
    [Cexpr.DEFVAR (header)]
end

(*****************************************************************************)
(* check if a OPER expresion is a array operator *)
let rec is_array_oper (oper_expr : Cexpr.expr_t) : bool = (
  match oper_expr with
  | Cexpr.OPER (typenode, _, expr1, _, _) -> (
    match typenode with
      | Cexpr.POINTER_TO typ -> (
          match expr1 with
          | Cexpr.VREF (new_ty, _, _, _) -> (
              match new_ty with
              | Cexpr.ARRAY_OF _ -> true
              | _ -> false)
          | Cexpr.PREF (new_ty, _, _, _) -> (
              match new_ty with
              | Cexpr.ARRAY_OF _ -> true
              | _ -> false)
          | Cexpr.OPER _ -> is_array_oper expr1
          | _ -> false)
      | _ -> false)
  | _ -> false
)

(*****************************************************************************)
let check_self_operator (op : Cexpr.op_t) : bool =
begin
    match op with
    | Cexpr.INCR_PRE
    | Cexpr.DECR_PRE
    | Cexpr.INCR_POST
    | Cexpr.DECR_POST
    | Cexpr.ADD
    | Cexpr.SUB
    | Cexpr.MUL
    | Cexpr.DIV
    | Cexpr.MOD
    | Cexpr.LSH
    | Cexpr.RSH
    | Cexpr.AND
    | Cexpr.IOR
    | Cexpr.XOR -> true
    | _ -> false
end

(** Classify variables. A variable is read variable or writed variable
*)
let rec classify_var_in_expr (current_expr : Cexpr.expr_t)
                             (read_vars : Cexpr.expr_t list)
                             (writed_vars : Cexpr.expr_t list)
                             (read_union_fields : Cexpr.expr_t list)
                             (written_union_fields : Cexpr.expr_t list)
                             (is_write : bool)
                             (is_in_pref : bool)
: ((Cexpr.expr_t list) * (Cexpr.expr_t list)
   * (Cexpr.expr_t list) (* fields of unions that read *)
   * (Cexpr.expr_t list) (* fields of unions that writen *)) =
begin
    let add_variable (current_expr : Cexpr.expr_t)
                     (read_vars : Cexpr.expr_t list)
                     (writed_vars : Cexpr.expr_t list)
                     (is_write : bool)
    : ((Cexpr.expr_t list) * (Cexpr.expr_t list)) = (
        if is_write then (
            if (Cexpr.is_mem_of_exprs current_expr writed_vars) then
                (read_vars, writed_vars)
            else
                (read_vars, ([current_expr] @ writed_vars))
        )
        else (
            if (Cexpr.is_mem_of_exprs current_expr read_vars) then
                (read_vars, writed_vars)
            else
                (([current_expr] @ read_vars), writed_vars)
        )
    ) in
    match current_expr with
    | Cexpr.CALL (ty, ex, exs, func_cond_op) ->
        classify_var_in_exprs exs
                              read_vars
                              writed_vars
                              read_union_fields
                              written_union_fields
                              is_write
                              is_in_pref
    | Cexpr.NIL
    | Cexpr.SIZEOF _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.PRAGMA _
    | Cexpr.PSEUDO_PROLOGUE 
    | Cexpr.PSEUDO_EPILOGUE
    | Cexpr.PSEUDO_NONDET_VALUE _
    | Cexpr.PROGN _ -> (read_vars, writed_vars, read_union_fields, written_union_fields)
    | Cexpr.VREF (tn, id, level, sequence) -> (
        let read_vars, written_vars =
            if (is_pointer tn) then (
                if is_in_pref then
                    add_variable current_expr read_vars writed_vars is_write
                else (read_vars, writed_vars))
            else add_variable current_expr read_vars writed_vars is_write
        in
        (read_vars, written_vars, read_union_fields, written_union_fields))
    | Cexpr.OPER (ty, op, expr1, expr2, expr3) ->
        classify_var_in_exprs [expr1; expr2; expr3]
                              read_vars writed_vars
                              read_union_fields written_union_fields
                              is_write is_in_pref
    | Cexpr.SETF (ty, op, expr1, expr2) -> (
        let (read_vars_1, writed_vars_1, read_union_fields_1, written_union_fields_1) =
            classify_var_in_expr expr1
                                 read_vars writed_vars
                                 read_union_fields written_union_fields
                                 true is_in_pref in
        let (read_vars_2, writed_vars_2, read_union_fields_2, written_union_fields_2) =
            if (check_self_operator op) then 
                classify_var_in_expr expr1
                                     read_vars_1 writed_vars_1
                                     read_union_fields_1 written_union_fields_1
                                     false is_in_pref
            else (read_vars_1, writed_vars_1, read_union_fields_1, written_union_fields_1) in
        classify_var_in_expr expr2
                             read_vars_2 writed_vars_2
                             read_union_fields_2 written_union_fields_2
                             is_write is_in_pref
    )
    | Cexpr.PREF (_, _, expr, _) -> (
        match expr with
        | Cexpr.OPER _ -> (
            if (is_array_oper expr) then
                classify_var_in_expr expr
                                     read_vars writed_vars
                                     read_union_fields written_union_fields
                                     is_write is_in_pref
            else
                classify_var_in_expr expr
                                     read_vars writed_vars
                                     read_union_fields written_union_fields
                                     is_write true)
        | _ ->
            classify_var_in_expr expr
                                 read_vars writed_vars
                                 read_union_fields written_union_fields
                                 is_write true
    )
    | Cexpr.SREF (_, _, expr, _) -> (
        (* if current_expr is get field of onion then add to read_vars or write_vars *)
        let read_union_fields, written_union_fields =
            if Cexpr.union_type_p (Cexpr.type_of_expr expr) then
                add_variable current_expr
                             read_union_fields
                             written_union_fields
                             is_write
            else (read_union_fields, written_union_fields)
        in
        classify_var_in_expr expr
                             read_vars writed_vars
                             read_union_fields written_union_fields
                             is_write is_in_pref

      )
    | Cexpr.ADDROF (_, expr)
    | Cexpr.INIT_FILL (_, expr)
    | Cexpr.PSEUDO_ACSL_RESULT (_, expr)
    | Cexpr.PSEUDO_ACSL_UNIQUE (_, expr)
    | Cexpr.PSEUDO_ACSL_VALID (_, expr)
    | Cexpr.ALL_FIELD (_, expr) ->
        classify_var_in_expr expr
                             read_vars writed_vars
                             read_union_fields written_union_fields
                             is_write is_in_pref
    | Cexpr.DSGN_AREF (_, expr1, expr2)
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (_, expr1, expr2)
    | Cexpr.PSEUDO_ACSL_OLD (_, expr1, expr2) ->
        classify_var_in_exprs [expr1; expr2]
                              read_vars writed_vars
                              read_union_fields written_union_fields
                              is_write is_in_pref
    | Cexpr.CLITERAL (_, init) ->
        classify_var_in_init init
                             read_vars writed_vars
                             read_union_fields written_union_fields
                             is_write is_in_pref
    | Cexpr.ASMSEQUENCE exprs ->
        classify_var_in_exprs exprs
                              read_vars writed_vars
                              read_union_fields written_union_fields
                              is_write is_in_pref
    | Cexpr.PHI (_, phi_lst) -> (
        let (_, exprs) = List.split phi_lst in 
        classify_var_in_exprs exprs
                              read_vars writed_vars
                              read_union_fields written_union_fields
                              is_write is_in_pref
    )
    | Cexpr.QUANTIFIER (_, exprs, expr) ->
        classify_var_in_exprs (exprs @ [expr])
                              read_vars writed_vars
                              read_union_fields written_union_fields
                              is_write is_in_pref
end

and classify_var_in_exprs (current_exprs : Cexpr.expr_t list)
                          (read_vars : Cexpr.expr_t list)
                          (writed_vars : Cexpr.expr_t list)
                          (read_union_fields : Cexpr.expr_t list)
                          (written_union_fields : Cexpr.expr_t list)
                          (is_write : bool)
                          (is_in_pref : bool)
: ((Cexpr.expr_t list) * (Cexpr.expr_t list)
   * (Cexpr.expr_t list) (* fields of unions that read *)
   * (Cexpr.expr_t list) (* fields of unions that writen *)) =
begin
    List.fold_left
        (fun (curr_r_vars, curr_w_vars, curr_r_union_fields, curr_w_union_fields)
             a_expr ->
            classify_var_in_expr a_expr curr_r_vars
                                 curr_w_vars
                                 curr_r_union_fields
                                 curr_w_union_fields
                                 is_write
                                 is_in_pref
        ) (read_vars, writed_vars, read_union_fields, written_union_fields)
          current_exprs
end

and classify_var_in_init (current_init : Cexpr.initializers_t)
                         (read_vars : Cexpr.expr_t list)
                         (writed_vars : Cexpr.expr_t list)
                         (read_union_fields : Cexpr.expr_t list)
                         (written_union_fields : Cexpr.expr_t list)
                         (is_write : bool)
                         (is_in_pref : bool)
: ((Cexpr.expr_t list) * (Cexpr.expr_t list)
   * (Cexpr.expr_t list) (* fields of unions that read *)
   * (Cexpr.expr_t list) (* fields of unions that writen *)) =
begin
    match current_init with
    | Cexpr.INIT_EXPR expr ->
        classify_var_in_expr expr
                             read_vars writed_vars
                             read_union_fields written_union_fields
                             is_write is_in_pref
    | Cexpr.INIT_LIST inits ->
        List.fold_left
        (fun (curr_r_vars, curr_w_vars, curr_r_union_fields, curr_w_union_fields)
             a_init ->
            classify_var_in_init a_init curr_r_vars
                                 curr_w_vars
                                 curr_r_union_fields curr_w_union_fields
                                 is_write
                                 is_in_pref
        ) (read_vars, writed_vars, read_union_fields, written_union_fields) inits
    | Cexpr.INIT_DSGN (exprs, init) -> (
        let (read_vars_1, writed_vars_1, read_union_fields_1, written_union_fields_1) =
            classify_var_in_exprs exprs
                                  read_vars writed_vars
                                  read_union_fields written_union_fields
                                  is_write is_in_pref in
        classify_var_in_init init read_vars_1
                             writed_vars_1
                             read_union_fields_1 written_union_fields_1
                             is_write
                             is_in_pref
    )
end
(*****************************************************************************)
(** Get the size of arary's true dimension
        @param oper_expr : an expresion
        @return
            Some expr if the expression is an array oper and None if others
*)
and get_array_size_oper (oper_expr : Cexpr.expr_t) : Cexpr.expr_t option =
begin
  match oper_expr with
  | Cexpr.OPER (typenode, _, expr1, _, _) -> (
    match typenode with
      | Cexpr.POINTER_TO typ -> (
          match expr1 with
          | Cexpr.VREF (new_ty, _, _, _) -> (
              match new_ty with
              | Cexpr.ARRAY_OF (ty, ex_op) -> ex_op
              | _ -> None)
          | Cexpr.PREF (new_ty, _, _, _) -> (
              match new_ty with
              | Cexpr.ARRAY_OF (ty, ex_op) -> ex_op
              | _ -> None)
          | Cexpr.OPER _ -> get_array_size_oper expr1
          | _ -> None)
      | _ -> None)
  | _ -> None
end

(******************** parse spec ********************)
let parse_spec (spec_string : string)
               (spec_pos : Sexpr.global_pos_t)
               (is_global_specs : bool)
               : Sexpr.lspec_t = (
    Sexpr.current_pos := spec_pos;
    Sexpr.entered_newline := false;
    Scommon.reset_typenames ();
    get_typenames_from_namescope ();
    let lexbuf  = (Lexing.from_string spec_string) in
    let old_is_in_cparser = !Cexpr.is_in_cparser in
    Cexpr.is_in_cparser := false;
    let result =
        if (is_global_specs) then
            (Sg.global_specs Slex.token lexbuf)
         else
            (Sg.local_specs Slex.token lexbuf)
    in
    Cexpr.is_in_cparser := old_is_in_cparser;
    result)
