/* parser/cg.mly 2009-01-07 */
/* Copyright (C) 2007-2009 Yonezawa Lab, The University of Tokyo */

/* C99 Grammer (with part of GCC extensions as of GCC 4.3.2) */

%{

(* MEMO: Rules become fragile in GCC attribute handling. *)

(* TODO: Fix initializers for temporaries introduced for variable
   length array types.  It is not set before use; OK for checking. *)

(* NOTE: Qualifiers in array parameters are ignored.  They are applied
   as qualifiers to pointers. *)

(* NOTE ON GCC ATTRIBUTES: (1) Specifiers named with {wattr} are ones
   with attributes at the tail but not at the head.  They are to avoid
   ambiguity between a <attribute>-<declarator> sequence and an
   <attribute>-<parameter_type_list> sequence in <direct_declarator>
   and <direct_abstract_declarator>.  (2) Attributes are stored at the
   toplevel of a type node in declaration and structure components,
   after handled by {Cexpr.attribute_handlers}.  They are not handled
   at all at the parser level. Pragmas are stored in the pragma stack
   and they are not handled at all at the parser level, either.  (3)
   Some keywords are ambiguous when keywords have alias strings, since
   they are not properly be back to the original identifiers.  *)

(* NOTE ON C89: <declaration_specifiers> and <declarator> are split
   into two variants for handling implicit integer of pre-ANSI C.  The
   ones with "_wt" mean "with a type", and the ones with "_nt" mean
   "not with a type".  It is necessary because a declarator can have a
   typedef-name as a variable, while one without a type cannot. *)

(* NOTE: NONSTANDARD EXTENSIONS: "$assert(logexp)", "$assume(logexp)",
   "$invariant(logexp)", "$requires(logexp)", "$ensures(logexp)",
   "=>", "<=>", "$forall", "$exists". *)

(* NOTE ON C99: (1) Compound literals are handled in flattening in
   "csyntax.ml".  (2) {__func__} is replaced by a string constant in
   the lexer. *)

(* NOTE ON C99 UNSUPPORTED: (1) Pragmas are ignored in the lexer.  (2)
   Pre-ANSI "implicit int" and "implicit function declaration" are
   allowed by intention. *)

(* GCC SUPPORTED EXTENSIONS: (1) Conditionals with Omitted Operands.
   (2) Statements and Declarations in Expressions (Statement
   expressions).  (3) Locally Declared Labels. (4) Referring to a
   Type with typeof.  (5) Case ranges. (6) Incomplete enum Types.  It
   does not check at all.  (7) Zero-sized structures. *)

(* GCC UNSUPPORTED: (1) ASM line.  (2) Nested functions.  (3)
   Transparent unions. (4) Labels as values and computed goto.   (8)
   Array ranges initializers.  AND MANY... *)

(* MEMO: Types in <declarator> may be a pseudo {NOTYPE} for pointer or
   function types, in which {NOTYPE} will be replaced with one in a
   specifier.  {id} in <abstract_declarator> is an empty string. *)

(* MEMO: <init_declarator> for variables takes effect on reducing each
   one to make an identifier visible (or shadowed) to the scanner at
   that point.  Specifiers are stored in
   {current_declaration_specifiers} for this purpose. *)

(* MEMO: <specifier_qualifier_list> is similar to
   <declaration_specifiers> but it does not include a storage class.
   It is used in structure declarations and type names.  See also
   comments to <declaration_specifiers>. *)

(* MEMO: <pointer> lvalue is a pointer reference.  It is of the form
   (POINTER_TO <subordinate>), where <subordinate> is of either a
   pointer reference or NOTYPE.  The NOTYPE will be replaced with a
   type. *)

(* MEMO: An identifier in <primary_expression> maybe an undeclared
   function name, so checking an error for it cannot be done at the
   rule of <primary_expression>.  <postfix_expression1> takes care of
   an undeclared function identifier.  An undeclared function
   identifier is allowed only in function call form.
   <postfix_expression1> may include undeclared one, but
   <postfix_expression> does not include any. *)

(* MEMO: Empty declarator in a declaration means a tag definition for
   structure, union, or enumeration.  The tag should shadow one in
   outer scope, and thus a new binding may be established in these
   cases.  Note that it is allowed for other declarations to have an
   empty declarator, for example, "int;" is a proper declaration. *)

(* MISC MEMO: "initializer" is an OCaml keyword, and cannot be used in
   the rules. *)

(* MISC MEMO: Structure declarators do not refer predecessors in the
   same declarator list. *)

(* MEMO: Assignment conversion calls for usual unary conversion.  It is not
clearly stated in [H&S5]. *)

open Cexpr

let nil_specifiers = (NOTYPE, NILSCLS, NILQUAL, [])

(* {current_declaration_specifiers} keeps track of the current
   defining specifiers.  It is used to define identifiers by referring
   to specifiers in declarators.  It is set in data declaration (but
   not in prototype parameters). *)

let current_declaration_specifiers = ref nil_specifiers

(* {current_enumeration_specifier} keeps track of a currently defining
   enumeration type.  It is used to define enumerators at seeing
   identifiers. *)

let current_enumeration_specifier = ref NOTYPE

(* {parse_error} overrides the default empty error function. *)

let parse_error = Cexpr.syntaxerror

let push_ref (r : 'a list ref) (e : 'a) : unit = (r := (e :: !r))
let pop_ref (r : 'a list ref) : unit = (r := (List.tl !r))
let tmp_stmts = ref []

%}

%start program
%type <string> program

%token Q_NONE
%token <string> Q_IDENTIFIER
%token <string * Cexpr.typenode_t> Q_TYPEDEFNAME
%token <string> Q_TYPESPEC Q_TYPEQUAL
%token <Cexpr.expr_t> Q_CONSTANTI Q_CONSTANTF
%token <Cexpr.expr_t> Q_STRING

%token Q_TYPEDEF Q_EXTERN Q_STATIC Q_AUTO Q_REGISTER
%token Q_CONST Q_VOLATILE Q_RESTRICT Q_INLINE

%token Q_SIZEOF Q_ENUM Q_STRUCT Q_UNION Q_IF Q_ELSE Q_WHILE Q_DO
%token Q_FOR Q_SWITCH Q_CASE Q_DEFAULT Q_BREAK Q_CONTINUE Q_RETURN Q_GOTO
%token Q_OFFSETOF Q_ASM
%token Q_TYPEOF Q_ALIGNOF Q_ATTRIBUTE Q_LABEL /*GCC*/
%token Q_TYPES_COMPATIBLE_P /*GCC*/
%token <string> Q_SPEC
%token Q_ELLIPSIS
%token Q_LPA Q_RPA Q_LBC Q_RBC Q_LBK Q_RBK
%token Q_DOT Q_ARW Q_AMP Q_STR
%token Q_EQL Q_QES Q_COL Q_COM Q_SEM Q_ATM Q_DOL
%token Q_PLS Q_MNS Q_DIV Q_MOD Q_BAR Q_HAT Q_TLD Q_BAN
%token Q_PLSPLS Q_MNSMNS Q_AMPAMP Q_BARBAR Q_LSH Q_RSH

/* LPA='(' RPA=')' LBC='{' RBC='}' LBK='[' RBK=']' */
/* AMP='&' STR='*' DOT='.' ARW='->' */
/* EQL='=' QES='?' COL=':' COM=',' SEM=';' ATM='@' DOL='$' */
/* PLS='+' MNS='-' DIV='/' MOD='%' BAR='|' HAT='^' TLD='~' BAN='!' */
/* PLSPLS='++' MNSMNS='--' AMPAMP='&&' BABAR='||' LSH='<<' RSH='>>' */

%token <Cexpr.op_t> Q_ASSIGN Q_EQUAL Q_COMPARE

/* Dummy tokens not used in the grammar. */

%token Q_Error
%token Q_Ignore

/*NONSTANDARD*/

%token Q_ASSERT Q_ASSUME Q_REQUIRES Q_ENSURES Q_INVARIANT
%token Q_CHECK
%token Q_IMP Q_IFF Q_FORALL Q_EXISTS

%nonassoc Q_IF
%nonassoc Q_ELSE

%right Q_ASSIGN Q_EQL
%right Q_QES Q_COL
%left Q_BARBAR
%left Q_AMPAMP
%left Q_BAR Q_HAT Q_AMP
%left Q_EQUAL
%left Q_COMPARE
%left Q_LSH Q_RSH
%left Q_PLS Q_MNS
%left Q_STR Q_DIV Q_MOD

%right Q_UNARY Q_PLSPLS Q_MNSMNS
%left Q_HYPERUNARY
%left Q_ARW Q_DOT Q_LPA Q_LBK

%%

program:
    /*empty*/ /*GCC*/
	{ (warn_pedantic_in_parse "Empty source file") ;
	  "" }
    | program_start translation_unit
	{ "" }
	;

program_start:
    /*empty*/
	{ (*(init_yylex ()) ;*)
	  Cexpr.name_scope_stack := [(Hashtbl.create 10)] ;
	  Cexpr.tag_scope_stack := [(Hashtbl.create 10)] ;
	  Cexpr.pragma_scope_stack := [(Vector.create 2 [])] ;
	  Cexpr.scope_block_counter := 2 ;
	  Cexpr.scope_block_stack := [1] ;
	  Cexpr.scope_stack_history := []
	  (*in_parameter_scanning_p := false*) }
	;

/* NOTE: {push_scope} and {pop_scope} are used in declarators, while
   {push_block} and {pop_block} are used in compound statements. */

push_scope:
    /*empty*/
	{ (push_ref Cexpr.name_scope_stack (Hashtbl.create 10)) ;
	  (push_ref Cexpr.tag_scope_stack (Hashtbl.create 10)) ;
	  (push_ref Cexpr.pragma_scope_stack (Vector.create 2 [])) ;
	  let level = ((List.hd !Cexpr.scope_block_stack) + 1) in
	  (push_ref Cexpr.scope_block_stack level) ;
	  (* For debugging. *)
	  (push_ref Cexpr.scope_stack_history
	      [!Cexpr.name_scope_stack ; !Cexpr.tag_scope_stack]) }
	;

get_location:
  { Cexpr.location_of_cur_fun :=
        Some (Sexpr.make_global_pos !Cexpr.current_file
                                    !Cexpr.current_line
                                    !Cexpr.current_cnum_from_bol) }
  ;

pop_scope:
    /*empty*/
	{ (pop_ref Cexpr.name_scope_stack) ;
	  (pop_ref Cexpr.tag_scope_stack) ;
	  (pop_ref Cexpr.pragma_scope_stack) ;
	  (pop_ref Cexpr.scope_block_stack) }
	;

push_block:
    /*empty*/
	{ (push_ref Cexpr.name_scope_stack (Hashtbl.create 10)) ;
	  (push_ref Cexpr.tag_scope_stack (Hashtbl.create 10)) ;
	  (push_ref Cexpr.pragma_scope_stack (Vector.create 2 [])) ;
	  (incr Cexpr.scope_block_counter) ;
	  (push_ref Cexpr.scope_block_stack !Cexpr.scope_block_counter) ;
	  (* For debugging. *)
	  (push_ref Cexpr.scope_stack_history
	      [!Cexpr.name_scope_stack ; !Cexpr.tag_scope_stack]) }
	;

pop_block:
    /*empty*/
	{ (pop_ref Cexpr.name_scope_stack) ;
	  (pop_ref Cexpr.tag_scope_stack) ;
	  (pop_ref Cexpr.pragma_scope_stack) ;
	  (pop_ref Cexpr.scope_block_stack) }
	;

update_func_scope:
    /*empty*/
        { Ccommon.current_func_scope :=
              (Ccommon.compact_scope !Cexpr.name_scope_stack,
               Ccommon.compact_scope !Cexpr.tag_scope_stack) }
        ;

update_loop_scope:
    /*empty*/
        { Ccommon.current_loop_scope :=
              (Ccommon.compact_scope !Cexpr.name_scope_stack,
               Ccommon.compact_scope !Cexpr.tag_scope_stack) }
        ;

push_stmt_spec_scopes_in_a_block:
    /*empty*/
        { Stack.push (Queue.create ()) Ccommon.stmt_spec_scopes; }
        ;

pop_stmt_spec_scope_in_a_block:
    /*empty*/
        { let _ = Stack.pop Ccommon.stmt_spec_scopes in () }
        ;

translation_unit:
    external_declaration
	{ tmp_stmts := (!tmp_stmts @ $1); 
    (List.iter !Cexpr.process_declaration $1) ;
	  Cexpr.scope_block_counter := 2 ;
	  Cexpr.scope_block_stack := [1] }
    | translation_unit external_declaration
	{ (Ccommon.check_using_var_declarationvar_following_var_spec !tmp_stmts $2);
    (List.iter !Cexpr.process_declaration $2) ;
    Cexpr.scope_block_counter := 2 ;
    Cexpr.scope_block_stack := [1];
    tmp_stmts := (!tmp_stmts @ $2) }
	;

external_declaration:
    get_location function_definition
        { Ccommon.push_func_scope ();
          $2 }
    | get_location variable_declaration
	{ List.iter (fun stmt ->
              match stmt with
              | Cexpr.DEFVAR decl -> (
                  if (Cexpr.function_type_p decl.Cexpr.decl_tn)
                      or (Cexpr.function_pointer_type_p decl.Cexpr.decl_tn) then
                      ()
                  else (
                      Ccommon.current_func_scope :=
                          (Ccommon.compact_scope !Cexpr.name_scope_stack,
                           Ccommon.compact_scope !Cexpr.tag_scope_stack)
                  );
                  Ccommon.push_func_scope ();
                )
              | _ -> ()
          ) $2;
          $2 }
    | /*empty*/ get_location Q_SEM
	{ (warn_pedantic_in_parse "Empty declaration") ;
	  [] }
    | get_location gcc_asm_simple Q_SEM
	{ [$2] }
    /* ANLAB:
        + function contacts
        + data invariants
        + (global) ghost code */
    | Q_SPEC
        { let pos = Sexpr.make_global_pos !Cexpr.current_file
                                          !Cexpr.spec_startpos_ln
                                          !Cexpr.spec_startpos_cnum_from_bol
          in
          let specs = Ccommon.parse_spec ($1) pos true in
          (Ccommon.push_global_spec_scopes ());
          [!Cexpr.note_of_curr_spec; Cexpr.SPEC specs] }
	;

function_definition:
    function_declaration_maybe_traditional update_func_scope
    push_stmt_spec_scopes_in_a_block compound_statement
    pop_stmt_spec_scope_in_a_block pop_scope
	{ assert ((List.length !Cexpr.name_scope_stack) == 1) ;
	  Cexpr.current_function := Cexpr.nil_declaration ;
          let (h, (_, p2)) = $1 in
          let normalized_return_func =
              Ccommon.normalize_return Cexpr.NOTYPE (DEFUN (h, $4)) in
          (p2 :: normalized_return_func) }
        ;

function_declaration_maybe_traditional: /*Added*/
    function_declaration
    { $1 }
    | function_declaration traditional_declaration_list
	{ assert ((List.length !Cexpr.name_scope_stack) == 2) ;
          let p1, _ = $1 in
(*	  (check_traditional_parameters $1 $2) ;*)
	  (check_traditional_parameters p1 $2) ;
	  $1 }
	;

function_declaration: /*Added*/
    declaration_specifiers_setcontext_wt declarator_define_wt_wattr push_scope
	{ assert ((List.length !Cexpr.name_scope_stack) == 2) ;
	  (add_binding_of_parameters $2 !Cexpr.name_scope_stack) ;
	  (check_function_declaration $2 !Cexpr.name_scope_stack) ;
	  ($2, $1) }
    | declaration_specifiers_setcontext_nt declarator_define_nt_wattr push_scope
	{ assert ((List.length !Cexpr.name_scope_stack) == 2) ;
	  (add_binding_of_parameters $2 !Cexpr.name_scope_stack) ;
	  (check_function_declaration $2 !Cexpr.name_scope_stack) ;
	  ($2, $1) }
    | declaration_specifiers_setcontext_empty declarator_define_nt_wattr push_scope
	{ assert ((List.length !Cexpr.name_scope_stack) == 2) ;
	  (warn_pedantic_in_parse "Implicit int in declaration") ;
	  (add_binding_of_parameters $2 !Cexpr.name_scope_stack) ;
	  (check_function_declaration $2 !Cexpr.name_scope_stack) ;
          let nil_specifier =(Cexpr.NOTYPE, Cexpr.NILSCLS, Cexpr.NILQUAL, []) in
          let note = make_linenote () in
          let temp = (nil_specifier, note) in
	  ($2, temp) }
	;

/* Non-function declaration (variable, structure, union, enum). */

variable_declaration:
    declaration_specifiers_setcontext_wt init_declarator_list_wt_wattr Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
          let (_, note) = $1 in
	  [note] @ (List.map (fun d -> (DEFVAR d)) $2) }
    | declaration_specifiers_setcontext_nt init_declarator_list_nt_wattr Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
          let (_ , note) = $1 in
	  [note] @ (List.map (fun d -> (DEFVAR d)) $2) }
    | declaration_specifiers_setcontext_wt Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
	  let ((tn, scls, quals, attrs), note) = $1 in
	  let newtype = (make_structure_tag_on_current_scope tn) in
	  let x = (combine_declaration true
		      (newtype, scls, quals, attrs) empty_declarator) in
	  [note; (DEFVAR x)] }
    | declaration_specifiers_setcontext_nt Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
	  let ((tn, scls, quals, attrs), note) = $1 in
	  let newtype = (make_structure_tag_on_current_scope tn) in
	  let x = (combine_declaration false
		      (newtype, scls, quals, attrs) empty_declarator) in
	  [note; (DEFVAR x)] }
	;

variable_declaration_wattr:
    declaration_specifiers_setcontext_wt_wattr init_declarator_list_wt_wattr Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
	  (List.map (fun d -> (DEFVAR d)) $2) }
    | declaration_specifiers_setcontext_nt_wattr init_declarator_list_nt_wattr Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
	  (List.map (fun d -> (DEFVAR d)) $2) }
    | declaration_specifiers_setcontext_wt_wattr Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
	  let (tn, scls, quals, attrs) = $1 in
	  let newtype = (make_structure_tag_on_current_scope tn) in
	  let x = (combine_declaration true
		      (newtype, scls, quals, attrs) empty_declarator) in
	  [(DEFVAR x)] }
    | declaration_specifiers_setcontext_nt_wattr Q_SEM
	{ (current_declaration_specifiers := nil_specifiers) ;
	  let (tn, scls, quals, attrs) = $1 in
	  let newtype = (make_structure_tag_on_current_scope tn) in
	  let x = (combine_declaration false
		      (newtype, scls, quals, attrs) empty_declarator) in
	  [(DEFVAR x)] }
	;

/* Traditional parameter declaration (cannot have GCC attributes). */

traditional_declaration_list:
    variable_declaration_wattr
	{ $1 }
    | traditional_declaration_list variable_declaration_wattr
	{ ($1 @ $2) }
	;

declaration_specifiers_setcontext_wt:
    attr0 declaration_specifiers_wt_wattr
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $2 $1)) in
          let note = (make_linenote ()) in
	  (current_declaration_specifiers := spec) ;
	  (spec, note) }
	;

declaration_specifiers_setcontext_nt:
    attr0 declaration_specifiers_nt_wattr
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $2 $1)) in
          let note = (make_linenote ()) in
	  (current_declaration_specifiers := spec) ;
	  (spec, note) }
	;

declaration_specifiers_setcontext_wt_wattr:
    declaration_specifiers_wt_wattr
	{ let spec = (make_proper_specifier $1) in
	  (current_declaration_specifiers := spec) ;
	  spec }
	;

declaration_specifiers_setcontext_nt_wattr:
    declaration_specifiers_nt_wattr
	{ let spec = (make_proper_specifier $1) in
	  (current_declaration_specifiers := spec) ;
	  spec }
	;

declaration_specifiers_setcontext_empty:
    attr0
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier
			     (make_simple_specifier NOTYPE) $1)) in
	  (current_declaration_specifiers := spec) ;
	  spec }
	;

declaration_specifiers_wt_wattr:
    type_specifier attr0
	{ (append_attributes_to_specifier $1 $2) }
    | type_specifier attr0 declaration_specifiers_typespecs_wattr
	{ (append_attributes_to_specifier (merge_specifiers $1 $3) $2) }
    | declaration_specifiers_nt_wattr type_specifier attr0
	{ (append_attributes_to_specifier (merge_specifiers $1 $2) $3) }
    | declaration_specifiers_nt_wattr type_specifier attr0
	declaration_specifiers_typespecs_wattr
	{ (append_attributes_to_specifier
	      (merge_specifiers $1 (merge_specifiers $2 $4)) $3) }
	;

declaration_specifiers_nt_wattr:
    storage_class_specifier attr0
	{ (make_specifier NOTYPE $1 NILQUAL $2) }
    | declaration_specifiers_nt_wattr storage_class_specifier attr0
	{ (merge_specifiers $1 (make_specifier NOTYPE $2 NILQUAL $3)) }
    | type_qualifier attr0
	{ (make_specifier NOTYPE NILSCLS $1 $2) }
    | declaration_specifiers_nt_wattr type_qualifier attr0
	{ (merge_specifiers $1 (make_specifier NOTYPE NILSCLS $2 $3)) }
    | function_specifier attr0
	{ (append_attributes_to_specifier $1 $2) }
    | declaration_specifiers_nt_wattr function_specifier attr0
	{ (append_attributes_to_specifier (merge_specifiers $1 $2) $3) }
	;

declaration_specifiers_typespecs_wattr:
    storage_class_specifier attr0
	{ (make_specifier NOTYPE $1 NILQUAL $2) }
    | declaration_specifiers_typespecs_wattr storage_class_specifier attr0
	{ (merge_specifiers $1 (make_specifier NOTYPE $2 NILQUAL $3)) }
    | type_qualifier attr0
	{ (make_specifier NOTYPE NILSCLS $1 $2) }
    | declaration_specifiers_typespecs_wattr type_qualifier attr0
	{ (merge_specifiers $1 (make_specifier NOTYPE NILSCLS $2 $3)) }
    | Q_TYPESPEC attr0
	{ (make_specifier (SPECIFR_TYPE [$1]) NILSCLS NILQUAL $2) }
    | declaration_specifiers_typespecs_wattr Q_TYPESPEC attr0
	{ (merge_specifiers $1
	      (make_specifier (SPECIFR_TYPE [$2]) NILSCLS NILQUAL $3)) }
    | function_specifier attr0
	{ (append_attributes_to_specifier $1 $2) }
    | declaration_specifiers_typespecs_wattr function_specifier attr0
	{ (append_attributes_to_specifier (merge_specifiers $1 $2) $3) }
	;

storage_class_specifier:
    Q_TYPEDEF
	{ TYPEDEF }
    | Q_EXTERN
	{ EXTERN }
    | Q_STATIC
	{ STATIC }
    | Q_AUTO
	{ AUTO }
    | Q_REGISTER
	{ REGISTER }
	;

type_qualifier:
    Q_CONST
	{ CONSTANT }
    | Q_VOLATILE
	{ VOLATILE }
    | Q_RESTRICT
	{ RESTRICT }
	;

function_specifier:
	Q_INLINE
	{ (make_specifier NOTYPE NILSCLS INLINE []) }
	;

type_specifier:
    Q_TYPESPEC
	{ (make_simple_specifier (SPECIFR_TYPE [$1])) }
    | Q_TYPEDEFNAME
	{ (make_simple_specifier (snd $1)) }
    | struct_or_union_specifier
	{ (Ccommon.current_list_components := []);$1 }
    | enum_specifier
	{ $1 }
    | typeof_specifier /*GCC*/
	{ $1 }
	;

/* NOTE: GCC attributes at tails in <struct_or_union_specifier> and
   <enum_specifier> are omitted because attributes follow in
   <declaration_specifiers>.  Handling attributes is there. */

struct_or_union_specifier:
    struct_or_union attr0 identifier_or_typedef_name
	{ let tn = (find_or_make_structure_type $1 $3
		       !Cexpr.tag_scope_stack false) in
	  (make_specifier tn NILSCLS NILQUAL $2) }
    | struct_or_union_specifier1 Q_LBC struct_declaration_list Q_RBC
	{ (define_structure_type $1 $3) ;
	  (make_structuredef_specifier $1 $3) }
    | struct_or_union_specifier1 Q_LBC Q_RBC /*GCC*/
	{ (warn_pedantic_in_parse "Empty structure/union") ;
	  (define_structure_type $1 []) ;
	  (make_structuredef_specifier $1 []) }
    | struct_or_union_specifier0 Q_LBC struct_declaration_list Q_RBC
	{ (define_structure_type $1 $3) ;
	  (make_structuredef_specifier $1 $3) }
    | struct_or_union_specifier0 Q_LBC Q_RBC /*GCC*/
	{ (warn_pedantic_in_parse "Empty structure/union") ;
	  (define_structure_type $1 []) ;
	  (make_structuredef_specifier $1 []) }
	;

struct_or_union_specifier0: /*Added*/
    struct_or_union attr0
	{ let tn = (find_or_make_structure_type $1 ""
		       !Cexpr.tag_scope_stack true) in
	  (make_specifier tn NILSCLS NILQUAL $2) }
	;

struct_or_union_specifier1: /*Added*/
    struct_or_union attr0 identifier_or_typedef_name
	{ Sexpr.current_name_struct := $3;
    let tn = (find_or_make_structure_type $1 $3
		       !Cexpr.tag_scope_stack true) in
	  (make_specifier tn NILSCLS NILQUAL $2) }
	;

struct_or_union:
    Q_STRUCT
	{ true }
    | Q_UNION
	{ false }
	;

struct_declaration_list:
    struct_declaration
	{ $1 }
    | struct_declaration_list struct_declaration
	{ ($1 @ $2 ) }
	;

struct_declaration:
    /*empty*/ Q_SEM /*GCC*/
	{ (warn_pedantic_in_parse "Empty structure declaration") ;
	  [] }
    | specifier_qualifier_list_wt_wattr struct_declarator_list_wt_wattr Q_SEM
	{ let spec = $1 (*(append_attributes_to_specifier $1 $2)*) in
     (* added by anlab - get params of function-pointer field*)
          let scope_stacks = !Ccommon.current_func_scope in
          Sinstrument.check_curr_component $2 $1 scope_stacks;
	  (List.map
	      (fun (declarator : declarator_t) ->
		(get_structure_components_in_declaration
		    (combine_declaration true spec declarator)))
	      $2) }
    | specifier_qualifier_list_nt_wattr struct_declarator_list_nt_wattr Q_SEM
	{ let spec = $1 (*(append_attributes_to_specifier $1 $2)*) in
	  (List.map
	      (fun (declarator : declarator_t) ->
		(get_structure_components_in_declaration
		    (combine_declaration false spec declarator)))
	      $2) }
    /* ANLAB:
         + model fields in struct declarations */
    | Q_SPEC
        { Cexpr.is_in_cparser := false;
          let result = (Sinstrument.process_model_field $1) in
          Cexpr.is_in_cparser := true;
          result }
	;

typeof_specifier: /*GCC*/
    Q_TYPEOF Q_LPA expression Q_RPA
	{ (make_simple_specifier (type_of_expr $3)) }
    | Q_TYPEOF Q_LPA type_name Q_RPA
	{ (make_simple_specifier $3) }
	;

specifier_qualifier_list_wt_wattr:
    specifier_qualifier_list_wt_wattr1
	{ (make_proper_specifier $1) }
	;

specifier_qualifier_list_nt_wattr:
    type_qualifier_list attr0
	{ (make_proper_specifier
	      (make_specifier NOTYPE NILSCLS $1 $2)) }
	;

specifier_qualifier_list_wt_wattr1:
      type_specifier attr0
	{ (append_attributes_to_specifier $1 $2) }
    | type_specifier attr0 qualifier_typespec_list_wattr
	{ (append_attributes_to_specifier (merge_specifiers $1 $3) $2) }
    | type_qualifier_list type_specifier attr0 qualifier_typespec_list_wattr
	{ (merge_specifiers
	      (make_specifier NOTYPE NILSCLS $1 $3)
	      (merge_specifiers $2 $4)) }
    | type_qualifier_list type_specifier attr0
	{ (merge_specifiers (make_specifier NOTYPE NILSCLS $1 $3) $2) }
	;

qualifier_typespec_list_wattr:
    type_qualifier attr0
	{ (make_specifier NOTYPE NILSCLS $1 $2) }
    | qualifier_typespec_list_wattr type_qualifier attr0
	{ (merge_specifiers $1
	      (make_specifier NOTYPE NILSCLS $2 $3)) }
    | Q_TYPESPEC attr0
	{ (make_specifier (SPECIFR_TYPE [$1]) NILSCLS NILQUAL $2) }
    | qualifier_typespec_list_wattr Q_TYPESPEC attr0
	{ (merge_specifiers $1
	      (make_specifier (SPECIFR_TYPE [$2]) NILSCLS NILQUAL $3)) }
	;

struct_declarator_list_wt_wattr:
    struct_declarator_wt_wattr
	{ [$1] }
    | struct_declarator_list_wt_wattr Q_COM attr0 struct_declarator_wt_wattr
	{ ($1 @ [(append_attributes_to_declarator $4 $3)]) }
	;

struct_declarator_list_nt_wattr:
    struct_declarator_nt_wattr
	{ [$1] }
    | struct_declarator_list_nt_wattr Q_COM attr0 struct_declarator_nt_wattr
	{ ($1 @ [(append_attributes_to_declarator $4 $3)]) }
	;

struct_declarator_wt_wattr:
    /*empty*/ /*GCC*/
	{ (*(warn_in_parse "Unnamed structure member")*)
	  empty_declarator }
    | abstract_declarator_wattr /*GCC*/
	{ (*(warn_in_parse "Unnamed structure member")*)
	  $1 }
    | declarator_wt attr0
	{ (append_attributes_to_declarator $1 $2) }
    | declarator_wt Q_COL constant_expression attr0
	{ let decl = (append_attributes_to_declarator $1 $4) in
	  match decl with
	    (NOTYPE, varid, params, attrs) ->
	      (make_declarator
		  (make_bitfield_type NOTYPE $3) varid params attrs)
	  | _ ->
	      (syntaxerror "Bitfield has illegal type") }
    | Q_COL constant_expression attr0
	{ (make_declarator
	      (make_bitfield_type NOTYPE $2) ("", 0) [] $3) }
	;

struct_declarator_nt_wattr:
    declarator_nt attr0
	{ (append_attributes_to_declarator $1 $2) }
    | declarator_nt Q_COL constant_expression attr0
	{ let decl = (append_attributes_to_declarator $1 $4) in
	  match decl with
	    (NOTYPE, varid, params, attrs) ->
	      (make_declarator
		  (make_bitfield_type NOTYPE $3) varid params attrs)
	  | _ -> (syntaxerror "Bitfield has illegal type") }
    | Q_COL constant_expression attr0
	{ (make_declarator
	      (make_bitfield_type NOTYPE $2) ("", 0) [] $3) }
	;

enum_specifier:
    Q_ENUM attr0 identifier_or_typedef_name
	{ current_enumeration_specifier := NOTYPE ;
	  let tn = (find_or_make_enumeration_type
		       $3 !Cexpr.tag_scope_stack false) in
	  (make_specifier tn NILSCLS NILQUAL $2) }
    | enum_specifier0 Q_LBC enumerator_list Q_RBC
	{ current_enumeration_specifier := NOTYPE ;
	  ignore (define_enumeration_type $1 $3) ;
	  (make_enumeratordef_specifier $1 $3) }
    | enum_specifier0 Q_LBC enumerator_list Q_COM Q_RBC
	{ current_enumeration_specifier := NOTYPE ;
	  ignore (define_enumeration_type $1 $3) ;
	  (make_enumeratordef_specifier $1 $3) }
	;

enum_specifier0: /*Added*/
    Q_ENUM attr0
	{ let tn = (find_or_make_enumeration_type
		      "" !Cexpr.tag_scope_stack true) in
	  (current_enumeration_specifier := tn) ;
	  (make_specifier tn NILSCLS NILQUAL $2) }
    | Q_ENUM attr0 identifier_or_typedef_name
	{ let tn = (find_or_make_enumeration_type
		       $3 !Cexpr.tag_scope_stack true) in
	  (current_enumeration_specifier := tn) ;
	  (make_specifier tn NILSCLS NILQUAL $2) }
	;

enumerator_list:
    enumerator
	{ [$1] }
    | enumerator_list Q_COM enumerator
	{ ($1 @ [$3]) }
	;

enumerator:
    Q_IDENTIFIER
	{ (define_enumerator
	      !current_enumeration_specifier
	      $1 !Cexpr.name_scope_stack) ;
	  ($1, NIL) }
    | Q_IDENTIFIER Q_EQL constant_expression
	{ (define_enumerator
	      !current_enumeration_specifier
	      $1 !Cexpr.name_scope_stack) ;
	  ($1, $3) }
	;

init_declarator_list_wt_wattr:
    init_declarator_wt_wattr
	{ [$1] }
    | init_declarator_list_wt_wattr Q_COM init_declarator_wt_wattr
	{ ($1 @ [$3]) }
	;

init_declarator_list_nt_wattr:
    init_declarator_nt_wattr
	{ [$1] }
    | init_declarator_list_nt_wattr Q_COM init_declarator_nt_wattr
	{ ($1 @ [$3]) }
	;

init_declarator_wt_wattr:
    declarator_define_wt_wattr
	{ $1 }
    | declarator_define_wt_wattr Q_EQL initializers
	{ (merge_inits_to_declaration $1 $3) }
	;

init_declarator_nt_wattr:
    declarator_define_nt_wattr
	{ $1 }
    | declarator_define_nt_wattr Q_EQL initializers
	{ (merge_inits_to_declaration $1 $3) }
	;

declarator_define_wt_wattr: /*Added*/
    declarator_wt asm0 attr0
	{ let spec = (append_attributes_to_specifier
			 !current_declaration_specifiers $3) in
	  let decl = (combine_declaration true spec $1) in
	  (define_identifier_in_scope decl) }
	;

declarator_define_nt_wattr: /*Added*/
    declarator_nt asm0 attr0
	{ let spec = (append_attributes_to_specifier
			 !current_declaration_specifiers $3) in
	  let decl = (combine_declaration false spec $1) in
	  (define_identifier_in_scope decl) }
	;

declarator_wt:
    pointer_wattr direct_declarator_wt
	{ let (pt, at) = $1 in
	  let (tn, varid, params, attrs) = $2 in
	  ((substitute_basetype_of tn pt), varid, params, (at @ attrs)) }
    | direct_declarator_wt
	{ $1 }
	;

declarator_nt:
    pointer_wattr direct_declarator_nt
	{ let (pt, at) = $1 in
	  let (tn, varid, params, attrs) = $2 in
	  ((substitute_basetype_of tn pt), varid, params, (at @ attrs)) }
    | direct_declarator_nt
	{ $1 }
	;

abstract_declarator_wattr:
    pointer_wattr
	{ let (pt, at) = $1 in
	  (pt, ("", 0), [], at) }
    | pointer_wattr direct_abstract_declarator_wattr
	{ let (pt, at) = $1 in
	  let (tn, varid, params, attrs) = $2 in
	  assert ((string_equal (fst varid) "") && params == []) ;
	  ((substitute_basetype_of tn pt), varid, params, attrs) }
    | direct_abstract_declarator_wattr
	{ $1 }
	;

/* NOTE: {TYPEDEFNAME} can be used here just as {IDENTIFIER}. */

direct_declarator_wt:
    Q_TYPEDEFNAME
	{ let level = (List.hd !Cexpr.scope_block_stack) in
	  (make_simple_declarator ((fst $1), level)) }
    | Q_IDENTIFIER
	{ let level = (List.hd !Cexpr.scope_block_stack) in
	  (make_simple_declarator ($1, level)) }
    | Q_LPA attr0 declarator_nt Q_RPA
	{ (append_attributes_to_declarator $3 $2) }
    | direct_declarator_wt Q_LBK type_qualifier_list
	assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of tn
		       (make_array_type NOTYPE (Some $4))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK type_qualifier_list Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE None)) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some $3))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK Q_STATIC type_qualifier_list
	assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (add_type_qualifiers
		       (substitute_basetype_of
			   tn (make_array_type NOTYPE (Some $5)))
		       $4) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK Q_STATIC assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some $4))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK type_qualifier_list Q_STATIC
	assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (add_type_qualifiers
		       (substitute_basetype_of
			   tn (make_array_type NOTYPE (Some $5)))
		       $3) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK type_qualifier_list Q_STR Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (add_type_qualifiers
		       (substitute_basetype_of
			   tn (make_array_type NOTYPE (Some NIL)))
		       $3) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK Q_STR Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some NIL))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LBK Q_RBK
	{ let (tn, _, _, _) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE None)) in
	  (modify_declarator $1 tx) }
    | direct_declarator_wt Q_LPA parameter_type_list Q_RPA
	{ let (tn, varid, _, attrs) = $1 in
	  let (params, types) = (split_parameter_types_and_ids $3) in
	  let tx = (substitute_basetype_of
		       tn (make_function_type NOTYPE types)) in
	  (make_declarator tx varid params attrs) }
    | direct_declarator_wt Q_LPA push_scope identifier_list pop_scope Q_RPA
	{ let (tn, varid, _, attrs) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_function_type_traditional NOTYPE)) in
	  (make_declarator tx varid $4 attrs) }
    | direct_declarator_wt Q_LPA update_func_scope Q_RPA
	{ let (tn, _, _, _) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_function_type_traditional NOTYPE)) in
	  (modify_declarator $1 tx) }
	;

direct_declarator_nt:
    Q_IDENTIFIER
	{ let level = (List.hd !Cexpr.scope_block_stack) in
	  (make_simple_declarator ($1, level)) }
    | Q_LPA attr0 declarator_nt Q_RPA
	{ (append_attributes_to_declarator $3 $2) }
    | direct_declarator_nt Q_LBK type_qualifier_list
	assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of tn
		       (make_array_type NOTYPE (Some $4))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK type_qualifier_list Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE None)) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some $3))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK Q_STATIC type_qualifier_list
	assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (add_type_qualifiers
		       (substitute_basetype_of
			   tn (make_array_type NOTYPE (Some $5)))
		       $4) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK Q_STATIC assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some $4))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK type_qualifier_list Q_STATIC
	assignment_expression Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (add_type_qualifiers
		       (substitute_basetype_of
			   tn (make_array_type NOTYPE (Some $5)))
		       $3) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK type_qualifier_list Q_STR Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (add_type_qualifiers
		       (substitute_basetype_of
			   tn (make_array_type NOTYPE (Some NIL)))
		       $3) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK Q_STR Q_RBK /*C99*/
	{ let (tn, _, _, _) = $1 in
	  (* Qualifiers ignored. *)
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some NIL))) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LBK Q_RBK
	{ let (tn, _, _, _) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE None)) in
	  (modify_declarator $1 tx) }
    | direct_declarator_nt Q_LPA parameter_type_list Q_RPA
	{ let (tn, varid, _, attrs) = $1 in
	  let (params, types) = (split_parameter_types_and_ids $3) in
	  let tx = (substitute_basetype_of
		       tn (make_function_type NOTYPE types)) in
	  (make_declarator tx varid params attrs) }
    | direct_declarator_nt Q_LPA push_scope identifier_list pop_scope Q_RPA
	{ let (tn, varid, _, attrs) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_function_type_traditional NOTYPE)) in
	  (make_declarator tx varid $4 attrs) }
    | direct_declarator_nt Q_LPA update_func_scope Q_RPA
	{ let (tn, _, _, _) = $1 in
	  let tx = (substitute_basetype_of
		       tn (make_function_type_traditional NOTYPE)) in
	  (modify_declarator $1 tx) }
	;

direct_abstract_declarator_wattr:
    Q_LPA attr0 abstract_declarator_wattr Q_RPA
	{ (append_attributes_to_declarator $3 $2) }
    | direct_abstract_declarator_wattr
	Q_LBK assignment_expression Q_RBK /*C99*/
	{ let (tn, varid, params, _) = $1 in
	  assert ((string_equal (fst varid) "") && params == []) ;
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some $3))) in
	  (modify_declarator $1 tx) }
    | direct_abstract_declarator_wattr Q_LBK Q_STR Q_RBK /*C99*/
	{ let (tn, varid, params, _) = $1 in
	  assert ((string_equal (fst varid) "") && params == []) ;
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE (Some NIL))) in
	  (modify_declarator $1 tx) }
    | direct_abstract_declarator_wattr Q_LBK Q_RBK
	{ let (tn, varid, params, _) = $1 in
	  assert ((string_equal (fst varid) "") && params == []) ;
	  let tx = (substitute_basetype_of
		       tn (make_array_type NOTYPE None)) in
	  (modify_declarator $1 tx) }
    | Q_LBK assignment_expression Q_RBK /*C99*/
	{ let tx = (make_array_type NOTYPE (Some $2)) in
	  (make_declarator tx ("", 0) [] []) }
    | Q_LBK Q_STR Q_RBK /*C99*/
	{ let tx = (make_array_type NOTYPE (Some NIL)) in
	  (make_declarator tx ("", 0) [] []) }
    | Q_LBK Q_RBK
	{ let tx = (make_array_type NOTYPE None) in
	  (make_declarator tx ("", 0) [] []) }
    | direct_abstract_declarator_wattr Q_LPA parameter_type_list Q_RPA
	{ let (tn, varid, params, _) = $1 in
	  assert ((string_equal (fst varid) "") && params == []) ;
	  let (ids, types) = (split_parameter_types_and_ids $3) in
	  let tx = (substitute_basetype_of
		       tn (make_function_type NOTYPE types)) in
	  (modify_declarator $1 tx) }
    | direct_abstract_declarator_wattr Q_LPA update_func_scope Q_RPA
	{ let (tn, varid, params, _) = $1 in
	  assert ((string_equal (fst varid) "") && params == []) ;
	  let tx = (substitute_basetype_of
		       tn (make_function_type_traditional NOTYPE)) in
	  (modify_declarator $1 tx) }
    | Q_LPA parameter_type_list Q_RPA
	{ let (params, types) = (split_parameter_types_and_ids $2) in
	  let tx = (make_function_type NOTYPE types) in
	  (make_declarator tx ("", 0) params []) }
    | Q_LPA update_func_scope Q_RPA
	{ let tx = (make_function_type_traditional NOTYPE) in
	  (make_declarator tx ("", 0) [] []) }
	;

pointer_wattr:
    Q_STR attr0
	{ ((make_pointer_type NOTYPE), $2) }
    | Q_STR attr0 type_qualifier_list attr0
	{ ((make_qualified_type (make_pointer_type NOTYPE) $3), ($2 @ $4)) }
    | Q_STR attr0 pointer_wattr
	{ let (pt, at) = $3 in
	  ((substitute_basetype_of pt (make_pointer_type NOTYPE)), ($2 @ at)) }
    | Q_STR attr0 type_qualifier_list attr0 pointer_wattr
	{ let (pt, at) = $5 in
	  ((substitute_basetype_of pt
	       (make_qualified_type (make_pointer_type NOTYPE) $3)),
	  ($2 @ $4 @ at)) }
	;

type_qualifier_list:
    type_qualifier
	{ $1 }
    | type_qualifier_list type_qualifier
	{ (merge_qualifiers true $1 $2) }
	;

/* Parameter prototypes. */

parameter_type_list:
    parameter_list update_func_scope pop_scope
	{ $1 }
    | parameter_list Q_COM Q_ELLIPSIS update_func_scope pop_scope
	{ let binding_of_ellipsis = { binding_name = (ellipsis_string, 0) ;
                               binding_tn = ELLIPSIS_TYPE ;
                               binding_scls = NILSCLS
                             } in
	  Hashtbl.add (fst !Ccommon.current_func_scope) ellipsis_string
               binding_of_ellipsis;
	  ($1 @ [pseudo_ellipsis]) }
	;

parameter_list:
    parameter_declaration_head
	{ [$1] }
    | parameter_list Q_COM parameter_declaration
	{ ($1 @ [$3]) }
	;

parameter_declaration_head:
    attr0 push_scope declaration_specifiers_wt_wattr declarator_wt attr0
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $3 ($1 @ $5))) in
	  let decl = (combine_declaration true spec $4) in
	  (define_identifier_in_scope decl) }
    | attr0 push_scope declaration_specifiers_wt_wattr
	abstract_declarator_wattr
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $3 $1)) in
	  (combine_declaration true spec $4) }
    | attr0 push_scope declaration_specifiers_wt_wattr
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $3 $1)) in
	  (combine_declaration true spec empty_declarator) }
	;

parameter_declaration:
    attr0 declaration_specifiers_wt_wattr declarator_wt attr0
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $2 ($1 @ $4))) in
	  let decl = (combine_declaration true spec $3) in
	  (define_identifier_in_scope decl) }
    | attr0 declaration_specifiers_wt_wattr abstract_declarator_wattr
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $2 $1)) in
	  (combine_declaration true spec $3) }
    | attr0 declaration_specifiers_wt_wattr
	{ let spec = (make_proper_specifier
			 (append_attributes_to_specifier $2 $1)) in
	  (combine_declaration true spec empty_declarator) }
	;

identifier_list:
    Q_IDENTIFIER
	{ [$1] }
    | identifier_list Q_COM Q_IDENTIFIER
	{ ($1 @ [$3]) }
	;

initializers:
    assignment_expression
	{ (INIT_EXPR $1) }
    | Q_LBC initializer_list Q_RBC
	{ (INIT_LIST $2) }
    | Q_LBC initializer_list Q_COM Q_RBC
	{ (INIT_LIST $2) }
    | Q_LBC Q_RBC /*GCC*/
	{ (warn_pedantic_in_parse "Empty initializer") ;
	  (INIT_LIST []) }
	;

/* C99: initializer_list is defined using designation in C99. */

/* NOTE: The rules below accept strange mix of old-style and C99
   designators: (such as {slot: = 10}). */

initializer_list:
    initializers
	{ [$1] }
    | initializer_list Q_COM initializers
	{ ($1 @ [$3]) }
    | designator_list Q_EQL initializers /*C99*/
	{ [(INIT_DSGN ($1, $3))] }
    | initializer_list Q_COM designator_list Q_EQL initializers /*C99*/
	{ ($1 @ [(INIT_DSGN ($3, $5))]) }
    | designator_list initializers /*GCC*/
	{ (warn_pedantic_in_parse "Old-style designator") ;
	  [(INIT_DSGN ($1, $2))] }
    | initializer_list Q_COM designator_list initializers /*GCC*/
	{ (warn_pedantic_in_parse "Old-style designator") ;
	  ($1 @ [(INIT_DSGN ($3, $4))]) }
	;

designator_list: /*C99*/
    designator
	{ [$1] }
    | designator_list designator
	{ ($1 @ [$2]) }
	;

designator: /*C99*/
    Q_LBK constant_expression Q_RBK
	{ (DSGN_AREF (NOTYPE, $2, NIL)) }
    | Q_DOT identifier_or_typedef_name
	{ (SREF (NOTYPE, false, NIL, $2)) }
    | identifier_or_typedef_name Q_COL /*GCC*/
	{ (warn_pedantic_in_parse "Old-style designator") ;
	  (SREF (NOTYPE, false, NIL, $1)) }
    | Q_LBK constant_expression Q_ELLIPSIS constant_expression Q_RBK /*GCC*/
	{ (warn_pedantic_in_parse "Array range initializer") ;
	  (DSGN_AREF (NOTYPE, $2, $4)) }
	;

type_name:
    attr0 specifier_qualifier_list_wt_wattr abstract_declarator_wattr
	{ let spec = (append_attributes_to_specifier $2 $1) in
	  let decl = (combine_declaration true spec $3) in
	  (decl.decl_tn) }
    | attr0 specifier_qualifier_list_nt_wattr abstract_declarator_wattr
	{ let spec = (append_attributes_to_specifier $2 $1) in
	  let decl = (combine_declaration false spec $3) in
	  (decl.decl_tn) }
    | attr0 specifier_qualifier_list_wt_wattr
	{ let spec = (append_attributes_to_specifier $2 $1) in
	  let decl = (combine_declaration true spec empty_declarator) in
	  (decl.decl_tn) }
    | attr0 specifier_qualifier_list_nt_wattr
	{ let spec = (append_attributes_to_specifier $2 $1) in
	  let decl = (combine_declaration false spec empty_declarator) in
	  (decl.decl_tn) }
	;

statement:
    labeled_statement
	{ $1 }
    | push_stmt_spec_scopes_in_a_block compound_statement
      pop_stmt_spec_scope_in_a_block
	{ [$2] }
    | attr0 expression_statement
	{ if ($1 != []) then
	    (warn_in_parse ("Attributes for an expression ignored: "
			     ^ (Cexpr.attribute_names $1 ";"))) ;
	  $2 }
    | push_block selection_statement pop_block /*C99*/
	{ $2 }
    | push_block iteration_statement pop_block /*C99*/
	{ Ccommon.push_loop_scope (); $2 }
    | jump_statement
	{ $1 }
    | gcc_asm_statement Q_SEM
	{ [$1] }
    | assertion_statement
	{ [$1] }
    /* ANLAB:
        + statement annotations
        + ghost code */
    | Q_SPEC
        { let pos = Sexpr.make_global_pos !Cexpr.current_file
                                          !Cexpr.spec_startpos_ln
                                          !Cexpr.spec_startpos_cnum_from_bol
          in
          let specs = Ccommon.parse_spec ($1) pos false in
          Ccommon.push_stmt_spec_scopes ();
          [!Cexpr.note_of_curr_spec;Cexpr.SPEC specs]
        }
    ;

labeled_statement:
    attr0 identifier_or_typedef_name ln0 Q_COL statement
	{ let s = (make_named_label $2 $1 !name_scope_stack) in
	  ($3 :: s :: $5) }
    | Q_CASE ln0 constant_expression Q_COL statement
	{ ($2 :: (LABEL (CASE_LABEL ($3, NIL))) :: $5) }
    | Q_CASE ln0 constant_expression Q_ELLIPSIS constant_expression Q_COL
	statement /*GCC*/
	{ ($2 :: (LABEL (CASE_LABEL ($3, $5))) :: $7) }
    | Q_DEFAULT ln0 Q_COL statement
	{ ($2 :: (LABEL (DFLT_LABEL)) :: $4) }
	;

expression_statement:
    expression ln0 Q_SEM
	{ [$2 ; (STMT $1)] }
    | ln0 Q_SEM
	{ [$1] }
	;

compound_statement: /*C99*/
    Q_LBC push_block ln0 block_item_list
    pop_block ln0 Q_RBC
	{ Cexpr.is_in_cparser := false;
          let stmts = (Sinstrument.spec_instrument $4 (ref [])
                                                      (Hashtbl.create 2)
                                                      (Hashtbl.create 2))
          in
          Cexpr.is_in_cparser := true;
          (BEGIN ($3 :: (stmts @ [$6]))) }
    | Q_LBC Q_RBC
	{ (BEGIN []) }
	;

block_item_list: /*C99*/
    block_item
	{ $1 }
    | block_item_list block_item
  { (Ccommon.check_using_var_declarationvar_following_var_spec $1 $2);
    ($1 @ $2) }
	;
/* NOTE: (gcc-attributes are added for statement to avoid conflict) */

block_item: /*C99*/
/*    ln0 variable_declaration
	{ [$1] @ $2 }*/
    variable_declaration
	{ $1 }
    | statement
	{ (*[make_linenote_1()]@*)$1 }
    | Q_LABEL label_declarator_list Q_SEM /*GCC*/
	{ [] }
	;

label_declarator_list: /*GCC*/
    identifier_or_typedef_name
	{ (define_local_label $1 !Cexpr.name_scope_stack) }
    | label_declarator_list Q_COM identifier_or_typedef_name
	{ (define_local_label $3 !Cexpr.name_scope_stack) }
	;

selection_statement:
    Q_IF ln0 Q_LPA expression Q_RPA push_block ln0 statement pop_block
	{ [$2 ; (IF ($4, ($7 :: $8), []))] }
    | Q_IF ln0 Q_LPA expression Q_RPA push_block ln0 statement pop_block
	Q_ELSE push_block ln0 statement pop_block
	{ [$2 ; (IF ($4, ($7 :: $8), ($12 :: $13)))] }
    | Q_SWITCH ln0 Q_LPA expression Q_RPA push_block
	ln0 statement pop_block
	{ [$2 ; (SWITCH ($4, ($7 :: $8)))] }
	;

/* Note: No checks are included that the declaration in the 2nd-form
   "for" only includes auto (or register) variables. */

iteration_statement:
    Q_WHILE ln0 Q_LPA expression Q_RPA push_block update_loop_scope
    ln0 statement pop_block
	{ [$2 ; (DOWHILE (true, $4, ($8 :: $9)))] }
    | Q_DO ln0 push_block update_loop_scope ln0 statement pop_block
	Q_WHILE Q_LPA expression Q_RPA Q_SEM
	{ [$2 ; (DOWHILE (false, $10, ($5 :: $6)))] }
    | Q_FOR ln0 Q_LPA expression_opt Q_SEM expression_opt Q_SEM
	expression_opt Q_RPA push_block update_loop_scope ln0 statement pop_block
	{ [$2 ; (FOR ([(STMT $4)], $6, $8, ($12 :: $13)))] }
    | Q_FOR ln0 Q_LPA variable_declaration expression_opt Q_SEM
	expression_opt Q_RPA push_block update_loop_scope ln0 statement pop_block
	{ [$2 ; (FOR ($4, $5, $7, ($11 :: $12)))] }
	;

expression_opt:
    /*empty*/
	{ NIL }
    | expression
	{ $1 }
	;

jump_statement:
    Q_GOTO ln0 identifier_or_typedef_name Q_SEM
	{ let s = (make_goto_stmt $3 !Cexpr.name_scope_stack) in
	  [$2 ; s] }
    | Q_CONTINUE ln0 Q_SEM
	{ [$2 ; (CONTINUE)] }
    | Q_BREAK ln0 Q_SEM
	{ [$2 ; (BREAK)] }
    | Q_RETURN ln0 expression Q_SEM
	{ [$2 ; (make_return_expression (Some $3))] }
    | Q_RETURN ln0 Q_SEM
	{ [$2 ; (make_return_expression None)] }
	;

expression:
    assignment_expression
	{ $1 }
    | assignment_expression Q_COM expression
	{ (OPER ((type_of_expr $3), COMMA, $1, $3, NIL)) }
	;

assignment_expression:
    conditional_expression
	{ (make_pseudo_value_simpler $1) }
    | unary_expression assignment_operator assignment_expression
	{ (make_assignment $2 $1 $3) }
	;

assignment_operator:
    Q_EQL
	{ SET }
    | Q_ASSIGN
	{ $1 }
	;

conditional_expression:
    logical_fol_expression
	{ $1 }
    | logical_fol_expression Q_QES expression Q_COL conditional_expression
	{ (make_conditional_expr COND $1 $3 $5) }
    | logical_fol_expression Q_QES Q_COL conditional_expression /*GCC*/
	{ (make_conditional_expr COND $1 NIL $4 ) }
	;

constant_expression:
    conditional_expression
	{ $1 }
	;

/*NONSTANDARD*/
logical_fol_expression:
    logical_imp_expression
	{ $1 }
    | Q_FORALL push_block variable_declaration logical_fol_expression pop_block
	{ (make_quantified_expression "forall" $3 $4) }
    | Q_EXISTS push_block variable_declaration logical_fol_expression pop_block
	{ (make_quantified_expression "exists" $3 $4) }
	;

/*NONSTANDARD*/
logical_imp_expression:
    logical_or_expression
	{ $1 }
    | logical_imp_expression Q_IMP logical_or_expression
	{ (make_conditional_expr IMP $1 $3 NIL) }
    | logical_imp_expression Q_IFF logical_or_expression
	{ (make_conditional_expr IFF $1 $3 NIL) }
	;

logical_or_expression:
    logical_and_expression
	{ $1 }
    | logical_or_expression Q_BARBAR logical_and_expression
	{ (make_conditional_expr OROR $1 $3 NIL) }
	;

logical_and_expression:
    ior_expression
	{ $1 }
    | logical_and_expression Q_AMPAMP ior_expression
	{ (make_conditional_expr ANDAND $1 $3 NIL) }
	;

ior_expression:
    xor_expression
	{ $1 }
    | ior_expression Q_BAR xor_expression
	{ (make_binary_operation IOR $1 $3) }
	;

xor_expression:
    and_expression
	{ $1 }
    | xor_expression Q_HAT and_expression
	{ (make_binary_operation XOR $1 $3) }
	;

and_expression:
    equality_expression
	{ $1 }
    | and_expression Q_AMP equality_expression
	{ (make_binary_operation AND $1 $3) }
	;

equality_expression:
    relational_expression
	{ $1 }
    | equality_expression Q_EQUAL relational_expression
	{ (make_binary_operation $2 $1 $3) }
	;

relational_expression:
    shift_expression
	{ $1 }
    | relational_expression Q_COMPARE shift_expression
	{ (make_binary_operation $2 $1 $3) }
	;

shift_expression:
    additive_expression
	{ $1 }
    | shift_expression Q_LSH additive_expression
	{ (make_binary_operation LSH $1 $3) }
    | shift_expression Q_RSH additive_expression
	{ (make_binary_operation RSH $1 $3) }
	;

additive_expression:
    multiplicative_expression
	{ $1 }
    | additive_expression Q_PLS multiplicative_expression
	{ (make_binary_operation ADD $1 $3) }
    | additive_expression Q_MNS multiplicative_expression
	{ (make_binary_operation SUB $1 $3) }
	;

multiplicative_expression:
    cast_expression
	{ $1 }
    | multiplicative_expression Q_STR cast_expression
	{ (make_binary_operation MUL $1 $3) }
    | multiplicative_expression Q_DIV cast_expression
	{ (make_binary_operation DIV $1 $3) }
    | multiplicative_expression Q_MOD cast_expression
	{ (Cexpr.CALL (Cexpr.UNSIGNED_INTEGER,
                       Cexpr.VREF (Cexpr.UNSIGNED_INTEGER,
                                   "clap_builtin_modulo",
                                   Cexpr.global_scope_level,
                                   0),
                       [$1; $3],
                       None)) }
	;

cast_expression:
    unary_expression
	{ $1 }
    | Q_LPA type_name Q_RPA cast_expression
	{ (make_cast $2 $4) }
	;

unary_expression:
    postfix_expression
	{ $1 }
    | Q_PLSPLS unary_expression
	{ (make_unary_operation INCR_PRE $2) }
    | Q_MNSMNS unary_expression
	{ (make_unary_operation DECR_PRE $2) }
    | unary_operator cast_expression
	{ (make_unary_operation $1 $2) }
    | Q_SIZEOF unary_expression
	{ (make_sizeof_operation (type_of_expr $2)) }
    | Q_SIZEOF Q_LPA type_name Q_RPA
	{ (make_sizeof_operation $3) }
    | Q_OFFSETOF Q_LPA type_specifier Q_COM
	offsetof_member Q_RPA /*C99*/
	{ (make_offsetof_operation $3 $5) }
    | Q_ALIGNOF unary_expression /*GCC*/
	{ (make_alignof_operation (type_of_expr $2)) }
    | Q_ALIGNOF Q_LPA type_name Q_RPA /*GCC*/
	{ (make_alignof_operation $3) }
    | Q_AMPAMP identifier_or_typedef_name /*GCC*/
	{ (warn_pedantic_in_parse "Taking label values") ;
	  (make_label_value $2 !Cexpr.name_scope_stack) }
	;

offsetof_member:
    identifier_or_typedef_name
	{ (make_sref_pseudo false NIL $1) }
    | offsetof_member Q_DOT identifier_or_typedef_name
	{ (make_sref_pseudo false $1 $3) }
    | offsetof_member Q_ARW identifier_or_typedef_name
	{ (make_sref_pseudo true $1 $3) }

unary_operator:
    Q_AMP
	{ PSEUDO_AMP }
    | Q_STR
	{ PSEUDO_STR }
    | Q_PLS
	{ PSEUDO_PLS }
    | Q_MNS
	{ PSEUDO_MNS }
    | Q_TLD
	{ PSEUDO_TLD }
    | Q_BAN
	{ PSEUDO_BAN }
	;

postfix_expression:
    postfix_expression1
	{ (check_undeclared_reference $1 false) }
	;

postfix_expression1:
    primary_expression
	{ $1 }
    | postfix_expression Q_LBK expression Q_RBK
	{ (make_aref $1 $3) }
    | postfix_expression1 Q_LPA argument_expression_list Q_RPA
	{ (make_function_call (check_undeclared_reference $1 true) $3) }
    | postfix_expression1 Q_LPA Q_RPA
	{ (make_function_call (check_undeclared_reference $1 true) []) }
    | postfix_expression Q_DOT identifier_or_typedef_name
	{ (make_sref_dot $1 $3) }
    | postfix_expression Q_ARW identifier_or_typedef_name
	{ (make_sref_arw $1 $3) }
    | postfix_expression Q_PLSPLS
	{ (make_unary_operation INCR_POST $1) }
    | postfix_expression Q_MNSMNS
	{ (make_unary_operation DECR_POST $1) }
    | Q_LPA type_name Q_RPA Q_LBC initializer_list Q_RBC /*C99*/
	{ (make_compound_literal $2 (INIT_LIST $5)) }
    | Q_LPA type_name Q_RPA Q_LBC initializer_list Q_COM Q_RBC /*C99*/
	{ (make_compound_literal $2 (INIT_LIST $5)) }
    | Q_TYPES_COMPATIBLE_P Q_LPA attr0 declaration_specifiers_wt_wattr Q_COM
	attr0 declaration_specifiers_wt_wattr Q_RPA
	{ let u0 = (append_attributes_to_specifier $4 $3) in
	  let u1 = (append_attributes_to_specifier $7 $6) in
	  (builtin_types_compatible_p u0 u1) }
	;

primary_expression:
    Q_IDENTIFIER
	{ (make_variable_reference $1) }
    | Q_CONSTANTI
	{ $1 }
    | Q_CONSTANTF
	{ $1 }
    | string_literal_x
	{ $1 }
    | Q_LPA expression Q_RPA
	{ $2 }
    | Q_LPA push_stmt_spec_scopes_in_a_block compound_statement
      pop_stmt_spec_scope_in_a_block Q_RPA /*GCC*/
	{ (make_statement_expression $3) }
	;

string_literal_x:
    Q_STRING
	{ $1 }
    | string_literal_x Q_STRING
	{ match ($1, $2) with
	    (STRING (w0, s0, content0), STRING (w1, s1, content1)) ->
	      if (w0 != w1) then
		(syntaxerror "mixing wide and narrow characters") ;
	      let newcontent =
		(Array.concat
		    [(Array.sub content0 0 ((Array.length content0) - 1)) ;
		     content1]) in
	      (STRING (w0, (s0 ^ s1), newcontent))
	  | _ -> (assert false) }
	;

identifier_or_typedef_name: /*Added*/
    Q_IDENTIFIER
	{ $1 }
    | Q_TYPEDEFNAME
	{ (fst $1) }
	;

argument_expression_list:
    assignment_expression
	{ [$1] }
    | argument_expression_list Q_COM assignment_expression
	{ ($1 @ [$3]) }
	;

/*NONSTANDARD*/

assertion_statement:
    Q_ASSERT Q_LPA logical_formula Q_RPA
	{ (make_annotation "$assert" $3) }
    | Q_ASSUME Q_LPA logical_formula Q_RPA
	{ (make_annotation "$assume" $3) }
    | Q_REQUIRES Q_LPA logical_formula Q_RPA
	{ (make_annotation "$requires" $3) }
    | Q_ENSURES Q_LPA logical_formula Q_RPA
	{ (make_annotation "$ensures" $3) }
    | Q_INVARIANT Q_LPA logical_formula Q_RPA
	{ (make_annotation "$invariant" $3) }
    | Q_CHECK Q_LPA logical_formula Q_RPA
	{ (make_annotation "$check" $3) }
	;

logical_formula:
    expression
	{ $1 }
	;

asm0:
    /*empty*/
	{ [] }
    | gcc_asm_simple
	{ [$1] }
	;

gcc_asm_simple:
    | Q_ASM Q_LPA string_literal_x Q_RPA
	{ (ASM $3) }
	;

/* MEMO: const or restrict may not appear after ASM. */

gcc_asm_statement:
    Q_ASM Q_LPA gcc_asm_argument Q_RPA
	{ (ASM $3) }
    | Q_ASM Q_VOLATILE Q_LPA gcc_asm_argument Q_RPA
	{ (ASM $4) }
	;

gcc_asm_argument:
    string_literal_x
	{ $1 }
    | string_literal_x Q_COL gcc_asm_operand_list0
	{ (Cexpr.ASMSEQUENCE [$1 ; $3]) }
    | string_literal_x Q_COL gcc_asm_operand_list0
	Q_COL gcc_asm_operand_list0
	{ (Cexpr.ASMSEQUENCE [$1 ; $3 ; $5])}
    | string_literal_x Q_COL gcc_asm_operand_list0
	Q_COL gcc_asm_operand_list0 Q_COL gcc_asm_string_literal_list
	{ (Cexpr.ASMSEQUENCE [$1 ; $3 ; $5 ; $7]) }
	;

gcc_asm_string_literal_list:
    string_literal_x
	{ (Cexpr.ASMSEQUENCE [$1]) }
    | gcc_asm_string_literal_list Q_COM string_literal_x
	{ (Cexpr.append_to_sequnce $1 [$3]) }
	;

gcc_asm_operand_list0:
    /*empty*/
	{ (Cexpr.ASMSEQUENCE []) }
    | gcc_asm_operand_list
	{ (Cexpr.ASMSEQUENCE $1) }
	;

gcc_asm_operand_list:
    gcc_asm_operand
	{ [$1] }
    | gcc_asm_operand Q_COM gcc_asm_operand_list
	{ ($1 :: $3) }
	;

gcc_asm_operand:
    string_literal_x Q_LPA expression Q_RPA
	{ (Cexpr.ASMSEQUENCE [$1 ; $3]) }
    | Q_LBK identifier_or_typedef_name Q_RBK
	string_literal_x Q_LPA expression Q_RPA
	{ let content = (Misc.string_to_int_array $2) in
	  (Cexpr.ASMSEQUENCE [(Cexpr.STRING (false, $2, content)) ; $4 ; $6]) }
	;

/* Line note.  It adds note on following statements. */

ln0:
    /*empty*/
	{ (make_linenote ()) }
	;

/* GCC attributes.  This part is taken from the GCC attribute
   description. */

attr0:
    /*empty*/
	{ [] }
    | attr0 gcc_attributes
	{ ($1 @ $2) }
	;

gcc_attributes:
    Q_ATTRIBUTE attribute_start Q_LPA Q_LPA attribute_list
	attribute_end Q_RPA Q_RPA
	{ $5 }
	;

attribute_start:
    /*empty*/
	{ Cexpr.in_attribute_scanning_p := true }
	;

attribute_end:
    /*empty*/
	{ Cexpr.in_attribute_scanning_p := false }
	;

attribute_list:
    attribute_entry
	{ $1 }
    | attribute_list Q_COM attribute_entry
	{ ($1 @ $3) }
	;

attribute_entry:
    /*empty*/
	{ [] }
    | gcc_anyword
	{ [(ATTR ($1, NIL))] }
    | gcc_anyword Q_LPA attribute_expression Q_RPA
	{ [(ATTR ($1, $3))] }
	;

attribute_expression:
    expression
	{ $1 }
	;

gcc_anyword:
    Q_TYPESPEC { $1 }
    | Q_TYPEDEFNAME { (fst $1) }
    | Q_IDENTIFIER { $1 }
    | Q_TYPEDEF { "typedef" }
    | Q_EXTERN { "extern" }
    | Q_STATIC { "static" }
    | Q_AUTO { "auto" }
    | Q_REGISTER { "register" }
    | Q_INLINE { "inline" }
    | Q_CONST { "const" }
    | Q_VOLATILE { "volatile" }
    | Q_RESTRICT { "restrict" }
    /*| Q_THREAD { () }*/
    /*| Q_DFLOAT32 { () }*/
    /*| Q_DFLOAT64 { () }*/
    /*| Q_DFLOAT128 { () }*/
    /*| Q_FRACT { () }*/
    /*| Q_ACCUM { () }*/
    /*| Q_SAT { () }*/
    /* Belows are not allowed in GCC 4.3.2 yet. */
    | Q_TYPEQUAL { $1 }
    | Q_SIZEOF { "sizeof" }
    | Q_ENUM { "enum" }
    | Q_STRUCT { "struct" }
    | Q_UNION { "union" }
    | Q_IF { "if" }
    | Q_ELSE { "else" }
    | Q_WHILE { "while" }
    | Q_DO { "do" }
    | Q_FOR { "for" }
    | Q_SWITCH { "switch" }
    | Q_CASE { "case" }
    | Q_DEFAULT { "default" }
    | Q_BREAK { "break" }
    | Q_CONTINUE { "continue" }
    | Q_RETURN { "return" }
    | Q_GOTO { "goto" }
    | Q_ASM { "asm" }
    | Q_TYPEOF { "__typeof__" }         /*ambiguous*/
    | Q_ALIGNOF { "__alignof__" }       /*ambiguous*/
    | Q_ATTRIBUTE { "__attribute__" }   /*ambiguous*/
    | Q_LABEL { "__label__" }           /*ambiguous*/
	;

%%

(*
yyabort :=
  (fun state yym yyn ->
    (Printf.printf "abort: state: %s, yym: %s, yyn: %s\n"
	state yym yyn))
*)
