(* parser/specs/sinstrument.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num

let hashtbl_size = 10

(******************** Type *****************)
type tuple_type_expr = Cexpr.typenode_t * Cexpr.expr_t

(*********************************************)
(* map string -> typenode_t *)
let typedef_info_hashtbl : (string, Cexpr.typenode_t) Hashtbl.t = 
    Hashtbl.create hashtbl_size

(* map string -> string list : 
   name of type invariant -> list of except function *)
let hash_except_function_type_invariant : (string, string) Hashtbl.t = 
    Hashtbl.create hashtbl_size

let (fun_header_infos_hashtbl : (string, (Cexpr.declaration_t))Hashtbl.t)
        = Hashtbl.create hashtbl_size

let (fun_names_is_assigned : string list ref) = ref [] 

let insert_func_contract_callee = ref true

(************** Get infos *****************)
let update_global_variable 
    (global_vars : (string, Cexpr.expr_t) Hashtbl.t)
    (stmt : Cexpr.stmt_t) : unit = 
    (begin
    match stmt with
    | Cexpr.DEFVAR decl -> (
        let (id, lev) = decl.Cexpr.decl_id in
        let typ = decl.Cexpr.decl_tn in
        let expr = Cexpr.VREF(typ, id, lev, 0) in
        Hashtbl.replace global_vars id expr
        )
    | _ -> assert false
    end)

(*  *)
let import_parameters (local_vars : (string, Cexpr.expr_t) Hashtbl.t)
                      (fun_header : Cexpr.declaration_t) : unit = (
    let fun_type = fun_header.Cexpr.decl_tn in 
    let param_names = fun_header.Cexpr.decl_params in
    (match (Ccommon.make_return_variable fun_header) with
     | Some ex -> (Hashtbl.add local_vars Ccommon.return_expr_name ex)
     | None -> ()
    );
    match fun_type with
    | Cexpr.FUNCTION_OF(_, param_types) -> (
        match param_types with
        | None -> ()
        | Some types -> 
            (begin
            assert ((List.length types) == (List.length param_names));
            List.iter2 (fun _type name -> 
                            let expr = Cexpr.VREF(_type, name, 2, 0)
                            in Hashtbl.replace local_vars name expr)
                        types param_names
            end)
        )
    | _ -> ()
        (*assert false*)
)

(* get true type *)
let find_true_type (typ : Cexpr.typenode_t) : Cexpr.typenode_t = (
   match typ with
   | Cexpr.STRUCTURE_TYPE(struct_type) -> (
       let id = struct_type.Cexpr.strc_tag in
       (Ccommon.get_struct_type_from_binding id)
   )
   | Cexpr.ENUMERATION_TYPE (enum_type) -> (
       let name = enum_type.Cexpr.enum_tag in
       (Ccommon.get_struct_type_from_binding name)
   )
   | Cexpr.SPECIFR_TYPE(ls_id) -> (
            match ls_id with
            | [id] -> (
                try Hashtbl.find typedef_info_hashtbl id 
                with
                | Not_found -> typ
              )
            | _ -> assert false
         )
   | _ -> typ
)     
(* *)
let rec get_param_types (fun_type : Cexpr.typenode_t) = (
    match fun_type with
    | Cexpr.POINTER_TO ty -> get_param_types ty
    | Cexpr.FUNCTION_OF (_ , s) -> (
        match s with
        | Some types -> types
        | None -> []
    )
    | Cexpr.ATTRIBUTED_TYPE (tn, _) ->
        (get_param_types tn)
    | _ -> assert false
)
(* *)
let collect_header (defs : Cexpr.stmt_t list) : unit = (
    List.iter(
        fun def -> 
            match def with
            | Cexpr.DEFVAR(header) -> (
                match (Ccommon.get_function_name header) with
                | Some name -> 
                    (Hashtbl.replace fun_header_infos_hashtbl 
                                     name 
                                     (header))
                | None -> ())
            | Cexpr.DEFUN(header,_) -> (
                match (Ccommon.get_function_name header) with
                | Some name -> 
                    (Hashtbl.replace fun_header_infos_hashtbl 
                                     name 
                                     (header))
                | None -> ())          
            | _ -> ()
    )defs
)

(**************** Process model field - call by cg.ml **************)

(* make type invariant for model field *)
(* (init_fun_name,    (struct_name, field_name, field_type)) *)
let (const_field_of_initializer_hashtbl :
        (string, (string * Sexpr.lexpr * Cexpr.typenode_t)) Hashtbl.t)
            = Hashtbl.create 100
(*
let make_type_inv_for_field (struct_name : string) 
                            (field_name : string)
                            (term : Sexpr.term)
                            (init_fun_name : string) 
                            (const_or_notnull : bool)
                            (position : Sexpr.global_pos_t) : unit = (
    let invariant_name =
        if const_or_notnull then 
            struct_name ^ "_" ^ field_name ^ "_is_constant" 
        else 
            struct_name ^ "_" ^ field_name ^ "_is_not_null" 
    in
    (* var_name, var_type: parameter of the invariant function *)
    let var_name = "var_temp_" ^ struct_name in
    let term1 = (Sexpr.change_term_to_sref var_name term) in
    (*print_endline "hello:";print_endline (Scommon.string_of_term term);
    print_endline (Scommon.string_of_term term1);Scommon.pause();*)
    let var_type =
        Cexpr.STRUCTURE_TYPE { 
            Cexpr.strc_struct_or_union_p = true;
            Cexpr.strc_tag = struct_name;
            Cexpr.strc_unique_id = struct_name;
            Cexpr.strc_typebody_table = (Hashtbl.create 10)
        }
    in
    (* if (var_name.__initialized__ == 1)
            check_const(var_name.field_name);
         or check_not_null(var_name.feal_name); *)
    let pred_sub = 
        if const_or_notnull then 
            Sexpr.PRED_IS_CONST (term1
                (*Sexpr.TERM_DEF_FIELD (
                    Sexpr.ID (var_name), 
                    field_name)*))
        else 
            Sexpr.PRED_IS_NOT_NULL (term1
                (*Sexpr.TERM_DEF_FIELD (
                    Sexpr.ID (var_name), 
                    field_name)*))
    in
    let preds = [
        Sexpr.PRED_IF(
            Sexpr.PRED_REL ( 
                Sexpr.TERM_DEF_FIELD (
                    (Sexpr.ID (var_name), 
                     "__initialized__" )),
                Sexpr.REL_EQ, 
                Sexpr.TERM_INTEGER (1)),
            [pred_sub],
            None)]
    in
    let type_invariant = (
        invariant_name,
        var_type, 
        var_name,
        preds,
        Some([(Sexpr.EXCEPT_WHOLE (init_fun_name))]),
        position
    )
    in
    Hashtbl.add
        (Scommon.implicit_type_invariants_hashtbl) 
        struct_name type_invariant 
)
*)
(* make function contracts for constant model field *)
(*
let make_func_contract_for_const_field (struct_name : string)
                                       (term : Sexpr.term)
                                       (field_type : Cexpr.typenode_t) 
                                       (init_fun_name : string) : unit = (
    Hashtbl.add const_field_of_initializer_hashtbl 
                    init_fun_name 
                    (struct_name, term, field_type)
)
*)
let struct_modelfields_hashtbl : 
    (string , string list) Hashtbl.t  = Hashtbl.create hashtbl_size

let add_modelfields (struct_name : string) 
                    (fields_tobe_added : string list) = (
    let model_fields = 
        try 
            Hashtbl.find struct_modelfields_hashtbl struct_name
        with
        | Not_found -> []
    in
    let model_fields = (model_fields @ fields_tobe_added) in
    Hashtbl.replace struct_modelfields_hashtbl struct_name model_fields
)

let (contract_for_init_fun_hashtbl : (string, (string*bool)) Hashtbl.t)
    = Hashtbl.create hashtbl_size    

let make_fun_contract (struct_name : string)
                      (fun_init_name : string)
                      (no_contract : bool) : unit = (
    Hashtbl.replace contract_for_init_fun_hashtbl
                        fun_init_name
                        (struct_name, no_contract)
)

let process_model_field (raw_spec : string):
                        (string * Cexpr.typenode_t) list = (
    let position = (Sexpr.make_global_pos !Cexpr.current_file
                                          !Cexpr.spec_startpos_ln
                                          !Cexpr.spec_startpos_cnum_from_bol) in
    let spec_data = Ccommon.parse_spec raw_spec position true in
    match spec_data with
    | Sexpr.SPEC_MODEL_FIELD model_field -> (
        match model_field with
        | Sexpr.INIT_MODEL_FIELD (struct_name, fun_init_name, _) -> (
            add_modelfields struct_name ["__initialized__"];
            make_fun_contract struct_name 
                              fun_init_name 
                              !Scommon.no_check_initializer;
            [( "__initialized__", Cexpr.INTEGER)] 
         )
        | Sexpr.DECL_FIELD (struct_name, str, typ) -> (
            let ctyp = Stranslate.translate_logictype typ in
            add_modelfields struct_name [(str)];
            [(struct_name ^ "_" ^ str ^ "_model_field"), ctyp]
         )
        | Sexpr.RELEASE_MODEL_FIELD  _ -> []
(*        | Sexpr.NOT_NULL_MODEL_FIELD (struct_name, init_fun_name, term) ->*)
(*          assert false (* temporary do not process *)*)
(*
            (
            let field_name = (Scommon.get_name_of_term term) in
            (*let field_type = (Scommon.get_type_of_term term) in*)
            (make_type_inv_for_field struct_name 
                                     field_name
                                     term
                                     init_fun_name
                                     false
                                     position);
            []
        )
*)
(*        | Sexpr.CONSTANT_MODEL_FIELD (struct_name, init_fun_name, term) ->*)
(*          assert false (* temporary do not process *)                     *)
(*
 (
            let field_name = (Scommon.get_name_of_term term) in
            let field_type = (Scommon.get_type_of_term term) in
            (make_type_inv_for_field struct_name 
                                     field_name
                                     term
                                     init_fun_name
                                     true);
            (make_func_contract_for_const_field struct_name 
                                                term
                                                field_type 
                                                init_fun_name) ;
            add_modelfields struct_name 
                            [(field_name ^ "_compare_constant")];
            [(field_name ^ "_compare_constant"), field_type]
        )
*)
    )
    | Sexpr.SPEC_FUN t -> (
            Sexpr.current_fun_contract := t; [])
    | _ ->
        raise (Sexpr.Specs_error ((Sexpr.string_of_lspecs spec_data)
                                   ^ " unsupported in model field declaration",
		                  None))
)

(******************* data invariant*********************)

                                            (* global invariant *)

(* name of all global invariant function *)
let global_invariant_names = ref []
let global_invariant_stmts = ref []
let type_invariant_names = ref []
let type_invariant_stmts = Hashtbl.create 2

let is_invariant (header : Cexpr.declaration_t) : bool = (
    let func_name_op = Ccommon.get_function_name header in
    match func_name_op with
    | Some func_name -> (
        (List.mem func_name (!global_invariant_names @ !type_invariant_names));)
    | None -> false
)

(* make ghost checking global invariant variables*)
let make_check_data_invariant_var (d_name : string)
: unit = (
    let v_name = ("checking_"^d_name) in
    let d_var = Cexpr.VREF(Cexpr.INTEGER, v_name, 1, 0) in
    let inits = (
        [Cexpr.SETF(Cexpr.INTEGER, 
                    Cexpr.SET, 
                    d_var, 
                    Cexpr.CONST_I(Cexpr.INTEGER, Num.Int(0), "0"))]
    )in
    let decl = {
                   Cexpr.decl_id = (v_name, 1);
                   Cexpr.decl_tn = Cexpr.INTEGER ;
                   Cexpr.decl_scls = Cexpr.NILSCLS;
                   Cexpr.decl_inits = inits ;
                   Cexpr.decl_params = [] ;
                   Cexpr.decl_struct_defined_p = false ;
                   Cexpr.decl_specs = None;
                   Cexpr.decl_location = None;
    }in
    global_invariant_stmts := (Cexpr.DEFVAR(decl)) :: (!global_invariant_stmts)
)
(* make a call of global invariant *)
let make_call_global_invariant (name_fun : string) = ( 
   Cexpr.STMT(
         Cexpr.CALL(
                    Cexpr.VOID_TYPE,
                    Cexpr.OPER (
                               Cexpr.POINTER_TO ( 
                                     Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE, None)
                               ),
                               Cexpr.CAST(false),
                               Cexpr.VREF (
                                     Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE,None),
                                     name_fun,1,0
                               ),
                               Cexpr.NIL,
                               Cexpr.NIL
                    ),
                    [],
                    None  
         )  
   )
)
(* make a list call of global invariants *)
let make_calls_global_invariant (fun_names : string list) 
                                : Cexpr.stmt_t list = (
    List.map make_call_global_invariant fun_names
)

let rec instrument_strong_global_invariant_stmt 
                                       (stmt: Cexpr.stmt_t)
                                       : Cexpr.stmt_t list = (
    let calls_global_invariant = 
        (make_calls_global_invariant (!global_invariant_names))
    in
    match stmt with
    | Cexpr.DEFUN (header, body) -> (
        let func_name_op = Ccommon.get_function_name header in
        if (is_invariant header) then [stmt]
        else (
            let new_body = (
                match (instrument_strong_global_invariant_stmt body) with
                | [Cexpr.BEGIN(stmts)] -> Cexpr.BEGIN(stmts)
                | _ -> assert false
            ) in
            [Cexpr.DEFUN(header, new_body)])
    )
    | Cexpr.DEFVAR header -> (
        let _, level = header.Cexpr.decl_id in
        if (level == 1) then [stmt]
        else (calls_global_invariant @ [stmt])
    ) 
    | (Cexpr.STMT _) | (Cexpr.ASSUME _)
    | (Cexpr.ASGN _) | (Cexpr.RESET _)
    | (Cexpr.CALLASGN _) -> (calls_global_invariant @ [stmt])
    | Cexpr.BEGIN(stmts) -> (
        [Cexpr.BEGIN(instrument_strong_global_invariant_stmts stmts)]
    )
    | Cexpr.IF (ex, stmts1, stmts2) -> (
        [Cexpr.IF(ex,
                 (instrument_strong_global_invariant_stmts stmts1),
                 (instrument_strong_global_invariant_stmts stmts2))]
    )
    | Cexpr.SWITCH(ex, stmts) -> (
        [Cexpr.SWITCH(ex,
                     (instrument_strong_global_invariant_stmts stmts))]
    )
    | Cexpr.DOWHILE(boo, ex, stmts) -> (
        [Cexpr.DOWHILE(boo,
                      ex,
                      (instrument_strong_global_invariant_stmts stmts))]
    )
    | Cexpr.FOR (stmts1, ex1, ex2, stmts2) -> (
        [Cexpr.FOR((instrument_strong_global_invariant_stmts stmts1),
                   ex1, 
                   ex2,
                   (instrument_strong_global_invariant_stmts stmts2))]
    )
    | Cexpr.BREAK  -> (calls_global_invariant @ [stmt])
    | Cexpr.CONTINUE -> (calls_global_invariant @ [stmt])
    | Cexpr.LABEL _ -> [stmt]
    | Cexpr.GOTO  _ -> (calls_global_invariant @ [stmt])
    | Cexpr.RETURN  _ -> (calls_global_invariant @ [stmt])
    | Cexpr.ANNOTATION _ -> [stmt]
    | Cexpr.ASM  _ -> (calls_global_invariant @ [stmt])
    | Cexpr.NOTE _ -> [stmt]
    | Cexpr.SPEC _ -> [stmt]
    (*| _ -> assert false*)
    
)
and instrument_strong_global_invariant_stmts 
                                       (stmts: Cexpr.stmt_t list)
                                       : Cexpr.stmt_t list = (
    List.flatten(List.map instrument_strong_global_invariant_stmt stmts)
    
)
                                               (* type invariant *)
(* get_exprs_with_types_in_expr *)
let rec get_exprs_with_types_in_expr 
            (expr : Cexpr.expr_t) 
            : ((Cexpr.typenode_t * Cexpr.expr_t)list)= ( 
    match expr with
    | Cexpr.OPER (_, _, ex1, ex2, ex3) -> 
        (get_exprs_with_types_in_expr ex1)
        @ (get_exprs_with_types_in_expr ex2)
        @ (get_exprs_with_types_in_expr ex3)
    | Cexpr.SETF(_, _, _, ex) -> 
        (get_exprs_with_types_in_expr ex)
    | Cexpr.VREF(typ, _, _, _) -> 
        let _type = Ccommon.get_type_from_typenode typ in
        [( _type , expr)]
    | Cexpr.ADDROF (_, ex) -> 
        (get_exprs_with_types_in_expr ex)
    | Cexpr.PREF (typ, _, ex, ex1) -> 
        (get_exprs_with_types_in_expr ex)
        @ (get_exprs_with_types_in_expr ex1)
    | Cexpr.SREF (typ, _, ex, _) -> 
        (get_exprs_with_types_in_expr ex) 
    | Cexpr.CALL(_, _, exs, _) ->  
        (get_exprs_with_types exs)
    | _ -> []
)

(* get_list_type_and_expr_from_exprs *)
and get_exprs_with_types (exprs : Cexpr.expr_t list) 
                         : ((Cexpr.typenode_t * Cexpr.expr_t)list)= (
    let temp = List.map get_exprs_with_types_in_expr exprs in
    List.flatten temp
)

(* is typedef ?*) 
let find_key_type_from_hash (typ : Cexpr.typenode_t) = (
   let res = ref None in
   Hashtbl.iter(
      fun key value ->
        if (value == typ) then res := Some key
   )typedef_info_hashtbl;   
   !res)

(* get function type invariant name *)
let get_type_inv_fun_name_from_type 
        (typ : Cexpr.typenode_t)
        (hash_type_invariants : (Cexpr.typenode_t, string) Hashtbl.t)
        : string list = ( 
   let typ_nd = ( 
      match (find_key_type_from_hash typ) with
      | None ->  typ
      | Some name -> Cexpr.SPECIFR_TYPE ([name]))
   in
   if (Hashtbl.mem hash_type_invariants typ) then
      (Hashtbl.find_all hash_type_invariants typ)
   else  
      if (Hashtbl.mem hash_type_invariants typ_nd) then
         (Hashtbl.find_all hash_type_invariants typ_nd)
      else []
)

(* make a call of type invariant function *)
let make_call_type_invariant (list_type_param : Cexpr.typenode_t list) 
                             (name_fun : string) 
                             (list_param : Cexpr.expr_t list) = ( 
   Cexpr.STMT(
         Cexpr.CALL(
                    Cexpr.VOID_TYPE,
                    Cexpr.OPER (
                               Cexpr.POINTER_TO ( 
                                     Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE, 
                                                        Some(list_type_param))
                               ),
                               Cexpr.CAST(false),
                               Cexpr.VREF (
                                     Cexpr.FUNCTION_OF (Cexpr.VOID_TYPE,
                                                        Some(list_type_param)),
                                     name_fun,1,0
                               ),
                               Cexpr.NIL,
                               Cexpr.NIL
                    ),
                    list_param,
                    None  
         )  
   ))

(* make a call of type invariant function *)
let rec make_calls_type_invariant 
            (list_type_param : Cexpr.typenode_t list) 
            (name_funs : string list) 
            (list_param : Cexpr.expr_t list) = ( 
   match name_funs with
   | [] -> []
   | head :: tail -> 
        (make_call_type_invariant list_type_param head list_param)::
        (make_calls_type_invariant list_type_param tail list_param))

(* except by type invariant - get name*)
let process_except_function_type_invariant (fun_names : string list) 
                                           (fun_except : string list) = ( 
   let res = (
       List.map (
           fun fun_name -> 
               let list_except_function = 
                   try
                       Hashtbl.find_all  
                           hash_except_function_type_invariant
                           fun_name
                   with
                   | Not_found -> []
               in 
               let check_init_fun = 
                   Scommon.check_include_list_string 
                               fun_except 
                               list_except_function in
               if ((check_init_fun != true)) then
                   [fun_name]
               else []
       )fun_names) in
   List.flatten res)

let instrument_type_invariants 
        (exprs_with_types : tuple_type_expr list)  
        (fun_except : string list) 
        (hash_type_invariants : (Cexpr.typenode_t, string) Hashtbl.t):
        Cexpr.stmt_t list = (
   let res = (
       List.flatten (
           List.map(
               fun tuple -> 
                   let typ, expr1 = tuple in
                   let typ_temp, expr = Ccommon.change_vref_expr expr1 in
                   let fun_names = (get_type_inv_fun_name_from_type 
                                        typ 
                                        hash_type_invariants )
                   in
                   let fun_names_processed = 
                       (process_except_function_type_invariant 
                            fun_names fun_except ) in
                   let call_types = (
                       match fun_names_processed with
                       | [] -> []
                       | _ -> 
                            (make_calls_type_invariant []
                                                       fun_names_processed
                                                       [expr]))
                   in
                   call_types
           )exprs_with_types))
   in
   let temps = ref [] in
   List.flatten (
       List.map (
           fun t -> 
               if (List.mem t !temps) = false then (
                  temps := (t)::(!temps);
                  [t]
               )
               else []
       )res))

let rec instrument_type_invariant_stmt 
        (stmt : Cexpr.stmt_t)
        (hash_type_invariants : (Cexpr.typenode_t, string) Hashtbl.t)
        : Cexpr.stmt_t list = (
    match stmt with
    | Cexpr.DEFUN (header, body) -> 
        let func_name_op = Ccommon.get_function_name header in
        if (is_invariant header) then [stmt]
        else (
            let new_body = (
                match (instrument_type_invariant_stmt body 
                                                      hash_type_invariants) with
                | [Cexpr.BEGIN(stmts)] -> Cexpr.BEGIN(stmts)
                | _ -> assert false
            ) in
            [Cexpr.DEFUN(header, new_body)])
    | Cexpr.DEFVAR header -> (
        let inits = header.Cexpr.decl_inits in
        match inits with
        | [] -> [stmt]
        | _ -> (
            let exprs_types = get_exprs_with_types inits in
            let calls = (instrument_type_invariants exprs_types [] 
                                                    hash_type_invariants) in
            calls @ [stmt]
        )
    ) 
    | Cexpr.STMT expr -> (
        let exprs_types = get_exprs_with_types_in_expr expr in
        let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
        (calls @ [stmt]))
    | Cexpr.ASGN (_, expr) -> (
        let exprs_types = get_exprs_with_types_in_expr expr in
        let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
        (calls @ [stmt]))
    | Cexpr.CALLASGN _ -> [stmt]
    | Cexpr.BEGIN(stmts) -> (
        [Cexpr.BEGIN(instrument_type_invariant_stmts stmts 
                                                     hash_type_invariants)]
    )
    | Cexpr.IF (ex, stmts1, stmts2) -> (
        let exprs_types = get_exprs_with_types_in_expr ex in
        let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
        (calls @
        [Cexpr.IF(ex,
                 (instrument_type_invariant_stmts stmts1 
                                                     hash_type_invariants),
                 (instrument_type_invariant_stmts stmts2 
                                                     hash_type_invariants))])
    )
    | Cexpr.SWITCH(ex, stmts) -> (
        let exprs_types = get_exprs_with_types_in_expr ex in
        let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
        (calls @
        [Cexpr.SWITCH(ex,
                     (instrument_type_invariant_stmts stmts 
                                                     hash_type_invariants))])
    )
    | Cexpr.DOWHILE(boo, ex, stmts) -> (
        let exprs_types = get_exprs_with_types_in_expr ex in
        let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
        (calls @
        [Cexpr.DOWHILE(boo,
                      ex,
                      (instrument_type_invariant_stmts stmts 
                                                     hash_type_invariants))])
    )
    | Cexpr.FOR (stmts1, ex1, ex2, stmts2) -> (
        let exprs_types = get_exprs_with_types [ex1;ex2] in
        let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
        (calls @
        [Cexpr.FOR((instrument_type_invariant_stmts stmts1 
                                                     hash_type_invariants),
                   ex1, 
                   ex2,
                   (instrument_type_invariant_stmts stmts2 
                                                     hash_type_invariants))])
    )
    | (Cexpr.BREAK)   
    | (Cexpr.CONTINUE)  
    | (Cexpr.LABEL _) | (Cexpr.ASSUME _) | (Cexpr.RESET _)
    | (Cexpr.GOTO  _) -> ([stmt])
    | Cexpr.RETURN  ex_o -> (
        match ex_o with
        | None ->  [stmt]
        | Some ex -> (
            let exprs_types = get_exprs_with_types_in_expr ex in
            let calls = (instrument_type_invariants exprs_types [] 
                                                hash_type_invariants) in
            (calls @ [stmt])
        )
    )
    | Cexpr.ANNOTATION _ -> [stmt]
    | Cexpr.ASM  _ -> ([stmt])
    | Cexpr.NOTE _ -> [stmt]
    | Cexpr.SPEC _ -> [stmt]
    (*| _ -> assert false*)
    
)

and instrument_type_invariant_stmts 
        (stmts : Cexpr.stmt_t list)
        (hash_type_invariants : (Cexpr.typenode_t, string) Hashtbl.t)
        : Cexpr.stmt_t list = (
    List.flatten(
        List.map(
            fun stmt -> 
                (instrument_type_invariant_stmt stmt 
                                                hash_type_invariants)
        )stmts)
    
)
(* data invariant *)
let instrument_data_invariant 
        (decls : Cexpr.stmt_t list)
        (hash_type_invariants : (Cexpr.typenode_t, string) Hashtbl.t)
        : Cexpr.stmt_t list = (
    let defs = (instrument_strong_global_invariant_stmts decls) in
    (instrument_type_invariant_stmts defs hash_type_invariants)
)


(******************** instrument function contracts ***************************)

(* Inserts stmts for checking pre-conditions at 
   beginning of function *)
let insert_pre_stmts (stmts_source : Cexpr.stmt_t list)
                     (stmts_insert : Cexpr.stmt_t list) 
                     : Cexpr.stmt_t list = (begin
    (stmts_insert @ stmts_source)
end)

(* Inserts stmts for checking post-conditions at 
   end of function *)
let insert_post_stmts (stmts_source : Cexpr.stmt_t list)
                      (stmts_insert : Cexpr.stmt_t list) 
                      : Cexpr.stmt_t list = (begin
    let len = 
        List.length stmts_source in
    let head, tail = 
        Ccommon.split_stmts stmts_source (len-1) 0 [] [] in
    head @ stmts_insert @ tail
end)

let stmts_of_triple (triple : Stranslate.insert_triple) 
                     : Cexpr.stmt_t list = (
    let (_, _, stmts) = triple in
    stmts
)
let rec stmts_of_triples (triples : Stranslate.insert_triple list) 
                     : Cexpr.stmt_t list = (
    let temp = (List.map stmts_of_triple triples) 
    in
    (List.flatten temp)
)
(*  *)
let instrument_function_contracts_stmt
        (triples : Stranslate.insert_triple list)
        (stmt : Cexpr.stmt_t)
        : Cexpr.stmt_t  = (
    let stmts = (
        match stmt with
        | Cexpr.BEGIN(stmts) -> stmts
        | _ -> assert false
    ) in 
    let pre_triples = 
        List.filter (fun triple ->
                        match triple with
                        | (_, Stranslate.PRE_FUN, _) -> true
                        | _ -> false
                    ) triples
    in
    let pre_triples = 
        pre_triples @  
        (List.filter (fun triple -> 
                        match triple with
                        | (_, Stranslate.PRE_FUN_COND, _) -> true
                        | _ -> false
                     ) triples)
    in
    let pre_triples = 
        pre_triples @ 
        (List.filter (fun triple -> 
                        match triple with
                        | (_, Stranslate.PRE_FUN_BEHAV, _) -> true
                        | _ -> false
                     ) triples)
    in
    let post_triples = 
        (List.filter (fun triple -> 
                        match triple with
                        | (_, Stranslate.POST_FUN_BEHAV, _) -> true
                        | _ -> false
                     ) triples)
    in
    let post_triples = 
        post_triples @ 
        (List.filter (fun triple -> 
                        match triple with
                        | (_, Stranslate.POST_FUN, _) -> true
                        | _ -> false
                     ) triples)
    in
    let post_triples = 
        post_triples @ 
        (List.filter (fun triple ->
                        match triple with
                        | (_, Stranslate.POST_RESULT, _) -> true
                        | _ -> false
                     ) triples)
    in
    let stmts_pre = (stmts_of_triples pre_triples) in
    let stmts_post = (stmts_of_triples post_triples) in
    let res = (insert_pre_stmts stmts stmts_pre) in
    let res = (insert_post_stmts res stmts_post) in
    Cexpr.BEGIN(res)
)

let collect_vars_from_scope_stack
    (scope_stacks : (string, Cexpr.binding_t) Hashtbl.t *
                   (string, Cexpr.binding_t) Hashtbl.t)
: (string, Cexpr.expr_t) Hashtbl.t = (
    let collect_vars (scope_stack : (string, Cexpr.binding_t) Hashtbl.t)
                     (vars : (string, Cexpr.expr_t) Hashtbl.t)
    : unit = (
        Hashtbl.iter (fun str binding ->
            let (id, lev) = binding.Cexpr.binding_name in
            let typ = binding.Cexpr.binding_tn in
            let expr = Cexpr.VREF (typ, id, lev, 0) in
            Hashtbl.replace vars id expr
        ) scope_stack
    ) in
    let vars = Hashtbl.create 2 in
    collect_vars (fst scope_stacks) vars;
    (* temporarily left out tag_scope_stack as it contains
       information about structs, unions, and enums *)
(*     collect_vars (snd scope_stacks) vars; *)
    vars
)

let check_curr_component (s : Cexpr.declarator_t list)
                         (specifiers : Cexpr.specifier_t)
                         (scope_stacks) : unit = (
   (match s with
    | [a] -> (
        let _ , _, param_names, _ = a in
        let name, ty = Cexpr.get_structure_components_in_declaration
                          (Cexpr.combine_declaration true specifiers a) in
        Ccommon.current_components := (name, ty);
        Ccommon.current_list_components := (!Ccommon.current_components)::
                                           (!Ccommon.current_list_components);
        if ((Ccommon.is_pointer ty) == true) then (begin
            if ((Ccommon.is_function_type ty) == true) then (begin
                if (!Sexpr.current_fun_contract != Sexpr.empty_fun_contract) then (begin
                    let local_vars = collect_vars_from_scope_stack scope_stacks in
                    let params =
                        List.fold_left (fun params_ param_name ->
                            try params_ @ [Hashtbl.find local_vars param_name]
                            with Not_found -> params_
                        ) [] param_names in
                    let translated_current_fun_contract =
                        Stranslate.cfunccontract_of_lfunccontract
                            !Sexpr.current_fun_contract local_vars in
                    Hashtbl.add Ccommon.contract_of_field_hashtbl
                                (!Sexpr.current_name_struct, name)
                                (translated_current_fun_contract, params)
                end)
            end)
        end)
        )
    | _ -> ());
    (Sexpr.current_fun_contract := Sexpr.empty_fun_contract)
)

let get_ghost_stmts (spec : Sexpr.lspec_t) (fun_name : string) = (
    match spec with
    | Sexpr.SPEC_FUN {
                       Sexpr.fspec_conditions = _;
                       Sexpr.fspec_behaviors = _;
                       Sexpr.fspec_assigns = _;
                       Sexpr.fspec_prohibits = prohibits;
                       Sexpr.fspec_local_invs = _;
                       Sexpr.fspec_context = _;
                       Sexpr.fspec_is_handler = _
                     } -> (
         (List.map (
             fun name ->
                 let name_var = "dont_call_"^name^"_in_"^fun_name in
                 let expr = Cexpr.VREF(
                                Cexpr.INTEGER,
                                name_var, 1, 0) in
                 Cexpr.STMT(
                     Cexpr.SETF(
                         Cexpr.INTEGER,
                         Cexpr.INCR_POST,
                         expr,
                         Cexpr.NIL))
         ) prohibits),
         (List.map (
             fun name ->
                 let name_var = "dont_call_"^name^"_in_"^fun_name in
                 let expr = Cexpr.VREF(
                                Cexpr.INTEGER,
                                name_var, 1, 0) in
                 Cexpr.STMT(
                     Cexpr.SETF(
                         Cexpr.INTEGER,
                         Cexpr.DECR_POST,
                         expr,
                         Cexpr.NIL))
         )prohibits)
     )
    | _ -> ([],[])
)

(* *)
let instrument_function_contracts (defun : Cexpr.stmt_t) 
                                  (triples : Stranslate.insert_triple list)
                                  : Cexpr.stmt_t = ( 
    match defun with
    | Cexpr.DEFUN (fun_header, body) ->
        let new_body = 
            (instrument_function_contracts_stmt triples body)
        in
        Cexpr.DEFUN (fun_header, new_body)
    | _ -> defun
)

(******************** insert statement annotation ***************************)
(*  *)
let rec make_variable_for_statement_contracts 
                      (name_behavs : string list) : Cexpr.expr_t  = (   
   match name_behavs with 
   | [name] -> Cexpr.VREF(Cexpr.INTEGER, "spec_var_"^name^"_temp",3,0)
   | head :: tail -> Cexpr.OPER (
                           Cexpr.INTEGER,
                           Cexpr.OROR, 
                           Cexpr.VREF(
                                 Cexpr.INTEGER,
                                 "spec_var_"^head^"_temp",3,0), 
                           make_variable_for_statement_contracts tail,
                           Cexpr.NIL)
   | [] -> Cexpr.NIL)


(* *)
let make_insert_stmts_for_jump_statement (behavs : string list) 
                                         (stmts : Cexpr.stmt_t list) 
                                         : Cexpr.stmt_t list = (
   let expr_cond = make_variable_for_statement_contracts behavs in
   let res = (
     match expr_cond with
     | Cexpr.NIL -> stmts
     | Cexpr.VREF _ 
     | Cexpr.OPER _  -> [Cexpr.IF(expr_cond, [Cexpr.BEGIN(stmts)], [])]
     | _ -> assert false
   ) 
   in
   res
)

(* *)
let make_insert_stmts_for_stmt_contracts (behavs : string list) 
                                         (lexpr : Sexpr.lexpr) 
                                         local_scope
                                         : Cexpr.stmt_t list = (
   let expr_cond = (make_variable_for_statement_contracts behavs) in
   let range = (Sexpr.make_range  (fst lexpr.Sexpr.lexpr_loc)
                                  (snd lexpr.Sexpr.lexpr_loc))
   in
   let range_note = Stranslate.make_range_note range in
   let stmts = 
       range_note
       :: (Stranslate.make_list_assert 
              [Stranslate.translate_lexpr lexpr local_scope])
   in
   let res = (
     match expr_cond with
     | Cexpr.NIL -> stmts
     | Cexpr.VREF _ 
     | Cexpr.OPER _  -> [Cexpr.IF(expr_cond, [Cexpr.BEGIN(stmts)], [])]
     | _ -> assert false
   ) in
   res)

(* assertion *)
let insert_assert_spec (ids : Sexpr.identifier_t list)
                       (lexpr : Sexpr.lexpr)
                       local_scope 
                       : Cexpr.stmt_t list = ( 
    (make_insert_stmts_for_stmt_contracts ids lexpr 
                                          local_scope))

(* loop invariant *)
let insert_loop_spec (insert_stmts : Cexpr.stmt_t list) 
                     (stmt : Cexpr.stmt_t) 
                     : Cexpr.stmt_t list = (
   let res = (
      match stmt with
      | Cexpr.FOR (stmts1, expr1, expr2, stmts2) ->(
           [Cexpr.FOR (stmts1, expr1, expr2, 
                      (insert_stmts @ 
                       stmts2))]
       )    
     | Cexpr.DOWHILE (boo, expr, stmts) ->(
           (* while loops evaluate guard expressions at the beginning
	      whereas do-while loops evaluate guard exprs at the end *)
           if boo then
               [Cexpr.DOWHILE(boo, 
                              expr, 
                              (insert_stmts @
                               stmts))]
	   else
	       [Cexpr.DOWHILE(boo, 
                              expr, 
                              (stmts @ 
                               insert_stmts))]
      )    
     | _ -> assert false 
   ) in
   res)
let rec insert_loop_specs (insert_stmts : Cexpr.stmt_t list) 
                          (stmts : Cexpr.stmt_t list) 
                          : Cexpr.stmt_t list= (
    match stmts with
    | [] -> []
    | head::tail -> (insert_loop_spec insert_stmts head) @
                    (insert_loop_specs insert_stmts tail)
)
(*  *)

let rec find_statement (stmts_insert : Cexpr.stmt_t list) 
                       (curr_stmt : Cexpr.stmt_t)
                       (sample : string) : Cexpr.stmt_t list= (
    let return_label = Ccommon.return_label in
    if return_label = "" then
        (* Wrong return label *)
	assert false;
    match curr_stmt, sample with 
    | Cexpr.BEGIN(stmts_begin), _ ->
                let temp_stmt = 
                    Cexpr.BEGIN(find_statements stmts_insert 
                                                stmts_begin 
                                                sample) in
                [temp_stmt]             
        | (Cexpr.IF(expr, stmts_t, stmts_f)), _ ->           
                let find_stmt_t = find_statements 
                                                stmts_insert
                                                stmts_t 
                                                sample in
                let find_stmt_f = find_statements 
                                                stmts_insert
                                                stmts_f
                                                sample in
                [Cexpr.IF(expr, find_stmt_t, find_stmt_f)]
   | Cexpr.BREAK , "break" -> (stmts_insert @ [curr_stmt])
   | Cexpr.CONTINUE , "continue"-> (stmts_insert @ [curr_stmt])
   | Cexpr.GOTO(return_label) , "return" -> (stmts_insert @ [curr_stmt])
   | _ -> [curr_stmt]
)
and find_statements (stmts_insert : Cexpr.stmt_t list)
                    (stmts : Cexpr.stmt_t list)
                    (sample: string) : Cexpr.stmt_t list = ( 
    match stmts with
    | [] -> []
    | head::tail -> (find_statement stmts_insert head sample) @
                    (find_statements stmts_insert tail sample)
)

let insert_stmt_jump_return 
            (stmts_return : Cexpr.stmt_t list)
            (stmts : Cexpr.stmt_t list) = (
    (find_statements stmts_return
                     stmts
                     "return")
                     
)                
let insert_stmt_jump_continue 
            (stmts_continue : Cexpr.stmt_t list)
            (stmts : Cexpr.stmt_t list) = (
    (find_statements stmts_continue
                     stmts
                     "continue")
                     
)                
let insert_stmt_jump_break 
            (stmts_break : Cexpr.stmt_t list)
            (stmts : Cexpr.stmt_t list) = (
    (find_statements stmts_break
                     stmts
                     "break")
                     
)                
(* break/continue/return contracts *)
let insert_stmt_jump (stmts_break : Cexpr.stmt_t list) 
                     (stmts_continue : Cexpr.stmt_t list) 
                     (stmts_return : Cexpr.stmt_t list) 
                     (stmts : Cexpr.stmt_t list) 
                      : Cexpr.stmt_t list = ( 
    (insert_stmt_jump_break 
        stmts_break
        (insert_stmt_jump_continue 
            stmts_continue
            (insert_stmt_jump_return 
                stmts_return
                stmts)))
)

let rec flatten_stmts list_temp res = (
    match list_temp with
    | [] -> res
    | head::tail -> (
        let (s1,s2,s3,s4) = head in
        let (t1,t2,t3,t4) = res in
        let temp = (s1@t1, s2@t2, s3@t3, s4@t4) in
        flatten_stmts tail temp
     )
)

let spec_instrument_annos0 
         (stmt_loops : Sexpr.lloop_statement_t list)
         (local_scope : (string, Cexpr.expr_t) Hashtbl.t)
         : Cexpr.stmt_t list * 
           Cexpr.stmt_t list *
           Cexpr.stmt_t list *
           Cexpr.stmt_t list = ( 
(* looop_invariants * breaks * continues * returns *)
   let list_triple = 
           (Stranslate.translate_stmt_loops stmt_loops local_scope) in
   let temp_list = (
       List.map ( 
           fun triple ->
               let (behavs, position, stmts) = triple in
               let stmts = make_insert_stmts_for_jump_statement 
                                                behavs stmts in
               match position with
               | Stranslate.PRE_LOOP -> (
                   (stmts, [], [], [])            
                )
               | Stranslate.POS_BREAK -> (
                   ([], stmts, [], [])            
                )
               | Stranslate.POS_CONTINUE -> (
                   ([], [], stmts, [])            
                )
               | Stranslate.POS_RETURN -> (
                   ([], [], [], stmts)            
                )
               | _ ->
                   raise (Sexpr.Specs_error("Wrong specification !!!", None))
      ) list_triple) in
    flatten_stmts temp_list ([],[],[],[])

)

let spec_instrument_annos stmts_insert 
                          (stmt : Cexpr.stmt_t) 
                          : Cexpr.stmt_t list = (
    let (stmts_loop, stmts_break, stmts_continue, stmts_return) =
        stmts_insert in
    let stmts = (
        match stmt with
        | Cexpr.FOR(stmts1, ex1, ex2, stmts2) -> (
             [Cexpr.FOR(stmts1, ex1, ex2, 
                        (insert_stmt_jump stmts_break 
                                          stmts_continue 
                                          stmts_return 
                                          stmts2))]
         )
        | Cexpr.DOWHILE(boo, ex1, stmts1) -> (
              [Cexpr.DOWHILE(boo, ex1, 
                             (insert_stmt_jump stmts_break 
                                               stmts_continue 
                                               stmts_return 
                                               stmts1))]
         )
        | _ -> assert false
       
    ) in
    (insert_loop_specs stmts_loop stmts)
)


let add_loop_specs (specs : Sexpr.lloop_statement_t list)
                   (stmts : Cexpr.stmt_t list)
                   (loop_scope_stack : (string, Cexpr.binding_t) Hashtbl.t *
                                       (string, Cexpr.binding_t) Hashtbl.t)
: (Cexpr.stmt_t list) * (Cexpr.stmt_t list) = (
    let rec find_loop_stmt (stmts : Cexpr.stmt_t list) = (
        match stmts with
        | [] -> raise (Sexpr.Specs_error ("Wrong specification!", None))
        | head :: tails -> (begin
            match head with
            | Cexpr.FOR _
            | Cexpr.BEGIN _
            | Cexpr.DOWHILE _
            | Cexpr.SWITCH _   -> 
                (head, tails)
            | Cexpr.NOTE _ -> (
                find_loop_stmt tails
              )               
            | _ -> (assert false)
          end)
    ) in
    let vars = collect_vars_from_scope_stack loop_scope_stack in
    let loop_stmt, remain_stmts = find_loop_stmt stmts in 
    let list_stmt = (spec_instrument_annos0 specs vars) in
    let stmts = (spec_instrument_annos list_stmt loop_stmt) in
    (stmts, remain_stmts)
)

let translate_stmt_contract contract local_scope =
    match contract with
    | Sexpr.STMT_RESET reset ->
        (Stranslate.translate_stmt_reset reset local_scope)
    | Sexpr.STMT_ASSUME assumes ->
        (Stranslate.translate_stmt_assume assumes local_scope)
    | Sexpr.STMT_ASSERT (ids, expr) ->
        (insert_assert_spec ids expr local_scope)
    | Sexpr.STMT_INIT strs -> (
        List.map (
            fun str ->
                let var = Cexpr.VREF (Cexpr.INTEGER, ("checking_" ^ str), 1, 0)
                in
                Cexpr.STMT (Cexpr.SETF (Cexpr.INTEGER,
                                        Cexpr.SET,
                                        var,
                                        Cexpr.CONST_I (Cexpr.INTEGER,
                                                       Num.Int(1),
                                                       "1"))))
            strs)
    | Sexpr.STMT_RELEASE strs -> (
        List.map (
            fun str ->
                let var = Cexpr.VREF (Cexpr.INTEGER, ("checking_" ^ str), 1, 0)
                in
                Cexpr.STMT (Cexpr.SETF (Cexpr.INTEGER,
                                        Cexpr.SET,
                                        var,
                                        Cexpr.CONST_I (Cexpr.INTEGER,
                                                       Num.Int(0),
                                                       "0"))))
            strs)

let rec instrument_invariant (var_spec : Sexpr.lspec_t)
                             (stmt_dvar : Cexpr.stmt_t)
                             (vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.stmt_t list =
begin
    match stmt_dvar with
    | Cexpr.DEFVAR header -> (
        let tyn = header.Cexpr.decl_tn in
        if (Ccommon.is_function_type tyn) then (
            let pos_op = Sexpr.get_first_loc_var_spec var_spec in
            raise (Sexpr.Specs_error
                  ("Variable spec without a variable", pos_op)));
        (* Add new variable into local_scope *)
        let (id, lev) = header.Cexpr.decl_id in
        let new_expr = Cexpr.VREF (tyn, id, lev, 0) in
        Hashtbl.replace vars id new_expr;

        let new_specs = Stranslate.cspec_of_lspec var_spec vars in
        let new_header = { header with Cexpr.decl_specs = Some new_specs } in
        let new_decl = Cexpr.DEFVAR new_header in
        [new_decl]
    )
    | _ -> assert false (* DODO process for multiple $reader and $writer *)
end

let rec insert_var_spec (var_spec : Sexpr.lspec_t)
                        (stmts : Cexpr.stmt_t list)
                        (vars : (string, Cexpr.expr_t) Hashtbl.t)
: ((Cexpr.stmt_t list) * (Cexpr.stmt_t list)) =
begin
    match stmts with
    | [] -> (
        let pos_op = Sexpr.get_first_loc_var_spec var_spec in
        raise (Sexpr.Specs_error
                  ("Variable spec without a variable", pos_op)))
    | head :: tails -> (
        match head with
        | Cexpr.DEFVAR header -> (
            let processed_stmts =
                instrument_invariant var_spec head vars in
            (processed_stmts, tails)
        )
        | Cexpr.NOTE _ -> (
            let processed_stmts, remain_stmts =
                insert_var_spec var_spec tails vars in
            ([head] @ processed_stmts, remain_stmts)
        )
        | _ -> (
            let pos_op = Sexpr.get_first_loc_var_spec var_spec in
            raise (Sexpr.Specs_error
                      ("Variable spec without a variable", pos_op)))
               
    )
end

let rec spec_instrument (stmts : Cexpr.stmt_t list)
                    (loop_specs : Sexpr.lloop_statement_t list ref)
                    (local_ghost_scope : (string, Cexpr.expr_t) Hashtbl.t)
                    (modelfields : (string, string list) Hashtbl.t)
: Cexpr.stmt_t list = (
    match stmts with
    | [] -> (
        assert (!loop_specs = []);
        ([]))
    | head :: tails -> (
        let stmts1, loop_specs1, remain_stmts1 = (
            match head with
            | Cexpr.SPEC (Sexpr.SPEC_STMT stmt_contracts) -> (
                assert (!loop_specs = []);
                let scope_stack = Ccommon.pop_stmt_spec_scope () in
                let local_scope = collect_vars_from_scope_stack scope_stack in
                let translated_stmts =
                    List.fold_left 
                        (fun stmts s_contract ->
                             stmts @ (translate_stmt_contract s_contract
                                                              local_scope))
                        [] stmt_contracts
                in
                (translated_stmts, [], tails))
            | Cexpr.SPEC (Sexpr.STMT_LOOP stmt_loops) ->
                ([], !loop_specs @ stmt_loops, tails)
            | Cexpr.SPEC Sexpr.SPEC_VAR var_spec -> (
                assert (!loop_specs = []);
                let scope_stack = Ccommon.pop_stmt_spec_scope () in
                let local_scope = collect_vars_from_scope_stack scope_stack in
                let proccessed_stmts, remain_stmts =
                    insert_var_spec (Sexpr.SPEC_VAR var_spec) tails local_scope
                in
                (proccessed_stmts, [], remain_stmts)
            )
            | Cexpr.SPEC Sexpr.SPEC_FUN _
            | Cexpr.SPEC Sexpr.SPEC_GLOBAL_INV _
            | Cexpr.SPEC Sexpr.SPEC_TYPE_INV _
            | Cexpr.SPEC Sexpr.SPEC_MODEL_FIELD _
            | Cexpr.SPEC Sexpr.SPEC_ENUM _
            | Cexpr.SPEC Sexpr.SPEC_ANYTHING -> (
                raise (Sexpr.Specs_error
                    ("Global specifications are not allowed inside function bodies.",
                     None))
              )
            | Cexpr.NOTE _ -> (
                ([head], !loop_specs, tails)
              )
            | _ -> (
                if (!loop_specs <> []) then (
                    let scope_stack = Ccommon.pop_loop_scope () in
                    let procced_stmts, remain_stmts =
                        add_loop_specs !loop_specs stmts scope_stack in
                    (procced_stmts, [], remain_stmts))
                else ([head], [], tails)
              )
          ) in
          loop_specs := loop_specs1;
          stmts1 @ (spec_instrument remain_stmts1
                                    loop_specs
                                    local_ghost_scope
                                    modelfields)
      )
)

(******************** specification process ***************************)

let import_info_invariant (spec : Sexpr.lspec_t) : unit = (
    match spec with
    |Sexpr.SPEC_TYPE_INV (type_inv_spec) ->(begin
        let {
              Sexpr.tinv_strength = _;
              Sexpr.tinv_name = id1;
              Sexpr.tinv_type = typ;
              Sexpr.tinv_param_name = id2;
              Sexpr.tinv_expr = preds;
              Sexpr.tinv_exceptions = except
            } = type_inv_spec in
        (match except with
        | None -> ()
        | Some(excepts) -> (
             let list_except_type = excepts in
             List.iter(
                 fun except_type -> 
                     match except_type with
                     | Sexpr.EXCEPT_WHOLE str 
                     | Sexpr.EXCEPT_HEAD str 
                     | Sexpr.EXCEPT_TAIL str -> 
                         Hashtbl.add  
                            hash_except_function_type_invariant id1 str
             )list_except_type;))       
     end)
    | Sexpr.SPEC_GLOBAL_INV(global_inv_spec) -> ()
    | _ -> assert false)
let import_ghost_var (spec : Sexpr.lspec_t)
                     (local_scope : (string, Cexpr.expr_t) Hashtbl.t)
                     : unit = (
    match spec with
    | Sexpr.SPEC_GLOBAL_INV _ -> ()
    | Sexpr.SPEC_TYPE_INV {
                            Sexpr.tinv_strength = _;
                            Sexpr.tinv_name = _;
                            Sexpr.tinv_type = typ;
                            Sexpr.tinv_param_name = id;
                            Sexpr.tinv_expr = _;
                            Sexpr.tinv_exceptions = _
                          } -> (
          let ctyp = Stranslate.translate_logictype typ in
          let new_type =  find_true_type ctyp in
          let vref = Cexpr.VREF(new_type, id, 2, 0) in
          Hashtbl.replace local_scope id vref
     )
    | _ -> ()
)

let instrument_global_spec 
        (spec : Sexpr.lspec_t)
        (global_vars : (string, Cexpr.expr_t) Hashtbl.t)
        (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
        (modelfields :(string, Sexpr.identifier_t list) Hashtbl.t) = (begin
    let triples = (
        match spec with
        | Sexpr.SPEC_GLOBAL_INV _ 
        | Sexpr.SPEC_TYPE_INV _  -> 
             import_info_invariant spec;
             let local_scope = Hashtbl.copy global_vars in
             let local_ghost_scope = Hashtbl.copy global_ghost_vars in
             import_ghost_var spec local_ghost_scope;
            (Stranslate.translate spec local_scope local_ghost_scope 
                                  1 "" None false)
(*        | Sexpr.SPEC_GHOST _ -> (begin                      *)
(*             (Stranslate.translate spec global_vars         *)
(*                                   global_ghost_vars 1 ""   *)
(*                                   (Some modelfields) false)*)
(*         end)                                               *)
        | _ -> []
    )in
    stmts_of_triples triples
end)
(* If fun_name is a struct's initializer, we insert additional assignments 
   which set field __initialized__, and save initialized values of
   declared constant fields*)
let get_post_stmts_of_initializer (fun_name : string)  
                                 (fun_header : Cexpr.declaration_t) = (
    let post_stmts1 = (begin
    try 
        let s_t_fs = 
            Hashtbl.find_all const_field_of_initializer_hashtbl fun_name in
        List.map(
            fun s_t_f -> 
                let struct_name, lexpr, field_type = s_t_f in
            
                (* get name and type of the relevant parameter *)
                let param_name, param_type = 
                    Ccommon.get_struct_param_info fun_header struct_name in
                let term_ref = (Sexpr.change_term_to_sref param_name lexpr)
                        (*Cexpr.SREF(field_type,
                                   true, 
                                   Cexpr.VREF(param_type, param_name, 2, 0),
                                   field_name) in*)
                in
                let var = Cexpr.VREF(param_type, param_name, 2, 0) in
                let vars = Hashtbl.create 10 in
                (Hashtbl.add vars param_name var);
                let field_ref = (Stranslate.translate_lexpr term_ref vars) in
                let field_name = (Sexpr.get_name_of_term lexpr) in
                Cexpr.STMT(
                    Cexpr.SETF (
                        field_type, Cexpr.SET,
                        (Ccommon.get_constant_field_of field_name
                                                       field_type 
                                                       field_ref),
                        field_ref))
              )s_t_fs
    with
    | Not_found -> []
    end) 
    in
    let post_stmts2 = (begin
        try
            let struct_name, no_contract = 
                Hashtbl.find contract_for_init_fun_hashtbl fun_name
            in
            let param_name, param_type = 
                Ccommon.get_struct_param_info fun_header struct_name in
            let init_field_ref = Cexpr.SREF(Cexpr.INTEGER, 
                                            true, 
                                            Cexpr.VREF(
                                                param_type, 
                                                param_name, 2, 0),
                                            "__initialized__") in
            [Cexpr.STMT(
                 Cexpr.SETF (
                     Cexpr.INTEGER, Cexpr.SET,
                     init_field_ref,
                     Cexpr.CONST_I (Cexpr.INTEGER, 
                                    Num.Int(1) , "1" )))]
        with
        | Not_found -> []
    end) 
    in
    (post_stmts1 @ post_stmts2)
)

let get_ghost_stmts (spec : Sexpr.lspec_t) (fun_name : string) = (
    match spec with
    | Sexpr.SPEC_FUN {
                       Sexpr.fspec_conditions = _;
                       Sexpr.fspec_behaviors = _;
                       Sexpr.fspec_assigns = _;
                       Sexpr.fspec_prohibits = prohibits;
                       Sexpr.fspec_local_invs = _;
                       Sexpr.fspec_context = _;
                       Sexpr.fspec_is_handler = _
                     } -> (
         (List.map (
             fun name ->
                 let name_var = "dont_call_"^name^"_in_"^fun_name in
                 let expr = Cexpr.VREF(
                                Cexpr.INTEGER,
                                name_var, 1, 0) in
                 Cexpr.STMT(
                     Cexpr.SETF(
                         Cexpr.INTEGER,
                         Cexpr.INCR_POST,
                         expr,
                         Cexpr.NIL))
         ) prohibits),
         (List.map (
             fun name ->
                 let name_var = "dont_call_"^name^"_in_"^fun_name in
                 let expr = Cexpr.VREF(
                                Cexpr.INTEGER,
                                name_var, 1, 0) in
                 Cexpr.STMT(
                     Cexpr.SETF(
                         Cexpr.INTEGER,
                         Cexpr.DECR_POST,
                         expr,
                         Cexpr.NIL))
         )prohibits)
     )
    | _ -> ([],[])
)

let func_contract_all_field : (string, Cexpr.declaration_t) Hashtbl.t =
    Hashtbl.create 10

(* instrument function *)
let instrument_function (header : Cexpr.declaration_t)
                        (body : Cexpr.stmt_t)
                        (function_contract : Sexpr.lspec_t)
                        (scope_stack : (string, Cexpr.binding_t) Hashtbl.t *
                                       (string, Cexpr.binding_t) Hashtbl.t)
                        (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
                        type_invariant_hashtbl
                        modelfields
                        : Cexpr.stmt_t = (begin
    let local_ghost_scope = Hashtbl.copy global_ghost_vars in
    let fun_name = match (Ccommon.get_function_name header) with
                   | Some name -> name
                   | None -> assert false
    in
    let new_fun = 
    try
        let scope = Hashtbl.copy local_ghost_scope in
        let local_scope = collect_vars_from_scope_stack scope_stack in
        Hashtbl.iter (fun key value -> Hashtbl.replace scope key value)
                      local_scope;
        let _ = match (Ccommon.make_return_variable header) with
                | Some vref ->
                    Hashtbl.add scope Ccommon.return_expr_name vref
               | None -> ()
        in
        let new_specs =
            Stranslate.cspec_of_lspec function_contract scope
        in
        let structs_get_field =
            Sexpr.struct_get_field Sexpr.struct_get_field_in_pred
                                   function_contract in
        let new_header =
            if (structs_get_field = []) then (
                match header with
                  {
                    Cexpr.decl_id = _;
                    Cexpr.decl_tn = _;
                    Cexpr.decl_scls = _;
                    Cexpr.decl_inits = _;
                    Cexpr.decl_params = _;
                    Cexpr.decl_struct_defined_p = _;
                    Cexpr.decl_specs = specsopt;
                  } ->
                    assert (specsopt = None);
                    { header with Cexpr.decl_specs = Some new_specs })
            else (
                let func_name_op = Ccommon.get_function_name header in
                (match func_name_op with
                | None -> ()
                | Some func_name ->
                    Hashtbl.add func_contract_all_field func_name
                                { header with Cexpr.decl_specs = Some new_specs });
                header)
        in
        let pre_ghost_stmts, post_ghost_stmts = 
                get_ghost_stmts function_contract fun_name in
        (* additional assignments in case of a struct's initializer *)
        let additional_asgns =  
            (get_post_stmts_of_initializer fun_name header) in
        let new_body = 
            match body with
            | Cexpr.BEGIN(stmts) -> (
                let new_stmts = insert_post_stmts stmts 
                                                  additional_asgns
                in
                let len = List.length new_stmts in
                let head, tail = 
                    Ccommon.split_stmts new_stmts (len-1) 0 [] [] in
                Cexpr.BEGIN (pre_ghost_stmts @ 
                             head @ 
                             post_ghost_stmts @ tail)
                )
            | _ -> 
                assert false
        in
        let new_fun = Cexpr.DEFUN (new_header, new_body) in
        if (!insert_func_contract_callee && (structs_get_field = [])) then (
            let _ =
                match (Ccommon.make_return_variable new_header) with
                | Some vref ->
                    Hashtbl.add local_scope Ccommon.return_expr_name vref
                | None ->
                    ()
            in
            let boo = (List.mem fun_name !fun_names_is_assigned) in
            let triples = 
                (Stranslate.translate function_contract local_scope 
                                      local_ghost_scope 1 fun_name
                                     (Some modelfields) boo) in
           (instrument_function_contracts new_fun triples ))
        else new_fun
    with 
    | Not_found -> Cexpr.DEFUN(header,body)
    in
    new_fun
end)

let add_enum_fun (header : Cexpr.declaration_t)
                 (cspec : Cexpr.cspec_t)
                 (scope : (string, Cexpr.expr_t) Hashtbl.t)
: unit = (
    match cspec with
    | Sexpr.SPEC_FUN func_spec -> (
        let f_name, level = header.Cexpr.decl_id in
        let tyn = header.Cexpr.decl_tn in
        let param_names = header.Cexpr.decl_params in
        let params =
            List.map (fun param_name ->
                try Hashtbl.find scope param_name
                with Not_found -> assert false
            ) param_names in
        if (Ccommon.is_pointer tyn) then
            Hashtbl.add Ccommon.contract_of_fun_pointer f_name (func_spec, params)
      )
    | _ -> assert false
)

let instrument_defvar (stmt : Cexpr.stmt_t)
                      (spec : Sexpr.lspec_t)
                      (scope_stack: (string, Cexpr.binding_t) Hashtbl.t *
                                    (string, Cexpr.binding_t) Hashtbl.t)
                      (implicit_type_invariants :
                          (string, Sexpr.ltype_invariant_t) Hashtbl.t)
                      (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.stmt_t list = (
    let modelfields = Hashtbl.create 2 in
    match stmt with
    | Cexpr.DEFVAR header as decl -> (begin         
        let local_scope = collect_vars_from_scope_stack scope_stack in
        let scope = Hashtbl.copy global_ghost_vars in
            Hashtbl.iter (fun key value -> Hashtbl.replace scope key value)
                local_scope;
        let _ = match (Ccommon.make_return_variable header) with
                | Some vref ->
                    Hashtbl.add scope Ccommon.return_expr_name vref
                | None -> ()
        in
        let new_specs = Stranslate.cspec_of_lspec spec scope in
        let header = { header with Cexpr.decl_specs = Some new_specs } in
        let decl = Cexpr.DEFVAR header in
        add_enum_fun header new_specs scope;
        match (Ccommon.get_struct_or_union_name header) with
        | Some name -> (begin
            let relevant_type_invs =
                try Hashtbl.find_all implicit_type_invariants name
                with 
                | Not_found -> []
            in
            let new_stmts = (begin
                List.fold_left (
                    fun stmts spec ->
                        stmts @ (instrument_global_spec 
                                    (Sexpr.SPEC_TYPE_INV spec)
                                    scope
                                    global_ghost_vars
                                    modelfields))
                        [] relevant_type_invs
            end)
            in 
            [decl] @ new_stmts
          end)
        | None -> 
            [decl]
      end)
    | _ -> assert false
)

let create_file_loc_from_note (note_op : Cexpr.note_t option)
: Sexpr.file_location_t =
begin
    let filename = 
        match note_op with
        | None -> ""
        | Some note -> (
            match note with
            | Cexpr.LINE (_, fn, _, _)
            | Cexpr.RANGE (fn, _, _, _,_) -> fn 
        ) in 
    Sexpr.make_file_location filename
end

(* find and insert function_contracts to body of function *)
let rec insert_function_contract1 (func_contract : Sexpr.lspec_t)
    (stmts : Cexpr.stmt_t list)
    (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
    type_invariant_hashtbl
    modelfields
    (note_op : Cexpr.note_t option) 
: Cexpr.stmt_t list * Cexpr.stmt_t list = (
    match stmts with
    | [] ->
        raise (Sexpr.Specs_error ("Function contract without funtion", None))
    | head :: tails -> (
        match head with
        | Cexpr.DEFVAR header -> (
            if (is_invariant header) then
                (raise (Sexpr.Specs_error (
                           "Function contract without function", None)));
            let scope_stack = Ccommon.pop_func_scope () in
            let new_stmts = (instrument_defvar head
                                 func_contract
                                 (scope_stack)
                                 (Hashtbl.create 2)
                                 (Hashtbl.create 2))
            in
            (new_stmts, tails)
          )
        | Cexpr.DEFUN (header, body) -> (
            let scope_stack = Ccommon.pop_func_scope () in
            let func_name = 
                match (Ccommon.get_function_name header) with
                | Some name -> name
                | None -> assert false
            in
            let _ =
                try
                    let fl = create_file_loc_from_note note_op in
                    let is_in_cutlst =
                        Sexpr.is_in_cutlist !Sexpr.cut_functions fl func_name 
                    in
                    if (not is_in_cutlst) then
                        Sexpr.check_validity_of_function_contract func_contract
                with Not_found ->
                    Sexpr.check_validity_of_function_contract func_contract
            in
            let func = instrument_function header
                            body
                            (func_contract)
                            (scope_stack)
                            (Hashtbl.create 1)
                            (Hashtbl.create 1)
                            (Hashtbl.copy struct_modelfields_hashtbl) in
             ([func], tails)
          ) 
        | Cexpr.NOTE a_note -> (
            let processed_stmts, remain_stmts =
                (insert_function_contract1 func_contract
                                               tails
                                               global_ghost_vars
                                               type_invariant_hashtbl
                                               modelfields
                                               (Some a_note)) in
            (head :: processed_stmts, remain_stmts)
          )
        | _ ->
            raise (Sexpr.Specs_error ("Function contract without function", None))
    )
)

let find_insert_function_contract (stmts : Cexpr.stmt_t list)
                             (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
                             type_invariant_hashtbl
                             modelfields
: Cexpr.stmt_t list = (
    let rec find_insert_function_contract1 (stmts: Cexpr.stmt_t list)
                             (func_contract_op : Sexpr.lspec_t option)
                             (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
                             type_invariant_hashtbl
                             modelfields
    : Cexpr.stmt_t list = (
        match stmts with
        | [] -> []
        | head :: tails -> (
            match func_contract_op, head with
            | Some old_func_contract,
              Cexpr.SPEC (Sexpr.SPEC_FUN curr_func_contract) -> (
                let func_contract = Some (Sexpr.merge_fun_contracts
                                        [old_func_contract;
                                        Sexpr.SPEC_FUN curr_func_contract]) in
                (find_insert_function_contract1 tails func_contract
                                      global_ghost_vars
                                      type_invariant_hashtbl
                                      modelfields)
              )
            | _, Cexpr.SPEC (Sexpr.SPEC_FUN curr_func_contract) -> (
                let func_contract = Some (Sexpr.SPEC_FUN curr_func_contract) in
                (find_insert_function_contract1 tails func_contract
                                      global_ghost_vars
                                      type_invariant_hashtbl
                                      modelfields)
              )
            | Some old_func_contract,
              _ -> (
                let processed_stmts, remain_stmts =
                    (insert_function_contract1 old_func_contract
                                    stmts
                                    global_ghost_vars
                                    type_invariant_hashtbl
                                    modelfields 
                                    None) in
                processed_stmts @ (find_insert_function_contract1 remain_stmts
                                                         None
                                                         global_ghost_vars
                                                         type_invariant_hashtbl
                                                         modelfields)
              )
            | None, Cexpr.SPEC (Sexpr.SPEC_VAR var_spec) -> (
                let scope_stack = Ccommon.pop_func_scope () in
                let local_scope = collect_vars_from_scope_stack scope_stack in
                let proccessed_stmts, remain_stmts =
                    insert_var_spec (Sexpr.SPEC_VAR var_spec) tails local_scope
                in
                proccessed_stmts @ (find_insert_function_contract1
                                       remain_stmts
                                       None
                                       global_ghost_vars
                                       type_invariant_hashtbl
                                       modelfields)
            )
            | None, _ -> (
                (match head with
                | Cexpr.DEFUN (header, _)
                | Cexpr.DEFVAR header -> (
                    if (is_invariant header) = false then
                        let _ = Ccommon.pop_func_scope () in ();)
                | _ -> ()
                );
                head :: (find_insert_function_contract1 tails
                                               None
                                               global_ghost_vars
                                               type_invariant_hashtbl
                                               modelfields)
              )
            | _ -> assert false
          )
    ) in
    (find_insert_function_contract1 stmts
                          None
                          global_ghost_vars
                          type_invariant_hashtbl
                          modelfields)
)

(************************* find and insert spec of var *************************)
let make_assert_read (var_spec : Cexpr.full_expr_t Sexpr.var_condition_t list)
: (Cexpr.stmt_t list * Cexpr.stmt_t list) =
begin
    let (read_pres, read_posts) =
        List.fold_left
            (fun (pre_conds, post_conds) condition ->
                let acc_post, expr = condition in
                match acc_post with
                | Sexpr.READ Sexpr.PRE_CONDITION ->
                    (expr.Cexpr.expr_node :: pre_conds, post_conds)
                | Sexpr.READ (Sexpr.POST_CONDITION _) ->
                    (pre_conds, expr.Cexpr.expr_node :: post_conds)
                | _ -> (pre_conds, post_conds)) ([], []) var_spec in
    (Stranslate.make_list_assert read_pres,
     Stranslate.make_list_assert read_posts)      
end
    
let make_assert_write (var_spec : Cexpr.full_expr_t Sexpr.var_condition_t list)
: (Cexpr.stmt_t list * Cexpr.stmt_t list) =
begin
    let (write_pres, write_posts) =
        List.fold_left
            (fun (pre_conds, post_conds) condition ->
                let acc_post, expr = condition in
                match acc_post with
                | Sexpr.WRITE Sexpr.PRE_CONDITION ->
                    (expr.Cexpr.expr_node :: pre_conds, post_conds)
                | Sexpr.WRITE (Sexpr.POST_CONDITION _) ->
                    (pre_conds, expr.Cexpr.expr_node :: post_conds)
                | _ -> (pre_conds, post_conds)) ([], []) var_spec in
    (Stranslate.make_list_assert write_pres,
     Stranslate.make_list_assert write_posts)      
end

let make_assert_var_specs (read_vars : Cexpr.expr_t list)
                          (write_vars : Cexpr.expr_t list)
                          (var_spec : (string, Cexpr.cspec_t option) Hashtbl.t)
                          (current_ofile : Cexpr.ofile_t)
                          (ofiles : Cexpr.ofile_t list)
: (Cexpr.stmt_t list * Cexpr.stmt_t list) =
begin
    let make_assert (make_fun : Cexpr.full_expr_t Sexpr.var_condition_t list ->
                            (Cexpr.stmt_t list * Cexpr.stmt_t list))
                    (vars : Cexpr.expr_t list)
                    (current_ofile : Cexpr.ofile_t)
                    (ofiles : Cexpr.ofile_t list)
    : (Cexpr.stmt_t list * Cexpr.stmt_t list) = (
        List.fold_left
            (fun (pres, posts) a_expr ->
                let name_var = Ccommon.get_name_of_expr a_expr in
                let spec_ops =
                    try
                        [Hashtbl.find var_spec name_var]
                    with Not_found -> (
                        let called_stmts =
                            Ccommon.find_called_stmts_in_ofiles ofiles
                                                                current_ofile
                                                                name_var in
                        List.fold_left
                            (fun tmp_specs a_stmt ->
                                match a_stmt with
                                | Cexpr.DEFUN (decl, st) ->
                                    tmp_specs
                                | Cexpr.DEFVAR (decl) ->
                                    (tmp_specs @ [decl.Cexpr.decl_specs])                
                                | _ ->
                                    assert false
                            ) [] called_stmts
                    ) in
                let specs =
                    List.fold_left
                         (fun tmp_specs a_sp_op ->
                             match a_sp_op with
                             | None -> tmp_specs
                             | Some spec -> (tmp_specs @ [spec])   
                         ) [] spec_ops in
                 
                let (tmp_pres, tmp_posts) =
                    match specs with
                    | [] -> ([], [])
                    | [spec] -> (
                        match spec with
                        | Sexpr.SPEC_VAR spec_var -> make_fun spec_var
                        | _ -> ([], [])
                    | _ -> (
                        LogUtil.log LogUtil.PARSER LogUtil.ERROR
                            ("Variable "
                              ^ name_var
                              ^ " has more than one spec");
                        exit 100
                    )
                ) in
                (pres @ tmp_pres, posts @ tmp_posts)
            ) ([], []) vars
    ) in

    let (pre_reads, post_reads) =
        make_assert make_assert_read read_vars current_ofile ofiles in
    let (pre_writes, post_writes) =
        make_assert make_assert_write write_vars current_ofile ofiles in
    (pre_reads @ pre_writes, post_reads @ post_writes)
end

let push_specs_of_var (specs_of_var : ((string, Cexpr.cspec_t option) Hashtbl.t) Stack.t)
: unit = (
    Stack.push (Hashtbl.create 10) specs_of_var
)

let pop_specs_of_var (specs_of_var : ((string, Cexpr.cspec_t option) Hashtbl.t) Stack.t)
: unit = (
    let _ = Stack.pop specs_of_var in ()
)

let collect_specs_of_var (specs_of_var : ((string, Cexpr.cspec_t option) Hashtbl.t) Stack.t)
: (string, Cexpr.cspec_t option) Hashtbl.t = (
    let result = Hashtbl.create (10) in
    Stack.iter (fun a_hash ->
        Hashtbl.iter (fun name spec_op ->
            if (Hashtbl.mem result name) = false then
                Hashtbl.add result name spec_op
        ) a_hash;
    ) specs_of_var;
    result
)

(** Each "continue" construction in a loop will transform into "goto"
    instruction which appear before var-conditions *)
let rec change_direct_of_continue_stmt (stmt : Cexpr.stmt_t)
                                       (new_label : string)
: Cexpr.stmt_t =
begin
    match stmt with
    | Cexpr.DEFUN (header, body) -> (
        let new_body = change_direct_of_continue_stmt body new_label in
        Cexpr.DEFUN(header, new_body))
    | Cexpr.DEFVAR _ -> stmt
    | Cexpr.BEGIN stmts -> (
        let new_stmts = change_direct_of_continue_stmts stmts new_label in
        Cexpr.BEGIN new_stmts)
    | Cexpr.IF (cond, then_stmts, else_stmts) -> (
        let new_then_stmts =
            change_direct_of_continue_stmts then_stmts new_label
        and new_else_stmts =
            change_direct_of_continue_stmts else_stmts new_label
        in
        Cexpr.IF (cond, new_then_stmts, new_else_stmts))
    | Cexpr.SWITCH (e, stmts) -> (
        let new_stmts =
            change_direct_of_continue_stmts stmts new_label in
        Cexpr.SWITCH (e, new_stmts))
    | Cexpr.CALLASGN (e, stmts, es) -> (
        let new_stmts =
            change_direct_of_continue_stmts stmts new_label in
        Cexpr.CALLASGN (e, new_stmts, es)
    )
    | Cexpr.CONTINUE -> Cexpr.GOTO new_label
    | _ -> stmt
end

and change_direct_of_continue_stmts (stmts : Cexpr.stmt_t list)
                                    (new_label : string)
: Cexpr.stmt_t list =
begin
    List.map
        (fun a_stmt ->
            change_direct_of_continue_stmt a_stmt new_label) stmts 
end


let spec_of_var_in_global (var_name : string)
                          (current_ofile : Cexpr.ofile_t)
                          (ofiles : Cexpr.ofile_t list)
: Cexpr.cspec_t list =
begin
    let stmts =
        Ccommon.find_called_stmts_in_ofiles ofiles 
                                            current_ofile
                                            var_name in
    let cspec_lst =
        List.fold_left
            (fun tmp_spec_lst a_stmt ->
                match a_stmt with
                | Cexpr.DEFUN (decl, st) -> tmp_spec_lst 
                | Cexpr.DEFVAR (decl) -> (
                    match decl.Cexpr.decl_specs with
                    | None -> tmp_spec_lst
                    | Some spec -> tmp_spec_lst @ [spec]
                )
                | _ -> assert false
        ) [] stmts in
    (cspec_lst)
end

(** Suppose u is an union var
    union {int a; int b;} u;
    When write u.a: Set flag permit_u.a = 1 and u.b = 0
    when read u.a: Check flag permit_u.a = 1
*)
let rec find_pos_of_field ?(pos : int = 0)
                          (field_name : string)
                          (components : (string * Cexpr.typenode_t) list)
: int = (
    match components with
    | [] -> assert false
    | (name, _) :: remain_components -> (
        if (String.compare name field_name) = 0 then
            (pos + 1)
        else find_pos_of_field ~pos:(pos+1) field_name remain_components
      )
)

let set_permit_union_field (expr : Cexpr.expr_t)
: Cexpr.expr_t = (
    match expr with
    | Cexpr.SREF (_, _, union_expr, field_name) -> (
        let tn = Cexpr.type_of_expr union_expr in
        assert (Cexpr.union_type_p tn);
        let _, components = Ccommon.get_name_and_component tn in
        let pos_of_field = find_pos_of_field field_name components in
        Cexpr.make_assignment Cexpr.SET
                              (Cexpr.SREF (Cexpr.INTEGER,
                                           true,
                                           union_expr,
                                           "permit"))
                              (Cexpr.CONST_I (Cexpr.INTEGER,
                                              num_of_int pos_of_field,
                                              (string_of_int pos_of_field)))
      )
    | _ -> assert false
)

let set_permit_union_fields (exprs : Cexpr.expr_t list)
: Cexpr.expr_t list = (
        List.map (set_permit_union_field) exprs
)

let check_permit_union_field (expr : Cexpr.expr_t)
: Cexpr.stmt_t = (
    match expr with
    | Cexpr.SREF (_, _, union_expr, field_name) -> (
        let tn = Cexpr.type_of_expr union_expr in
        assert (Cexpr.union_type_p tn);
        let _, components = Ccommon.get_name_and_component tn in
        let pos_of_field = find_pos_of_field field_name components in
        Stranslate.make_assert (
            Cexpr.make_binary_operation Cexpr.EQ
                                        (Cexpr.SREF (Cexpr.INTEGER,
                                                     true,
                                                     union_expr,
                                                     "permit"))
                                        (Cexpr.CONST_I (Cexpr.INTEGER,
                                                        num_of_int pos_of_field,
                                                        (string_of_int pos_of_field))));
      )
    | _ -> assert false
)

let check_permit_union_fields (exprs : Cexpr.expr_t list)
: Cexpr.stmt_t list = (
    List.map (fun expr ->
        check_permit_union_field expr
    ) exprs
)

let rec find_insert_specs_of_var_in_global
    (stmt : Cexpr.stmt_t)
    (specs_of_var : ((string, Cexpr.cspec_t option) Hashtbl.t) Stack.t)
    (current_ofile : Cexpr.ofile_t)
    (ofiles : Cexpr.ofile_t list)
: Cexpr.stmt_t =
begin
    match stmt with
    | Cexpr.DEFUN (decl, stmt1) -> (
        push_specs_of_var specs_of_var;
        List.iter (fun param ->
            Hashtbl.replace (Stack.top specs_of_var) param None
        ) decl.Cexpr.decl_params;
        let new_stmts =
            find_insert_specs_of_var_in_stmt stmt1 specs_of_var
                                             current_ofile ofiles in

        pop_specs_of_var specs_of_var;
        match new_stmts with
        | [new_stmt] -> Cexpr.DEFUN (decl, new_stmt)
        | _ -> assert false;
      )
    | Cexpr.DEFVAR decl -> (
        if (Cexpr.function_type_p decl.Cexpr.decl_tn) then ()
        else (
            let specs_op = decl.Cexpr.decl_specs in
            let (var_name, _) = decl.Cexpr.decl_id in
            match specs_op with
            | None -> ()
            | Some spec -> (
                (* This var is global var *)
                let glo_spec_vars =
                    spec_of_var_in_global var_name current_ofile ofiles in
                if ((List.length glo_spec_vars) > 1) then (
                    LogUtil.log LogUtil.PARSER LogUtil.ERROR
                        ("Variable "
                          ^ var_name
                          ^ " has more than one spec");
                    exit 100
                );
                Hashtbl.replace (Stack.top specs_of_var) var_name specs_op;)
        );
        let (_, _, _, written_union_fields) =
            Ccommon.classify_var_in_exprs decl.Cexpr.decl_inits
                                          [] [] [] [] false false in
        let set_permit_union_fields =
            set_permit_union_fields written_union_fields in
        let decl = {decl with Cexpr.decl_inits = decl.Cexpr.decl_inits
                                                 @ set_permit_union_fields}
        in
        (Cexpr.DEFVAR decl)
      )
    | Cexpr.ANNOTATION _ -> (stmt)
    | _ -> assert false
end

(** Use a stack to save var_specs in each scope *)
and find_insert_specs_of_var_in_stmt
    (stmt : Cexpr.stmt_t)
    (specs_of_var : ((string, Cexpr.cspec_t option) Hashtbl.t) Stack.t)
    (current_ofile : Cexpr.ofile_t)
    (ofiles : Cexpr.ofile_t list)
: Cexpr.stmt_t list = (
    match stmt with
    | Cexpr.DEFUN (decl, stmt1) -> (
        push_specs_of_var specs_of_var;
        List.iter (fun param ->
            Hashtbl.replace (Stack.top specs_of_var) param None
        ) decl.Cexpr.decl_params;
        let new_stmts =
            find_insert_specs_of_var_in_stmt stmt1 specs_of_var
                                             current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        match new_stmts with
        | [new_stmt] -> [Cexpr.DEFUN (decl, new_stmt)]
        | _ -> assert false
      )
    | Cexpr.DEFVAR decl -> (
        if (Cexpr.function_type_p decl.Cexpr.decl_tn) then ([stmt])
        else (
            let specs_op = decl.Cexpr.decl_specs in
            let (var_name, _) = decl.Cexpr.decl_id in
            let (read_vars, write_vars, read_union_fields, written_union_fields) =
                Ccommon.classify_var_in_exprs decl.Cexpr.decl_inits
                                              [] [] [] [] false false in
            let set_permit_union_fields =
                set_permit_union_fields written_union_fields in
            let decl = {decl with Cexpr.decl_inits = decl.Cexpr.decl_inits
                                                     @ set_permit_union_fields}
            in
            let check_permit_union_fields =
                check_permit_union_fields read_union_fields in
            let spec_var = collect_specs_of_var specs_of_var in
            let (pre_conds, pos_conds) =
                make_assert_var_specs read_vars write_vars spec_var
                                      current_ofile ofiles in
            (match specs_op with
            | None -> ()
            | Some spec -> (
                (* This var isn't global var *)
                match decl.Cexpr.decl_scls with
                | Cexpr.EXTERN -> (
                    let glo_spec_vars =
                        spec_of_var_in_global var_name current_ofile ofiles
                    in
                    if ((List.length glo_spec_vars) > 0) then (
                        LogUtil.log LogUtil.PARSER LogUtil.ERROR
                            ("Variable "
                              ^ var_name
                              ^ " has more than one spec");
                        exit 100)
                )
                | _ -> ();
                Hashtbl.replace (Stack.top specs_of_var) var_name specs_op;
            ));
            (check_permit_union_fields @ pre_conds @ [Cexpr.DEFVAR decl]
             @ pos_conds)
        )
      )
    | Cexpr.CALLASGN (expr1, stmts1, expr2) -> (
        push_specs_of_var specs_of_var;
        let new_stmts =
            find_insert_specs_of_var_in_stmts stmts1 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        [Cexpr.CALLASGN (expr1, new_stmts, expr2)]
      )
    | Cexpr.BEGIN stmts1 -> (
        push_specs_of_var specs_of_var;
        let new_stmts =
            find_insert_specs_of_var_in_stmts stmts1 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        [Cexpr.BEGIN new_stmts]
      )
    | Cexpr.IF (expr1, stmts1, stmts2) -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_expr expr1 [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields = check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        assert (List.length pos_conds = 0);
        push_specs_of_var specs_of_var;
        let new_stmts1 =
            find_insert_specs_of_var_in_stmts stmts1 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        push_specs_of_var specs_of_var;
        let new_stmts2 =
            find_insert_specs_of_var_in_stmts stmts2 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        (check_permit_union_fields
         @ pre_conds @ [Cexpr.IF (expr1, new_stmts1, new_stmts2)]
         @ set_permit_union_fields)
      )
    | Cexpr.SWITCH (expr1, stmts1) -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_expr expr1 [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields =
            check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        assert (List.length pos_conds = 0);
        push_specs_of_var specs_of_var;
        let new_stmts1 =
            find_insert_specs_of_var_in_stmts stmts1 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        (check_permit_union_fields
         @ pre_conds @ [Cexpr.SWITCH (expr1, new_stmts1)]
         @ set_permit_union_fields)
      )
    | Cexpr.DOWHILE (b, expr1, stmts1) -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_expr expr1 [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields =
            check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        assert (List.length pos_conds = 0);
        push_specs_of_var specs_of_var;
        let new_stmts1 =
            find_insert_specs_of_var_in_stmts stmts1 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        if (b && (List.length pre_conds > 0)) then ( 
            let new_label_name = Cexpr.make_new_label () in
            let new_stmts1 =
                change_direct_of_continue_stmts new_stmts1 new_label_name in
            let new_goto = [Cexpr.GOTO new_label_name] in
            let new_stmts1 = new_stmts1 @ new_goto in
            let new_label = [(Cexpr.LABEL (Cexpr.NAME_LABEL new_label_name))] in
            (new_label @ check_permit_union_fields
             @ pre_conds @ [Cexpr.DOWHILE (b, expr1, new_stmts1)]
             @ set_permit_union_fields))
        else (
            let new_stmts1 = new_stmts1 @ pre_conds in
            [Cexpr.DOWHILE (b, expr1, new_stmts1)]  
        ) 
      )
    | Cexpr.FOR (stmts1, expr1, expr2, stmts2) -> (
        push_specs_of_var specs_of_var;
        let new_stmts1 =
            find_insert_specs_of_var_in_stmts stmts1 specs_of_var
                                              current_ofile ofiles in
        let (read_vars1, write_vars1, read_union_fields_1, written_union_fields_1) =
            Ccommon.classify_var_in_expr expr1 [] [] [] [] false false in
        let set_permit_union_fields_1 =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields_1) in
        let check_permit_union_fields_1 =
            check_permit_union_fields read_union_fields_1 in
        let (read_vars2, write_vars2, read_union_fields_2, written_union_fields_2) =
            Ccommon.classify_var_in_expr expr2 [] [] [] [] false false in
        let set_permit_union_fields_2 =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields_2) in
        let check_permit_union_fields_2 =
            check_permit_union_fields read_union_fields_2 in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds1, pos_conds1) =
            make_assert_var_specs read_vars1 write_vars1 spec_var
                                  current_ofile ofiles in
        let (pre_conds2, pos_conds2) =
            make_assert_var_specs read_vars2 write_vars2 spec_var
                                  current_ofile ofiles in
        assert ((List.length pos_conds1 = 0) && (List.length pos_conds2 = 0));
        pop_specs_of_var specs_of_var;
        push_specs_of_var specs_of_var;
        let new_stmts2 =
            find_insert_specs_of_var_in_stmts stmts2 specs_of_var
                                              current_ofile ofiles in
        pop_specs_of_var specs_of_var;
        [Cexpr.FOR ((check_permit_union_fields_1 @ new_stmts1 @ pre_conds1
                     @ set_permit_union_fields_2),
                    expr1, expr2,
                    (check_permit_union_fields_1 @ new_stmts2 @ pre_conds2
                     @ set_permit_union_fields_2))]
      )
    | Cexpr.STMT expr1 -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_expr expr1 [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields =
            check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        (check_permit_union_fields
         @ pre_conds @ [stmt] @ pos_conds
         @ set_permit_union_fields)
      )
    | Cexpr.ASGN (expr1, expr2) -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_exprs [expr1; expr2] [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields =
            check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        (check_permit_union_fields
         @ pre_conds @ [stmt] @ pos_conds
         @ set_permit_union_fields) 
      )
    | Cexpr.LABEL label -> (
        match label with
        | Cexpr.NAME_LABEL _
        | Cexpr.DFLT_LABEL -> [stmt]
        | Cexpr.CASE_LABEL (expr1, expr2) -> (
            let (read_vars, write_vars, read_union_fields, written_union_fields) =
                Ccommon.classify_var_in_exprs [expr1; expr2]
                                              [] [] [] [] false false in
            let set_permit_union_fields =
                List.map (fun ex -> Cexpr.STMT ex)
                         (set_permit_union_fields written_union_fields) in
            let check_permit_union_fields =
                check_permit_union_fields read_union_fields in
            let spec_var = collect_specs_of_var specs_of_var in
            let (pre_conds, pos_conds) =
                make_assert_var_specs read_vars write_vars spec_var
                                      current_ofile ofiles in
            (check_permit_union_fields
             @ pre_conds @ [stmt] @ pos_conds
             @ set_permit_union_fields)    
        )
      )
    | Cexpr.RETURN expr_op -> [stmt]
    | Cexpr.ANNOTATION (_, expr1)
    | Cexpr.ASM expr1
    | Cexpr.ASSUME expr1 -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_expr expr1 [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                     (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields =
            check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        (check_permit_union_fields
         @ pre_conds @ [stmt] @ pos_conds
         @ set_permit_union_fields)
      )
    | Cexpr.RESET exprs1 -> (
        let (read_vars, write_vars, read_union_fields, written_union_fields) =
            Ccommon.classify_var_in_exprs exprs1 [] [] [] [] false false in
        let set_permit_union_fields =
            List.map (fun ex -> Cexpr.STMT ex)
                (set_permit_union_fields written_union_fields) in
        let check_permit_union_fields =
            check_permit_union_fields read_union_fields in
        let spec_var = collect_specs_of_var specs_of_var in
        let (pre_conds, pos_conds) =
            make_assert_var_specs read_vars write_vars spec_var
                                  current_ofile ofiles in
        (check_permit_union_fields
         @ pre_conds @ [stmt] @ pos_conds
         @ set_permit_union_fields)
      )
    | _ -> [stmt]
)

and find_insert_specs_of_var_in_stmts (stmts : Cexpr.stmt_t list)
    (specs_of_var : ((string, Cexpr.cspec_t option) Hashtbl.t) Stack.t)
    (current_ofile : Cexpr.ofile_t)
    (ofiles : Cexpr.ofile_t list)
: Cexpr.stmt_t list = (
    List.flatten (
        List.map (fun stmt ->
            find_insert_specs_of_var_in_stmt stmt specs_of_var
                                             current_ofile ofiles
        ) stmts)
)

and find_insert_specs_of_var (ofiles : Cexpr.ofile_t list)
: Cexpr.ofile_t list =
begin
    let specs_of_var = Stack.create () in
    push_specs_of_var specs_of_var;
    let inserted_ofiles =
        List.map
            (fun a_ofile ->
                let new_ofile_defs =
                    List.map
                        (fun a_def ->
                            let (name_def, def) = a_def in
                            let new_def =
                                find_insert_specs_of_var_in_global def
                                                                   specs_of_var
                                                                   a_ofile
                                                                   ofiles in
                            (name_def, new_def)
                        ) !(a_ofile.Cexpr.ofile_defs)
                in {a_ofile with Cexpr.ofile_defs = ref new_ofile_defs}
            ) ofiles in
    
    Stack.clear specs_of_var;
    (inserted_ofiles)
end

(************************* Collects *******************************************)
(*  Does collecting global_vars, function_contracts, global_invariants, and
    type_invariants from decls.
*)
(* collect global ghost var*)
(*
let collect_global_ghost_vars 
         (decls : Cexpr.stmt_t list) 
         (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
         =(begin
    List.iter (
        function
        | Cexpr.SPEC (spec) ->(
            match spec with
            | Sexpr.SPEC_GHOST ghosts ->(
                List.iter (
                    function
                    | Sexpr.GHOST_DEF(typ, id, _) ->
                        let expr = 
                            Cexpr.VREF(
                                (Scommon.convert_Ctype_to_typenode_t typ),
                                id, 1, 0)
                        in
                        Hashtbl.replace global_ghost_vars id expr;
                    | _ ->
                        raise (Sexpr.Specs_error("Assignment outside a function", None))
                )ghosts;
             )
            | _ -> ())
        | _ -> ()
    )decls
end)
*)

(* Collects global invariants *)
let collect_global_invariants (decls : Cexpr.stmt_t list)
    (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t)
: Cexpr.stmt_t list = (begin
    let scope_stacks = Queue.create () in
    let result =
    List.map (function stmt ->
        match stmt with
            | Cexpr.SPEC spec as spec_stmt -> (begin
                let scope_stack = Ccommon.pop_global_spec_scope () in
                let vars = collect_vars_from_scope_stack scope_stack in
                match spec with
                | Sexpr.SPEC_GLOBAL_INV global_invariant ->(begin
                    let {
                          Sexpr.ginv_strength = _;
                          Sexpr.ginv_name = id;
                          Sexpr.ginv_expr = lexpr;
                          Sexpr.ginv_exceptions = _
                        } = global_invariant in
                    global_invariant_names := (id)::
                                              (!global_invariant_names);
                    make_check_data_invariant_var id;
                    import_info_invariant spec;
                    import_ghost_var spec global_ghost_vars;
                    let triples = (Stranslate.translate_global_invariant 
                                       global_invariant vars) in
                    (* get and add position of invariant to declaration of it *)
                    let pos = fst lexpr.Sexpr.lexpr_loc in
                    let invariant_pos =
                        Sexpr.make_global_pos pos.Lexing.pos_fname
                                              pos.Lexing.pos_lnum
                                              pos.Lexing.pos_cnum in
                    let (_, _, new_stmt) = triples in
                    (Ccommon.add_location_to_stmt invariant_pos new_stmt)
                  end)
                | Sexpr.SPEC_TYPE_INV type_invariant -> (begin
                    let {
                          Sexpr.tinv_strength = _;
                          Sexpr.tinv_name = id;
                          Sexpr.tinv_type = typ;
                          Sexpr.tinv_param_name = _;
                          Sexpr.tinv_expr = lexpr;
                          Sexpr.tinv_exceptions = _
                        } = type_invariant in
                    type_invariant_names := id :: (!type_invariant_names);
                    let ctyp = Stranslate.translate_logictype typ in
                    Hashtbl.add type_invariant_stmts (find_true_type ctyp) id;
                    import_info_invariant spec;
                    import_ghost_var spec global_ghost_vars;
                    let triples = (Stranslate.translate_type_invariant 
                                       type_invariant vars) in
                    (* get and add position of invariant to declaration of it *)
                    let pos = fst lexpr.Sexpr.lexpr_loc in
                    let invariant_pos =
                        Sexpr.make_global_pos pos.Lexing.pos_fname
                                              pos.Lexing.pos_lnum
                                              pos.Lexing.pos_cnum in
                    let (_, _, new_stmt) = triples in
                    (Ccommon.add_location_to_stmt invariant_pos new_stmt)
                  end)
                | Sexpr.SPEC_FUN {
                                   Sexpr.fspec_conditions = _;
                                   Sexpr.fspec_behaviors = _;
                                   Sexpr.fspec_assigns = _;
                                   Sexpr.fspec_prohibits = _;
                                   Sexpr.fspec_local_invs = invs;
                                   Sexpr.fspec_context = _;
                                   Sexpr.fspec_is_handler = _
                                 } -> (begin
                    Queue.push scope_stack scope_stacks;
                    List.iter (
                        fun inv -> 
                            let id, _ = inv in
                            global_invariant_names := (id)::
                                                      (!global_invariant_names);
                            make_check_data_invariant_var id;
                    ) invs;
                    spec_stmt
                  end)
                | _ -> (
                    Queue.push scope_stack scope_stacks;
                    spec_stmt) 
              end)
          | _ -> stmt
  ) decls in
    Queue.iter (fun a_mem ->
        Ccommon.push_global_spec_scopes ()
    ) scope_stacks;
    result
end)

let prohibit_ghost_vars = ref [];;

let collect_prohibit (fun_name : string) 
                     (prohibits_location : string list * Sexpr.global_pos_t option)
                     (function_contracts : (string, Sexpr.lspec_t) Hashtbl.t)
                     (global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t) = (
    let make_pre_cond (fun_name : string) (prohibit :string)= (
        let var_name = ("dont_call_"^prohibit^"_in_"^fun_name) in
        let lexpr1 = Sexpr.create_lexpr_without_loc (Sexpr.PLvar var_name) in
        let lexpr2 = Sexpr.create_lexpr_without_loc (
                         Sexpr.PLconstant (Sexpr.IntConstant "0")) in
        let lexpr = Sexpr.create_lexpr_without_loc
                        (Sexpr.PLrel (lexpr1, Sexpr.Eq, lexpr2)) in
        Sexpr.SPEC_FUN
            (Sexpr.make_function_contract
                 [(Sexpr.PRE_CONDITION, lexpr)]
                 []
                 []
                 []
                 []
                 Sexpr.ANY
                 false)
    ) in
    let make_expr_global_ghost_var (fun_name : string) (prohibit: string) =(
        let var_name = ("dont_call_"^prohibit^"_in_"^fun_name) in
        Cexpr.VREF(Cexpr.INTEGER, 
                   var_name,
                   1, 0)
    ) in
    let make_global_ghost_var (fun_name : string) (prohibit: string)
                              (location : Sexpr.global_pos_t option) = (
        let var_name = ("dont_call_"^prohibit^"_in_"^fun_name) in
        let init_expr = Cexpr.SETF( 
                            Cexpr.INTEGER,
                            Cexpr.SET,
                            (make_expr_global_ghost_var fun_name prohibit),
                            Cexpr.CONST_I(Cexpr.INTEGER, Int(0), "0"))
        in
        let defvar = {
              Cexpr.decl_id = (var_name, 1);
              Cexpr.decl_tn = Cexpr.INTEGER ;
              Cexpr.decl_scls = Cexpr.NILSCLS;
              Cexpr.decl_inits = [init_expr] ;
              Cexpr.decl_params = [];
              Cexpr.decl_struct_defined_p = true;
              Cexpr.decl_specs = None;
              Cexpr.decl_location = location;
        } in
        Cexpr.DEFVAR(defvar)
    ) in
    let (prohibits, location) = prohibits_location in
    let res = (
        List.map(
            fun prohibit ->
                Hashtbl.add function_contracts
                            prohibit
                            (make_pre_cond fun_name prohibit);
                Hashtbl.replace global_ghost_vars 
                               ("dont_call_"^prohibit^"_in_"^fun_name)
                                (make_expr_global_ghost_var fun_name prohibit);
                (make_global_ghost_var fun_name prohibit location)
        )prohibits)
    in
    prohibit_ghost_vars := res @ (!prohibit_ghost_vars)
)

(* collect function contract from model field *)
let rec collect_fun_contract_model_field 
                (decls : Cexpr.stmt_t list)
                (function_contracts : (string, Sexpr.lspec_t) Hashtbl.t)
                = (begin
    match decls with
    | [] -> 
        ()
    | head::tails -> (begin
        (match head with
        | Cexpr.DEFUN (d, _) -> (begin
            let fun_name = 
                match Ccommon.get_function_name d with
                | Some name -> name
                | None -> assert false
            in
            let fun_contract =  
                (collect_fun_contract_model_field_1 fun_name d) in
            if fun_contract <> 
                (Sexpr.SPEC_FUN(Sexpr.empty_fun_contract)) then            
                    Hashtbl.add function_contracts fun_name fun_contract
            end)
        | _ -> ()
        );
        collect_fun_contract_model_field tails function_contracts
        end)
end)

and collect_fun_contract_model_field_1 
                            (fun_name : string)  
                            (fun_header : Cexpr.declaration_t) = (
    try 
        let struct_name, no_contract = 
                Hashtbl.find contract_for_init_fun_hashtbl fun_name
        in
        if no_contract = false then (begin
            (* get name and type of the relevant parameter *)
            let param_name, param_type = 
                Ccommon.get_struct_param_info fun_header struct_name in
            (* param_name.__initialized__ == 0 *)
            (*
            let preds_with_loc = 
	        [(Sexpr.PRED_REL ( 
                            Sexpr.TERM_DEF_FIELD (
                                Sexpr.ID (param_name), 
                                "__initialized__" ), 
                            Sexpr.REL_EQ,
                            Sexpr.TERM_INTEGER (0)), None)]
            in
            *)
            (* unknown TERM_INTEGER *)
            let lexpr1 = Sexpr.create_lexpr_without_loc 
                             (Sexpr.PLdot ((Sexpr.create_lexpr_without_loc
                                              (Sexpr.PLvar param_name)),
                                           "__initialized__" )) in
            let lexpr2 = Sexpr.create_lexpr_without_loc (
                             Sexpr.PLconstant (Sexpr.IntConstant "0")) in
            let lexpr = Sexpr.create_lexpr_without_loc
                            (Sexpr.PLrel (lexpr1, Sexpr.Eq, lexpr2)) in
	    (* FIXME: DzungDK *)
            let pre_cond =
                Sexpr.SPEC_FUN (Sexpr.make_function_contract
                                    [(Sexpr.PRE_CONDITION, lexpr)]
                                    [] [] [] [] Sexpr.ANY false) 
            in
            pre_cond
        end)
        else (Sexpr.SPEC_FUN(Sexpr.empty_fun_contract))
    with
    | Not_found -> (Sexpr.SPEC_FUN(Sexpr.empty_fun_contract))
)

(*  NOTE: Could be done by traversing through the whole decl at once. 
          But for now, program clearness is more preferred than performance.*)
(*
let collect_local_invariant (decls : Cexpr.stmt_t list) 
                        (local_scope : (string, Cexpr.expr_t) Hashtbl.t)
                        (local_ghost_scope : (string, Cexpr.expr_t) Hashtbl.t)
                        : Cexpr.stmt_t list = (
    let res = (
        List.map(
            function
            | Cexpr.SPEC (spec) -> (
                match spec with
                | Sexpr.SPEC_FUN (_,_,_,_, invss, _, _, _, _) -> (
                  let temp = (
                    List.map ( fun invs -> 
                    let name, lexpr = invs in
                    let name_var = ("checking_"^name) in
                    let var = Cexpr.VREF(Cexpr.INTEGER, name_var, 1, 0) in
                    Hashtbl.replace local_ghost_scope name_var var;
                    let pred = Sexpr.PRED_REL(
                                   Sexpr.ID(name_var),
                                   Sexpr.REL_GT,
                                   Sexpr.TERM_INTEGER(0))
                    in
                    let if_pred = Sexpr.PRED_IF(pred, preds, None) in
                    let spec_data = 
                        Sexpr.SPEC_GLOBAL_INV (name, [if_pred], None) in
                    let triples = 
                        (Stranslate.translate spec_data local_scope 
                                        local_ghost_scope 1 name None false) in
                    let init_expr = Cexpr.SETF(
                                        Cexpr.INTEGER,
                                        Cexpr.SET,
                                        var, 
                                        Cexpr.CONST_I(
                                            Cexpr.INTEGER, Int(0), "0")) in
                    let defvar = {
                        Cexpr.decl_id = (name_var, 1);
                        Cexpr.decl_tn = Cexpr.INTEGER ;
                        Cexpr.decl_scls = Cexpr.NILSCLS;
                        Cexpr.decl_inits = [init_expr] ;
                        Cexpr.decl_params = [];
                        Cexpr.decl_struct_defined_p = false;
                        Cexpr.decl_specs = None
                        Cexpr.decl_location = None;
                    } in
                    [Cexpr.DEFVAR(defvar)] @ (stmts_of_triples triples)
                    ) invss) in
                    (List.flatten temp)
                 )
                | _ -> []
             )
            | _ -> []
        ) decls
    ) in
    (List.flatten res) @ decls
)
*)

let collect_type_invariant_from_model_field 
          (spec_type_inv : Sexpr.ltype_invariant_t)
          (type_invariant_hashtbl : (Cexpr.typenode_t, string) Hashtbl.t)
                             : unit = (
    let {
          Sexpr.tinv_strength = _;
          Sexpr.tinv_name = id;
          Sexpr.tinv_type = typ;
          Sexpr.tinv_param_name = _;
          Sexpr.tinv_exceptions = _
        } = spec_type_inv in
    let ctyp = Stranslate.translate_logictype typ in
    Hashtbl.add type_invariant_hashtbl (find_true_type ctyp) id
)
(**********************************)
(* struct_var_name option * field_name(var_name) --> fun_name *)
let (fp_to_f_hashtbl : ((string option * string),
                         (Cexpr.expr_t * Cexpr.note_t))Hashtbl.t) =
        Hashtbl.create 10

let current_note = ref (Cexpr.LINE (false, "", -1, -1))

let f_pointer_field_hashtbl : (string * string, string) Hashtbl.t =
    Hashtbl.create 10

(* find and map function pointer with real function *)
let assign_include_function_pointer_expr
            (expr : Cexpr.expr_t)
            (function_contracts : (string, Sexpr.lspec_t) Hashtbl.t)
            : unit = (
    match expr with
    | Cexpr.SETF(_, Cexpr.SET, ex1, ex2) -> (
        match (Ccommon.right_expr_is_function_call ex2
                                                   fun_header_infos_hashtbl),
              (Ccommon.left_expr_is_field ex1),
              (Ccommon.left_expr_is_function_pointer ex1) with
        | Some f_name, Some(s_name, fp_f_name, var_name), None -> (
            (Hashtbl.add fp_to_f_hashtbl
                         (Some var_name,fp_f_name)
                         (ex2, !current_note));
            Hashtbl.add f_pointer_field_hashtbl (var_name, fp_f_name) (s_name);
            if (List.mem f_name !fun_names_is_assigned) then ()
            else (fun_names_is_assigned := f_name ::(!fun_names_is_assigned));
        )
        | Some (f_name), None, Some(fp_name) -> (
            (Hashtbl.add fp_to_f_hashtbl
                         (None,fp_name)
                         (ex2, !current_note));
            if (List.mem f_name !fun_names_is_assigned) then ()
            else (fun_names_is_assigned := f_name ::(!fun_names_is_assigned));)
        | _, Some(s_name, fp_f_name, var_name), _ -> (
            Hashtbl.add f_pointer_field_hashtbl (var_name, fp_f_name) (s_name);
        )
        | _, Some _, Some _ -> assert false
        | _ -> ()
    )
    | _ -> ()
)
(* find and map function pointer with real function *)
let rec assign_include_function_pointer_stmt 
            (stmt : Cexpr.stmt_t)
            (function_contracts : (string, Sexpr.lspec_t) Hashtbl.t)
            : unit = (
    match stmt with
    | Cexpr.STMT(ex) -> ( 
        assign_include_function_pointer_expr ex function_contracts
    )
    | Cexpr.DEFVAR(header) -> (
        let inits = header.Cexpr.decl_inits in
        List.iter(
            fun init ->
                assign_include_function_pointer_expr init function_contracts
        ) inits
    )
    | Cexpr.ASGN(ex1, ex2) -> (
        (* later *)    
    )
    | Cexpr.BEGIN(stmts) -> (
        assign_include_function_pointer_stmts stmts function_contracts
    )
    | Cexpr.DEFUN(_,body) ->  (
        assign_include_function_pointer_stmt body function_contracts
    )
    | Cexpr.FOR(stmts1, _, _, stmts2) ->  ( (* compound expression don't care *)
        assign_include_function_pointer_stmts stmts1 function_contracts;
        assign_include_function_pointer_stmts stmts2 function_contracts
    )
    | Cexpr.DOWHILE(_,_,stmts)-> (
        assign_include_function_pointer_stmts stmts function_contracts
    )
    | Cexpr.SWITCH(_,stmts)-> (
        assign_include_function_pointer_stmts stmts function_contracts
    )
    | Cexpr.NOTE note ->
        current_note := note
    | _ -> ()
)
(* find and map function pointer with real function *)
and assign_include_function_pointer_stmts 
        (stmts : Cexpr.stmt_t list)
        (function_contracts : (string, Sexpr.lspec_t) Hashtbl.t)
        :unit = (
    List.iter( 
        fun stmt -> 
            assign_include_function_pointer_stmt stmt function_contracts
    )stmts
)

(**********************************)
let import_into_post_conditions_hashtbl 
        (function_contracts : (string, Sexpr.lspec_t) Hashtbl.t)
        (global_vars : (string, Cexpr.expr_t) Hashtbl.t) : unit = (
    let visited_fun_name = ref [] in
    Hashtbl.iter(
        fun fun_name _ -> (
            let header = (
                try Hashtbl.find fun_header_infos_hashtbl fun_name
            with
            | Not_found -> assert false) 
            in
            let specs = 
                if not (List.mem fun_name !visited_fun_name) then (begin
                    visited_fun_name := fun_name :: (!visited_fun_name);
                    try Hashtbl.find_all function_contracts fun_name
                    with
                    | Not_found -> []
                end)
                else []
            in
            let spec = Sexpr.merge_fun_contracts specs in
            let (vars :(string, Cexpr.expr_t) Hashtbl.t) = Hashtbl.create 10 in
            let param_names = header.Cexpr.decl_params in
            let fun_type = header.Cexpr.decl_tn in
            let param_types = get_param_types fun_type in
            (try 
                List.iter2 (
                    fun name tyn ->
                        let expr = Cexpr.VREF(tyn, name, 2, 0) in
                        Hashtbl.add vars name expr
                ) param_names param_types;
                
            with
            | Invalid_argument _ -> ());
            (Hashtbl.iter(fun k v -> Hashtbl.add vars k v)global_vars);
            let _ = (Stranslate.translate spec vars 
                                          (Hashtbl.create 10)
                                          1 fun_name
                                          None false) in
            ()
        )      
    )function_contracts;

)
(********************************************************)
let collect_implicit_type_invariant 
        (implicit_type_invariants : (string, Sexpr.ltype_invariant_t) Hashtbl.t)
        (type_invariants : (Cexpr.typenode_t, string) Hashtbl.t) : unit = (
    Hashtbl.iter(
        fun key value -> 
            (collect_type_invariant_from_model_field value 
                                                     type_invariants);
    )implicit_type_invariants
)

(*(** Find cspec that contain struct_var.* of function func_var.
    Replace variables in cspec that are parameter of func_var by args.
    Replace variables struct_var.* by every field of struct_var
*)
let get_spec_all_field (func_var : Cexpr.expr_t)
                       (args : Cexpr.expr_t list) : Cexpr.cspec_t option = (
    let func_name = Ccommon.get_name_of_expr func_var in
    try
        let decl = Hashtbl.find func_contract_all_field func_name in
        let param_types =
            match decl.Cexpr.decl_tn with
            | Cexpr.FUNCTION_OF (typenode, typenodes_op) -> (
                match typenodes_op with
                | None -> []
                | Some typenodes -> typenodes
              )
            | _ -> [] in
        let params =
            List.map2 (fun param_type param_name ->
                Cexpr.VREF (param_type,
                            param_name,
                            Cexpr.parameter_scope_level,
                            0)
            ) param_types decl.Cexpr.decl_params
        in
        let spec_op = decl.Cexpr.decl_specs in
        (* replace params by arguments *)
        match spec_op with
        | None -> None
        | Some spec -> (
            let spec_replaced_params =
               (Ccommon.replace_exprs_in_spec spec params args) in
            let structs_get_field =
                Sexpr.struct_get_field Cexpr.struct_get_field_in_full_expr
                                       spec_replaced_params in
            assert ((List.length structs_get_field) = 1);
            let struct_var = List.hd structs_get_field in
            let spec_replaced_fields =
                replace_fields_of_struct struct_var
                                         spec_replaced_params in
            Some spec_replaced_fields)
    with Not_found -> None
) *)

(*(** Find Cexpr.CALL in expr, find and insert information of function_contract
    of called_function if this function_contract contains struct_var.*
*)
let rec process_spec_all_field_in_expr (expr : Cexpr.expr_t) : Cexpr.expr_t = (
    match expr with
    | Cexpr.OPER (tn, op, expr1, expr2, expr3) ->
        Cexpr.OPER (tn, op, 
              process_spec_all_field_in_expr expr1,
              process_spec_all_field_in_expr expr2,
              process_spec_all_field_in_expr expr3)
    | Cexpr.SETF (tn, op, expr1, expr2) ->
        Cexpr.SETF (tn, op,
              process_spec_all_field_in_expr expr1,
              process_spec_all_field_in_expr expr2)
    | Cexpr.CALL (tn, expr1, exprs1, cspec_op) -> (
        let spec = get_spec_all_field expr1 exprs1 in
        Cexpr.CALL (tn,
              process_spec_all_field_in_expr expr1,
              process_spec_all_field_in_exprs exprs1,
              spec))
    | Cexpr.PREF (tn, b, expr1) ->
        Cexpr.PREF (tn,
                    b,
                    process_spec_all_field_in_expr expr1)
    | Cexpr.SREF (tn, b, expr1, str) ->
        Cexpr.SREF (tn,
                    b,
                    process_spec_all_field_in_expr expr1,
                    str)
    | Cexpr.DSGN_AREF (tn, expr1, expr2) ->
        Cexpr.DSGN_AREF (tn,
                         process_spec_all_field_in_expr expr1,
                         process_spec_all_field_in_expr expr2)
    | Cexpr.ADDROF (tn, expr1) ->
        Cexpr.ADDROF (tn,
                      process_spec_all_field_in_expr expr1)
    | Cexpr.PROGN (tn, stmts) ->
        Cexpr.PROGN (tn,
                      process_spec_all_field_in_stmts stmts)
    | Cexpr.ASMSEQUENCE exprs1 ->
        Cexpr.ASMSEQUENCE (process_spec_all_field_in_exprs exprs1)
    | Cexpr.PHI (tn, str_exprs) ->
        let new_str_exprs =
            List.map (fun (str, expr1) ->
                (str, process_spec_all_field_in_expr expr1)
            ) str_exprs in
        Cexpr.PHI (tn, new_str_exprs)
    | Cexpr.INIT_FILL (tn, expr1) ->
        Cexpr.INIT_FILL (tn,
                         process_spec_all_field_in_expr expr1)
    | Cexpr.QUANTIFIER (str, exprs1, expr1) ->
        Cexpr.QUANTIFIER (str,
                          process_spec_all_field_in_exprs exprs1,
                          process_spec_all_field_in_expr expr1)
    | Cexpr.PSEUDO_ACSL_RESULT (tn, expr1) ->
        Cexpr.PSEUDO_ACSL_RESULT (tn,
                                  process_spec_all_field_in_expr expr1)
    | Cexpr.PSEUDO_ACSL_OLD (tn, expr1, expr2) ->
        Cexpr.PSEUDO_ACSL_OLD (tn,
                               process_spec_all_field_in_expr expr1,
                               process_spec_all_field_in_expr expr2)
    | Cexpr.PSEUDO_ACSL_UNIQUE (tn, expr1) ->
        Cexpr.PSEUDO_ACSL_UNIQUE (tn,
                                  process_spec_all_field_in_expr expr1)
    | Cexpr.PSEUDO_ACSL_VALID (tn, expr1) ->
        Cexpr.PSEUDO_ACSL_VALID (tn,
                                  process_spec_all_field_in_expr expr1)
    | Cexpr.ALL_FIELD _ ->
        assert false
    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE
    | Cexpr.NIL
    | Cexpr.PSEUDO_NONDET_VALUE _
    | Cexpr.VREF _
    | Cexpr.SIZEOF _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.CLITERAL _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.PRAGMA _ -> expr
)

(** Find Cexpr.CALL in expr list, find and insert information of function_contract
    of called_function if this function_contract contains struct_var.*
*)
and process_spec_all_field_in_exprs (exprs : Cexpr.expr_t list)
: Cexpr.expr_t list = (
    List.map process_spec_all_field_in_expr exprs
)

(** Find Cexpr.CALL in stmt, find and insert information of function_contract
    of called_function if this function_contract contains struct_var.*
*)
and process_spec_all_field_in_stmt (stmt : Cexpr.stmt_t) : Cexpr.stmt_t = (
    match stmt with
    | Cexpr.DEFUN (decl, stmt1) ->
        Cexpr.DEFUN (decl, process_spec_all_field_in_stmt stmt1)
    | Cexpr.STMT expr1 ->
        Cexpr.STMT (process_spec_all_field_in_expr expr1)
    | Cexpr.ASGN (expr1, expr2) ->
        Cexpr.ASGN (process_spec_all_field_in_expr expr1,
              process_spec_all_field_in_expr expr2)
    | Cexpr.CALLASGN (expr1, stmts1, exprs1) ->
        Cexpr.CALLASGN (process_spec_all_field_in_expr expr1,
                  process_spec_all_field_in_stmts stmts1,
                  process_spec_all_field_in_exprs exprs1)
    | Cexpr.BEGIN (stmts1) ->
        Cexpr.BEGIN (process_spec_all_field_in_stmts stmts1)
    | Cexpr.IF (expr1, stmts1, stmts2) ->
        Cexpr.IF (process_spec_all_field_in_expr expr1,
            process_spec_all_field_in_stmts stmts1,
            process_spec_all_field_in_stmts stmts2)
    | Cexpr.SWITCH (expr1, stmts1) ->
        Cexpr.SWITCH (process_spec_all_field_in_expr expr1,
                process_spec_all_field_in_stmts stmts1)
    | Cexpr.DOWHILE (b, expr1, stmts1) ->
        Cexpr.DOWHILE (b,
                 process_spec_all_field_in_expr expr1,
                 process_spec_all_field_in_stmts stmts1)
    | Cexpr.FOR (stmts1, expr1, expr2, stmts2) ->
      Cexpr.FOR (process_spec_all_field_in_stmts stmts1,
           process_spec_all_field_in_expr expr1,
           process_spec_all_field_in_expr expr2,
           process_spec_all_field_in_stmts stmts2)
    | Cexpr.RETURN (expr_op) ->
        (match expr_op with
        | None -> Cexpr.RETURN None
        | Some expr1 -> Cexpr.RETURN (Some (process_spec_all_field_in_expr expr1)))
    | Cexpr.ANNOTATION (str, expr1) ->
        Cexpr.ANNOTATION (str,
                    process_spec_all_field_in_expr expr1)
    | Cexpr.ASM (expr1) ->
        Cexpr.ASM (process_spec_all_field_in_expr expr1)
    | Cexpr.ASSUME (expr1) ->
        Cexpr.ASSUME (process_spec_all_field_in_expr expr1)
    | Cexpr.RESET (exprs1) ->
        Cexpr.RESET (process_spec_all_field_in_exprs exprs1)
    | Cexpr.DEFVAR _
    | Cexpr.BREAK
    | Cexpr.CONTINUE
    | Cexpr.LABEL _
    | Cexpr.GOTO _
    | Cexpr.NOTE _
    | Cexpr.SPEC _ -> stmt
)

(** Find Cexpr.CALL in stmt list, find and insert information of function_contract
    of called_function if this function_contract contains struct_var.*
*)
and process_spec_all_field_in_stmts (stmts : Cexpr.stmt_t list)
: Cexpr.stmt_t list = (
   List.map process_spec_all_field_in_stmt stmts
) *)

(******************************* Instrument Main *******************************************)
(* Main function for instrumenting specifications into source code.
        - implicit_type_invariants: type invariants generated by 
          constantness ans non-nullability of structs' fields 
*)        
let instrument_main 
         (decls : Cexpr.stmt_t list) 
         (implicit_type_invariants : (string, Sexpr.ltype_invariant_t) Hashtbl.t)
         : (Cexpr.stmt_t list) = ( 
    
    (*  - global_vars:          variable's name -> expression
        - function_contracts:   function's name -> function's contract
        - type_invariants:      type -> names of its invariants*)
    let global_vars : (string, Cexpr.expr_t) Hashtbl.t = 
        Hashtbl.create hashtbl_size 
    and global_ghost_vars : (string, Cexpr.expr_t) Hashtbl.t = 
        Hashtbl.create hashtbl_size 
    and function_contracts : (string, Sexpr.lspec_t) Hashtbl.t = 
        Hashtbl.create hashtbl_size 
    and type_invariants : (Cexpr.typenode_t, string) Hashtbl.t = 
        Hashtbl.create hashtbl_size in
    let modelfields = Hashtbl.copy struct_modelfields_hashtbl in
(*    read_line ();*)
    let decls = collect_global_invariants decls 
                                          (Hashtbl.create 2) in
    let decls = find_insert_function_contract decls
                             (global_ghost_vars)
                             (Hashtbl.create 2)
                             (modelfields) in
    (* let decls = process_spec_all_field_in_stmts decls in *)
    (* list os function *)
    let function_names = (
        List.flatten(
          List.map (
            fun decl ->
                match decl with
                | Cexpr.DEFUN(h, _) -> (
                    match (Ccommon.get_function_name h) with
                    | Some n -> [n]
                    | None -> []
                )
                | Cexpr.DEFVAR(h) -> (
                    match (Ccommon.get_function_name h) with
                    | Some n -> [n]
                    | None -> []
                )
                | _ -> []
        )decls)
    ) in
    (* First, collects all information about global_vars, 
       function contracts, etc *)

    (collect_implicit_type_invariant implicit_type_invariants type_invariants);
    (* *)
    let locs_of_decls : (string, Cexpr.note_t) Hashtbl.t = Hashtbl.create 10 in
    (* Check existence of construct '$result' in function contracts
       of void functions *)
    Ccommon.check_result_in_specs_of_void_funcs function_contracts decls;
    (* get information for cutlist_assumes *)
    (* find and map function pointer with real function *)
    (assign_include_function_pointer_stmts decls function_contracts);
    (*let defs = decls in*)
(*    let ghost_var_data_invariant =                                    *)
(*            make_check_data_invariant_var ((!global_invariant_names) @*)
(*                                           (!type_invariant_names)) in*)
    (* instrument data invariant (global + type) *);
    let defs = instrument_data_invariant decls type_invariant_stmts in
    (* add ghost variables for data invariant *)
    let defs = !global_invariant_stmts @ defs in
    (* add ghost variables for prohibit *)
    let defs =  ((!prohibit_ghost_vars) @ defs) in 
    (* tempolary do not process *)
(*
    let defs = (collect_local_invariant defs global_vars 
                                        global_ghost_vars) in
*)

    (* 
     * Then, does instrumenting specs
     *)
    let res = 
    List.map (
        function
        | Cexpr.DEFVAR header as decl -> (begin
            (*
             * NOTE: not so sure if this case is useful.
                     Anyways, global assignments are ignored at the moment
             *)
            match (Ccommon.get_struct_or_union_name header) with
            | Some name -> (begin
                let relevant_type_invs =
                    try Hashtbl.find_all implicit_type_invariants name
                    with 
                    | Not_found -> []
                in
                let new_stmts = (begin
                    List.fold_left (
                        fun stmts spec ->
                            stmts @ (instrument_global_spec 
                                                (Sexpr.SPEC_TYPE_INV spec)
                                                global_vars global_ghost_vars
                                                modelfields))
                            [] relevant_type_invs
                    end)
                in 
                [decl] @ new_stmts
                end)
            | None -> 
                [decl]
            end)
        | Cexpr.DEFUN (header, body) as func -> [func]
        | Cexpr.SPEC (spec) -> (
            (instrument_global_spec spec global_vars global_ghost_vars
                                    modelfields))
        | Cexpr.NOTE _ -> []
        | _ -> 
            assert false
        ) defs
    in
    (import_into_post_conditions_hashtbl function_contracts global_vars);
    (List.flatten res)
 )       
(********************************** Remove specs *******************************************)
(* removes all specifications inside programs *)

let rec remove_specs_in_stmt (s : Cexpr.stmt_t) : Cexpr.stmt_t = (
    match s with
    | Cexpr.SPEC _ -> Cexpr.NOTE(Cexpr.LINE(false, "Spec", 0, 0))               
    | Cexpr.DEFUN (d, l) ->
        let body = remove_specs_in_stmt l in
        (Cexpr.DEFUN (d, body))
    | Cexpr.BEGIN l ->
        (Cexpr.BEGIN (List.map remove_specs_in_stmt l))
    | Cexpr.IF (e0, l1, l2) ->
        (Cexpr.IF (e0, (List.map remove_specs_in_stmt l1),
                       (List.map remove_specs_in_stmt l2)))
    | Cexpr.SWITCH (e0, l1) ->     
        (Cexpr.SWITCH (e0, (List.map remove_specs_in_stmt l1)))
    | Cexpr.DOWHILE (b, e0, l1) ->
        (Cexpr.DOWHILE (b, e0, (List.map remove_specs_in_stmt l1)))
    | Cexpr.FOR (i0, e1, e2, l3) ->
        (Cexpr.FOR ((List.map remove_specs_in_stmt i0),
                    e1, e2, 
                    (List.map remove_specs_in_stmt l3)))
    | _ -> s)

let remove_specs (defs : Cexpr.stmt_t list) : Cexpr.stmt_t list = (
    let defs = List.filter (fun decl -> match decl with
                             | Cexpr.SPEC _ -> false
                             | _ -> true 
                           ) defs
    in
    List.map remove_specs_in_stmt defs
)
(*****************************************************************************)
(* replace function pointer by real function - call after flatten *)
let rec remove_function_pointer_called_stmts (stmts : Cexpr.stmt_t list)
                                             : Cexpr.stmt_t list = (
     List.flatten( List.map remove_function_pointer_called_stmt stmts)
)
and remove_function_pointer_called_stmt (stmt : Cexpr.stmt_t) 
                                        : Cexpr.stmt_t list = (
    match stmt with
    | Cexpr.GOTO _
    | Cexpr.LABEL _
    | Cexpr.IF _
    | Cexpr.ASGN _ 
    | Cexpr.STMT _
    | Cexpr.NOTE _
    | Cexpr.DEFVAR _ -> [stmt]
    | Cexpr.ANNOTATION _ -> [stmt]
    | Cexpr.BEGIN stmts -> (
        [Cexpr.BEGIN(remove_function_pointer_called_stmts stmts)]
    )
    | Cexpr.DEFUN (header, body) -> (
        let new_body = 
            match (remove_function_pointer_called_stmt body) with
            | [Cexpr.BEGIN (stmts)] -> Cexpr.BEGIN (stmts)
            | _ -> assert false
        in
        [Cexpr.DEFUN(header, new_body)]
    )
    | Cexpr.CALLASGN (Cexpr.CALL (ty, expr, exs, spec_op), stmts, te) -> (
        match (Ccommon.left_expr_is_field expr),
              (Ccommon.left_expr_is_function_pointer expr) with
        | Some (_, fp_f_name, var_name) , None -> (
            try 
                let ex_new, _ = 
                    Hashtbl.find fp_to_f_hashtbl (Some var_name, fp_f_name) in
                let new_ex = Cexpr.CALL(ty, ex_new, exs, spec_op) in
                let stmts_new = remove_function_pointer_called_stmts stmts in
                [Cexpr.CALLASGN(new_ex, stmts_new, te)]
            with
            | Not_found -> [stmt]
        )
        | None, Some (fp_name) -> (
            try 
                let ex_new, _ = 
                    Hashtbl.find fp_to_f_hashtbl (None, fp_name) in
                let new_ex = Cexpr.CALL(ty, ex_new, exs, spec_op) in
                let stmts_new = remove_function_pointer_called_stmts stmts in
                [Cexpr.CALLASGN(new_ex, stmts_new, te)]
            with
            | Not_found -> [stmt]
        )
        | None, None -> [stmt]
        | _ -> assert false
    )
    | _ -> ((Cexpr.dump_stmts [stmt]);assert false)
)
(*****************************************************************************)
(* Create new var and assign caller to this var. Then insert pre-condition
   before this new var and post-condition after this new var *)

(** Insert pre-condition before call function and post-condition
    after call function
        @param func_conds : function conditions to insert
        @param stmt_caller : function caller (Cexpr.stmt_t type)
        @param new_var : variable is assigned to function caller 
*)
let insert_func_cont_into_caller
        (func_cont : ('pred, 'locs) Sexpr.function_contract_t)
        (stmt_caller : Cexpr.stmt_t)
        (new_var : Cexpr.expr_t)
: (Cexpr.stmt_t list) =
begin
   let old_stmts = Ccommon.make_stmt_old_expr_from_funct_cont func_cont in
   
   (* replace all old-vars *) 
   let replaced_old_func_cont =
       Ccommon.replace_old_term_in_function_contract func_cont in

   (* DODO Now only process for function conditions of function contract *)
   let func_conds = replaced_old_func_cont.Sexpr.fspec_conditions in
   let ((pre_conds, post_conds) : ((Cexpr.stmt_t list) * (Cexpr.stmt_t list))) = 
       List.fold_left
           (fun (curr_pre_conds, curr_post_conds) func_cond ->
		           let (fun_type, full_expr) = func_cond in
               let expr = full_expr.Cexpr.expr_node in
               (* Replace $result by new_var *)
               let result_expr_op = Cexpr.get_result_expr expr in
               let new_expr =
                   match result_expr_op with
                   | None -> expr
                   | Some re_ex ->
                       Ccommon.replace_expr_by_expr [re_ex]
                                                    [new_var]
                                                    expr in

		           let stmts = Stranslate.make_list_assert [new_expr] in
               if (fun_type == Sexpr.PRE_CONDITION) then
                   ((curr_pre_conds @ stmts), curr_post_conds)
               else (curr_pre_conds, (curr_post_conds @ stmts))
       ) ([], []) func_conds in
    (old_stmts @ pre_conds @ [stmt_caller] @ post_conds)
end

let rec make_stmt_caller_stmt (current_stmt : Cexpr.stmt_t)
: ((Cexpr.stmt_t list) * Cexpr.stmt_t) =
begin
    match current_stmt with
    | Cexpr.DEFUN (decl, st) -> (
        let (stmt_vars, new_stmt) = make_stmt_caller_stmt st in
        (stmt_vars, Cexpr.DEFUN (decl, new_stmt))
    )
    | Cexpr.DEFVAR _
    | Cexpr.BREAK
    | Cexpr.GOTO _
    | Cexpr.NOTE _
    | Cexpr.SPEC _
    | Cexpr.CONTINUE ->
        ([], current_stmt)
    | Cexpr.STMT ex -> (
        let (stmt_vars, new_expr) = make_stmt_caller_expr ex in 
        (stmt_vars, Cexpr.STMT new_expr)
    )
    | Cexpr.ASGN (ex1, ex2) -> (
        let (stmt_vars_1, new_expr_1) = make_stmt_caller_expr ex1 in
        let (stmt_vars_2, new_expr_2) = make_stmt_caller_expr ex2 in
        ((stmt_vars_1 @ stmt_vars_2), Cexpr.ASGN (new_expr_1, new_expr_2))
    )
    | Cexpr.CALLASGN (ex, sts, exs) -> (
        let (stmt_vars_1, new_ex) = make_stmt_caller_expr ex in
        let new_sts = make_stmt_caller_stmts sts in
        let (stmt_vars_2, new_exs) = make_stmt_caller_exprs exs in
        ((stmt_vars_1 @ stmt_vars_2), Cexpr.CALLASGN (new_ex, new_sts, new_exs))
    )
    | Cexpr.BEGIN sts -> (
        let new_sts = make_stmt_caller_stmts sts in
        ([], Cexpr.BEGIN new_sts)
    )
    | Cexpr.IF (ex, sts_1, sts_2) -> (
        let (stmt_vars, new_ex) = make_stmt_caller_expr ex in
        let new_sts_1 = make_stmt_caller_stmts sts_1 in
        let new_sts_2 = make_stmt_caller_stmts sts_2 in
        (stmt_vars, Cexpr.IF (new_ex, new_sts_1, new_sts_2))
    )

    | Cexpr.SWITCH (ex, sts) -> (
        let (stmt_vars, new_ex) = make_stmt_caller_expr ex in
        let new_sts = make_stmt_caller_stmts sts in
        (stmt_vars, Cexpr.SWITCH (new_ex, new_sts))
    )
    | Cexpr.DOWHILE (bo, ex, sts) -> (
        let (stmt_vars, new_ex) = make_stmt_caller_expr ex in
        let new_sts = make_stmt_caller_stmts sts in
        (stmt_vars, Cexpr.DOWHILE (bo, new_ex, new_sts))
    )
    | Cexpr.FOR (sts_1, ex1, ex2, sts_2) -> (
        let new_sts_1 = make_stmt_caller_stmts sts_1 in
        let (stmt_vars_1, new_ex1) = make_stmt_caller_expr ex1 in
        let (stmt_vars_2, new_ex2) = make_stmt_caller_expr ex2 in
        let new_sts_2 = make_stmt_caller_stmts sts_2 in
        ((stmt_vars_1 @ stmt_vars_2), Cexpr.FOR (new_sts_1, new_ex1, new_ex2, new_sts_2))
    )
    | Cexpr.RETURN ex_op -> (
        match ex_op with
        | None -> ([], current_stmt)
        | Some ex -> (
            let (stmt_vars, new_ex) = make_stmt_caller_expr ex in 
            (stmt_vars, Cexpr.RETURN (Some new_ex))
        )
    )
    | Cexpr.ANNOTATION (str, ex) -> (
        let (stmt_vars, new_ex) = make_stmt_caller_expr ex in
        (stmt_vars, Cexpr.ANNOTATION (str, new_ex))
    )
    | Cexpr.ASM ex -> (
        let (stmt_vars, new_ex) = make_stmt_caller_expr ex in
        (stmt_vars, Cexpr.ASM new_ex)
    )
    | Cexpr.ASSUME ex -> (
        let (stmt_vars, new_ex) = make_stmt_caller_expr ex in
        (stmt_vars, Cexpr.ASSUME new_ex)
    )
    | Cexpr.RESET exs -> (
        let (stmt_vars, new_exs) = make_stmt_caller_exprs exs in
        (stmt_vars, Cexpr.RESET new_exs)
    )    
    | Cexpr.LABEL label -> (
        match label with
        | Cexpr.NAME_LABEL _
        | Cexpr.DFLT_LABEL -> ([], current_stmt)
        | Cexpr.CASE_LABEL (ex1, ex2) -> (
            let (stmt_vars_1, new_ex1) = make_stmt_caller_expr ex1 in
            let (stmt_vars_2, new_ex2) = make_stmt_caller_expr ex2 in
            ((stmt_vars_1 @ stmt_vars_2), Cexpr.LABEL (Cexpr.CASE_LABEL (new_ex1, new_ex2)))
        )
    ) 
end

and make_stmt_caller_stmts (sts : Cexpr.stmt_t list) : (Cexpr.stmt_t list) =
begin
    match sts with
    | [] -> []
    | head :: tails -> (
        let (stmt_vars, new_stmt) = make_stmt_caller_stmt head in
        (stmt_vars @ [new_stmt] @ (make_stmt_caller_stmts tails))
    )
end

(** Create stmt from caller and insert function conditions for caller
        1. Create a new variable and assign it to function caller
        2. Insert function condition list for new varable (is correspondingly
           function caller)
*)
and make_stmt_caller_expr (current_expr : Cexpr.expr_t)
: ((Cexpr.stmt_t list) * (Cexpr.expr_t)) =
begin
    match current_expr with
    | Cexpr.CALL (ty, func_var, args, spec_op) -> (
        let (sts, _) = make_stmt_caller_exprs args in
        let (inserted_stmts, new_var_op) =
            match spec_op with
            | None -> ([], None)
            | Some spec -> (
                let ty_op = Ccommon.get_function_type ty in
                let func_ty =
                    match ty_op with
                    | Some tyn ->
                        Ccommon.get_return_function_type tyn
                    | None -> ty
                in
                let new_var = (Cexpr.VREF (func_ty, (Cexpr.gensym "V"), 3, 0)) in
                let var_name = Ccommon.get_name_of_expr new_var in
                let init_expr = Cexpr.SETF (ty, Cexpr.SET, new_var, current_expr) in
                let (decl : Cexpr.declaration_t) =
                  {
                    Cexpr.decl_id = (var_name, 3);
                    Cexpr.decl_tn = func_ty;
                    Cexpr.decl_scls = Cexpr.NILSCLS;
                    Cexpr.decl_inits = [init_expr];
                    Cexpr.decl_params = [];
                    Cexpr.decl_struct_defined_p = false;
                    Cexpr.decl_specs = None;
                    Cexpr.decl_location = None;
                  } in
                let new_stmt = Cexpr.DEFVAR (decl) in
                let func_cont = (
                    match spec with
                    | Sexpr.SPEC_FUN f_cont -> f_cont
                    | _ -> assert false) in
                (insert_func_cont_into_caller func_cont new_stmt new_var,
                 Some new_var)) in
        match new_var_op with
        | None -> ((sts @ inserted_stmts), current_expr)
        | Some new_var -> ((sts @ inserted_stmts), new_var)
    )
    | Cexpr.NIL
    | Cexpr.VREF _
    | Cexpr.SIZEOF _
    | Cexpr.CONST_I _
    | Cexpr.CONST_F _
    | Cexpr.STRING _
    | Cexpr.PSEUDO_MEMORY _
    | Cexpr.PRAGMA _
    | Cexpr.PSEUDO_PROLOGUE
    | Cexpr.PSEUDO_EPILOGUE
    | Cexpr.PSEUDO_NONDET_VALUE _ -> ([], current_expr)
    | Cexpr.OPER (ty, op, expr1, expr2, expr3) -> (
        let (sts_1, new_expr1) = make_stmt_caller_expr expr1 in
        let (sts_2, new_expr2) = make_stmt_caller_expr expr2 in
        let (sts_3, new_expr3) = make_stmt_caller_expr expr3 in
        (sts_1 @ sts_2 @ sts_3, Cexpr.OPER (ty, op, new_expr1, new_expr2, new_expr3))
    )
    | Cexpr.SETF (ty, op, expr1, expr2) -> (
        let (sts_1, new_expr1) = make_stmt_caller_expr expr1 in
        let (sts_2, new_expr2) = make_stmt_caller_expr expr2 in
        (sts_1 @ sts_2, Cexpr.SETF (ty, op, new_expr1, new_expr2))
    )
    | Cexpr.PREF (ty, bo, expr, expr1) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.PREF (ty, bo, new_expr, expr1))
    )
    | Cexpr.SREF (ty, bo, expr, str) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.SREF (ty, bo, new_expr, str))
    )
    | Cexpr.DSGN_AREF (ty, expr1, expr2) -> (
        let (sts_1, new_expr1) = make_stmt_caller_expr expr1 in
        let (sts_2, new_expr2) = make_stmt_caller_expr expr2 in
        (sts_1 @ sts_2, Cexpr.DSGN_AREF (ty, new_expr1, new_expr2))
    )
    | Cexpr.ADDROF (ty, expr) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.ADDROF (ty, new_expr))
    )
    | Cexpr.PROGN (ty, sts) -> (* DODO process sts (Cexpr.stmt_t list) later *)
        ([], current_expr)
    | Cexpr.CLITERAL (ty, init) -> (
        let (sts, new_init) = make_stmt_caller_init init in
        (sts, Cexpr.CLITERAL (ty, new_init))
    )
    | Cexpr.ASMSEQUENCE exprs -> (
        let (sts, new_exprs) = make_stmt_caller_exprs exprs in
        (sts, Cexpr.ASMSEQUENCE new_exprs)
    )
    | Cexpr.PHI (ty, phi_lst) -> (
        let tmp =
            List.map
                (fun a_phi ->
                    let (str, expr) = a_phi in
                    let (sts, new_expr) = make_stmt_caller_expr expr in
                    (sts, (str, new_expr))
                )
                phi_lst
        in
        let (sts, new_phi_lst) = List.split tmp in
        (List.flatten sts, Cexpr.PHI (ty, new_phi_lst))
    )
    | Cexpr.INIT_FILL (ty, expr) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.INIT_FILL (ty, new_expr))
    )
    | Cexpr.QUANTIFIER (str, exprs, expr) -> (
        let (sts_1, new_exprs) = make_stmt_caller_exprs exprs in
        let (sts_2, new_expr) = make_stmt_caller_expr expr in
        (sts_1 @ sts_2, Cexpr.QUANTIFIER (str, new_exprs, new_expr))
    )
    | Cexpr.PSEUDO_ACSL_RESULT (ty, expr) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.PSEUDO_ACSL_RESULT (ty, new_expr))
    )
    | Cexpr.PSEUDO_ACSL_OLD (ty, expr1, expr2) -> (
        let (sts_1, new_expr1) = make_stmt_caller_expr expr1 in
        let (sts_2, new_expr2) = make_stmt_caller_expr expr2 in
        (sts_1 @ sts_2, Cexpr.PSEUDO_ACSL_OLD (ty, new_expr1, new_expr2))
    )
    | Cexpr.PSEUDO_ACSL_UNIQUE (ty, expr) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.PSEUDO_ACSL_UNIQUE (ty, new_expr))
    )
    | Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (ty, expr1, expr2) -> (
        let (sts_1, new_expr1) = make_stmt_caller_expr expr1 in
        let (sts_2, new_expr2) = make_stmt_caller_expr expr2 in
        (sts_1 @ sts_2, Cexpr.PSEUDO_ACSL_UNIQUEBLOCK (ty, new_expr1, new_expr2))
    )
    | Cexpr.PSEUDO_ACSL_VALID (ty, expr) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.PSEUDO_ACSL_VALID (ty, new_expr))
    )
    | Cexpr.ALL_FIELD (ty, expr) -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.ALL_FIELD (ty, new_expr))
    )
end

and make_stmt_caller_exprs (exs : Cexpr.expr_t list)
: (Cexpr.stmt_t list * Cexpr.expr_t list) =
begin
    let tmp =
        List.map
            (fun a_ex ->
                make_stmt_caller_expr a_ex 
            ) exs in
    let (stss, new_exs) = List.split tmp in
    (List.flatten stss, new_exs)
end

and make_stmt_caller_init (init : Cexpr.initializers_t)
: (Cexpr.stmt_t list * Cexpr.initializers_t) =
begin
    match init with
    | Cexpr.INIT_EXPR expr -> (
        let (sts, new_expr) = make_stmt_caller_expr expr in
        (sts, Cexpr.INIT_EXPR (new_expr))
    )
    | Cexpr.INIT_LIST inits -> (
        let tmp =
            List.map
                (fun a_init ->
                    make_stmt_caller_init a_init
                ) inits
        in
        let (stss, new_inits) = List.split tmp in
        (List.flatten stss, (Cexpr.INIT_LIST new_inits))
    )
    | Cexpr.INIT_DSGN (exprs, init) -> (
        let (sts_1, new_exprs) = make_stmt_caller_exprs exprs in
        let (sts_2, new_init) = make_stmt_caller_init init in
        (sts_1 @ sts_2, Cexpr.INIT_DSGN (new_exprs, new_init))
    )
end

and make_stmt_all_caller (ofiles : Cexpr.ofile_t list) : Cexpr.ofile_t list =
begin
    List.map 
        (fun a_ofile ->
            let new_ofile_defs =
                List.map
                    (fun a_def ->
                        let (name_def, def) = a_def in
                        let (stmt_vars, new_def) = make_stmt_caller_stmt def in
                        assert ((List.length stmt_vars) == 0);
                        (name_def, new_def) 
                    )
                    !(a_ofile.Cexpr.ofile_defs)
            in {a_ofile with Cexpr.ofile_defs = ref new_ofile_defs}                
        ) ofiles
end

(*****************************************************************************)
(* insert pre_main_function which call to main_function *)
(** is_ofile
    @param ofile : Cexpr.ofile_t list
    @param file_loc : Sexpr.file_location_t
*)
let rec is_ofile (ofile : Cexpr.ofile_t)
                 (file_loc : Sexpr.file_location_t)
: bool = (
    let string_of_file_loc = Sexpr.string_of_file_location file_loc in
    let current_ofile_loc =
        Sexpr.make_file_location ofile.Cexpr.ofile_file_name in
    if (String.compare (Sexpr.string_of_file_location current_ofile_loc)
                       string_of_file_loc) == 0 then
        true
    else false
)

(* Name of pre_main_func which calls to main function *)
let prefix_of_pre_main = "pre_main_" ^ "_of_"

let pre_main_func_id : string = (
    Random.self_init ();
    string_of_int (Random.bits ()))

let make_pre_main_func_name (func_main_name : string) : string = (
    prefix_of_pre_main ^ func_main_name
)

(** Make a call to a function
    @param header : declaration of called function
    @return expr of function call
*)
let make_call_to_func (header : Cexpr.declaration_t)
: Cexpr.expr_t = (
    let func_type = header.Cexpr.decl_tn in
    let return_type = match Ccommon.get_function_type func_type with
                      | Some typ -> typ
                      | _ -> assert false in
    let func_name, func_level = header.Cexpr.decl_id in
    let func_expr = Cexpr.VREF (return_type, func_name, func_level, 0) in
    let param_names = header.Cexpr.decl_params in
    let param_types = get_param_types func_type in
    let param_exprs =
        (try 
            List.fold_left2 (
                fun param_exprs_ name tyn ->
                    let expr = Cexpr.VREF(tyn, name, 2, 0) in
                    (param_exprs_ @ [expr])
            ) [] param_names param_types;
        with Invalid_argument _ -> assert false)
    in
    (Cexpr.make_function_call func_expr param_exprs);
)

(** Used for inserting a function calling to main func
*)
let insert_pre_main_func (ofile : Cexpr.ofile_t)
                         (main_func_name : string)
: Cexpr.ofile_t = (
    (* file functions havings name as in main_entry *)
    let main_funcs =
        Ccommon.find_called_stmts_in_ofile ofile
                                           main_func_name
    in
    match main_funcs with
    | [] -> ofile
    | main_func :: _ -> (
        (* get a main_fun and create a func call to main func *)
        let decl = match main_func with
                   | Cexpr.DEFUN (decl, _)
                   | Cexpr.DEFVAR (decl) -> decl
                   | _ -> assert false
        in
        let pre_main_func_name = make_pre_main_func_name main_func_name in
        let decl_id_of_pre_main_func =
            (pre_main_func_name, snd decl.Cexpr.decl_id)
        in
        let decl_of_pre_main_func =
            {decl with Cexpr.decl_id = decl_id_of_pre_main_func} in
        let expr_of_pre_main_func = decl_of_pre_main_func.Cexpr.decl_id in
        let func_calling_to_main =
            make_call_to_func decl in
        let return_type =
            match Ccommon.get_function_type decl_of_pre_main_func.Cexpr.decl_tn with
            | Some typ -> typ
            | _ -> assert false
        in
        let body_of_pre_main_func =
            Cexpr.BEGIN [(Cexpr.RETURN (Some func_calling_to_main))]
        in 
        let stmt = Cexpr.DEFUN (decl_of_pre_main_func, body_of_pre_main_func) in
        let stmts = Ccommon.normalize_return return_type stmt in
        match stmts with
        | [stmt] -> (
            ofile.Cexpr.ofile_defs := !(ofile.Cexpr.ofile_defs)
                                      @ [pre_main_func_name, stmt];
            ofile)
        | _ -> assert false
    )
)

(** Insert a funtion which calls to main function
    @param ofiles : Cexpr.ofile_t list
    @param main_entry : Sexpr.function_entry_t of main function
    @return (Cexpr.ofile_t list * (ofile inserted func calling to main func)
             Sexpr.function_entry_t (entry of func calling to main func))
*)
let insert_pre_main_func (ofiles : Cexpr.ofile_t list)
                         (main_entry : Sexpr.function_entry_t)
: Cexpr.ofile_t list * Sexpr.function_entry_t = (
    let new_ofiles =
        List.map (fun ofile ->
            match main_entry.Sexpr.fe_filelocation_opt with
            | Some file_loc -> (
                if is_ofile ofile file_loc then (
                    insert_pre_main_func ofile main_entry.Sexpr.fe_funcname
                )
                else ofile
              )
            | None -> insert_pre_main_func ofile main_entry.Sexpr.fe_funcname
        ) ofiles
    in
    let pre_main_func_name =
        make_pre_main_func_name main_entry.Sexpr.fe_funcname
    in
    let pre_main_entry =
      { main_entry with Sexpr.fe_funcname = pre_main_func_name }
    in
    (new_ofiles, pre_main_entry)
)

(** Insert functions that call to main functions
    @param ofiles : Cexpr.ofile_t list
    @param main_entries : Sexpr.function_entry_t vector of main funcs
    @return (Cexpr.ofile_t list * (ofiles inserted funcs calling to main funcs)
             Sexpr.function_entry_t Vector (vector entry of funcs calling to main funcs))
*)
let insert_pre_main_funcs (ofiles : Cexpr.ofile_t list)
                          (main_entries : Sexpr.function_entry_t Vector.t)
: Cexpr.ofile_t list * Sexpr.function_entry_t Vector.t = (
    List.fold_left (fun (new_ofiles, pre_main_entries) main_entry ->
        let new_ofiles_, pre_main_entry_ =
            insert_pre_main_func new_ofiles main_entry in
        Vector.add pre_main_entries pre_main_entry_;
        (new_ofiles_, pre_main_entries)
    ) (ofiles, Vector.create main_entries.Vector.size
                             Sexpr.nil_function_entry)
      (Vector.to_list main_entries)
)

(*****************************************************************************)
(* insert assume clauses after called function in cut-list - call after flatten *)
let rec insert_assume (stmt : Cexpr.stmt_t) : Cexpr.stmt_t list = (
    match stmt with
    | Cexpr.GOTO _
    | Cexpr.LABEL _
    | Cexpr.IF _
    | Cexpr.ASGN _ 
    | Cexpr.STMT _
    | Cexpr.NOTE _
    | Cexpr.DEFVAR _ -> [stmt]
    | Cexpr.BEGIN stmts -> [Cexpr.BEGIN(insert_assumes stmts)]
    | Cexpr.DEFUN (header, body) -> (
        let new_body = 
            match (insert_assume body) with
            | [Cexpr.BEGIN (stmts)] -> Cexpr.BEGIN (stmts)
            | _ -> assert false
        in
        [Cexpr.DEFUN(header, new_body)]
    )
    | Cexpr.CALLASGN (Cexpr.CALL (te, ex, exprs, spec_op), stmts, _) -> (begin
        let fun_name = Ccommon.get_name_of_expr(
                           Ccommon.get_deep_expr_from_expr ex) in
(*         let boo = (List.mem fun_name (!Scommon.except_functions)) in *)
        let boo = false in
        if ((boo == true)) then (begin
            let curr_exprs = (
                try 
                    List.flatten (
                        Hashtbl.find_all (Ccommon.post_condition_hashtbl)
                                         fun_name)
                with
                | Not_found -> (
                    Scommon.emerg ("Function " ^ fun_name ^ " has no post-condition");
                    assert false
                 )
            )in
            let param_exprs =  exprs in
            let ret_expr = (
                match stmts with
                | head::tail -> (
                    match head with 
                    | Cexpr.ASGN(ex1, _) -> ex1
                    | Cexpr.STMT(Cexpr.NIL) -> (Cexpr.NIL)
                    | _ -> assert false
                )
                | [] -> (assert false)
            )in
            let header = (
                try
                    Hashtbl.find fun_header_infos_hashtbl fun_name
                with
                | Not_found -> assert false
            )in
            let param_names = header.Cexpr.decl_params in
            let fun_type = header.Cexpr.decl_tn in            
            let param_types = get_param_types fun_type in
            let param_expr_orgs = (
                List.map2(
                    fun name ty ->
                        Cexpr.VREF(ty, name, 2, 0)
                )param_names param_types
            )in
            let exprs_new = (Ccommon.replace_exprs_in_exprs curr_exprs
                                                            param_exprs 
                                                            param_expr_orgs
                                                            ret_expr)
            in

            let annos = (
                List.map(fun ex -> Cexpr.ASSUME(ex))exprs_new
            )in
            let expr_assigns = (
                try 
                    Hashtbl.find Ccommon.assign_in_func fun_name
                with 
                | Not_found -> []
            )in
            let new_expr_assigns  = (Ccommon.replace_exprs_in_exprs expr_assigns
                                                            param_exprs 
                                                            param_expr_orgs
                                                            ret_expr)
            in
            let new_stmt = Cexpr.CALLASGN (
                               Cexpr.CALL (te, ex, exprs, spec_op), 
                               stmts, 
                               new_expr_assigns) in
            ([new_stmt] @ annos)
        end)
        else ([stmt])
     end)
    | Cexpr.ANNOTATION _ -> [stmt]
    | _ -> ((Cexpr.dump_stmts [stmt]);assert false)
)
and insert_assumes (stmts : Cexpr.stmt_t list) : Cexpr.stmt_t list = (
    List.flatten( List.map insert_assume stmts)
)
(*****************************************************************************)
(* Main function of the specification reader.
     + if specs are to be instrumented, reads and instrument them
     + otherwise, remove them from code
*)
let main 
     (defs : Cexpr.stmt_t list) 
     (implicit_type_invariants : (string, Sexpr.ltype_invariant_t) Hashtbl.t)
     : (Cexpr.stmt_t list) = (
    match !Scommon.process_specs with
    | true -> 
        (collect_header defs);
        (Ccommon.get_note_of_entry_function defs);
        instrument_main defs implicit_type_invariants
    | false ->
        (remove_specs defs)
)

(** Insert pre_main_funcs that call to main funcs
    Insert specs of vars
    Insert func contracts in caller side
    @param ofiles : Cexpr.ofile_t list
    @param main_entries : Vector Sexpr.function_entry_t of main funcs
    @return Cexpr.ofile_t list * Sexpr.function_entry_t Vector.t
*)
let process_ofiles (stmt_ofiles : Cexpr.ofile_t list)
                   (main_entries : Sexpr.function_entry_t Vector.t)
: Cexpr.ofile_t list * Sexpr.function_entry_t Vector.t = (
    let stmt_ofiles_inserted_pre_main_func, new_opt_entries =
        insert_pre_main_funcs stmt_ofiles
                              main_entries
    in
    let stmt_ofiles_inserted_var_specs =
        find_insert_specs_of_var stmt_ofiles_inserted_pre_main_func
    in
    let processed_stmt_ofiles_1 =
        Ccommon.insert_conds_when_calling_func stmt_ofiles_inserted_var_specs
    in
    let processed_stmt_ofiles_2 =
        make_stmt_all_caller processed_stmt_ofiles_1 in
    (Ccommon.check_standard_quantity_func_cont_of_a_func
        processed_stmt_ofiles_2);
    (processed_stmt_ofiles_2, new_opt_entries)
)
