(* slicing/constructACFG.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num
open Slicingstructure

module Sat = Minisat

(*================ Global Variables ====================*)

let number_func = ref 0

(* thu tu cau lenh Goto hien tai *)
let number_goto = ref 0
(*
 thu tu cau lenh hien tai *)
let number_statement = ref 0

(* thu tu Label hien tai *)
let number_label = ref 0

(* thu tu NOTE hien tai *)
let num_note = ref 0

(* thu tu cau lenh Return hien tai *)
let number_return = ref 0

(* thu tu canh ACFG_CDG hien tai *)
let number_edge = ref 0


(* thu tu cau lenh assign hien tai *)
let number_assign = ref 0

(* thu tu cau lenh assign (them vao) hien tai *)
let number_v_assign = ref 0

(* thu tu cau lenh if hien tai *)
let number_if = ref 0


(* thong tin ve stmt *)
let stats = Hashtbl.create 1000

(* thong tin ve function header *)
let funcs = Hashtbl.create 1000


(* ds Node ACFG *)
let list_node_acfg = ref ([] : node list)

(* ds Node ACFG dung cho xay dung DDG *)
let list_node_acfg_for_ddg = ref ([] : node list)

(* ds Edge ACFG  *)
let list_edge_acfg  = ref ([] : edge list)

(* ds Edge ACFG dunf cho xay dung DDG *)
let list_edge_acfg_for_ddg  = ref ([] : edge list)

(* Thong tin ve Note*)
let note_infos = Hashtbl.create 1000


(* Ten ham hien tai *)
let current_function_name = ref ""

(* ds bien global *)
let list_global_variable = ref []

(*Danh sach cac cap vertex va edge ung voi moi ham - dung cho xay dung CDG *)
let list_couple_acfg = ref ([] : graph_type_slicing list )

(*Danh sach cac cap vertex va edge ung voi moi ham - dung cho xay dung DDG *)
let list_couple_acfg_for_ddg = ref ([] : graph_type_slicing list )

(*Danh sach input cho slicing *)
let list_input_slicing = ref []

(* Kiem tra co phai ham main khong? *)
let is_main_function = ref false


(* ds ten ham *)
let list_function = ref []

(* danh sach canh lien ket giua cac ham *)
let list_edge_connect = ref ([] : edge list)

(* Nodes toan chuong trinh *)
let list_node_all = ref ([] : node list )

let list_couple_all = ref ([] : graph_type_slicing list )

(* danh sach tham so cua ham hien tai *)
let list_curr_param = ref [] 

(* Thong tin ve Gen *)
let htbl_gen_infos = Hashtbl.create 1000

(* Thong tin ve Use *)
let htbl_use_infos = Hashtbl.create 1000

(* Thong tin ve Stmt - gan ten node voi statement *)
let htbl_statement_infos = Hashtbl.create 1000

(* danh sach statement dung xay dung DDG *)
let list_statement_for_ddg = ref []

(* Current function's node for each function *)
let current_function_node = ref (NodeV("","",NODE_OTHER))

(* dang sach cac ham co the link den node error *)
let list_fun_contain_error_label = ref ([] : string list)
(*========================== construct Augment Control Flow Graph(ACFG) for program ================================*)



(*In kieu cua node*)
let print_t (t : edge_type) = 
   match t with 
    | ACFG_FALSE -> print_string " : ACFG False : "
    | ACFG_TRUE -> print_string " : ACFG True : "
    | PDT -> print_string " : PDT : "
    | CDG -> print_string " : CDG : "
    | DDG -> print_string " : DDG : "
    | _ -> print_string " : Other : "

let print_type couble  = 
   let (vers,edges) = couble in
   print_edges edges

let rec print_type_edges list_couple = 
   match list_couple with 
    | [] -> ()
    | couble::rest -> print_type couble;print_type_edges rest

let string_of_type_node (node_type : type_node) = 
   match node_type with
    | NODE_STMT -> "stmt"
    | NODE_IF -> "if"
    | NODE_ASGN -> "assign"
    | NODE_LABEL -> "label"
    | NODE_GOTO -> "goto"
    | NODE_OTHER -> "other"
    | _ -> assert false

(* Dua ra ten kieu cua node*)
let find_node_type (stmt : Cexpr.stmt_t) = 
   match stmt with 
    | Cexpr.STMT(_) -> NODE_STMT
    | Cexpr.IF(_,_,_) -> NODE_IF
    | Cexpr.ASGN(_,_) -> NODE_ASGN
    | Cexpr.LABEL(_) -> NODE_LABEL
    | Cexpr.GOTO(_) -> NODE_GOTO
    | _ -> NODE_OTHER

(*Kiem tra bang nhau cua 2 node *)
let is_equal_nodes (node1 : node) (node2 : node) = 
   match node1,node2 with 
    | NodeV(str1,str2,t1),NodeV(str3,str4,t2) -> (
        if (str1 = str3) && (str2 = str4) && (t1 = t2) then true
        else false
        )
    | _ -> assert false

(* Dua ra thong tin ve Note cua moi node *)
let find_note_infos (stmt : Cexpr.stmt_t) = 
   let res = ref "" in
   Hashtbl.iter (fun key value -> if value = stmt then res := key) note_infos;
   !res

(*Import thong tin vao Hashtable chua function header*)   
let import_into_functions (key : string) (value : Cexpr.declaration_t) = 
   if Hashtbl.mem funcs key then ()
   else Hashtbl.add funcs key value

(*Import thong tin vao Hashtable chua stmt*)
let import_into_hash_stats (key:string) (value : Cexpr.stmt_t) = 
   if Hashtbl.mem stats key then ()
   else Hashtbl.add stats key value

(* Import thong tin vao htbl_statement_info *)
let import_into_statement_info (name : string) (stmt : Cexpr.stmt_t) = 
   Hashtbl.replace htbl_statement_infos (name) (stmt,!number_statement);
   number_statement := !number_statement + 1

(*In stmt*)
let print_stmt (stmt : Cexpr.stmt_t) = 
   match stmt with
     |Cexpr.LABEL(label_type) -> (match label_type with 
                                   |Cexpr.NAME_LABEL(label) -> print_string ("Label:"^label^"\n")
                                   | _ -> print_string ("Label:\n")
                                 )
     |Cexpr.GOTO(goto) -> print_string ("Goto:"^goto^"\n")
     |Cexpr.ASGN(_,_) -> print_string "Assign:\n"
     |Cexpr.STMT(_) -> print_string "STMT:\n"
     |Cexpr.IF(_,_,_) -> print_string "if:\n"
     |Cexpr.NOTE(tf,str,num1,num2) -> print_string ("note:"^str^(string_of_int num1)^"\n")
     | _ -> print_string "other:\n"

(*In danh sach cac stmt*) 
let rec print_stmts (stmts : Cexpr.stmt_t list) = 
   match stmts with 
    | [] -> ()
    | stmt::rest -> print_stmt stmt;print_stmts rest

(*In cac declaration *) 
let print_decl ( decl : Cexpr.declaration_t ) = 
   let (func_name, func_level) = decl.Cexpr.decl_id in
   print_string (func_name^"\n")

(* In cac definition *)
let print_def ( def : Cexpr.stmt_t ) = 
   match def with
    | Cexpr.DEFVAR (decl) -> print_decl decl
    | Cexpr.DEFUN (decl, stmts) -> print_string "\nHam: "; 
                                   print_decl decl ; 
                                   print_string " co: "; print_int (List.length stmts); print_string " stats\n";
                                   print_stmts stmts
    | _ ->  assert false

(*In thong tin trong hashtable chua cac stmt*)
let print_hash () = 
   Hashtbl.iter (fun value key -> print_string (value^"-->"); print_stmt key; print_string "\n") stats


(* Dua thong tin vao USE *)
let import_into_use (key : string) (value : string list) = 
   Hashtbl.replace htbl_use_infos key value

(* Dua thong tin vao GEN *)
let import_into_gen (key : string) (value : couple_node_var list) = 
   Hashtbl.replace htbl_gen_infos key value

(* Kiem tra 2 canh co bang nhau khong*)
let edge_equal (e1 : edge) (e2 : edge) = 
   let Edge(s1,d1,t1,n1), Edge(s2,d2,t2,n2) = e1, e2 in
   if (s1 = s2) && (d1 = d2) then true
   else false

let check_mem_edges (e : edge) = 
   let res = ref false in
   List.iter (fun e1 -> if edge_equal e e1 then res := true) (!list_edge_acfg);
   ();
   !res
   
(*Tao edge cho ACFG*)
let make_edge_acfg (source : node) (dest : node) (type_edge : edge_type)=
   let e = Edge(source, dest,type_edge,!number_edge) in
   if (check_mem_edges e) = true then ()
   else begin 
           list_edge_acfg := e::!list_edge_acfg;
           list_edge_acfg_for_ddg := e::!list_edge_acfg_for_ddg;           
           number_edge := !number_edge+1
        end 									

(*Tao edge cho ACFG dung cho xay dung DDG *)
let make_edge_acfg_for_ddg (source : node) (dest : node) (type_edge : edge_type)=
    list_edge_acfg_for_ddg := Edge(source, dest, type_edge, 0)::!list_edge_acfg_for_ddg           

(*Tao node cho ACFG *)
let insert_node_acfg (n : node) = 
	if List.mem n !list_node_acfg then ()
        else 
           begin
              list_node_acfg := n::!list_node_acfg;
              list_node_acfg_for_ddg := n::!list_node_acfg_for_ddg
           end


(*Tao node cho ACFG dung cho xay dung DDG *)
let insert_node_acfg_for_ddg (n : node) = 
	if List.mem n !list_node_acfg_for_ddg then ()
        else list_node_acfg_for_ddg := n::!list_node_acfg_for_ddg

(*=====================================================*)

(*Tao stmt ASGN cho mot parameter hoac global variable*)
let make_assign (str : string)  =
   Cexpr.ASGN(Cexpr.VREF(Cexpr.INTEGER,str,0,0),Cexpr.VREF(Cexpr.INTEGER,str,0,0))

(*Tao stmt ASGN cho mot list parameters hoac global variables*)
let rec make_assigns (strs : string list) = 
   match strs with 
    | [] -> []
    | str::rest -> (make_assign str)::(make_assigns rest)

(*=====================================================*)

(*Lay ten cua bien dc truyen vao ham *)
let get_name_param (e : Cexpr.expr_t) = 
   match e with 
     | Cexpr.VREF(t,str,n1,n2) -> str
     | _ -> assert false

(*Lay danh sach (thu tu,ten) cua cac bien dc truyen vao ham *)
let rec get_list_pointer (exs : Cexpr.expr_t list) = 
  let list_pointer = ref [] in
	for i = 0 to (List.length exs) - 1 do
		 let ex = List.nth exs i in
		  match ex with 
                     |Cexpr.ADDROF(t,e) -> list_pointer := (i,get_name_param e)::!list_pointer
                     | _ -> ()
	done;
	();	 
   !list_pointer

(*Kiem tra mot expr co phai la loi goi ham khong?
    Neu dung thi tra ve (true, danh sach cac bien truyen vao theo kieu tham bien - co the rong)
    Neu sai thi tra ve (false, danh sach rong)
*)
let rec check_function_call (expr : Cexpr.expr_t) = 
   match expr with
    | Cexpr.CALL(t,ex,exs) -> let fun_call = (
                                match ex with
                                 | Cexpr.OPER(tyn,op,ex1,ex2,ex3) -> (match ex1 with 
                                                                      | Cexpr.VREF(typ,str,num1,num2) -> str
                                                                      | _ -> assert false
                                                                     )
                                 | _ -> assert false              
                              ) in
                              (true,fun_call,get_list_pointer exs)
    | _ -> (false,"",[])

(*Tao node ASGN sau cac loi goi ham *)
let make_assign_node_after_call (param_num : (int*string)) = 
   match param_num with
      | (num,param) -> make_assign param
      | _ -> assert false

(*Tao ds node ASGN sau cac loi goi ham *)
let rec make_assign_node_after_calls (list_param : (int*string) list) = 
   match list_param with 
     | [] -> []
     | couple::rest -> (make_assign_node_after_call couple)::(make_assign_node_after_calls rest)

(*=====================================================*)

(*=====================================================*)
(* tao node tai moi khoi dau ham *)
let make_node_function_header (ls_param : string list) ( curr_node : node )= 
   let pa_node = ref curr_node in
   
   for i = 0 to (List.length ls_param) - 1 do
      let name = List.nth ls_param i in
      let node_name = ("assign_"^(!current_function_name)^"_in_"^name^"_"^( string_of_int i )) in
      let temp_node =  NodeV( node_name, "", NODE_ASGN_V ) in
      import_into_gen node_name [(temp_node,name)];
      import_into_use node_name [name];
      insert_node_acfg_for_ddg (temp_node);
      make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
      pa_node := temp_node;
      number_v_assign := !number_v_assign + 1
   done;
   for j = 0 to (List.length !list_global_variable) - 1 do
         let g_var = ((List.nth !list_global_variable j)^"_global") in
         let node_name = ("assign_gvar_"^(!current_function_name)^"_in_"^g_var) in
         let temp_node =  NodeV( node_name, "", NODE_ASGN_V ) in
         import_into_gen node_name [(temp_node,g_var)];
         import_into_use node_name [g_var];   
         insert_node_acfg_for_ddg (temp_node);
         make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
         pa_node := temp_node;
         number_v_assign := !number_v_assign + 1 
   done;
   
   !pa_node

(* tao node tai moi ket thuc ham *)
let make_node_end_function (ls_param : string list) ( curr_node : node )= 
   let pa_node = ref curr_node in
   for i = 0 to (List.length ls_param) - 1 do
      let name = List.nth ls_param i in
      let node_name = ("assign_"^(!current_function_name)^"_out_"^name^"_"^( string_of_int i )) in
      let temp_node =  NodeV( node_name, "", NODE_ASGN_V ) in
      import_into_gen node_name [(temp_node,name)];
      import_into_use node_name [name];
      insert_node_acfg_for_ddg (temp_node);
      make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
      pa_node := temp_node;
      number_v_assign := !number_v_assign + 1
   done;
   for j = 0 to (List.length !list_global_variable) - 1 do
         let g_var = ((List.nth !list_global_variable j)^"_global") in
         let node_name = ("assign_gvar_"^(!current_function_name)^"_out_"^g_var) in
         let temp_node =  NodeV( node_name, "", NODE_ASGN_V ) in
         import_into_gen node_name [(temp_node,g_var)];
         import_into_use node_name [g_var];
         insert_node_acfg_for_ddg (temp_node);
         make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
         pa_node := temp_node;
         number_v_assign := !number_v_assign + 1 
   done;
   ();
   !pa_node
   
(*Tham function header*)	 
let visit_function_header (decl:Cexpr.declaration_t) = 
   let (func_name, func_level) = decl.Cexpr.decl_id in
   let curr_node = NodeV(func_name,"",NODE_OTHER) in
   if func_name = "main" then is_main_function := true
   else is_main_function := false;
   current_function_node := curr_node;
   insert_node_acfg  curr_node ;
   current_function_name := func_name ;
   import_into_functions func_name decl ;

   if (List.mem !current_function_name !list_function) = false then list_function := (!current_function_name)::(!list_function)
   else  ();

   list_curr_param := decl.Cexpr.decl_params ;

   make_node_function_header !list_curr_param curr_node 
   (* *)

(*=====================================================*)
  
(*Xu ly truong hop co bien toan cuc *)	
let visit_global_variable (decl : Cexpr.declaration_t) = 
   let (var_name, var_level) = decl.Cexpr.decl_id in
   let tn = decl.Cexpr.decl_tn in
   match tn with
    | Cexpr.FUNCTION_OF(_) -> list_function := (var_name)::(!list_function)
    | _ -> list_global_variable := var_name::!list_global_variable


(*=====================================================*)
(* tao canh lien ket tu node khoi dau ham cho den cac node trong ham *)
let make_edge_cdg (source : node) (dest : node) = 
   let e = Edge(source,dest,CDG,0) in
   if (List.mem e !list_edge_connect ) then ()
   else  list_edge_connect := e::(!list_edge_connect)

(* tao canh lien ket giua cac ham *)
let make_connect_edge ( source : node ) ( dest : node ) = 
   let e = Edge(source,dest,SDG,0) in
   if (List.mem e !list_edge_connect ) then ()
   else  list_edge_connect := e::(!list_edge_connect)

      
(* tham node co loi goi ham *)
let visit_function_call ( expr : Cexpr.expr_t ) ( note_info : string ) ( curr_node : node ) = 
   let (check,fun_call,lsparam) =  check_function_call expr in
   let pa_node = ref curr_node in
   if check then
       begin
           let len = (List.length lsparam) - 1 in
           for i = 0 to len do
               let (num,name) = (List.nth lsparam i) in
               let node_name = ("assign_"^fun_call^"_in_"^name^"_"^( string_of_int !number_v_assign )) in
               let temp_node =  NodeV( node_name, note_info, NODE_ASGN_V ) in
      	       import_into_gen node_name [(temp_node,name)];
               import_into_use node_name [name];
               insert_node_acfg_for_ddg (temp_node);
               make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
               let param_node = NodeV(("assign_"^fun_call^"_in_"^(string_of_int num)), "", NODE_ASGN_V) in
               make_connect_edge temp_node param_node;
               pa_node := temp_node;
               number_v_assign := !number_v_assign + 1 
           done;
           for j = 0 to (List.length !list_global_variable) - 1 do
               let g_var = ((List.nth !list_global_variable j)^"_global") in
               let node_name = ("assign_"^(!current_function_name)^"_"^fun_call^"_in_"^g_var^"_"^( string_of_int !number_v_assign )) in
               let temp_node =  NodeV( node_name, note_info, NODE_ASGN_V ) in
               import_into_gen node_name [(temp_node,g_var)];
               import_into_use node_name [g_var]; 
               insert_node_acfg_for_ddg (temp_node);
               make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
               let g_node = NodeV(("assign_gvar_"^fun_call^"_in_"^g_var), "", NODE_ASGN_V) in
               make_connect_edge temp_node g_node;
               pa_node := temp_node;
               number_v_assign := !number_v_assign + 1 
           done;
           for i = 0 to len do
               let (num,name) = (List.nth lsparam i) in
               let node_name = ("assign_"^fun_call^"_out_"^name^"_"^( string_of_int !number_v_assign )) in
               let temp_node =  NodeV( node_name, note_info, NODE_ASGN_V ) in
               import_into_gen node_name [(temp_node,name)];
               import_into_use node_name [name];
               insert_node_acfg_for_ddg (temp_node);
               make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
               let param_node = NodeV(("assign_"^fun_call^"_out_"^(string_of_int num)), note_info, NODE_ASGN_V) in
               make_connect_edge param_node temp_node;
               pa_node := temp_node;
               number_v_assign := !number_v_assign + 1 
           done;
           for j = 0 to (List.length !list_global_variable) - 1 do
               let g_var = ((List.nth !list_global_variable j)^"_global") in
               let node_name = ("assign_"^(!current_function_name)^"_"^fun_call^"_out_"^g_var^"_"^( string_of_int !number_v_assign )) in
               let temp_node =  NodeV( node_name, note_info, NODE_ASGN_V ) in
               import_into_gen node_name [(temp_node,g_var)];
               import_into_use node_name [g_var];
               insert_node_acfg_for_ddg (temp_node);
               make_edge_acfg_for_ddg !pa_node temp_node ACFG_TRUE;
               let g_node = NodeV(("assign_gvar_"^fun_call^"_out_"^g_var), "", NODE_ASGN_V) in
               make_connect_edge g_node temp_node;
               pa_node := temp_node;
               number_v_assign := !number_v_assign + 1 
           done;
       end
   else ();
   ();
   !pa_node
   
(* Tao danh sach cac couple_node_var *)
let rec process_couple_node_var (names : string list) (n : node) = 
   match names with
    | [] -> []
    | name::rest -> (n,name)::(process_couple_node_var rest n)


(* Tham moi stmt cua ham *)
let visit_stmt (pre_node : node ) (stmt : Cexpr.stmt_t) = 
         let NodeV(l_pre,c_pre,ty_pre) = pre_node in
         let res_node = ref pre_node in
	 (match stmt with
                | Cexpr.STMT(expr) -> (
                     match expr with                       
                      | Cexpr.PSEUDO_PROLOGUE -> (
                        let node_name = ("begin_"^(!current_function_name)) in 
			import_into_statement_info node_name stmt;
                        let curr_node = NodeV(node_name,find_note_infos stmt,NODE_STMT) in
                        import_into_gen node_name [];
                        import_into_use node_name [];
                        insert_node_acfg (curr_node);
                        import_into_hash_stats node_name stmt;
                        make_edge_acfg  pre_node curr_node ACFG_TRUE;
                        make_edge_cdg !current_function_node curr_node;
                        make_edge_cdg curr_node !current_function_node ;
			current_function_node := curr_node;
                        res_node := curr_node
                        )
                         
                      | Cexpr.PSEUDO_EPILOGUE -> (
                        let node_name = ("end_"^(!current_function_name)) in
			import_into_statement_info node_name stmt;
                        let curr_node = NodeV(node_name,find_note_infos stmt,NODE_STMT) in
                        import_into_gen node_name [];
                        import_into_use node_name [];
                        import_into_hash_stats (node_name) stmt;
                        let pa_node = make_node_end_function !list_curr_param pre_node in 
                        make_edge_acfg pa_node curr_node ACFG_TRUE;
                        make_edge_cdg !current_function_node curr_node ;
                        make_edge_cdg curr_node !current_function_node;
                        make_edge_cdg pre_node curr_node ;
                        res_node := curr_node;
			insert_node_acfg curr_node;
                       )
                      | Cexpr.CALL(_,_,_) -> (
                         let (check,fun_call,lsparam) =  check_function_call expr in
                         let node_name = ( "function_call_"^fun_call^"_"^(string_of_int !number_func)^"_"^(!current_function_name)) in
			 number_func := !number_func + 1;
       			 import_into_statement_info node_name stmt;
                         let curr_node = NodeV( node_name , "", NODE_STMT) in
			 insert_node_acfg curr_node;
                         if ty_pre <> NODE_GOTO then make_edge_acfg pre_node curr_node ACFG_TRUE;
                         make_edge_cdg !current_function_node curr_node;
                         import_into_gen node_name [];
                         import_into_use node_name (Slicingstructure.get_list_var_from_expr expr);
			 import_into_hash_stats node_name stmt;
                         if (List.mem fun_call !list_function) = true then 
                             begin
                         	let link_node = NodeV(fun_call,"",NODE_OTHER) in
                                let link_node_end = NodeV(("end_"^fun_call),"",NODE_LABEL) in
                         	make_connect_edge link_node curr_node;
				if (List.mem fun_call !list_fun_contain_error_label) = true then make_connect_edge curr_node link_node_end;
	                 	res_node := visit_function_call expr (find_note_infos stmt) curr_node;
                             end
                         else res_node := curr_node
                         
                       ) 
                      | _ -> ()
                      
                 )
		| Cexpr.GOTO(goto) -> (
                        number_goto := !number_goto+1;
                        let node_name = ("goto_"^(string_of_int !number_goto)^"_"^goto) in
			import_into_statement_info node_name stmt;
                        let curr_node = NodeV(node_name, find_note_infos stmt, NODE_GOTO) in
			make_edge_cdg !current_function_node curr_node;
			if goto = "#:return" then 
			    begin
				import_into_use node_name ["#:ACC"];
				make_edge_cdg curr_node !current_function_node;
			    end
			else  import_into_use node_name [];

                        import_into_gen node_name [];
                        insert_node_acfg (curr_node);
                        import_into_hash_stats node_name stmt;
                        if ty_pre <> NODE_GOTO then make_edge_acfg pre_node curr_node ACFG_TRUE;
                        let label = Cexpr.LABEL(Cexpr.NAME_LABEL(goto)) in
			let name_label = ref "" in
                        if goto = "#:return" then name_label := ("label_"^(!current_function_name)^"_"^goto)
                        else name_label := ("label_"^goto);

                        let label_node =  NodeV(!name_label,find_note_infos label,NODE_LABEL) in
                        insert_node_acfg label_node;
                        make_edge_acfg curr_node label_node ACFG_TRUE;
                        make_edge_cdg label_node curr_node;
                        make_edge_cdg curr_node pre_node;
                        res_node := curr_node
                 )
		| Cexpr.LABEL(label) -> ( match label with 
			                  | Cexpr.NAME_LABEL(label_name) -> (
                                            let node_name = ref "" in
				            if label_name = "#:return" then 
                                                node_name :=  ("label_"^(!current_function_name)^"_"^label_name)
                                            else 
                                                node_name := ("label_"^label_name);
   			                    import_into_statement_info !node_name stmt;
                                            import_into_gen !node_name [];
                                            import_into_use !node_name [];
                                            import_into_hash_stats !node_name stmt;
                                            let curr_node = NodeV(!node_name, find_note_infos stmt,NODE_LABEL) in
					    if (String.length label_name) >= 9 then
 					       begin
					          let sub_string = (String.sub label_name 0 9) in
   					          if sub_string = "__ERROR__" then 
						      list_input_slicing := curr_node::(!list_input_slicing);
					       end;
                                            insert_node_acfg (curr_node);
					    if (ty_pre <> NODE_GOTO) then  make_edge_acfg pre_node curr_node ACFG_TRUE;
                                            res_node := curr_node
                                            )
                                          | _ -> assert false 
		    )	 
		| Cexpr.IF(expr,[Cexpr.GOTO(l1)],[Cexpr.GOTO(l2)]) -> 
		             begin 
                                 let node_name = ("if:"^(string_of_int !number_if)^(Cexpr.expr_to_string_body 0 expr)) in
			         import_into_statement_info node_name stmt;
                                 let curr_node = (NodeV(node_name,find_note_infos stmt,NODE_IF)) in
                                 import_into_gen node_name [];
                                 import_into_use node_name (Slicingstructure.get_list_var_from_expr expr);
				 insert_node_acfg curr_node;
		                 import_into_hash_stats node_name stmt;
				 if ty_pre <> NODE_GOTO then make_edge_acfg pre_node curr_node ACFG_TRUE;
	                         make_edge_cdg !current_function_node curr_node;
		                 let label1 = Cexpr.LABEL(Cexpr.NAME_LABEL(l1)) in
		                 let label2 = Cexpr.LABEL(Cexpr.NAME_LABEL(l2)) in
                                 let name_label1 = ("label_"^l1) in
                                 let name_label2 = ("label_"^l2) in
				 let label_node1 = NodeV(name_label1, find_note_infos label1,NODE_LABEL) in
				 let label_node2 = NodeV(name_label2, find_note_infos label2,NODE_LABEL) in
				 make_edge_cdg label_node1 curr_node;
				 make_edge_cdg label_node2 curr_node;
				 make_edge_acfg curr_node label_node1 ACFG_TRUE;
				 make_edge_acfg curr_node label_node2 ACFG_FALSE;
				 number_if := !number_if+1;
                                 res_node := curr_node
		            end                      
		| Cexpr.ASGN(left,right) -> 
		       begin 
                           let node_name = ("assign"^(string_of_int !number_assign)) in
			   import_into_statement_info node_name stmt;
                           let curr_node = NodeV(node_name, find_note_infos stmt, NODE_ASGN) in
                           import_into_gen node_name (process_couple_node_var (Slicingstructure.get_list_var_from_expr left) curr_node);
                           import_into_use node_name (Slicingstructure.get_list_var_from_expr right);
			   insert_node_acfg curr_node;
			   if ty_pre <> NODE_GOTO then  make_edge_acfg pre_node curr_node ACFG_TRUE;
			   make_edge_cdg !current_function_node curr_node;
			   import_into_hash_stats ("assign"^(string_of_int !number_assign)) stmt;
			   number_assign := !number_assign+1;
		           let (check,fun_call,lsparam) =  check_function_call right in
	                   if ( check ) && ( List.mem fun_call !list_function ) then
	                       begin
			           let link_node = NodeV(fun_call,"",NODE_OTHER) in
                                   let link_node_end = NodeV(("end_"^fun_call),"",NODE_LABEL) in
			           make_connect_edge link_node curr_node;
				   if (List.mem fun_call !list_fun_contain_error_label) = true then make_connect_edge curr_node  link_node_end;
	                           res_node := visit_function_call right (find_note_infos stmt) curr_node;
	                       end
	                   else res_node := curr_node
		      end
		| Cexpr.NOTE(_,_,_,_) -> ()
		| _ -> assert false
        );
         if (String.length l_pre) >= 15 then 
            begin
               let sub_string_label = (String.sub l_pre 0 15) in
               
               if (ty_pre = NODE_LABEL) && (sub_string_label = "label___ERROR__") then
                   make_edge_cdg !res_node pre_node;
            end;
        !res_node

(* Tham ds cac stmt cua moi ham *)			
let rec visit_list_stmt (pre_node : node) (stmtlist : Cexpr.stmt_t list) = 
    match stmtlist with
	| [] -> ()
	| stmt::rest -> let res_node = visit_stmt pre_node stmt in 
                        visit_list_stmt res_node rest


(* Tham tung ham mot*)                      
let visit_function ( decl : Cexpr.declaration_t ) ( stmtlist : Cexpr.stmt_t list ) = 
    let pre_node = visit_function_header decl in 
    visit_list_stmt pre_node ( stmtlist );
    list_node_all := ( !list_node_all )@( !list_node_acfg_for_ddg );
    ( ( !list_node_acfg, !list_edge_acfg ), ( !list_node_acfg_for_ddg, !list_edge_acfg_for_ddg ) )

(*In thong tin Note cua moi stmt*)	
let print_note_infos () = 
   Hashtbl.iter (fun key value -> print_string (key^"-->");print_stmt value) note_infos

(*Dua thong tin Note vao trong hashtable *)
let put_note_infos (str : string) (lsStat : Cexpr.stmt_t list) = 
     List.iter (fun stmt -> Hashtbl.add note_infos str stmt) lsStat


(*=====================================================*)
(* Tao comment cho node *)
let list_temp_stats = ref [] 

let old_number = ref 0 


let make_note_infos (stmtlist : Cexpr.stmt_t list) = 
  List.iter (fun stmt ->  
      match stmt with
       | Cexpr.NOTE(tf,str,num1,num2) -> (
                    let strs = (str^":"^(string_of_int !old_number)^":"^(string_of_int num1)) in 
                    let ls_stats = !list_temp_stats in
		    put_note_infos strs ls_stats;
                    old_number := num1;
                    list_temp_stats := [];
                   )                      
       | _ -> list_temp_stats := stmt::!list_temp_stats
       
      ) stmtlist

let reset () = 
   list_node_acfg := [];
   list_edge_acfg := [];
   list_node_acfg_for_ddg := [];
   list_edge_acfg_for_ddg := [];
   Hashtbl.clear htbl_use_infos;
   Hashtbl.clear htbl_gen_infos;
   Hashtbl.clear note_infos


(*=====================================================*)
let htbl_uses_infos = Hashtbl.create 1000
let htbl_gens_infos = Hashtbl.create 1000
let htbl_stmts_for_ddg_infos = Hashtbl.create 1000

(*Tham stmt ung voi 2 truong hop : DEFUN va DEFVAR*)
let visit_def (def : Cexpr.stmt_t) = 
	match def with
          | Cexpr.DEFUN(decl, stmtlist) -> (
             make_note_infos stmtlist;
             let couple_cdg,couple_ddg = visit_function decl stmtlist in
            

             list_couple_acfg_for_ddg := couple_ddg::!list_couple_acfg_for_ddg;
             Hashtbl.replace htbl_uses_infos (!current_function_name) (Hashtbl.copy htbl_use_infos);
             Hashtbl.replace htbl_gens_infos (!current_function_name) (Hashtbl.copy htbl_gen_infos);
            
             reset ();
            )
	| Cexpr.DEFVAR(decl) -> visit_global_variable decl
	| _ -> () 
(*================================================================================================*)
let print_strings ls = 
   List.iter (fun s -> print_string (s ^ " ")) ls

let print_use htbl = 
   Hashtbl.iter ( fun k v -> print_string (k^"->"); print_strings v ;print_string"\n") htbl

let print_uses htbl = 
   Hashtbl.iter ( fun k v -> print_string ("\nHam : "^k^":\n"); print_use v) htbl


let print_ls_couple couples = 
   List.iter (fun cou -> let (n,var) = cou in
                         print_string "( ";
                         print_node n; print_string (": "^var); 
                         print_string " ) ";
             ) couples

let print_gen gen = 
   Hashtbl.iter (fun k v -> print_string (k^"-->"); print_ls_couple v; print_string "\n") gen

let print_gens gens = 
   Hashtbl.iter (fun k v -> print_string ("\nHam : "^k^":\n"); print_gen v; print_string "\n") gens
(*================================================================================================*)
let print_stat_num (stat_num : ( Cexpr.stmt_t * int ) ) = 
   let (stmt,num) = stat_num in
   print_string ((string_of_int num)^":");
   print_stmt stmt

let print_statement_infos () = 
   Hashtbl.iter ( fun k v -> print_string ("\n"^k^"->"); print_stat_num v
                ) htbl_statement_infos 
   
(*================ =================== ====================*)
(*Tham danh sach dau vao*)    
let rec visit_defs (defs : Cexpr.stmt_t list) = 
	match defs with
		| [] -> ()
		| def::rest -> visit_def def; visit_defs rest

let construct_acfg (defs : Cexpr.stmt_t list) ( list_func_error : node list ) =
        List.iter (fun node_fun_error -> 
                       let NodeV(lb,cm,ty) = node_fun_error in
                       list_fun_contain_error_label := lb::(!list_fun_contain_error_label);
                  ) list_func_error;
        visit_defs defs;
        (*print_statement_infos ();
        pause();*)
        list_couple_all := [(!list_node_all,!list_edge_connect)];
        !list_couple_acfg_for_ddg


let get_uses () = 
   htbl_uses_infos

let get_gens () = 
   htbl_gens_infos


let get_couple_all () = 
   !list_couple_all

(* Lay danh sach list function *)
let get_list_func () = 
   !list_function

let get_input_slicing () =
   !list_input_slicing
(*=====================================================*)



