(* slicing/constructCDG.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num
open Form
open Smt
open Slicingstructure
open ConstructACFG

module Sat = Minisat

(* ***************************** *********************************** *)	
                           (* Xay dung CDG *)

(* Thong tin ve post-dominate cua moi node *)
let htbl_post_dominator = Hashtbl.create 1000

(* danh sach node va canh CDG cua cac ham*)
let list_couple_cdg = ref [] 

let list_couple_pdt = ref []

(* *****************************=  =*********************************** *)	

(* get post_dominator_tree *)
let get_list_couple_pdt () = 
   !list_couple_pdt

(* Dua thong tin vao htbl_post_dominator *)
let import_into_post_dominator_info (key : string) (value : node list) = 
   Hashtbl.replace htbl_post_dominator key value

(*init htbl_post_dominator *)
let init_a_pd (n : node) (list_node : node list) = 
   let NodeV(label,comment,typ) = n in
   import_into_post_dominator_info label list_node


let rec init_pd (list_node : node list) (list_node_1 : node list)= 
   match list_node with
     | [] -> ()
     | n::rest -> init_a_pd n list_node_1; init_pd rest list_node_1

(* ghep 2 ds node lai - bo di cac phan tu trung nhau *)
let append (l1 : node list) (l2 : node list) = 
   if (List.length l1) = 0 then l2
   else 
      begin
          if (List.length l2) = 0 then l1
          else begin
                  let temp = ref [] in 
                  temp := l1;
                  for i = 0 to (List.length l2) - 1 do
                     let n = List.nth l2 i in
                     if  List.mem n !temp then ()
                     else temp := n::!temp
                  done;
                  ();
                  !temp
               end
      end


(* Kiem tra xem l1 co phai la ds con cua l2 khong *)
let is_sub_list (l1 : node list) (l2 : node list) = 
   let res = ref true in
   List.iter (fun n -> if List.mem n l2 then () else res := false) l1;
   !res

(* Kiem tra xem l1 co bang l2  khong *)
let equal_lists (l1 : node list) (l2 : node list) = 
   if (is_sub_list l1 l2) && (is_sub_list l2 l1) then true
   else false


(* find SuccNode*)
let rec find_successors (n : node) (edges : edge list) = 
  let res = ref [] in
  let NodeV(l,c,t) = n in
  
  List.iter (fun e ->  
                       match e with
                        | Edge (s, d, typ, num) -> let NodeV(lbs,cms,tys),NodeV(lbd,cmd,tyd) = s,d in   
                                                  if lbs = l then  
                                                  begin
                                                     res := d::(!res);
						  end
                        | _ -> assert false
             ) edges;
  ();
  !res



(* In thong tin trong tap post-dominator*) 
let print_hash_pd hs = 
   print_string "\n Hash:  \n";
   Hashtbl.iter (fun k v -> 
			print_string(k^"->");
			List.iter (fun n -> 
				       print_node n;
				       print_string " ") v;
			print_string "\n" )
			hs

let rec find_common_list (l1 : node list) (l2 : node list) = 
   match l1 with
    | [] -> []
    | n::rest -> (if List.mem n l2  then [n] 
                  else []
                       )@(find_common_list rest l2)

(*Compute PD of each node *)
let compute_post_dominator (couple : (node list) * (edge list)) = 
   let (nodes, edges) = couple in
   let end_node = List.hd nodes in
   match end_node with
   | NodeV (label, comment, typ) -> (
           import_into_post_dominator_info label [end_node];
           init_pd (List.tl nodes) (List.tl nodes);
           let htbl_old = ref (Hashtbl.copy htbl_post_dominator) in
           let htbl_new = ref (Hashtbl.copy htbl_post_dominator) in 
           let is_changed = ref true in
	   let traverse_a_node = function
  	       | NodeV (label1, comment, typ) as a_node ->                      
			let old_post_dominator = try
						   Hashtbl.find !htbl_old label1
						    with
						    | Not_found -> assert false 
                        in
	 	        let successors = find_successors a_node edges in
                        let temp = ref [] in
                        if (List.length successors > 0) then 
                          begin
		                let node_succ = List.hd successors in
		                let NodeV(lb,cm,typ)  = node_succ in
		                temp:=(try
		                         Hashtbl.find !htbl_old lb
				       with
					| Not_found -> assert false
		                      );

		                 List.iter (fun n0 -> 
 				     let NodeV (label0, comment0, typ0) = n0 in 
				     let new_node = try
							Hashtbl.find !htbl_old label0
						    with
						    | Not_found -> assert false
				     in
			             temp := find_common_list (!temp) new_node
			          ) (List.tl successors);
                          end
                        else ();
                         temp := a_node::(!temp);	
		        Hashtbl.replace !htbl_new label1 (!temp);
		        if (equal_lists old_post_dominator !temp) = false then 
			   is_changed := true 
	                else ()
	       | _ -> assert false
	   in
           while !is_changed do
  	      is_changed := false;  
	      List.iter traverse_a_node (List.tl nodes);
              htbl_old := Hashtbl.copy !htbl_new;
	   done;
           Hashtbl.iter (fun k v -> Hashtbl.replace htbl_post_dominator k v) !htbl_old;
	   ()
       )
    | _ -> assert false

(* Xoa n ra khoi ds Node *) 
let rec remove_node_from_list_node (n : node) (nodes : node list) = 
   match nodes with
    | [] -> []
    | m::rest -> (if m = n  then [] else [m])@(remove_node_from_list_node n rest)

(* Xoa n ra khoi dang sach cua htbl_post_dominator(m) *)
let remove_node_from_post_dominator (n : node) (m : node) = 
   let NodeV(label,comment,typ) = n in
   let NodeV(label1,comment1,typ1) = m in
   let temp = Hashtbl.find htbl_post_dominator label1 in 
   Hashtbl.replace htbl_post_dominator label1 (remove_node_from_list_node n temp);
   ()

let make_edge_pdt (source : node) (dest : node) = 
   Edge(source,dest,PDT,0)


(* Construct PD-Tree *)   
let construct_post_dominator_tree (couple : graph_type_slicing) = 
   compute_post_dominator couple;
   let post_dominator_edges = ref [] in
   let (nodes,edges) = couple in
   let queue = Queue.create () in
   Queue.push (List.hd nodes) queue;
   for i = 0 to (List.length nodes) - 1 do 
      let nt = List.nth nodes i in
      remove_node_from_post_dominator nt nt
   done;
   while (Queue.is_empty queue) = false do
      let node_temp = Queue.pop queue in 
      for i =0 to (List.length nodes) -1 do
            let current_node = (List.nth nodes i) in
            let NodeV(label_current_node,comment,typ) =  current_node in
            if (List.mem node_temp (Hashtbl.find htbl_post_dominator label_current_node) ) then
            begin
               remove_node_from_post_dominator node_temp current_node;
                if (List.length (Hashtbl.find htbl_post_dominator label_current_node) = 0 ) then 
                  begin 
                     post_dominator_edges := (make_edge_pdt node_temp current_node)::(!post_dominator_edges);
                     Queue.push current_node queue
                  end
                else ();
            end
            else ();
      done
   done;
   !post_dominator_edges

(* find Parent *)
let find_parent_in_pdt (child : node) (edges_pdt : edge list)= 
    let NodeV(l,c,t) = child in
    let parent = ref child in 
    List.iter (fun e -> 
        let Edge(source,dest,typ,num) = e in
        let NodeV(ld,cd,td) = dest in
        if (ld = l) then parent := source 
    )edges_pdt;
    ();
    !parent

let is_equal_node (node1 : node) (node2 : node) = 
    match node1, node2 with
     | NodeV(lb1,cm1,t1),NodeV(lb2,cm2,t2) -> if (lb1 = lb2)  then true else false
     | _,_ -> assert false

(* Kiem tra xem ancestor co phai la "to tien" cua child khong? *)
let is_ancestor (child : node) (ancestor : node) (end_node : node) (edges : edge list) =
    let NodeV(lc,cc,tc),NodeV(la,ca,ta) = child,ancestor in
    
    let check = ref false in 
    let parent = ref child in
    let is_terminate = ref false in
    while (not (!parent = end_node)) && (!is_terminate = false) do
        parent := find_parent_in_pdt !parent edges;
        if (!parent = ancestor) then 
          begin 
             is_terminate := true; 
             check := true 
          end
        else ();
    done;
    ();
    !check


let find_least_common_ancestor (node1 : node) (node2 : node) (end_node : node) (edges : edge list) =  
    let res = ref node1 in
    let NodeV(l1,c1,t1), NodeV(l2,c2,t2) = node1,node2 in
    
    let ancestor1 = ref [] in
    let ancestor2 = ref [] in
    let parent1 = ref node1 in
    let parent2 = ref node2	in	
    ancestor1:= node1::!ancestor1;
    ancestor2:= node2::!ancestor2;
    let terminate = ref false in
    while (!terminate = false) do
        if (not (!parent1 = end_node)) then
          begin
            parent1 := find_parent_in_pdt !parent1 edges ;
          end;
        if (not (!parent2 = end_node)) then 
          begin
            parent2 := find_parent_in_pdt !parent2 edges ;
          end;
        if(List.mem !parent2 !ancestor1 ) then begin res := !parent2; terminate := true end
	else 
            if(List.mem !parent1 !ancestor2 ) then begin res := !parent1; terminate := true end 
            else begin
               ancestor1 := (!parent1)::(!ancestor1);
               ancestor2 := (!parent2)::(!ancestor2);
            end;
    done;
    ();
    !res

(* make a cdg's edge *)
let make_edge_cdg (source : node) (dest : node) = 
   Edge(source,dest,CDG,0)

(* Xay dung CDG cho mot ham*)
let construct_a_cdg (couple : graph_type_slicing) =
    let (nodes,edges) = couple in
    let edges_cdg = ref [] in
    let edges_pdt = ref [] in
    edges_pdt := construct_post_dominator_tree couple;
    let gr_pdt = (nodes,!edges_pdt) in
    list_couple_pdt := (nodes,!edges_pdt)::(!list_couple_pdt); 
    let end_node = List.hd nodes in
    List.iter (fun a_edge -> 
	let Edge(source,dest,typ,num) = a_edge in
	let NodeV(ls,cs,ts),NodeV(ld,cd,td) = source,dest in
	if (is_ancestor source dest end_node !edges_pdt) = false then
	begin
	    let least_common_ancestor = (find_least_common_ancestor source dest end_node !edges_pdt) in
	    if (least_common_ancestor = source) then 
	      begin
	        edges_cdg := ( make_edge_cdg source source )::(!edges_cdg)
	      end;
	    let parent = ref dest in
	    while (not (!parent = least_common_ancestor)) do 
	        edges_cdg := (make_edge_cdg source !parent)::(!edges_cdg);
	        parent := find_parent_in_pdt !parent !edges_pdt       
	    done;
	end;
    ) edges;
    list_couple_cdg := (nodes,!edges_cdg)::!list_couple_cdg;
    ()

(* Xay dung CDG cho chuong trinh *)
let construct_cdg list_couple_acfg =
   List.iter (fun couple -> construct_a_cdg couple) list_couple_acfg;
   ();
   !list_couple_cdg





