(* slicing/constructDDG.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num
open Slicingstructure
open ConstructACFG


(******************************= end 1 =************************************)	
(* Global vairiable *)
let list_couple_ddg = ref ([] : graph_type_slicing list)

(* kills infos *)
let htbl_kills : (string, (string, Slicingstructure.couple_node_var list) Hashtbl.t) Hashtbl.t = Hashtbl.create 1000

(* OUT infos *)
let htbl_outs : (string, (string, Slicingstructure.couple_node_var list) Hashtbl.t) Hashtbl.t = Hashtbl.create 1000

(* IN infos *)
let htbl_ins : (string, (string, Slicingstructure.couple_node_var list) Hashtbl.t) Hashtbl.t = Hashtbl.create 1000

let num_edge_ddg = ref 0

(*=====================================================*)	

let compare_list_couple (ls1 : couple_node_var list) (ls2 : couple_node_var list) = 
   match ls1,ls2 with
    |[c1],[c2] -> let (n1,v1),(n2,v2) = c1,c2 in
                  if v1 = v2 then true
                  else false
    | _ -> false

let compute_list_couple_node_var ( couple : couple_node_var list ) gen = 
    let res = ref [] in
    Hashtbl.iter ( fun k v -> if (compare_list_couple couple v) = true then res := v@(!res)
                 ) gen;
   ();
   !res  

let htbl_kill = Hashtbl.create 1000

let compute_kill gen = 
   Hashtbl.iter ( fun k v -> Hashtbl.replace htbl_kill k (compute_list_couple_node_var v gen)) gen;
   ()
let compute_a_kill (k : string) v = 
   compute_kill v; 
   Hashtbl.replace htbl_kills k (Hashtbl.copy htbl_kill); 
   Hashtbl.clear htbl_kill;
   ()

(* Tinh KILL dua vao GEN *)			
let compute_kills () = 
   Hashtbl.iter (compute_a_kill) htbl_gens_infos;
   ()

(******************************= end 2 =************************************)	


(* find PreNode*)
let rec find_predessors (n : node) (edges : edge list) = 
  let res = ref [] in
  let NodeV(l,c,t) = n in
  
  List.iter (fun e ->  
                       match e with
                        | Edge (s, d, typ, num) -> let NodeV(lbs,cms,tys),NodeV(lbd,cmd,tyd) = s,d in   
                                                  if lbd = l then  
                                                  begin
                                                     res := s::(!res);
						  end
                        | _ -> assert false
             ) edges;
  ();
  !res




let append_list_to_in_out (l1 : couple_node_var list) (l2 : couple_node_var list) = 
    let res = ref [] in
    if (List.length l1) = 0 then res := l2
    else
        begin
            res := l1@(!res);
            if (List.length l2) = 0 then ()
            else  
                begin
                     List.iter (fun coup -> if (List.mem coup l1) then ()
                                            else res := coup::(!res)
                     ) l2
                end
        end;
    ();
    !res

let remove_list_to_in_out (bitru : couple_node_var list) (tru : couple_node_var list) = 
    let res = ref [] in
    if (List.length tru) = 0 then res := bitru@(!res)
    else
        begin
            if (List.length bitru) = 0 then res := []
            else  
                begin
                     List.iter (fun coup -> if (List.mem coup tru) then ()
                                            else res := coup::(!res)
                     ) bitru
                end
        end;
    ();
    !res

(* Kiem tra xem l1 co phai la ds con cua l2 khong *)
let is_sub_list_couple_node_var (l1 : couple_node_var list) (l2 : couple_node_var list) = 
   let res = ref true in
   if (List.length l1) = 0 && (List.length l2) <> 0 then res := false
   else 
           if (List.length l2) = 0 && (List.length l1) <> 0 then res := false
	   else 
	      	   if (List.length l2) = 0 && (List.length l1) = 0 then res := true
		   else 
			 List.iter (fun coup -> if List.mem coup l2 then () else res := false) l1;

   !res

(* Kiem tra xem l1 co bang l2  khong *)
let is_equal_in_out (l1 : couple_node_var list) (l2 : couple_node_var list) = 
   if (is_sub_list_couple_node_var l1 l2) && (is_sub_list_couple_node_var l2 l1) then true
   else false


let compute_in_out ( nodes : node list)  (edges : edge list) gen kill = 

   let htbl_out : (string, Slicingstructure.couple_node_var list) Hashtbl.t = Hashtbl.create 1000 in
   let htbl_in :  (string, Slicingstructure.couple_node_var list) Hashtbl.t = Hashtbl.create 1000 in 
   Hashtbl.iter (fun k v -> Hashtbl.add htbl_out k v ) gen;
   let htbl_old_out = ref (Hashtbl.copy htbl_out) in
   let htbl_new_out = ref (Hashtbl.copy htbl_out) in
   let change = ref true in
   while !change do
       change := false;
       for i = 0 to (List.length nodes) - 1 do
          let n = List.nth nodes i in
          let NodeV(label,comment,typ) = n in
          let old_out = (try Hashtbl.find !htbl_old_out label
                         with 
                         | Not_found -> assert false
                         )
          in
          let predessors = find_predessors n edges in
          let curr_in = ref [] in
          List.iter (
              fun pre_node -> 
                 let NodeV (l1,c1,t1) = pre_node in
                 let temp =  (try Hashtbl.find !htbl_old_out l1
                             with 
                             | Not_found -> assert false
                             ) 
                 in
                 curr_in := append_list_to_in_out (!curr_in) temp;
          ) predessors;
          Hashtbl.replace htbl_in label !curr_in;
          let gen_node =  (try Hashtbl.find gen label
                       with 
                       | Not_found -> assert false
                      ) 
          in
          let kill_node =  (try Hashtbl.find kill label
                       with 
                       | Not_found -> assert false
                      ) 
          in
          let temp_out = append_list_to_in_out (gen_node) (remove_list_to_in_out (!curr_in) kill_node) in
          Hashtbl.replace !htbl_new_out label temp_out;
          if (is_equal_in_out temp_out old_out) = false then  change := true
          else ();
          ()
       done;
       htbl_old_out := Hashtbl.copy !htbl_new_out;
   done;
   Hashtbl.iter (fun k v -> Hashtbl.replace htbl_out k v ) !htbl_old_out;
   (htbl_out,htbl_in)

(* Tao canh ddg edge *)
let make_edge_ddg ( source : node ) ( dest : node ) = 
   Edge(source, dest, DDG, !num_edge_ddg)

(* Tao canh khi duyet moi phan tu trong danh sach use *)
let make_edge_use ( var : string ) (ls_couple : couple_node_var list) ( node_dest : node ) = 
   let res = ref [] in
   List.iter (fun coup -> let (n,v) = coup in
                          if (var = v) then res := (make_edge_ddg n node_dest)::(!res)
                          else ()
   ) ls_couple;
   !res

(* Tao tap canh ddg edge *)
let compute_edge_ddg (in_infos : (string, Slicingstructure.couple_node_var list) Hashtbl.t) ( nodes : node list )  use =
   let res = ref [] in
   List.iter (fun n -> let NodeV(label,comment,tyn) = n in
                       let list_use = ( try Hashtbl.find use label
                                        with
                                        | Not_found -> assert false
                       ) in
                       let list_couple_node_var =  ( try Hashtbl.find in_infos label
                                        with
                                        | Not_found -> assert false
                       ) in
                       List.iter (fun var -> res := (make_edge_use var list_couple_node_var n)@(!res)) list_use;
                       
             ) nodes;
   !res

(* xay dung ddg cho tung ham *)
let construct_a_ddg couple = 
    let nodes, edges = couple in
    let list_edge_ddg = ref [] in
    if (List.length edges) > 0 then
        begin
	    let end_node = List.nth nodes ((List.length nodes) -1) in
	    let NodeV(label,comment,typ) = end_node in
	    let gen = (try Hashtbl.find htbl_gens_infos label
		       with
		        | Not_found -> assert false
		      )
	    in
	    let kill = (try Hashtbl.find htbl_kills label
		       with
		        | Not_found -> assert false
		      )
	    in
	    let use = (try Hashtbl.find htbl_uses_infos label
		       with
		        | Not_found -> assert false
		      )
	    in
	    Hashtbl.add use label [];
	    Hashtbl.add gen label [];
	    Hashtbl.add kill label [];

	    let temp_out , temp_in = compute_in_out nodes edges gen kill in
	    Hashtbl.replace htbl_outs label (temp_out);
	    Hashtbl.replace htbl_ins label (temp_in);
	    list_edge_ddg := (compute_edge_ddg temp_in nodes use);
	    list_couple_ddg := (nodes,!list_edge_ddg)::(!list_couple_ddg);
        end
    else
    ()

let construct_ddg list_couple_acfg = 
   compute_kills ();

   List.iter (fun couple -> construct_a_ddg couple) list_couple_acfg;
   
   !list_couple_ddg






