(* slicing/slicing.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num


open Slicingstructure
open ConstructACFG
open ConstructCDG
open ConstructDDG


(******************************************************************)	

let enable_slicing = ref false

(*Danh sach cac cap vertex va edge ung voi moi ham*)
let list_couple_acfg = ref []

(* danh sach node va canh CDG cua cac ham*)
let list_couple_cdg = ref []

(* danh sach node va canh DDG cua cac ham*)
let list_couple_ddg = ref []

let list_couple_all = ref []

let list_couple_all_after_slicing = ref []

let list_couple_pdt = ref []

let list_input_slicing = ref []
	
(*let ls_keep_nodes = ref []*)

let curr_fun_name = ref ""

let number_statement = ref 0


(* htbl_stmts_for_ddg_infos  - constructACFG*)


	
(******************************************************************)	
  
(****************************** ************************************)	

let connect_edge ( cdgs : ((node list * edge list) list) )( ddgs : ((node list * edge list) list) ) ( alls : ((node list * edge list) list) ) = 
   let temp_edges = ref [] in
   let res = ref [] in
   List.iter (fun couple_cdg -> 
                                let (nodes_cdg,edges_cdg) = couple_cdg in
                                List.iter ( fun e -> if (List.mem e !temp_edges) = false then temp_edges := (e)::(!temp_edges)) edges_cdg;
             ) cdgs;
   List.iter (fun couple_ddg -> 
                                let (nodes_ddg,edges_ddg) = couple_ddg in
                                List.iter ( fun e -> if (List.mem e !temp_edges) = false then temp_edges := (e)::(!temp_edges)) edges_ddg;
             ) ddgs;
   (match alls with
    | [nodes,edges] -> res := [nodes,edges@(!temp_edges)]
    | _ -> assert false
   );
   !res



let get_depens_node_name (node_in : node) (edges : edge list) = 
   let NodeV(name,c,t) = node_in in
   let res = ref [] in
   List.iter (fun e -> 
                      let Edge(s,d,tye,ne) = e in
		      let NodeV(ls,cs,tyns),NodeV(ld,cd,tynd) = s,d in
		      if ld = name then 
                         begin
			      res:=  s::(!res);
                         end	
             ) edges;
   !res


let compute_keep_nodes ( inputs : node list) (list_all : (node list * edge list) list) = 
    let list_node_visited = ref [] in
    (*print_string "\n List all: \n";
    List.iter ( print_graph ) list_all;
    pause();*)
   (match list_all with
    | [nodes,edges] -> (
                         let queue = Queue.create () in
		         let temp_list_edges = ref [] in
			 let temp_list_edges_1 = ref [] in
                         List.iter (fun n -> Queue.push n queue) (inputs);
			 while (Queue.is_empty queue) = false do
 			     let curr_node = Queue.pop queue in
			     list_node_visited := curr_node::(!list_node_visited); 
                             let depens_nodes_names = (get_depens_node_name curr_node edges) in
			     List.iter (fun no -> if (is_mem_of_list_node no (!list_node_visited) ) = false then
							    begin
								Queue.push no queue
							    end
                                        ) depens_nodes_names;
                         done;
			 List.iter (fun e -> 
                                       let Edge(s,d,t,n) = e in
				       if (is_mem_of_list_node s !list_node_visited) = true && (is_mem_of_list_node d !list_node_visited) = true then
					   begin 
                                           	temp_list_edges := e::(!temp_list_edges);
						let e = Edge(s,d,KOBO,0) in
						temp_list_edges_1 := e::(!temp_list_edges_1);
					   end
				        else 
					   begin
						let e = Edge (s,d,BO,0) in 
						temp_list_edges_1 := e::(!temp_list_edges_1);
					   end
                                   ) edges;
			 list_couple_all_after_slicing := [nodes,!temp_list_edges_1];
                       )
    | _ -> assert false
    );      
   (!list_node_visited)

(* Kiem tra xem node nay co duoc giu lai hay khong? *)
let is_keeped_node ( node_name :  string  ) (ls_keep_nodes : node list) = 
   let res = ref false in
   List.iter ( fun n -> let NodeV(l,c,t) = n in
                         if ( l =  node_name )  then res := true  
             ) ls_keep_nodes;
   !res                  

(* Tham header cua moi function, gan ten ham hien tai cho cuur_fun_name *)
(*let visit_fun_header ( decl : Cexpr.declaration_t ) = 
   let (func_name, func_level) = decl.Cexpr.decl_id in
   curr_fun_name := func_name
   number_statement := !number_statement + 1*)

(* Tim node tu statement *)
let find_node_num ( stmt : Cexpr.stmt_t ) = 
   let res = ref "###:" in
   Hashtbl.iter ( fun k v -> match v with
                             | (stmt1,num) -> if ( Cexpr.stmt_eql stmt stmt1 ) && ( num = !number_statement ) then
                                                   res := k
                             | _ -> assert false 
                ) htbl_statement_infos;
   !res

(* Thuc hien slicing cho danh sach statement trong moi ham*)
let slicing_stmts ( stmts : Cexpr.stmt_t list ) (ls_keep_nodes : node list) = 
   let res = ref [] in
   List.iter (fun stmt -> 
                 match stmt with
                  | Cexpr.NOTE(_,_,_,_) -> res := (stmt)::(!res)
		  (*| Cexpr.STMT(Cexpr.PSEUDO_EPILOGUE) -> res := (stmt)::(!res)
		  | Cexpr.STMT(Cexpr.PSEUDO_PROLOGUE) -> res := (stmt)::(!res)*)
                  | _ -> begin
	                    let node_name = find_node_num stmt in
                            (*print_endline (""^(string_of_int (List.length ls_keep_nodes)));
		            print_endline ("\nnode: "^node_name); print_endline (", So thu tu cua statement:"^(string_of_int !number_statement));
                            print_endline "\n";
		            print_stmt stmt;
                            print_endline ("Co giu hay khong? -> "^(string_of_bool (is_keeped_node node_name ls_keep_nodes ) ));
			    pause();*)
                 
                            number_statement := !number_statement + 1;
	                    if (is_keeped_node node_name ls_keep_nodes ) then res := (stmt)::(!res)
                         end
             ) stmts;
   ( List.rev !res )


(* kiem tra ham co rong khong? *)
let check_null_fun (stmts : Cexpr.stmt_t list) = 
   let stmt = (List.hd stmts) in
   if (stmt = Cexpr.STMT(Cexpr.PSEUDO_PROLOGUE)) then false
   else true

(* Thuc hien slicing cho tung ham*)
let slicing_def ( def : Cexpr.stmt_t ) (ls_keep_nodes : node list) = 
   match def with
    | Cexpr.DEFVAR(_) -> def
    | Cexpr.DEFUN(decl,stmts) -> let stmts_sl = (slicing_stmts stmts ls_keep_nodes ) in    
                                 if (check_null_fun stmts_sl) = false then
                                    Cexpr.DEFUN(decl, stmts_sl)
                                 else Cexpr.STMT(Cexpr.NIL)

(* thuc hien slicing cho chuong trinh *)
let slicing_defs ( defs : Cexpr.stmt_t list ) = 
   let  (ls_keep_nodes : node list) = (compute_keep_nodes 
                                        (!list_input_slicing) 
                                        (!list_couple_all)
                                      )
   in 
   let res = ref ([] : Cexpr.stmt_t list) in
   List.iter (fun def ->  let def_sl = ( slicing_def def ls_keep_nodes ) in
                          (*print_string "\nHam:\n";
                          print_stmt def_sl;pause();*)
                          match def_sl with
                           |Cexpr.STMT(Cexpr.NIL) -> ()
                           | _ -> res := (def_sl)::(!res)
             ) (defs);
   !res

(* Xay dung do thi quan he giua cac ham trong chuong trinh *)
let construct_calling_graph (defs : Cexpr.stmt_t list) = 
   let nodes = ref ([] : node list) in
   let list_fun_contain_error = ref ([] : node list) in
   let edges = ref ([] : edge list) in
   List.iter ( fun def -> 
                  match def with 
                   | Cexpr.DEFVAR(decl) -> ()
                   | Cexpr.DEFUN(decl,stmts) -> 
                     begin
                         if (List.length stmts) > 0 then
                           begin
                               let curr_fun_name = (
                                   let (fun_name,fun_id) = decl.Cexpr.decl_id in
                                   fun_name
                                   )
                               in
                               let curr_node = NodeV(curr_fun_name,"",NODE_OTHER) in
                               if (List.mem curr_node !nodes) = false then nodes := (curr_node)::(!nodes);
                               List.iter ( fun stmt ->
                                              match stmt with
                                               | Cexpr.STMT(expr) -> 
                                                  (
						      let fun_call =  get_name_function expr in
                                                      if fun_call <> "###" then
                                                      begin
                                                         let call_node = NodeV(fun_call,"",NODE_OTHER) in
							 let edge_calling = Edge(curr_node,call_node,OTHER,0) in
							 if (List.mem edge_calling !edges) = false then edges := (edge_calling)::(!edges);
 						      end
                                                  )
                                               | Cexpr.ASGN(exprL,exprR) -> 
						  (
 								let fun_call =  get_name_function exprR in
								if fun_call <> "###" then 
								begin
                                                                   let call_node = NodeV(fun_call,"",NODE_OTHER) in
								   let edge_calling = Edge(curr_node,call_node,OTHER,0) in
       								   if (List.mem edge_calling !edges) = false then edges := (edge_calling)::(!edges);
                                                                end
						  )
                                               | Cexpr.LABEL(label) -> 
						
						  (
							match label with
							 | Cexpr.NAME_LABEL(name_label) -> 
							    (
								if (String.length name_label) >=9 then
							 	begin
								  let sub_string = String.sub name_label 0 9 in
								  if (sub_string = "__ERROR__") && (List.mem curr_node !list_fun_contain_error) = false then
								     list_fun_contain_error := (curr_node)::(!list_fun_contain_error);
								end
							    )
						  )
                                               | _ -> ()
                                         ) stmts;
                                
                           end
                         else ()
                     end
                   | _ -> assert false
             ) defs;
    let graph_fun = (!nodes,!edges) in
(*    print_string "\n fun:\n";
    print_graph graph_fun;
    print_string "\n list error fun init :\n";
    List.iter (fun n -> print_node n) (!list_fun_contain_error);
    pause("\n");*)
    let visited_node = ref ([] : node list) in
    let stack = Stack.create () in
    if (List.length !list_fun_contain_error) > 0 then
      begin
        List.iter ( fun node_fun_error -> Stack.push node_fun_error stack ) (!list_fun_contain_error);
        while(Stack.is_empty stack) = false do
           let cu_node = Stack.pop stack in
           if (List.mem cu_node !visited_node) = false then visited_node := cu_node::(!visited_node);
           let parent_nodes = (find_parent cu_node !edges) in
(*           print_string "\n curr node :\n";
           print_node cu_node;
           print_string "\n list parent node :\n";
           List.iter (fun n -> print_node n) (parent_nodes);
           pause("\n");         *)
           List.iter (fun parent_node -> 
                          if (List.mem parent_node !visited_node) = false then 
                              begin
                                 visited_node := parent_node::(!visited_node);
                                 Stack.push parent_node stack;
                              end
                      ) parent_nodes;
        done;
      end
    else enable_slicing := false ;
    !visited_node

(****************************** ************************************)	
(*ham main*)

let main (defs : Cexpr.stmt_t list) = 

  let fun_error = construct_calling_graph defs in

  (*print_string "\nDanh sach dc ham chua error \n";
  List.iter (fun n -> print_node n;print_string "\n") fun_error;
  pause("");*)
  
  list_couple_acfg := (construct_acfg defs fun_error);

  list_input_slicing := (get_input_slicing ());

  list_couple_all := (get_couple_all ());

  list_couple_cdg := (construct_cdg (!list_couple_acfg));
  
  list_couple_pdt := (get_list_couple_pdt ());

  list_couple_ddg := (construct_ddg (!list_couple_acfg));

  list_couple_all := (connect_edge 
                          (!list_couple_cdg) 
                          (!list_couple_ddg) 
                          (!list_couple_all) 
                     );

  slicing_defs defs (* ket qua tra ra *)

(****************************** ************************************)	

(* ===================== recycle ========================== *)
(*ls_keep_nodes := (
     let list_node_visited = ref [] in
      (match !list_couple_all with
	    | [] -> ()
	    | [nodes,edges] -> (
		                 let queue = Queue.create () in
				 let temp_list_edges = ref [] in
				 let temp_list_edges_1 = ref [] in
		                 List.iter (fun n -> Queue.push n queue) (!list_input_slicing);
				 while (Queue.is_empty queue) = false do
	 			     let curr_node = Queue.pop queue in
				     list_node_visited := curr_node::(!list_node_visited); 
		                     let depens_nodes_names = (get_depens_node_name curr_node edges) in
				     List.iter (fun no -> if (is_mem_of_list_node no (!list_node_visited) ) = false then
								    begin
									Queue.push no queue
								    end
		                                ) depens_nodes_names;
		                 done;
				 List.iter (fun e -> 
		                              let Edge(s,d,t,n) = e in
					      if (is_mem_of_list_node s !list_node_visited) = true && (is_mem_of_list_node d !list_node_visited) = true then
						   begin 
		                                   	temp_list_edges := e::(!temp_list_edges);
							let e = Edge(s,d,KOBO,0) in
							temp_list_edges_1 := e::(!temp_list_edges_1);
						   end
						else 
						   begin
							let e = Edge (s,d,BO,0) in 
							temp_list_edges_1 := e::(!temp_list_edges_1);
						   end
		                           ) edges;
				 list_couple_all_after_slicing := [nodes,!temp_list_edges_1];
		               )
	    | _ -> assert false
	   );
	   (!list_node_visited)
    ) ;*)
 	
	
