(* slicing/slicingstructure.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

open Num

(*================ Structure for Slicing====================*)
(* kieu node cua do thi *)
type type_node = 
    | NODE_ASGN
    | NODE_ASGN_V
    | NODE_IF
    | NODE_GOTO
    | NODE_LABEL
    | NODE_STMT
    | NODE_OTHER

(* node cua do thi *)
type node = NodeV of string * string * type_node
           

(* kieu canh cua do thi *)
type edge_type = 
	| ACFG_TRUE
	| ACFG_FALSE       
	| CDG
	| DDG
	| PDT
        | SDG
        | KOBO
        | BO
        | OTHER		

(* canh cua do thi *)
type edge = Edge of node * node * edge_type * int

(* graph *)
type graph_type_slicing = ((node list) * (edge list))

(* Tao kieu dung trong xay dung DDG *)
type couple_node_var = node * string 
(*================ ========= ====================*)

 
(* Ham pause chuong trinh*)
let pause (s : string) = 
   print_string ("\npause ti : \n"^s); 
   let t = read_int () in
   print_string s 

let equal_nodes (n1 : node) (n2 : node) = 
   match n1,n2 with
    | NodeV(l1,c1,t1),NodeV(l2,c2,t2) -> if l1 = l2 then true else false
    | _ -> assert false

let is_mem_of_list_node (n : node) (nodes : node list) = 
   let res = ref false in
   let NodeV(l1,c1,t1) = n in
   List.iter (fun no -> 
		   let NodeV(l2,c2,t2) = no in
                   if l1 = l2 then res := true
             ) nodes;
   !res

(* Ham tim cac cha cua mot node trong graph *)
let find_parent (n : node) (edges : edge list) = 
   let res = ref ([] : node list) in
   List.iter (fun e -> 
                  let Edge(n1,n2,ty,num) = e in
                  if (equal_nodes n2 n) = true && (List.mem n1 !res) = false then res := n1::(!res);                          
             ) edges;
   ();
   !res

let get_name_function (expr : Cexpr.expr_t) = 
   match expr with 
    | Cexpr.CALL(tyn,ex,exs) -> (let fun_call = (
                                match ex with
                                 | Cexpr.OPER(tyn,op,ex1,ex2,ex3) -> (match ex1 with 
                                                                      | Cexpr.VREF(typ,str,num1,num2) -> str
                                                                      | _ -> assert false
                                                                     )
                                 | _ -> assert false              
                              ) in
                              fun_call )
    | _ -> "###"

(* Lay danh sach bien tu mot expression *)
let rec get_list_var_from_expr (expr : Cexpr.expr_t) = 
   match expr with
    |Cexpr.NIL -> []
    |Cexpr.OPER(tyn,op,ex1,ex2,ex3) -> ( get_list_var_from_expr ex1 ) @ ( get_list_var_from_expr ex2 ) @ ( get_list_var_from_expr ex3 )
    |Cexpr.SETF(tyn,op,ex1,ex2) -> ( get_list_var_from_expr ex1 ) @ ( get_list_var_from_expr ex2 )
    |Cexpr.VREF(tyn,str,num1,num2) -> if num1 = 1 then [(str^"_global")] else [str]
    |Cexpr.PREF(tyn,ex) -> get_list_var_from_expr ex
    |Cexpr.SREF(tyn,boo,ex,str) -> get_list_var_from_expr ex
    |Cexpr.AREF(tyn,ex1,ex2,ex3) -> 
           ( get_list_var_from_expr ex1 ) 
	 @ ( get_list_var_from_expr ex2 )
	 @ ( get_list_var_from_expr ex3 )
    |Cexpr.CALL(tyn,ex,exs) -> get_list_var_from_list_expr exs
    |Cexpr.ADDROF(tyn,ex) -> get_list_var_from_expr ex
    |Cexpr.SIZEOF(tyn1,tyn2) -> []
    |Cexpr.CONST_I(tyn,num,str) -> []
    |Cexpr.CONST_F(tyn,fl,str) -> []
    |Cexpr.STRING(boo,str) -> []
    |Cexpr.INITS(inits) -> []
    |Cexpr.SEQUENCE(exs) -> get_list_var_from_list_expr exs
    |Cexpr.PHI(_,_) -> []
    |Cexpr.PSEUDO_MEMORY(tyn1,tyn2) ->  []
    |Cexpr.INIT_FILL _ -> []
    |Cexpr.PRAGMA(str) -> []
    |Cexpr.ASSUME(ex) -> get_list_var_from_expr ex
    |Cexpr.ASSERT(ex) -> get_list_var_from_expr ex
    |Cexpr.PSEUDO_PROLOGUE -> []
    |Cexpr.PSEUDO_EPILOGUE -> []
    | _ -> assert false

and get_list_var_from_list_expr (exprs : Cexpr.expr_t list) = 
   match exprs with
    | [] -> []
    | ex::rest -> ( get_list_var_from_expr ex ) @ ( get_list_var_from_list_expr rest )

(*In node*)
let print_node (node1 : node) = 
   match node1 with
    |NodeV(s1,s2,t) -> print_string (s1^" ")
    | _  -> assert false

let print_nodes (nodes : node list) = 
   List.iter (fun no -> print_node no; print_string "\n") nodes

(*In canh*)
let print_edge (e : edge) = 
   let Edge(s,d,t,n) = e in print_node s;print_string " --> ";print_node d;print_string ":";print_string "\n"

(*In canh*)
let print_edges (edges : edge list) = 
   List.iter ( print_edge ) edges


let print_graph (graph : (node list * edge list)) = 
   match graph with
    | (nodes,edges) -> begin print_string "\nIn node:\n";print_nodes nodes;print_string "\nIn node:\n";print_edges edges; end
    | _ -> assert false


