(* utils/logUtil.ml *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

module MyLogging = SimpleLogging
module MyLog = SimpleLogging.Log

type component_t = GENERAL | PARSER | SPECS | LA | NONE
type level_t = DEBUG | INFO | WARNING | ERROR | EMERG 

let verbose_components : (component_t, unit) Hashtbl.t = Hashtbl.create 0

let string_of_component (comp : component_t) : string = ( 
    match comp with 
    | GENERAL -> "GENERAL" 
    | PARSER -> "PARSER" 
    | SPECS -> "SPECS"
    | LA -> "LA" 
    | _ -> assert false
) 
 
let init () : unit = ( 
    MyLog.init (); 
    MyLog.set_new_log_level MyLogging.Debug; 
) 
 
let close () : unit = ( 
    MyLog.close (); 
) 

let log (comp : component_t) (level : level_t) (msg : string) : unit = ( 
    (* Only verbose components are allowed to log debug information *) 
    let green_light =
        if (comp = NONE) then
            false
        else
            match level with 
            | DEBUG -> (try let _ = Hashtbl.find verbose_components comp in true
                        with Not_found -> false)
            | _ -> true
    in 
    if (green_light) then 
        (begin 
            let message = ("<" ^ (string_of_component comp) ^ "> " ^ msg) in 
            match level with
            | DEBUG  -> MyLog.debug message
            | INFO -> MyLog.info message
            | WARNING -> MyLog.warn message
            | ERROR -> MyLog.error message
            | EMERG -> MyLog.emerg message
        end) 
)

let component_of_string (s : string) : component_t = (
    let upper_s = String.uppercase s in
    match upper_s with
    | "GENERAL" -> GENERAL
    | "PARSER" -> PARSER
    | "SPECS" -> SPECS
    | "LA" -> LA
    | _ -> 
        (log GENERAL ERROR ("ps option parsing error: Component \""
                            ^ upper_s 
                            ^ "\" not found");
        NONE)
) 

let add_verbose_component (s : string) : unit = (
    let comp = component_of_string s in
    Hashtbl.add verbose_components comp ()
)

let remove_verbose_component (s : string) : unit = (
    let comp = component_of_string s in
    Hashtbl.remove verbose_components comp
)

let enable_debug_mode () : unit = (
    Hashtbl.add verbose_components GENERAL ();
    Hashtbl.add verbose_components PARSER ();
    Hashtbl.add verbose_components SPECS ();
    Hashtbl.add verbose_components LA ())
 
(* an example *) 
let test_log () : unit = ( 
   log PARSER ERROR "Error_parser";
   log LA WARNING "Warning_la";
   log SPECS DEBUG "Debug_specs";
   log PARSER DEBUG "Debug_parser";
   log GENERAL DEBUG "Debug_General";
)
