(* utils/logicUtil.mli *)
(* Copyright (C) 2007-2010 Yonezawa Lab, The University of Tokyo *)

val enable_mathsat_debug_mode : unit -> unit
type solver_t = CSISAT | MATHSAT | HYBRID

module Logic :
  sig
    type symbol = string
    type mix_const = ((Num.num option) * (float option))
    type expression = 
        Constant of mix_const
      | Variable of symbol
      | Application of symbol * expression list
      | Sum of expression list
      | Coeff of mix_const * expression
    type atom_type = CsisatAst.atom_type
    type predicate =
        True
      | False
      | And of predicate list
      | Or of predicate list
      | Not of predicate
      | Eq of expression * expression
      | Lt of expression * expression
      | Leq of expression * expression
      | Atom of atom_type

    (* new features *)

    exception SAT_FORMULAS of (predicate list)

    val init : bool -> unit
    val set_solver : solver_t -> unit
    val string_of_predicate: predicate -> string 
    val string_of_expression : expression -> string

    val zero_mix_const : mix_const
    val create_mix_cst_from_num : Num.num -> mix_const
    val create_mix_cst_from_int : int -> mix_const
    val is_identical_mix_csts : mix_const -> mix_const -> bool

    val get_interpolant : predicate -> predicate -> predicate 
    val get_interpolants : predicate list -> predicate list

    val is_sat: predicate -> bool 
    val is_valid: predicate -> bool

    val approximate_division : predicate -> predicate
    val replace_var_in_expression : expression -> expression -> expression -> expression
    val replace_var_in_predicate: predicate -> expression -> expression -> predicate
    val get_weakest_precondition: predicate -> predicate -> predicate 
    val count_atom_pred_in_pred : predicate -> int
    val count_atom_pred_in_preds : predicate list -> int
    val get_total_logic_module_time : unit -> float
  end

