/*----------------------------------------------------------------------------*/
%{  

    open Printf
    open Lexing
    open MathsatUtil 
    (*FIXME open LogicUtil *)

(* ========================================================================== *)
    
    type expr_pred_t =
        | Expr of expression
        | Pred of predicate

    let expr_preds : (string, expr_pred_t) Hashtbl.t = Hashtbl.create 2
    let out_msat_test = ref 0
    
    let get_expr (e_p : expr_pred_t) : expression =
    begin
        match e_p with
        | Expr ex -> ex
        | _ -> assert false
    end
    
    let get_pred (e_p : expr_pred_t) : predicate =
    begin
        match e_p with
        | Pred pr -> pr
        | _ -> assert false
    end

    let get_const_of_expression (expr : expression) 
    : mix_const option =
    begin
        match expr with
        | Constant cons -> Some cons
        | Coeff (cons, ex) -> (
            let n_op, f_op = cons in
            match n_op, f_op with
            | None, None -> assert false
            | Some n, _ ->
                if (Num.eq_num n (Num.num_of_int (-1))) then
                    match ex with
                    | Constant c ->
                        Some (reverse_sign_mix_cst c)
                    | _ -> None
                else None
            | _, Some f ->  
                match (f, ex) with
                | -1., Constant c ->
                    Some (reverse_sign_mix_cst c)
                | _ -> None)
        | _ -> None
    end

%}

%token LPAREN RPAREN LBRACE RBRACE COLON COMMA
%token PLUS MINUS TIMES DIV 
%token AND_BIT OR_BIT NOT_BIT XOR_BIT RIGHT_S LEFT_S
%token EQ EQL DIFF NLESS NGREATER LESS GREATER
%token IF_ELSE FORMULA_D CONST_D VAR_D DEFINE_D
%token INT_T BOOLEAN_T WORD_T REAL_T OBJ_T
%token AND OR NOT IMPLY XOR EQV NAND
%token TRUE FALSE
%token <string> ID
%token <Num.num> NUM
%token <string> BIT_DEC
%token <float> REAL
%token EOF

%nonassoc prec_no_rel
%left LESS GREATER NLESS NGREATER DIFF EQ
%nonassoc prec_no_shift
%left RIGHT_S LEFT_S

%start all
%type < MathsatUtil.predicate > all

%%
				
all :
    |  clear_list msat_input EOF {
         get_pred (Hashtbl.find expr_preds "formula")
    }
;

clear_list :
    | /* empty */ {
        Hashtbl.clear expr_preds;
        out_msat_test := 0	
    }
;

msat_input : 
    | all_section { }
    | all_section msat_input { }
;

all_section :
    | var_section { }
    | define_section { }
    | const_section { }
    | formula_section { }
;

var_section :
    | VAR_D var_list { }
;

var_list :
    | var_decl { }
    | var_list var_decl { }
;

var_decl :
    | id_list COLON type_msat { }
;

define_section :
    | DEFINE_D define_list { 
    }
;

define_list :
    | define_decl { }
    | define_list define_decl { }
;

define_decl :
    | ID EQL formula {
        Hashtbl.add expr_preds $1 $3
    }
    | ID COLON simple_type EQL formula {
        Hashtbl.add expr_preds $1 $5 
    }
;

type_msat :
    | simple_type { }
    | compound_type { }
;

compound_type :
    | spltype_list IMPLY simple_type { }
;

spltype_list :
    | simple_type { }
    | spltype_list TIMES simple_type { }
;

simple_type :
    | BOOLEAN_T { }
    | WORD_T LBRACE NUM RBRACE { }
    | REAL_T { }
    | OBJ_T { }
    | INT_T { }
;

const_section :
    | CONST_D id_list { assert false }
;

id_list :
    | ID { }
    | id_list COMMA ID { }
;

formula_section :
    | FORMULA_D formula { 
        out_msat_test := !out_msat_test + 1;
        if ( !out_msat_test > 1 ) then (
            assert false
        )
        else Hashtbl.replace expr_preds "formula" $2           
    }
;

formula :
    | formula_ifelse { $1 }
;

formula_ifelse :
    | formula_eqv { $1 }
    | IF_ELSE LPAREN formula COMMA formula_ifelse COMMA formula_ifelse
      RPAREN { assert false } 
;

formula_eqv :
    | formula_imply { $1 }
    | formula_eqv EQV formula_imply {
        assert false
    }
;

formula_imply :
    | formula_or { $1 }
    | formula_imply IMPLY formula_or {
        assert false
    }
;

formula_or :
    | formula_or_list { 
        let len_tmp = List.length $1 in
        if ( len_tmp = 0 ) then (
            assert false
        )
        else if ( len_tmp = 1 ) then (
            List.hd $1
        )
        else (
            let preds = List.map (fun e_p -> get_pred e_p) $1 in
            Pred (Or preds)
        )
    }
;
formula_or_list :
    | formula_xor { [$1] }
    | formula_or_list OR formula_xor {
        $1 @ [$3]
    }
;

formula_xor :
    | formula_nand { $1 }
    | formula_xor XOR formula_nand {
        assert false
    }
;

formula_nand :
    | formula_nand_list {
        let len_tmp = List.length $1 in
        if ( len_tmp = 0 ) then (
            assert false
        )
        else if ( len_tmp = 1 ) then (
            List.hd $1
        )
        else (
            let preds = List.map (fun e_p -> get_pred e_p) $1 in
            Pred (Not (And preds))
        )
    }
;

formula_nand_list :
    | formula_and { [$1] }
    | formula_nand_list NAND formula_and {
        $1 @ [$3]
    }
;

formula_and :
    | formula_and_list {
        let len_tmp = ( List.length $1 ) in
        if ( len_tmp = 0 ) then (
            assert false
        )
        else if ( len_tmp = 1 ) then (
            List.hd $1
        )
        else (
            let preds = List.map (fun e_p -> get_pred e_p) $1 in
            Pred (And preds)
        )
    }
;

formula_and_list :
    | formula_not { [$1] }
    | formula_and_list AND formula_not {
        $1 @ [$3]
    }
;

formula_not :
    | math_atom { $1 }
    | NOT formula_not {
        Pred (Not (get_pred $2))
    }
    
;

math_atom :
    |  expression EQ expression {
        let expr1 = get_expr $1 in
        let expr2 = get_expr $3 in
        Pred (Eq (expr1, expr2))
    }
    | expression DIFF expression {
        let expr1 = get_expr $1 in
        let expr2 = get_expr $3 in
        Pred (Not (Eq (expr1, expr2)))
    }
    | expression LESS expression {
        let expr1 = get_expr $1 in
        let expr2 = get_expr $3 in
        Pred (Lt(expr1, expr2))
    }
    | expression NGREATER expression {
        let expr1 = get_expr $1 in
        let expr2 = get_expr $3 in
        Pred (Leq (expr1, expr2))
    }
    | expression NLESS expression {
        let expr1 = get_expr $1 in
        let expr2 = get_expr $3 in
        Pred (Not (Lt (expr1, expr2)))
    }
    | expression GREATER expression {
        let expr1 = get_expr $1 in
        let expr2 = get_expr $3 in
        Pred (Not (Leq (expr1, expr2)))
    }
    | expression %prec prec_no_rel { $1 }
;

expression :
    | express_orbit { $1 }
;

express_orbit :
    | orbit_list { 
        let len_tmp = List.length $1 in
        if ( len_tmp = 0 ) then (
            assert false
        )
        else if ( len_tmp = 1 ) then (
            List.hd $1
        )
        else (
            let exprs = List.map (fun e_p -> get_expr e_p) $1 in
            Expr (Application ("ior", exprs))
        )
    }
;

orbit_list : 
    | express_xorbit {  
        [$1]
    }   
    | orbit_list OR_BIT express_xorbit {
         $1 @ [$3]
    } 
;

express_xorbit :
    | xorbit_list { 
        let len_tmp = List.length $1 in
        if ( len_tmp = 0 ) then (
            assert false
        )
        else if ( len_tmp = 1 ) then (
            List.hd $1
        )
        else (
            let exprs = List.map (fun e_p -> get_expr e_p) $1 in           
            Expr (Application ( "xor", exprs ))
        )
    }
;

xorbit_list : 
    | express_andbit {  
        [$1]
    }   
    | xorbit_list XOR_BIT express_andbit {
         $1 @ [$3]
    } 
;

express_andbit :
    | andbit_list { 
        let len_tmp = List.length $1 in
        if ( len_tmp = 0 ) then (
            assert false
        )
        else if ( len_tmp = 1 ) then (
            List.hd $1
        )
        else (
            let exprs = List.map (fun e_p -> get_expr e_p) $1 in            
            Expr (Application ( "and", exprs ))
        )
    }
;

andbit_list : 
    | express_shift {  
        [$1]
    }   
    | andbit_list AND_BIT express_shift {
         $1 @ [$3]
    } 
;

shift_op :
    | RIGHT_S { "rsh" }
    | LEFT_S { "lsh" }    
;

express_shift :
    | express_add %prec prec_no_shift { $1 }
    | express_shift shift_op express_add {
        let expr2 = get_expr $3 in
        match expr2 with 
        | Constant _ -> 
            Expr (Application ($2, [(get_expr $1); expr2]))   
        | _ -> assert false
    }
;

express_add :
    | add_list {
        let len_tmp = List.length $1 in
        if ( len_tmp = 0 ) then
            assert false
        else if ( len_tmp = 1 ) then
            List.hd $1
        else (
            let exprs = List.map (fun e_p -> get_expr e_p) $1 in
            Expr (Sum exprs)
        )
    }
;

add_list : 
    | express_multi {  
        [$1]
    }   
    | add_list PLUS express_multi {
         $1 @ [$3]
    } 
    | add_list MINUS express_multi {
        $1 @ [Expr (Coeff ((create_mix_cst_from_int (- 1)), (get_expr $3)))]
    }
;

opt_times :
    | TIMES express_unary { Some (get_expr $2) }
    | /*empty*/ { None }
;

express_multi :
    | express_unary opt_times {
        match $2 with
        | None -> $1
        | Some expr2 -> (
            let expr1 = get_expr $1 in
            let opt_const1 = get_const_of_expression expr1 in
            let opt_const2 = get_const_of_expression expr2 in
            match opt_const1, opt_const2 with 
            | Some const1, _ -> 
                Expr (Coeff (const1, expr2))
            | _, Some const2 ->
                Expr (Coeff (const2, expr1))
            | _ -> assert false    
        ) 
    }
;

express_unary :
    | atom { $1 }    
    | MINUS express_unary {
        Expr (Coeff ((create_mix_cst_from_int (- 1)) , (get_expr $2)))
    }
    | NOT_BIT express_unary {
        Expr (Application ("bitnot" , [get_expr $2]))
    } 

;

express_have_para :
    | expression { [$1] }
    | express_have_para COMMA expression { 
        $1 @ [$3]   
    }
;
 
express_para :
    | /*empty*/ { [] }
    | express_have_para { 
        List.map (fun e_p -> get_expr e_p) $1 
    }
;

opt_apply :
    | LPAREN express_para RPAREN { Some $2 }
    | /*empty*/ { None }
;

const_num :
    | NUM { create_mix_cst_from_num $1 }
    | BIT_DEC { 
        let len = String.length $1 in
        let tmp = String.sub $1 5 ( len - 5 ) in
        let n = Num.num_of_string tmp in
        (Some n, None)  
    }
    | REAL { create_mix_cst_from_float $1 }
;

atom :
    | const_num { 
        Expr (Constant $1) 
    }
    | ID opt_apply {
        match $2 with
        | None -> (
            let lst = Hashtbl.find_all expr_preds $1 in
            let len = List.length lst in
            if ( len = 0 ) then
                Expr (Variable $1)
            else if ( len = 1 ) then
                List.hd lst
            else assert false
        )
        | Some para -> Expr (Application ($1, para))
    }
    | TRUE {
        Pred (True)
    }
    | FALSE {
        Pred (False) 
    }
    | LPAREN formula RPAREN { $2 }   
;

%%      
