{
    open Msat_dec
    open Lexing
} 

let digit     = ['0'-'9']
let ident     = ['a'-'z' 'A'-'Z' '_']
let ident_num = ['a'-'z' 'A'-'Z' '0'-'9' '_']

rule token = parse
    | [' ' '\t']   { token lexbuf }
    | '\n'	   { token lexbuf }
   (* ============================ *)
    | '('          { LPAREN }
    | ')'          { RPAREN } 
    | '['          { LBRACE }
    | ']'          { RBRACE }
    | ':'          { COLON }
    | ','          { COMMA }
   (* ============================ *)
    | '&'          { AND }
    | '|'          { OR }
    | '!'          { NOT } 
    | "->"         { IMPLY }
    | "<->"        { EQV }
   (* ============================ *)
    | ":="         { EQL }
    | "="          { EQ }
    | "!="         { DIFF }
    | ">="         { NLESS }
    | "<="         { NGREATER }
    | '<'          { LESS }
    | '>'          { GREATER }
   (* ============================ *)
    | '+'          { PLUS }
    | '-'          { MINUS }
    | '*'          { TIMES }
    | '/'          { DIV }
   (* ============================ *)
    | "&&"         { AND_BIT }
    | "||"         { OR_BIT }
    | "~~"         { NOT_BIT }  
    | "^^"         { XOR_BIT } 
    | ">>"         { RIGHT_S }
    | "<<"         { LEFT_S }
   (* ============================ *) 
    | digit+ as num 
                   { NUM ( Num.num_of_string num ) }      
    | "0d32_" digit+  as bit_dec 
                   { BIT_DEC (bit_dec)}
    | digit+ '.' digit+ as real 
                   { REAL (float_of_string real) }            
    | ident+ ident_num* as word
                   { match word with
                     | "integer" | "INTEGER" -> INT_T
                     | "boolean" | "BOOLEAN" -> BOOLEAN_T
                     | "word"    | "WORD"    -> WORD_T
                     | "real"    | "REAL"    -> REAL_T
                     | "object"  | "OBJECT"  -> OBJ_T
                    (* ==================================== *)
                     | "implies" | "IMPLIES" -> IMPLY
                     | "iff"     | "IFF"     -> EQV
                     | "and"     | "AND"     -> AND
                     | "or"      | "OR"      -> OR
                     | "xor"     | "XOR"     -> XOR
                     | "nand"    | "NAND"    -> NAND
                     | "not"     | "NOT"     -> NOT
                    (* ==================================== *)
                     | "ite"     | "ITE"     -> IF_ELSE
                    (* ==================================== *)
                     | "true"    | "TRUE"    -> TRUE
                     | "false"   | "FALSE"   -> FALSE
                    (* ==================================== *) 
                     | "formula" | "FORMULA" -> FORMULA_D
                     | "const"   | "CONST"   -> CONST_D
                     | "var"     | "VAR"     -> VAR_D
                     | "define"  | "DEFINE"  -> DEFINE_D
                    (* ==================================== *) 
                     | _                     -> ID (word)
                   }
    | "#"          { comment lexbuf }
    | _            { token lexbuf }
    | eof          { EOF }    
and comment = parse
  | '\n'	{ token lexbuf }
  | _		{ comment lexbuf }
  | eof		{ EOF }
