#ifdef __ONLY_FOR_MODELCHECKING
#include <assert.h>
#include <vmm-mc.h>
#else
#include <kernel/page_fault.h>
#include <kernel/guest.h>
#include <kernel/guest_paging.h>
#include <kernel/logger.h>
#include <kernel/panic.h>
#endif

void
page_fault(regs_t *regs)
{
	u32 cr2 = read_cr2();
	print_detail2("%p: page fault at %p (error_code=0x%08x)\n",(void *)regs->eip,(void *)cr2,regs->error_code);
	
	if ((regs->error_code & PF_ERR_P) == 0 && guest_paging_is_enabled() && update_guest_paging_at(cr2))
		/*
		 * ゲスト側にページ情報が存在したのでページング情報を更
		 * 新して再実行する。
		 */
		return;

	if (SEGSEL_RPL(regs->cs) == VMM_RING) {
		print_panic("********************************\n");
		call_panic(print_regs(regs));
		print_panic("cr2 = %p\n",(void *)cr2);
		PANIC("page fault in vmm kernel.\n");
#warning ページアウトされたゲストのページへのアクセスに対応すべき？
	}


#ifndef __ONLY_FOR_MODELCHECKING
	/*
	 * ゲストに配送する
	 */
	guest_assert(cr2 < VMM_TOP || VMM_END <= cr2);
	cpu.guest.cr[2] = cr2;
	setup_upcall(regs,regs->vector,true,regs->error_code);
#endif
}

