#include <assert.h>

#define VMM_TOP                 (0xfc000000U)
#define NUM_PDE                 (1024)
#define NUM_PTE                 (1024)
#define L1_PAGE_SHIFT           (12)
#define L2_PAGE_SHIFT           (22)
#define L1_PAGE_SIZE            (1 << L1_PAGE_SHIFT)
#define L2_PAGE_SIZE            (1 << L2_PAGE_SHIFT)
#define L1_PAGE_OFFSET_MASK     (L1_PAGE_SIZE - 1)
#define L2_PAGE_OFFSET_MASK     (L2_PAGE_SIZE - 1)

/*@ $requires ((pde_index >= 0) && (pde_index < NUM_PDE)); */
unsigned int get_guest_pde(unsigned int pde_index);

/*@ $requires ((pte_index >= 0) && (pte_index < NUM_PTE)); */
unsigned int get_guest_pte(unsigned int pde, unsigned int pte_index);

/*@ $requires ((vaddr >= 0) && (vaddr < VMM_TOP)); */
void update_guest_paging_at(unsigned int vaddr)
{
        unsigned int gpde, gpte;

        gpde = get_guest_pde(vaddr >> L2_PAGE_SHIFT);
        gpte = get_guest_pte(gpde, (vaddr & L2_PAGE_OFFSET_MASK) >> L1_PAGE_SHIFT);
}

/*@
  $ensures (0 <= $result && $result < VMM_TOP);
 */
unsigned int generate_vaddr(void);

int main()
{
	update_guest_paging_at(generate_vaddr());

	return 0;
}
